package examples.jmarkov;

import jmarkov.MarkovProcess;
import jmarkov.SimpleMarkovProcess;
import jmarkov.basic.Event;
import jmarkov.basic.EventsSet;
import jmarkov.basic.PropertiesState;
import jmarkov.basic.State;
import jmarkov.basic.States;
import jmarkov.basic.StatesSet;

/**
 * Sistema Kanban de una tarjeta.
 * 
 * @author Margarita Arana y Gloria Daz. Universidad de los Andes.
 * @author Germn Riao. Universidad de los Andes.
 */

/**
 * Este es un caso particular de PropertiesState.
 */

class KanbanState extends PropertiesState {
    /**
     * Construye un nuevo estado Kanban.
     * @param s Indica cuales de las estaciones esta activa.
     * @param x Indica cuantas piezas hay en cada buffer.
     * @param y Indica cuantas tarjetas hay en cada caja.
     */
    KanbanState(int[] status, int[] buffer, int[] kanbans) {
        super(status.length + (buffer.length - 1) + kanbans.length);
        int m = status.length;
        for (int i = 0; i < m; i++) {
            prop[i] = status[i];
        }
        for (int i = 0; i < m - 1; i++) {
            prop[i + m] = buffer[i + 1];
        }
        for (int i = 1; i < m; i++) {
            prop[i + (2 * (m - 1))] = kanbans[i - 1];
        }
    }

    @Override
    public void computeMOPs(MarkovProcess mp) {
        int m = getNumStations();
        for (int i = 0; i < m; i++) {
            setMOP(mp, "Utilizacion Servidor " + (i + 1), getStatus(i));
        }
        for (int i = 0; i < m - 1; i++) {
            setMOP(mp, "WIP en Kanbans " + (i + 1), getKanban(i));
        }
        for (int i = 1; i < m; i++) {
            setMOP(mp, "WIP en Buffer " + i, getBuffer(i));
        }
    }

    /**
     * @see jmarkov.basic.State#isConsistent()
     */
    @Override
    public boolean isConsistent() {
        // TODO Complete
        return true;
    }

    /**
     * Obtiene el status de la estacion i
     * @param i ndice de la estacin
     * @return 1 si la estacion esta activa.
     */
    public int getStatus(int i) { // Las estaciones estn numeradas
                                    // desde
        // cero.
        return prop[i];
    }

    /**
     * Obtiene la cantidad de material en el buffer i.
     * @param i ndice del buffer
     * @return Cantidad de material en el buffer i.
     */
    public int getBuffer(int i) { // Los buffers estn numeradas
                                    // desde uno.
        int m = (prop.length + 2) / 3;
        return prop[i + (m - 1)];
    }

    /**
     * Obtiene el nmero de tarjetas en la caja i.
     * @param i ndice de la caja
     * @return Cantidad de tarjetas en la caja i.
     */
    public int getKanban(int i) { // Las cajas de las tarjetas estn
                                    // numeradas
        // desde cero.
        int m = getNumStations();
        return prop[(i + 1) + 2 * (m - 1)];
    }

    /**
     * Obtiene el nmero de estaciones.
     * @return Nmero de estaciones.
     */
    public int getNumStations() {
        return (prop.length + 2) / 3;
    }

    @Override
    public String label() {
        String stg = "";
        int m = (prop.length + 1) / 3 + 1;
        for (int i = 0; i < m; i++) {
            stg += "S" + (i + 1) + "(" + getStatus(i) + ","
                    + ((i > 0) ? "" + getBuffer(i) : "*") + ","
                    + ((i < m - 1) ? "" + getKanban(i) : "*") + ")";
        }
        return stg;
    }

    @Override
    public String description() {
        String stg = "";
        int m = (prop.length + 1) / 3 + 1;
        for (int i = 0; i < m; i++) {
            stg += "STATION " + (i + 1) + "= (Status: " + getStatus(i)
                    + ", Buffer: " + ((i > 0) ? "" + getBuffer(i) : "*")
                    + ", Kanbans: " + ((i < m - 1) ? "" + getKanban(i) : "*")
                    + ")  ";
        }
        return stg;
    }

}

class KanbanEvent extends Event {
    private int station;

    /**
     * Creates an ending process at the given station.
     * @param station
     */
    public KanbanEvent(int station) {
        this.station = station;
    }

    /**
     * @return Returns the station.
     */
    public int getStation() {
        return station;
    }

    /**
     * @param m
     * @return a set with all Events
     */
    public static EventsSet<KanbanEvent> getAllEvents(int m) {
        KanbanEvent set[] = new KanbanEvent[m];
        for (int s = 0; s < m; s++)
            set[s] = new KanbanEvent(s);
        return new EventsSet<KanbanEvent>(set);
    }
}

/**
 * Esta clase implementa un Kanban. Extendiende la clase
 * SimpleMarkovProcess.
 * @author Margarita Arana y Gloria Daz. Universidad de los Andes.
 * @version 1.0a
 */

public class Kanban extends SimpleMarkovProcess<KanbanState, KanbanEvent> {
    private double[] mu;
    private int m;
    private int[] kanban;

    /**
     * Constructor de un Kanban.
     * @param m Nmero de estaciones.
     * @param mu Tasa de servicios de las estaciones.
     * @param kanban Nmero de tarjetas las estaciones.
     * @param s Estado de las estaciones.
     */
    public Kanban(int m, int[] kanban, double[] mu, int[] s) {
        super((new KanbanState(s, new int[m], kanban)), KanbanEvent
                .getAllEvents(m));
        this.m = m;
        this.mu = mu;
        this.kanban = kanban;
    }

    /**
     * Default Constructor for GUI
     */
    public Kanban() {
        this(3, new int[] { 2, 2 }, new double[] { 2.0, 2.0, 2.0 }, new int[] {
                1, 0, 0 });
    }

    /**
     * Describe el sisema
     * @see jmarkov.SimpleMarkovProcess#description()
     */
    @Override
    public String description() {
        String stg = "KANBAN SYSTEM \n" + m + " Stations.";
        for (int i = 0; i < m; i++) {
            stg += "\nStation "
                    + (i + 1)
                    + ": Rate = "
                    + mu[i]
                    + ((i < m - 1) ? ", Kanbans after station: " + kanban[i]
                            : "");
        }

        return stg;
    }

    /**
     * Determina cuando los eventos estn activos
     */
    @Override
    public boolean active(KanbanState i, KanbanEvent e) {
        return (i.getStatus(e.getStation()) == 1);
    }

    @Override
    public States<KanbanState> dests(KanbanState i, KanbanEvent ev) {
        int[] newBuff = new int[m];
        int e = ev.getStation();
        for (int j = 1; j < m; j++) {
            newBuff[j] = i.getBuffer(j);
        }

        int[] newKanban = new int[m - 1];
        for (int j = 0; j < m - 1; j++) {
            newKanban[j] = i.getKanban(j);
        }

        int[] newStatus = new int[m];
        for (int j = 0; j < m; j++) {
            newStatus[j] = i.getStatus(j);
        }

        newStatus[e] = 0;
        // Se desactiva la estacin donde se finaliz el trabajo, e.

        if (e== 0) { // Se termino el trabajo en la estacin 0?
            newBuff[1] = newBuff[1] + 1; // Se deja una pieza en el
                                            // buffer 1.
            if (m == 2 && newStatus[1] == 0) {
                newStatus[1] = 1;
                newKanban[0] = newKanban[0] + 1;
                newBuff[1] = newBuff[1] - 1;
            }
            if (m > 2 && newKanban[1] > 0 && newStatus[1] == 0) {
                // Se activa la estacin 1? (Por lo menos hay una
                // pieza en el
                // buffer para trabajar)
                newStatus[1] = 1;
                newKanban[0] = newKanban[0] + 1;
                newKanban[1] = newKanban[1] - 1;
                newBuff[1] = newBuff[1] - 1;
            }
            if (newKanban[0] > 0) { // Se puede activar la estacin
                                    // 0?
                newStatus[0] = 1;
                newKanban[0] = newKanban[0] - 1;
            }
            return new StatesSet<KanbanState>( new KanbanState(newStatus, newBuff, newKanban));
        } // end if(e==0)

        if (e == m - 1) { // Se termino el trabajo en la ltima
                            // estacin?
            if (newBuff[e] > 0) {
                // Se puede activar la ltima estacin?(solo importa
                // que halla
                // material).
                newStatus[e] = 1;
                newKanban[e - 1] = newKanban[e - 1] + 1;
                newBuff[e] = newBuff[e] - 1;
            }
        } // end if(e==m-1)
        else { // Se termino el trabajo en cualquier otra estacin?
            newBuff[e + 1] = newBuff[e + 1] + 1;
            // Se deja una pieza en el siguiente buffer.
            if ((e < m - 2) && (newKanban[e + 1] > 0)
                    && (newStatus[e + 1] == 0)) {
                // Se activa la estacin e+1? No considera la
                // penltima
                // estacin.(Por lo menos hay una pieza en el buffer
                // para
                // trabajar)
                newStatus[e + 1] = 1; // Activo la estacin e+1
                newKanban[e] = newKanban[e] + 1; // Se deja tarjeta
                                                    // en la
                // caja e
                newKanban[e + 1] = newKanban[e + 1] - 1;
                // Se quita tarjeta en la caja e+1
                newBuff[e + 1] = newBuff[e + 1] - 1;
                // Se toma la pieza del buffer e+1
            }
            if (e == m - 2 && newStatus[e + 1] == 0) {
                // Se termino el trabajo en la estacin m-2?
                newStatus[e + 1] = 1;
                // Activo la ltima estacin. Solo necesita material.
                newKanban[e] = newKanban[e] + 1; // Se deja tarjeta
                                                    // en la
                // caja e
                newBuff[e + 1] = newBuff[e + 1] - 1; // Se toma
                                                        // materioa
            }
            if (newKanban[e] > 0 && newBuff[e] > 0) {
                // Se activa la estacin e?(esta se acabo de
                // desactivar)
                newStatus[e] = 1; // Se activa la estacin e
                newKanban[e - 1] = newKanban[e - 1] + 1;
                // Se deja tarjeta en la caja e-1
                newKanban[e] = newKanban[e] - 1;
                // Se quita tarjeta de la caja e
                newBuff[e] = newBuff[e] - 1; // SE toma la pieza
                                                // del buffer e
            }
        } // end else (se termino en cualquier otra estacin).

        for (int j = e - 1; j >= 0; j--) {
            if (newStatus[j] == 1) { // Esta activo el estado j?
                return new StatesSet<KanbanState>(new KanbanState(newStatus, newBuff, newKanban));
            }

            if (j == 0) {
                if (newKanban[j] > 0) {
                    newStatus[j] = 1;
                    newKanban[j] = newKanban[j] - 1;
                }
                return new StatesSet<KanbanState>(new KanbanState(newStatus, newBuff, newKanban));
            } // end if(j==0)

            if (newBuff[j] == 0 || newKanban[j] == 0) { // Se puede
                                                        // activar el
                // estado j?
                return new StatesSet<KanbanState>(new KanbanState(newStatus, newBuff, newKanban));
            }
            newStatus[j] = 1;
            newKanban[j - 1] = newKanban[j - 1] + 1;
            newKanban[j] = newKanban[j] - 1;
            newBuff[j] = newBuff[j] - 1;
        } // end for

        return new StatesSet<KanbanState>(new KanbanState(newStatus, newBuff, newKanban));
    } // end de la funcin dest

    /**
     * Retorna la tasa de servicio de cada estacin.
     * @see SimpleMarkovProcess#rate(State, State, Event)
     */
    @Override
    public  double rate(KanbanState i, KanbanState j, KanbanEvent e) {
        return mu[e.getStation()];
    }

    /**
     * Main method
     * @param a Not used
     */
    public static void main(String[] a) {

        int[] tarj = { 2, 2 };
        double[] mu = { 2.0, 2.0, 2.0 };
        int[] s = { 1, 0, 0 };
        Kanban theKanban = new Kanban(s.length, tarj, mu, s);
        theKanban.showGUI();
        theKanban.generate();
        theKanban.printAll();
    }

} // class end
