/*
 * Created on 7/08/2005
 */
package examples.jmdp;

import java.io.PrintWriter;

import jmarkov.MarkovProcess;
import jmarkov.basic.Actions;
import jmarkov.basic.ActionsSet;
import jmarkov.basic.Event;
import jmarkov.basic.Events;
import jmarkov.basic.EventsSet;
import jmarkov.basic.PropertiesAction;
import jmarkov.basic.PropertiesState;
import jmarkov.basic.States;
import jmarkov.basic.StatesSet;
import jmarkov.basic.exceptions.SolverException;
import jmarkov.jmdp.CTMDPEvA;
import examples.jmdp.BankEvent.BEvent;

/**
 * The present example describes a Bank office with "maxServers" tellers. There
 * is a queue for clients for whom the bank wants to give the best service, and
 * another queue for users who have no commercial relation with the bank and are
 * not the service priority. At every transition, the bank can add or remove a
 * teller serving one of the queues and remove him or add him to the other
 * queue.
 * 
 * @author sarmiea
 * 
 */
public class BankWithEvents extends
        CTMDPEvA<BankQueues2, BankServers2, BankEvent> {

    double clientMinuteCost, userMinuteCost, clientProbability, lambda, mu;
    int maxCapacity, maxServers;

    /**
     * Builds a BankWithEvents
     * 
     * @param initial
     *            Initial state in the system
     * @param clientMinuteCost
     *            The cost that is charged to clients with commercial relation
     *            with the bank per minute they stay in queue
     * @param userMinuteCost
     *            The cost that is charged to users with no commercial relation
     *            with the bank per minute they stay in queue
     * @param clientProbability
     *            Probability that the next client has a commercial relation
     *            with the bank
     * @param lambda
     *            Arrival rate to the bank
     * @param mu
     *            Service rate of tellers
     * @param maxCapacity
     *            Queues capacity
     * @param maxServers
     *            Maximum number of tellers available in the bank
     */
    public BankWithEvents(States<BankQueues2> initial, double clientMinuteCost,
            double userMinuteCost, double clientProbability, double lambda,
            double mu, int maxCapacity, int maxServers) {
        super(initial);
        this.clientMinuteCost = clientMinuteCost;
        this.userMinuteCost = userMinuteCost;
        this.clientProbability = clientProbability;
        this.lambda = lambda;
        this.mu = mu;
        this.maxCapacity = maxCapacity;
        this.maxServers = maxServers;
    }

    @Override
    public Events<BankEvent> activeEvents(BankQueues2 i) {
        EventsSet<BankEvent> eventSet = new EventsSet<BankEvent>();
        if (i.getClients() + i.getUsers() < maxCapacity) {
            eventSet.add(new BankEvent(BEvent.CLIENTarrival));
            eventSet.add(new BankEvent(BEvent.USERarrival));
        }
        if (i.getClients() > 0)
            eventSet.add(new BankEvent(BEvent.CLIENTdeparture));
        if (i.getUsers() > 0)
            eventSet.add(new BankEvent(BEvent.USERDeparture));
        return eventSet;
    }

    @Override
    public double lumpCost(BankQueues2 i, BankServers2 a, BankEvent e) {
        return 0.0;
    }

    @Override
    public double continuousCost(BankQueues2 i, BankServers2 a, BankEvent e) {
        int waitingClients = Math.max(i.getClients() - i.getClientServers(), 0);
        int waitingUsers = Math.max(i.getUsers()
                - (maxServers - i.getClientServers()), 0);
        return clientMinuteCost * waitingClients + userMinuteCost
                * waitingUsers;
    }

    @Override
    public States<BankQueues2> reached(BankQueues2 i, BankServers2 a,
            BankEvent e) {
        StatesSet<BankQueues2> set = new StatesSet<BankQueues2>();
        int servers = i.getClientServers();
        int ac = a.getClientServers();
        switch (e.get()) {
        case CLIENTarrival: {
            int[] temp = { i.getClients() + 1, i.getUsers(), servers };
            set.add(new BankQueues2(temp));
            break;
        }
        case USERarrival: {
            int[] temp = { i.getClients(), i.getUsers() + 1, servers };
            set.add(new BankQueues2(temp));
            break;
        }
        case CLIENTdeparture: {
            int[] temp = { i.getClients() - 1, i.getUsers(), servers + ac };
            set.add(new BankQueues2(temp));
            break;
        }
        case USERDeparture: {
            int[] temp = { i.getClients(), i.getUsers() - 1, servers + ac };
            set.add(new BankQueues2(temp));
            break;
        }
        }
        return set;
    }

    @Override
    public double rate(BankQueues2 i, BankQueues2 j, BankServers2 a, BankEvent e) {
        int clients = i.getClients();
        int users = i.getUsers();
        int clientServers = i.getClientServers();
        switch (e.get()) {
        case CLIENTarrival: {
            if (a.getClientServers() == 0) // no change in servers
                return lambda * clientProbability;
        }
        case USERarrival: {
            if (a.getClientServers() == 0) // no change in servers
                return lambda * (1 - clientProbability);
        }
        case CLIENTdeparture: {
            if (a.getClientServers() <= 0)
                return Math.min(clients, clientServers) * mu;
        }
        case USERDeparture: {
            if (a.getClientServers() >= 0)
                return Math.min(users, maxServers - clientServers) * mu;
        }
        }
        return 0.0;
    }

    @Override
    public Actions<BankServers2> feasibleAct(BankQueues2 i) {
        ActionsSet<BankServers2> set = new ActionsSet<BankServers2>();
        if (i.getClients() > 0) {// client departures are possible
            if (i.getClientServers() > 1) { // possible to decrease client
                // servers
                int[] temp = { -1 };
                set.add(new BankServers2(temp));
            }
        }
        if (i.getUsers() > 0) {// user departures are possible
            if (i.getClientServers() < maxServers - 1) { // possible to
                // increase client
                // servers
                int[] temp = { +1 };
                set.add(new BankServers2(temp));
            }
        }// do nothing
        int[] temp = { 0 };
        set.add(new BankServers2(temp));

        return set;
    }

    /**
     * This method just tests the class.
     * 
     * @param args
     *            Not used
     * @throws SolverException
     */
    public static void main(String[] args) throws SolverException {
        int maxCapacity = 10, maxServers = 3;
        double clientMinuteCost = 10, userMinuteCost = 4, clientProbability = 0.4, lambda = maxServers * 2, mu = 2;
        int[] initState = { 0, 0, 1 };
        States<BankQueues2> init = new StatesSet<BankQueues2>(new BankQueues2(
                initState));

        BankWithEvents prob = new BankWithEvents(init, clientMinuteCost,
                userMinuteCost, clientProbability, lambda, mu, maxCapacity,
                maxServers);

        // ValueIterationSolver<BankQueues2,BankServers2> solv = new
        // ValueIterationSolver<BankQueues2,
        // BankServers2>(prob,0.9);
        // prob.getSolver().setGaussSeidel(false);

        prob.solve(0.05);
        // prob.printSolution();
        prob.getOptimalValueFunction().print(new PrintWriter(System.out, true));

        // System.out.println(solv.getProcessTime()+" milliseconds");
        // PrintWriter pw = new PrintWriter(System.out,true);
        // ProbabilitySolver<BankQueues2,BankServers2> solf =new
        // ProbabilitySolver<BankQueues2,BankServers2>(prob);
        // solf.setJacobi(true);
        // prob.setProbabilitySolver(solf);
        // ValueFunction<BankQueues2> probs =
        // prob.getSteadyStateProbabilities();
        // probs.print(pw,"%-12S", "%10.5f");
        // System.out.println("\n"+prob.length() +" average client queue
        // length");

    }

}

class BankQueues2 extends PropertiesState {
    /**
     * @param people
     *            State of the bank
     */
    public BankQueues2(int[] people) {
        super(people);
    }

    /**
     * @return Number of clients in the system
     */
    public int getClients() {
        return prop[0];
    }// in system

    /**
     * @return Number of Users in the system
     */
    public int getUsers() {
        return prop[1];
    }// in system

    /**
     * @return Number of servers available for queue 1.
     */
    public int getClientServers() {
        return prop[2];
    }// servers for queue1

    /**
     * @see jmarkov.basic.State#isConsistent()
     */
    @Override
    public boolean isConsistent() {
        // TODO Auto-generated method stub
        return true;
    }

    /**
     * @see jmarkov.basic.State#computeMOPs(MarkovProcess)
     */
    @Override
    public void computeMOPs(MarkovProcess mp) {
        // TODO Auto-generated method stub
    }

}

class BankServers2 extends PropertiesAction {
    /**
     * @param clientServers
     *            Parameters to construct a BankServers2 object
     */
    public BankServers2(int[] clientServers) {
        super(clientServers);
    }

    /**
     * @return Number of servers attending clients queue
     */
    public int getClientServers() {
        return getProperty(0);
    }
}

class BankEvent extends Event {
    /**
     * @author German Riano. Universidad de los Andes. (C) 2006
     * 
     */
    public enum BEvent {
        /** When a client arrives to the bank */
        CLIENTarrival,
        /** When a user arrives to the bank */
        USERarrival,
        /** When a client leaves the bank */
        CLIENTdeparture,
        /** When a user leaves the bank */
        USERDeparture;
    }

    private BEvent e;

    /**
     * @param e
     *            An event of type client arrival, user arrival, client
     *            departure, user departure
     */
    public BankEvent(BEvent e) {
        this.e = e;
    }

    /**
     * @return An event of type client arrival, user arrival, client departure,
     *         user departure
     */
    public BEvent get() {
        return e;
    }

    @Override
    public String label() {
        return e.name();
    }

    @Override
    public String description() {
        return "Event : " + e.name();
    }

    @Override
    public int compareTo(Event e1) {
        if (!(e1 instanceof BankEvent))
            return -1;
        BankEvent e2 = (BankEvent) e1;
        return e.compareTo(e2.e);
    }
}