package examples.jmdp;

import java.util.Iterator;

import jmarkov.MarkovProcess;
import jmarkov.basic.Actions;
import jmarkov.basic.ActionsSet;
import jmarkov.basic.PropertiesAction;
import jmarkov.basic.PropertiesState;
import jmarkov.basic.States;
import jmarkov.basic.StatesSet;
import jmarkov.jmdp.DTMDP;
import jmarkov.jmdp.solvers.ValueIterationSolver;

/**
 * 
 * This file models a multi item inventory problem. The demands are random and
 * independent in between them, with Poisson distribution with means "lambda[]".
 * 
 * @author Andres Sarmiento, Germn Riao - Universidad de Los Andes
 */
public class MultiItems extends DTMDP<MultiItemState, MultiItemOrder> {

    int maxInventory = 15;// Capacity

    int maxItemsPerOrder = 6;

    int typesOfItems = 1; // starting in 0. indicates the maximum index

    // of the types.

    int[] maxBackorders = { 2, 3 };// Capacity

    double K = 500;// Fixed cost per Order

    double cost[] = { 20000, 20000 }; // variable cost

    double price[] = { 23000, 22000 }; // variable cost

    double h = Math.pow(1.3, 1 / 52) - 1; // percentual cost of holding

    // inventory

    double holdingCost[] = { h * cost[0], h * cost[1] }; // holding cost

    // per item

    // per period.

    double demandProbability[][]; // Demand probability function P[i][t] =

    // P[D(t) = i]

    double demandCumulativeProbability[][]; // q(i) Q[i][t] = P[D(t) >= i]

    double[] theta = { 1, 3 }; // means of the Poisson demands per stage

    double[] backOrderCost = { 1000, 1000 }; // cost incurred when there is
    // no

    // stock to satisfy an order.
    // Charged only once.

    ActionsSet<MultiItemOrder> actions;

    // Constructor

    /**
     * @param initSet
     * @param maxInventory
     *            Capacity
     * @param maxItemsPerOrder
     *            maximum mumber of items per order
     * @param typesOfItems
     *            starting in 0. indicates the maximum index of the types.
     * @param maxBackorders
     *            Capacity
     * @param K
     *            Fixed cost per Order
     * @param cost
     *            variable cost
     * @param price
     *            variable cost
     * @param h
     *            percentual cost of holding inventory
     * @param holdingCost
     *            holding cost per item per period.
     * @param theta
     *            means of the Poisson demands per stage
     * @param backOrderCost
     *            cost incurred when there is no
     */
    public MultiItems(States<MultiItemState> initSet, int maxInventory, // Capacity
            int maxItemsPerOrder, int typesOfItems, // number of SKUs
            int[] maxBackorders,// Capacity
            double K,// Fixed cost per Order
            double cost[],// variable cost
            double price[],// variable cost
            double h, // percentual cost of holding inventory
            double holdingCost[], // holding cost per item per period.
            double[] theta,// means of the Poisson demands per stage
            double[] backOrderCost) // cost incurred when there is no
    {
        super(initSet);
        this.maxInventory = maxInventory;
        this.maxItemsPerOrder = maxItemsPerOrder;
        this.typesOfItems = typesOfItems;
        this.maxBackorders = maxBackorders;
        this.K = K;
        this.cost = cost;
        this.price = price;
        this.h = h;
        this.holdingCost = holdingCost;
        this.theta = theta;
        this.backOrderCost = backOrderCost;
        // initializeStates();
        initializeActions();
    }

    /*
     * (non-Javadoc)
     * 
     * @see jmdp.InfiniteMDP#getAllStates()
     */

    double fixedOrderCost(int totalOrderSize) { // Transport cost or some cost
        // associated with the whole
        // order.
        return (totalOrderSize > 0) ? Math.ceil(totalOrderSize
                / (double) maxItemsPerOrder)
                * K : 0.0;
    }

    double variableOrderCost(int itemOrderSize, int type) { // Order
        // cost due to the purchase cost of each item.
        return cost[type] * itemOrderSize;
    }

    double lostOrderCost(int x, int type) { // see documentation for the
        // explanation for this forumula
        int b = maxBackorders[type];
        double expectedBackorders = 0;
        for (int k = Math.max(x + 1, 0); k <= x + b; k++) {
            expectedBackorders += (k - x) * demandProbability[k][type];
        }

        double expectedLostDemand = demandCumulativeProbability[x + b][type]
                * (theta[type] - (x + b)) + (x + b)
                * demandProbability[x + b][type];

        double toReturn = (price[type] - cost[type]) * expectedLostDemand
                + backOrderCost[type] * expectedBackorders;
        return toReturn;
    }

    private void initializeProbabilities() {
        double[] p = new double[typesOfItems];
        double[] q = new double[typesOfItems];
        int maxBO = 0;
        for (int t = 0; t < typesOfItems; t++) {
            if (maxBO < maxBackorders[t])
                maxBO = maxBackorders[t];
        }
        demandProbability = new double[maxBO + maxInventory + 1][typesOfItems];
        demandCumulativeProbability = new double[maxBO + maxInventory + 1][typesOfItems];

        for (int t = 0; t < typesOfItems; t++) {
            p[t] = Math.exp(-theta[t]);
            q[t] = p[t];
            demandProbability[0][t] = p[t];
            demandCumulativeProbability[0][t] = 1; // P[demand t >= 0]
            for (int i = 1; i <= maxBO + maxInventory; i++) {
                p[t] = p[t] * theta[t] / i; // P[demand t = i]
                demandCumulativeProbability[i][t] = 1 - q[t]; // P[demand t >=
                // i]
                q[t] += p[t]; // P[demand t <= i]
                demandProbability[i][t] = p[t];
            }
        }
    }

    /**
     * This function initializes all the possible actions. Starts by creating
     * all the possible orders of size 0, then size 1, etc.
     * 
     */
    void initializeActions() {
        actions = new ActionsSet<MultiItemOrder>();
        int sumBackorders = 0;
        for (int k = 0; k < typesOfItems; k++) {
            sumBackorders += maxBackorders[k];
        }
        for (int sum = 0; sum <= maxInventory + sumBackorders; sum++) {
            int[] vector = new int[typesOfItems];
            actionCreator(typesOfItems - 1, sum, sum, 0, vector);
        }
    }

    /**
     * @param p
     *            the position in the vector that is being modified in this
     *            recursion
     * @param sum
     *            the total amount of items ordered in this Order
     * @param left
     *            the amount of items that have not yet been assigned in
     *            previous positions of the vector and must be assigned
     * @param vector
     *            the Order vector indicating the amount of items ordered per
     *            item.
     */
    void actionCreator(int p, int sum, int left, int acumInv, int[] vector) {
        if (p == 0) {
            acumInv += Math.max(left - maxBackorders[p], 0);
            if (acumInv <= maxInventory) {
                vector[p] = left;
                actions.add(new MultiItemOrder(vector, sum));
            }
        } else {
            for (int i = 0; i <= left; i++) {
                // acumInv += Math.max(i-maxBackorders[p],0);
                int newAcumInv = acumInv + Math.max(i - maxBackorders[p], 0);
                if (acumInv <= maxInventory) {
                    int[] a = new int[vector.length];
                    System.arraycopy(vector, 0, a, 0, vector.length);
                    a[p] = i;
                    actionCreator(p - 1, sum, left - i, newAcumInv, a);
                }
            }
        }
    }

    /**
     * This function initializes all the possible states. Starts by creating all
     * the possible states with total inventoty in 0 and iterates over the
     * possible total inventories. For each, it iterates growing the number of
     * backorders for the positions with no inventory.
     * 
     */
    // void initializeStates() {
    // states = new StatesCollection<MultiItemState>();
    // for (int inventory = 0; inventory <= maxInventory; inventory++) {
    // int[] vector = new int[typesOfItems ];
    // inventoryCreator(typesOfItems-1, inventory, inventory, vector);
    // }
    // }
    // void inventoryCreator(int p, int inventory, int left, int[] vector) {
    // if (p == 0) {
    // if (left == 0) {
    // for (int b = 0; b <= maxBackorders[p]; b++) {
    // vector[p] = -b;
    // states.add(new MultiItemState(vector, inventory));
    // }
    // }
    // vector[p] = left;
    // states.add(new MultiItemState(vector, inventory));
    // } else {
    // for (int i = 0; i <= left; i++) {
    // int[] a = new int[vector.length];
    // System.arraycopy(vector, 0, a, 0, vector.length);
    // a[p] = i;
    // if (i == 0) {
    // for (int b = 0; b <= maxBackorders[p]; b++) { //generate
    // // backorders
    // // for item p
    // int[] v = new int[vector.length];
    // System.arraycopy(vector, 0, v, 0, vector.length);
    // v[p] = -b;
    // inventoryCreator(p - 1, inventory, left - i, v);
    // //statesSet.add(new MultiItemState(v, inventory));
    // }
    // }
    // inventoryCreator(p - 1, inventory, left - i, a);
    // }
    // }
    // }
    @Override
    public States<MultiItemState> reachable(MultiItemState i, MultiItemOrder a) {
        StatesSet<MultiItemState> stSet = new StatesSet<MultiItemState>();
        int[] inventory = i.getStatus();
        int[] size = a.getSize();
        for (int n = -maxBackorders[0]; n <= inventory[0] + size[0]; n++) {
            for (int m = -maxBackorders[1]; n <= inventory[1] + size[1]; n++)
                stSet.add(new MultiItemState(new int[] { n, m }));
        }
        return stSet;

    }

    @Override
    public double prob(MultiItemState i, MultiItemState j, MultiItemOrder a) {
        int[] iLevel = i.getStatus();
        int[] jLevel = j.getStatus();
        int[] orderSize = a.getSize();
        int inventoryCapacity = maxInventory - i.totalInventory;
        int totalOrderSize = a.totalStock;

        boolean positiveDemand = true;

        for (int n = 0; n < iLevel.length; n++) {
            positiveDemand = positiveDemand
                    && (jLevel[n] <= orderSize[n] + iLevel[n]); // assures
            // positive demands for all items
        }
        double intersectionProbability = 0;
        if ((inventoryCapacity >= totalOrderSize) && positiveDemand) {
            intersectionProbability = 1;
            // feasable order doesn�t exceed the inventory capacity
            for (int n = 0; n < iLevel.length; n++) {
                int demand = orderSize[n] + iLevel[n] - jLevel[n];
                if (jLevel[n] == -maxBackorders[n])
                    intersectionProbability *= demandCumulativeProbability[demand][n];
                else
                    intersectionProbability *= demandProbability[demand][n];
            }
        }
        return intersectionProbability;
    }

    @Override
    public double immediateCost(MultiItemState i, MultiItemOrder a) {
        int[] iLevel = i.getStatus();
        double totalCost = fixedOrderCost(a.totalStock);
        for (int n = 0; n < iLevel.length; n++) {
            totalCost += lostOrderCost(iLevel[n] + a.getSize(n), n);
            // totalReward += variableOrderCost(orderSize[n], n);
            totalCost += holdingCost[n] * Math.max(iLevel[n], 0);
        }
        return totalCost;

    }

    void init() {
        initializeProbabilities();
        initializeActions();
    }

    @Override
    public Actions<MultiItemOrder> feasibleActions(MultiItemState i) {
        Actions<MultiItemOrder> toReturn = new ActionsSet<MultiItemOrder>(
                actions);
        for (Iterator<MultiItemOrder> a = toReturn.iterator(); a.hasNext();) {
            MultiItemOrder o = a.next();
            int newTotalInventory = 0;
            for (int t = 0; t < typesOfItems; t++) {
                newTotalInventory += Math.max(o.getSize(t) + i.getStatus()[t],
                        0);
            }
            if (newTotalInventory > maxInventory)
                a.remove();
        }
        return toReturn;
    }

    /**
     * @param a
     *            Not used
     * @throws Exception
     */
    public static void main(String a[]) throws Exception {
        int maxInventory = 15;// Capacity
        int maxItemsPerOrder = 6;
        int typesOfItems = 2; // starting in 0. indicates the maximum index
        // of the types.
        int[] maxBackorders = { 2, 3 };// Capacity
        double K = 500;// Fixed cost per Order
        double cost[] = { 20000, 20000 }; // variable cost
        double price[] = { 23000, 22000 }; // variable cost
        double h = Math.pow(1.3, 1 / 52) - 1; // percentual cost of holding
        // inventory
        double holdingCost[] = { h * cost[0], h * cost[1] }; // holding cost
        // per item
        // per period.
        double[] theta = { 1, 3 }; // means of the Poisson demands per stage
        double[] backOrderCost = { 1000, 1000 }; // cost incurred when there
        // is
        // no
        MultiItemState initialState = new MultiItemState(new int[] { 0, 0 });
        States<MultiItemState> initSet = new StatesSet<MultiItemState>(
                initialState);

        MultiItems prob = new MultiItems(initSet, maxInventory,
                maxItemsPerOrder, typesOfItems, maxBackorders, K, cost, price,
                h, holdingCost, theta, backOrderCost);

        prob.init();
        ValueIterationSolver<MultiItemState, MultiItemOrder> theSolver = null;
        theSolver = new ValueIterationSolver<MultiItemState, MultiItemOrder>(
                prob, 0.9);
        // PolicyIterationSolver theSolver = new PolicyIterationSolver(pro,
        // 0.9);
        // theSolver.setInitVal(-188000);
        // theSolver.setDelta(0.00001);
        theSolver.useErrorBounds(false);
        // theSolver.setGaussSeidel(false);
        theSolver.solve();
        prob.printSolution();
    }

}// class end

/*
 * Class InvLevel: Inventory Level
 */

class MultiItemState extends PropertiesState {
    int totalInventory = 0;

    MultiItemState(int[] k) {
        super(k);
        // this.totalInventory = totalInventory;
        for (int n = 0; n < prop.length; n++)
            totalInventory += prop[n];
    }

    /**
     * @return Inventory level
     */
    public int[] getStatus() {
        return prop;
    }

    /**
     * @see jmarkov.basic.State#computeMOPs(MarkovProcess)
     */
    @Override
    public void computeMOPs(MarkovProcess mp) {
        // TODO Auto-generated method stub
    }

    /**
     * @see jmarkov.basic.State#isConsistent()
     */
    @Override
    public boolean isConsistent() {
        // TODO Auto-generated method stub
        return true;
    }
}

/*
 * Class Order
 */

class MultiItemOrder extends PropertiesAction {
    int totalStock;

    MultiItemOrder(int[] orderSize, int stock) {
        super(orderSize);
        totalStock = stock;
    }

    @Override
    public String label() {
        return "Order " + super.label() + " Units";
    }

    /**
     * Returns an array with the size order for every item.
     * 
     * @return array of order sizes
     */
    public int[] getSize() {
        return getProperties();
    }

    /**
     * @param item
     *            Item index
     * @return Returns the size order for the given item
     */
    public int getSize(int item) {
        return getProperty(item);
    }

}