package examples.jmdp;

/*
 * Created on 25/11/2004
 *
 */
import java.util.ArrayList;
import java.util.List;

import jmarkov.MarkovProcess;
import jmarkov.basic.Action;
import jmarkov.basic.Actions;
import jmarkov.basic.ActionsSet;
import jmarkov.basic.PropertiesAction;
import jmarkov.basic.State;
import jmarkov.basic.StateC;
import jmarkov.basic.States;
import jmarkov.basic.StatesSet;
import jmarkov.jmdp.StochasticShortestPath;

/**
 * 
 * @author Juan F Redondo - Universidad de los Andes
 * 
 * This class extends StateC, you could name the state as you wish, and you can
 * include a parameter to order the state sequence.
 */
class Distance extends StateC {
    private int num;

    /**
     * Build up class for non-terminal states
     * 
     * @param k Max distance
     */
    public Distance(int k) {
        super();
        num = k;
    }

    /**
     * Build up class for the terminal state
     * 
     * @param k
     * @param t
     */
    public Distance(int k, boolean t) {
        super(t);
        num = k;
    }

    @Override
    public String label() {
        return "In the " + num + " distance(s) between Spider and the Fly";
    }

    /*
     * This method order the states by a criteria.
     */
    @Override
    public int compareTo(State state) {
        return num - ((Distance) state).getDist();
    }

    /**
     * This method return the distance between the spider and the fly.
     * 
     * @return Returns the num.
     */
    public int getDist() {
        return num;
    }

    /*
     * @see jmarkov.basic.State#computeMOPs(jmarkov.MarkovProcess)
     */
    @Override
    public void computeMOPs(MarkovProcess<?, ?> model) {

    }

    /**
     * @see jmarkov.basic.State#isConsistent()
     */
    @Override
    public boolean isConsistent() {
        return true;
    }

}

class SpiderAction extends PropertiesAction {

    /**
     * Creates an action for the fly. 0 == Move, 1 == Don't move.
     * 
     * @param action
     */
    public SpiderAction(int action) {
        super(1);
        setProperty(0, action);
    }

    @Override
    public String label() {
        return (getProperty(0) == 0) ? "Move" : "Don't move";
    }

}

/**
 * @author Juan F. Redondo , German Riao- Universidad de los Andes
 * 
 */
public class TheSpiderAndTheFly extends
        StochasticShortestPath<Distance, SpiderAction> {

    static SpiderAction move = new SpiderAction(0);
    static SpiderAction stay = new SpiderAction(1);
    static SpiderAction vector[] = { move, stay };
    static Actions<SpiderAction> actions = new ActionsSet<SpiderAction>(vector);

    // inicializate the actions

    double moveCost; // Move Cost of The Spider
    double stayCost; // Stay Cost of the Spider
    double p; // probability.
    private double probabilities[];
    private double probabilitiesMove[];
    private double probabilitiesDontMove[];

    /**
     * @param between
     * @param moveCost
     * @param p
     * @param stayCost
     */
    public TheSpiderAndTheFly(int between, double moveCost, double stayCost,
            double p) {
        super(initStates(between));
        this.moveCost = moveCost;
        this.p = p;
        this.stayCost = stayCost;

        // For a Fly approximation move, don't move and far localization move
        // for distances greather and equal than two

        probabilities = new double[] { p, 1 - 2 * p, p };

        // Transition probabilities for a Fly approximation move and don't move,
        // if the Spider choose a move action.
        probabilitiesMove = new double[] { 1 - 2 * p, 2 * p };// Transition

        // Transition probabilities for a Fly aproximation move, don't move and
        // far localization move.
        probabilitiesDontMove = new double[] { p, 1 - 2 * p, p };
    }

    /**
     * This method initializes all the state which can be reached in the graph.
     * 
     * @return All the states space.
     */
    private static States<Distance> initStates(int InitialDistanceBetween) {
        Distance[] vector = new Distance[InitialDistanceBetween + 1];
        Distance initial = new Distance(0, true);
        vector[0] = initial;
        for (int i = 1; i <= InitialDistanceBetween; i++) {
            initial = new Distance(i);
            vector[i] = initial;
        }
        return new StatesSet<Distance>(vector);

    }

    /**
     * This method is easy to design, the immediate reward is directly
     * associated with the incurred action, in that way, you have to associate a
     * reward with the action taken. However, you can discriminate by states, It
     * is very IMPORTANT: Associate a zero immediate reward to terminal state f.
     * 
     * @return The immediate reward o cost associate with an action and a state.
     * @see jmarkov.jmdp.DTMDP#immediateCost(State, Action)
     */
    @Override
    public double immediateCost(Distance i, SpiderAction a) {
        if (a == move) {
            return moveCost;
        } else if (i.isTerminal())
            return 0.0;
        else
            return stayCost;
    }

    /**
     * This method return the associate transition probability between two
     * states by taking an action.
     * 
     * @return A transition probability between two states associate to an
     *         action
     * @see jmarkov.jmdp.DTMDP#prob(State, State, Action)
     */
    @Override
    public double prob(Distance i, Distance j, SpiderAction a) {
        int n = i.getDist();
        int m = j.getDist();
        if (n >= 2) {
            if (n == m)
                return this.probabilities[0];
            else if (n == m - 1)
                return this.probabilities[1];
            else if (n == m - 2)
                return this.probabilities[2];
        } else {
            if (a == move) {
                if (n == 1 && m == 1)
                    return this.probabilitiesMove[1];
                else if (n == 1 && m == 0)
                    return this.probabilitiesMove[0];
            } else if (a == stay) {
                if (n == 1 && m == 0)
                    return this.probabilitiesDontMove[0];
                else if (n == 1 && m == 1)
                    return this.probabilitiesDontMove[1];
                else if (n == 1 && m == 2)
                    return this.probabilitiesDontMove[2];
            }

        }
        
        return 0.0;
    }

    /*
     * This method return all the reachables states, from a StateC i by taking
     * an action a. Be carefull, the terminal state must have the true parameter
     * I will recommend before this method implementation TO design the
     * problem's graph.
     * 
     * @return States
     * 
     * @see StochasticShortestPath#reachable(State, Action)
     */
    @Override
    public States<Distance> reachable(Distance i, SpiderAction a) {
        int l = i.getDist();
        if (l >= 2) {
            State next = new Distance(l - 2);
            if (l == 2) {
                next = new Distance(0, true);
            }
            State next2 = new Distance(l - 1);
            State next3 = new Distance(l);
            States<Distance> st = getAllStates();
            List<Distance> r = new ArrayList<Distance>();
            for (Distance j : st)
                if (j.equals(next) || j.equals(next2) || j.equals(next3))
                    r.add(j);
            return new StatesSet<Distance>(r);
        } else {
            if (a == move) {
                State next = new Distance(0, true);
                State next2 = new Distance(1);
                States<Distance> st = getAllStates();
                List<Distance> r = new ArrayList<Distance>();
                for (Distance j : st)
                    if (j.equals(next) || j.equals(next2))
                        r.add(j);
                return new StatesSet<Distance>(r);

            } else if (a == stay) {
                State next = new Distance(0, true);
                State next2 = new Distance(1);
                State next3 = new Distance(2);
                States<Distance> st = getAllStates();
                List<Distance> r = new ArrayList<Distance>();
                for (Distance j : st)
                    if (j.equals(next) || j.equals(next2) || j.equals(next3))
                        r.add(j);
                return new StatesSet<Distance>(r);

            }
        }

        return null;
    }

    /**
     * This method return all possible action which could be taken from a StateC
     * i, it is preferred that you create an action which makes the terminal
     * state becomes an absorbing state. This action return a self-transition
     * probability = 1.
     * 
     * @see jmarkov.jmdp.InfiniteMDP#feasibleActions(State)
     */
    @Override
    public Actions<SpiderAction> feasibleActions(Distance i) {
        if (i.isTerminal()) {
            SpiderAction[] vector = new SpiderAction[1];
            vector[0] = stay;
            return new ActionsSet<SpiderAction>(vector);
        } else {
            return actions;
        }
    }

    /**
     * In this method you have to generate the problem then insert it into the
     * solver and finally solve the problem.
     * 
     * @param args
     */
    @SuppressWarnings("unchecked")
    public static void main(String[] args) {
        int initialDistanceBetween = 6;
        double moveCost = 1; // Move Cost of The Spider
        double stayCost = 1; // Stay Cost of the Spider
        double p = 0.1; // probability.

        TheSpiderAndTheFly problem = new TheSpiderAndTheFly(
                initialDistanceBetween, moveCost, stayCost, p);
        // StochasticShortestPathSolver<?,?> theSolver = new
        // StochasticShortestPathSolver(
        // problem);
        problem.printSolution();
    }

}
