/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp;

/**
 * Partial implementation of BLAS
 */
public abstract class AbstractBLAS implements BLAS {

	/**
	 * @see jmp.BLAS#multAdd(double, jmp.Matrix, jmp.Vector, double, jmp.Vector, jmp.Vector)
	 */
	public Vector multAdd(
		double alpha,
		Matrix A,
		Vector x,
		double beta,
		Vector y,
		Vector z) {
		checkMultAddArguments(A, x, y, z);

		// Quick return if possible
		if (alpha == 0. && beta == 0.)
			set(0., z);
		else if (alpha == 0.)
			scaleCopy(beta, y, z);
		else
			multAddI(alpha, A, x, beta, y, z);

		return z;
	}

	protected abstract Vector multAddI(
		double alpha,
		Matrix A,
		Vector x,
		double beta,
		Vector y,
		Vector z);

	private void checkMultAddArguments(
		Matrix A,
		Vector x,
		Vector y,
		Vector z) {
		if (A.numColumns() != x.size())
			throw new ArrayIndexOutOfBoundsException("A.numColumns() != x.size()");
		else if (A.numRows() != y.size())
			throw new ArrayIndexOutOfBoundsException("A.numRows() != y.size()");
		else if (y.size() != z.size())
			throw new ArrayIndexOutOfBoundsException("y.size() != z.size()");
		else if (x == y)
			throw new IllegalArgumentException("x == y");
		else if (x == z)
			throw new IllegalArgumentException("x == z");
	}

	/**
	 * @see jmp.BLAS#norm(jmp.Matrix, int)
	 */
	public double norm(Matrix A, int normtype) {
		switch (normtype) {
			case BLAS.NORM1 :
				return norm1(A);
			case BLAS.NORMF :
				return normF(A);
			case BLAS.NORMINF :
				return normInf(A);
		}

		throw new IllegalArgumentException("normtype must be one of BLAS.NORM1, BLAS.NORMF or BLAS.NORMINF");
	}

	protected abstract double norm1(Matrix A);

	protected abstract double normF(Matrix A);

	protected abstract double normInf(Matrix A);

	/**
	 * @see jmp.BLAS#multAdd(double, jmp.Matrix, jmp.Vector, jmp.Vector, jmp.Vector)
	 */
	public Vector multAdd(
		double alpha,
		Matrix A,
		Vector x,
		Vector y,
		Vector z) {
		return multAdd(alpha, A, x, 1., y, z);
	}

	/**
	 * @see jmp.BLAS#multAdd(jmp.Matrix, jmp.Vector, double, jmp.Vector, jmp.Vector)
	 */
	public Vector multAdd(
		Matrix A,
		Vector x,
		double beta,
		Vector y,
		Vector z) {
		return multAdd(1., A, x, beta, y, z);
	}

	/**
	 * @see jmp.BLAS#multAdd(double, jmp.Matrix, jmp.Vector, double, jmp.Vector)
	 */
	public Vector multAdd(
		double alpha,
		Matrix A,
		Vector x,
		double beta,
		Vector y) {
		return multAdd(alpha, A, x, beta, y, y);
	}

	/**
	 * @see jmp.BLAS#multAdd(double, jmp.Matrix, jmp.Vector, jmp.Vector)
	 */
	public Vector multAdd(double alpha, Matrix A, Vector x, Vector y) {
		return multAdd(alpha, A, x, 1., y, y);
	}

	/**
	 * @see jmp.BLAS#multAdd(jmp.Matrix, jmp.Vector, double, jmp.Vector)
	 */
	public Vector multAdd(Matrix A, Vector x, double beta, Vector y) {
		return multAdd(1., A, x, beta, y, y);
	}

	/**
	 * @see jmp.BLAS#multAdd(jmp.Matrix, jmp.Vector, jmp.Vector, jmp.Vector)
	 */
	public Vector multAdd(Matrix A, Vector x, Vector y, Vector z) {
		return multAdd(1., A, x, 1., y, z);
	}

	/**
	 * @see jmp.BLAS#multAdd(jmp.Matrix, jmp.Vector, jmp.Vector)
	 */
	public Vector multAdd(Matrix A, Vector x, Vector y) {
		return multAdd(1., A, x, 1., y, y);
	}

	/**
	 * @see jmp.BLAS#mult(double, jmp.Matrix, jmp.Vector, jmp.Vector)
	 */
	public Vector mult(double alpha, Matrix A, Vector x, Vector y) {
		return multAdd(alpha, A, x, 0., y, y);
	}

	/**
	 * @see jmp.BLAS#mult(jmp.Matrix, jmp.Vector, jmp.Vector)
	 */
	public Vector mult(Matrix A, Vector x, Vector y) {
		return multAdd(1., A, x, 0., y, y);
	}

	/**
	 * @see jmp.BLAS#transMultAdd(double, jmp.Matrix, jmp.Vector, double, jmp.Vector, jmp.Vector)
	 */
	public Vector transMultAdd(
		double alpha,
		Matrix A,
		Vector x,
		double beta,
		Vector y,
		Vector z) {
		checkTransMultAddArguments(A, x, y, z);

		// Quick return if possible
		if (alpha == 0. && beta == 0.)
			set(0., z);
		else if (alpha == 0.)
			scaleCopy(beta, y, z);
		else
			transMultAddI(alpha, A, x, beta, y, z);

		return z;
	}

	protected abstract Vector transMultAddI(
		double alpha,
		Matrix A,
		Vector x,
		double beta,
		Vector y,
		Vector z);

	private void checkTransMultAddArguments(
		Matrix A,
		Vector x,
		Vector y,
		Vector z) {
		if (A.numRows() != x.size())
			throw new ArrayIndexOutOfBoundsException("A.numRows() != x.size()");
		else if (A.numColumns() != y.size())
			throw new ArrayIndexOutOfBoundsException("A.numColumns() != y.size()");
		else if (y.size() != z.size())
			throw new ArrayIndexOutOfBoundsException("y.size() != z.size()");
		else if (x == y)
			throw new IllegalArgumentException("x == y");
		else if (x == z)
			throw new IllegalArgumentException("x == z");
	}

	/**
	 * @see jmp.BLAS#transMult(double, jmp.Matrix, jmp.Vector, jmp.Vector)
	 */
	public Vector transMult(double alpha, Matrix A, Vector x, Vector y) {
		return transMultAdd(alpha, A, x, 0., y, y);
	}

	/**
	 * @see jmp.BLAS#transMult(jmp.Matrix, jmp.Vector, jmp.Vector)
	 */
	public Vector transMult(Matrix A, Vector x, Vector y) {
		return transMultAdd(1., A, x, 0., y, y);
	}

	/**
	 * @see jmp.BLAS#transMultAdd(double, jmp.Matrix, jmp.Vector, double, jmp.Vector)
	 */
	public Vector transMultAdd(
		double alpha,
		Matrix A,
		Vector x,
		double beta,
		Vector y) {
		return transMultAdd(alpha, A, x, beta, y, y);
	}

	/**
	 * @see jmp.BLAS#transMultAdd(double, jmp.Matrix, jmp.Vector, jmp.Vector, jmp.Vector)
	 */
	public Vector transMultAdd(
		double alpha,
		Matrix A,
		Vector x,
		Vector y,
		Vector z) {
		return transMultAdd(alpha, A, x, 1., y, z);
	}

	/**
	 * @see jmp.BLAS#transMultAdd(double, jmp.Matrix, jmp.Vector, jmp.Vector)
	 */
	public Vector transMultAdd(double alpha, Matrix A, Vector x, Vector y) {
		return transMultAdd(alpha, A, x, 1., y, y);
	}

	/**
	 * @see jmp.BLAS#transMultAdd(jmp.Matrix, jmp.Vector, double, jmp.Vector, jmp.Vector)
	 */
	public Vector transMultAdd(
		Matrix A,
		Vector x,
		double beta,
		Vector y,
		Vector z) {
		return transMultAdd(1., A, x, beta, y, z);
	}

	/**
	 * @see jmp.BLAS#transMultAdd(jmp.Matrix, jmp.Vector, double, jmp.Vector)
	 */
	public Vector transMultAdd(Matrix A, Vector x, double beta, Vector y) {
		return transMultAdd(1., A, x, beta, y, y);
	}

	/**
	 * @see jmp.BLAS#transMultAdd(jmp.Matrix, jmp.Vector, jmp.Vector, jmp.Vector)
	 */
	public Vector transMultAdd(Matrix A, Vector x, Vector y, Vector z) {
		return transMultAdd(1., A, x, 1., y, z);
	}

	/**
	 * @see jmp.BLAS#transMultAdd(jmp.Matrix, jmp.Vector, jmp.Vector)
	 */
	public Vector transMultAdd(Matrix A, Vector x, Vector y) {
		return transMultAdd(1., A, x, 1., y, y);
	}

	/**
	 * @see jmp.BLAS#transpose(jmp.Matrix, jmp.Matrix)
	 */
	public Matrix transpose(Matrix A, Matrix B) {
		checkTransposeArguments(A, B);
		return transposeI(A, B);
	}

	private void checkTransposeArguments(Matrix A, Matrix B) {
		if (A.numRows() != B.numColumns())
			throw new ArrayIndexOutOfBoundsException("A.numRows() != B.numColumns()");
		else if (A.numColumns() != B.numRows())
			throw new ArrayIndexOutOfBoundsException("A.numColumns() != B.numRows()");
		else if (A == B)
			throw new IllegalArgumentException("A == B");
	}

	protected abstract Matrix transposeI(Matrix A, Matrix B);

	/**
	 * @see jmp.BLAS#addDiagonal(double, jmp.Matrix)
	 */
	public Matrix addDiagonal(double alpha, Matrix A) {
		if (alpha == 0.)
			return A;
		else
			return addDiagonalI(alpha, A);
	}

	protected abstract Matrix addDiagonalI(double alpha, Matrix A);

	/**
	 * @see jmp.BLAS#scale(double, jmp.Matrix)
	 */
	public Matrix scale(double alpha, Matrix A) {
		if (alpha == 1.)
			return A;
		else if (alpha == 0.) {
			return zero(A);
		} else
			return scaleI(alpha, A);
	}

	protected abstract Matrix scaleI(double alpha, Matrix A);

	/**
	 * @see jmp.BLAS#copy(jmp.Matrix, jmp.Matrix)
	 */
	public Matrix copy(Matrix A, Matrix B) {
		checkCopyArguments(A, B);
		return copyI(A, B);
	}

	private void checkCopyArguments(Matrix A, Matrix B) {
		if (A.numRows() != B.numRows())
			throw new ArrayIndexOutOfBoundsException("A.numRows() != B.numRows()");
		else if (A.numColumns() != B.numColumns())
			throw new ArrayIndexOutOfBoundsException("A.numColumns() != B.numColumns()");
		else if (A == B)
			throw new IllegalArgumentException("A == B");
	}

	/**
	 * @see jmp.BLAS#copy(jmp.Matrix, jmp.Matrix, int, int, int, int)
	 */
	public Matrix copy(
		Matrix A,
		Matrix B,
		int startRow,
		int stopRow,
		int startColumn,
		int stopColumn) {
		checkCopyArguments(A, B, startRow, stopRow, startColumn, stopColumn);
		if (startRow == 0
			&& stopRow == A.numRows()
			&& startColumn == 0
			&& stopColumn == A.numColumns())
			return copyI(A, B);
		else
			return copyI(A, B, startRow, stopRow, startColumn, stopColumn);
	}

	private void checkCopyArguments(
		Matrix A,
		Matrix B,
		int startRow,
		int stopRow,
		int startColumn,
		int stopColumn) {
		if (startRow < 0)
			throw new ArrayIndexOutOfBoundsException("startRow < 0");
		else if (startColumn < 0)
			throw new ArrayIndexOutOfBoundsException("startColumn < 0");
		else if (stopRow > A.numRows())
			throw new ArrayIndexOutOfBoundsException("stopRow > A.numRows()");
		else if (stopColumn > A.numColumns())
			throw new ArrayIndexOutOfBoundsException("stopColumn > A.numColumns()");
		else if (B.numRows() != stopRow - startRow)
			throw new ArrayIndexOutOfBoundsException("B.numRows() != stopRow - startRow");
		else if (B.numColumns() != stopColumn - startColumn)
			throw new ArrayIndexOutOfBoundsException("B.numColumns() != stopColumn - startColumn");
		else if (A == B)
			throw new IllegalArgumentException("A == B");
	}

	protected abstract Matrix copyI(
		Matrix A,
		Matrix B,
		int startRow,
		int stopRow,
		int startColumn,
		int stopColumn);

	/**
	 * @see jmp.BLAS#rank1(double, jmp.Vector, jmp.Vector, jmp.Matrix)
	 */
	public Matrix rank1(double alpha, Vector x, Vector y, Matrix A) {
		// Quick return if possible
		if (alpha == 0.)
			return A;

		checkRankArguments(x, y, A);
		return rank1I(alpha, x, y, A);
	}

	protected abstract Matrix rank1I(
		double alpha,
		Vector x,
		Vector y,
		Matrix A);

	private void checkRankArguments(Vector x, Vector y, Matrix A) {
		if (x.size() != A.numRows())
			throw new ArrayIndexOutOfBoundsException("x.size() != A.numRows()");
		if (y.size() != A.numColumns())
			throw new ArrayIndexOutOfBoundsException("y.size() != A.numColumns()");
	}

	/**
	 * @see jmp.BLAS#rank1(jmp.Vector, jmp.Vector, jmp.Matrix)
	 */
	public Matrix rank1(Vector x, Vector y, Matrix A) {
		return rank1(1., x, y, A);
	}

	/**
	 * @see jmp.BLAS#rank1(double, jmp.Vector, jmp.Matrix)
	 */
	public Matrix rank1(double alpha, Vector x, Matrix A) {
		return rank1(alpha, x, x, A);
	}

	/**
	 * @see jmp.BLAS#rank1(jmp.Vector, jmp.Matrix)
	 */
	public Matrix rank1(Vector x, Matrix A) {
		return rank1(1., x, A);
	}

	/**
	 * @see jmp.BLAS#add(double, jmp.Vector, double, jmp.Vector, jmp.Vector)
	 */
	public Vector add(
		double alpha,
		Vector x,
		double beta,
		Vector y,
		Vector z) {
		checkAddArguments(x, y, z);

		if (alpha == 0. && beta == 0.)
			zero(z);
		else if (alpha == 0)
			if (y != z)
				z = scaleCopy(beta, y, z);
			else
				z = scale(beta, z);
		else if (beta == 0.)
			if (x != z)
				z = scaleCopy(alpha, x, z);
			else
				z = scale(alpha, z);
		else
			addI(alpha, x, beta, y, z);

		return z;
	}

	private void checkAddArguments(Vector x, Vector y, Vector z) {
		if (x.size() != y.size())
			throw new ArrayIndexOutOfBoundsException("x.size() != y.size()");
		else if (x.size() != z.size())
			throw new ArrayIndexOutOfBoundsException("x.size() != z.size()");
	}

	protected abstract Vector addI(
		double alpha,
		Vector x,
		double beta,
		Vector y,
		Vector z);

	/**
	 * @see jmp.BLAS#add(double, jmp.Vector, double, jmp.Vector)
	 */
	public Vector add(double alpha, Vector x, double beta, Vector y) {
		return add(alpha, x, beta, y, y);
	}

	/**
	 * @see jmp.BLAS#add(double, jmp.Vector, jmp.Vector, jmp.Vector)
	 */
	public Vector add(double alpha, Vector x, Vector y, Vector z) {
		return add(alpha, x, 1., y, z);
	}

	/**
	 * @see jmp.BLAS#add(jmp.Vector, double, jmp.Vector)
	 */
	public Vector add(Vector x, double beta, Vector y) {
		return add(1., x, beta, y, y);
	}

	/**
	 * @see jmp.BLAS#add(jmp.Vector, jmp.Vector, jmp.Vector)
	 */
	public Vector add(Vector x, Vector y, Vector z) {
		return add(1., x, 1., y, z);
	}

	/**
	 * @see jmp.BLAS#add(jmp.Vector, jmp.Vector)
	 */
	public Vector add(Vector x, Vector y) {
		return add(1., x, 1., y, y);
	}

	/**
	 * @see jmp.BLAS#add(jmp.Vector, double, jmp.Vector, jmp.Vector)
	 */
	public Vector add(Vector x, double beta, Vector y, Vector z) {
		return add(1., x, beta, y, z);
	}

	/**
	 * @see jmp.BLAS#add(double, jmp.Vector, jmp.Vector)
	 */
	public Vector add(double alpha, Vector x, Vector y) {
		return add(alpha, x, 1., y, y);
	}

	/**
	 * @see jmp.BLAS#dot(jmp.Vector, jmp.Vector)
	 */
	public double dot(Vector x, Vector y) {
		checkDotArguments(x, y);
		return dotI(x, y);
	}

	protected abstract double dotI(Vector x, Vector y);

	private void checkDotArguments(Vector x, Vector y) {
		if (x.size() != y.size())
			throw new ArrayIndexOutOfBoundsException("x.size() != y.size()");
	}

	/**
	 * @see jmp.BLAS#norm(jmp.Vector, int)
	 */
	public double norm(Vector x, int normtype) {
		switch (normtype) {
			case BLAS.NORM1 :
				return norm1(x);
			case BLAS.NORM2 :
				return norm2(x);
			case BLAS.NORMINF :
				return normInf(x);
		}

		throw new IllegalArgumentException("normtype must be one of BLAS.NORM1, BLAS.NORM2 or BLAS.NORMINF");
	}

	protected abstract double norm1(Vector x);

	protected abstract double norm2(Vector x);

	protected abstract double normInf(Vector x);

	/**
	 * @see jmp.BLAS#scaleCopy(double, jmp.Vector, jmp.Vector)
	 */
	public Vector scaleCopy(double alpha, Vector x, Vector y) {
		return scale(alpha, copy(x, y));
	}

	/**
	 * @see jmp.BLAS#scale(double, jmp.Vector)
	 */
	public Vector scale(double alpha, Vector x) {
		if (alpha == 1.)
			return x;
		else if (alpha == 0.)
			return zero(x);
		else
			return scaleI(alpha, x);
	}

	protected abstract Vector scaleI(double alpha, Vector x);

	/**
	 * @see jmp.BLAS#set(double, jmp.Vector)
	 */
	public Vector set(double alpha, Vector x) {
		if (alpha == 0.)
			return zero(x);
		else
			return setI(alpha, x);
	}

	protected abstract Vector setI(double alpha, Vector x);

	/**
	 * @see jmp.BLAS#copy(jmp.Vector, jmp.Vector)
	 */
	public Vector copy(Vector x, Vector y) {
		checkCopyArguments(x, y);
		return copyI(x, y);
	}

	protected abstract Vector copyI(Vector x, Vector y);

	private void checkCopyArguments(Vector x, Vector y) {
		if (x.size() != y.size())
			throw new ArrayIndexOutOfBoundsException("x.size() != y.size()");
		else if (x == y)
			throw new IllegalArgumentException("x == y");
	}

	/**
	 * @see jmp.BLAS#copy(jmp.Vector, jmp.Vector, int, int)
	 */
	public Vector copy(Vector x, Vector y, int start, int stop) {
		checkCopyArguments(x, y, start, stop);
		return copyI(x, y, start, stop);
	}

	protected abstract Vector copyI(Vector x, Vector y, int start, int stop);

	private void checkCopyArguments(Vector x, Vector y, int start, int stop) {
		if (start < 0)
			throw new ArrayIndexOutOfBoundsException("start < 0");
		else if (stop > x.size())
			throw new ArrayIndexOutOfBoundsException("stop > x.size()");
		else if (y.size() != stop - start)
			throw new ArrayIndexOutOfBoundsException("y.size() != stop - start");
		else if (x == y)
			throw new IllegalArgumentException("x == y");
	}

	protected abstract Matrix copyI(Matrix A, Matrix B);

}
