/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp;

/**
 * Basic linear algebra subroutines. The methods present here are
 * suitable for sparse matrices and iterative solvers, as such not all
 * the <a href="http://www.netlib.org/blas">Fortran BLAS</a> methods
 * have been implemented. Of particular note is the absence of
 * matrix/matrix multiplication and triangular solvers.
 */
public interface BLAS {

	//==========================================================================
	//==== Level 3, Matrix operations ==========================================
	//==========================================================================

	/**
	 * <code>B = A<sup>T</sup></code>.
	 * @return B
	 */
	Matrix transpose(Matrix A, Matrix B);

	/**
	 * <code>A = A + alpha*I</code>.
	 * @return A
	 */
	Matrix addDiagonal(double alpha, Matrix A);

	/**
	 * Computes the specified norm of the matrix.
	 * @param normtype Any of BLAS.NORM1, BLAS.NORMF or BLAS.NORMINF
	 */
	double norm(Matrix A, int normtype);

	/**
	 * <code>A = alpha*A</code>.
	 * @return A
	 */
	Matrix scale(double alpha, Matrix A);

	/**
	 * <code>B = A</code>.
	 * @return B
	 */
	Matrix copy(Matrix A, Matrix B);

	/**
	 * <code>B = A(startRow:stopRow,startColumn:stopColumn)</code>.
	 * Start-indices are inclusive, stop-indices are exclusive.
	 * @return B
	 */
	Matrix copy(
		Matrix A,
		Matrix B,
		int startRow,
		int stopRow,
		int startColumn,
		int stopColumn);

	/**
	 * <code>A = 0</code>. The non-zero structure is preserved for sparse
	 * matrices, speeding up later reconstructions.
	 * @return A
	 */
	Matrix zero(Matrix A);

	/**
	 * Number of non-zero entries of A
	 */
	int cardinality(Matrix A);

	//==========================================================================
	//==== Level 2, Matrix-Vector operations ===================================
	//==========================================================================

	/**
	 * <code>z = alpha*A*x + beta*y</code>. x can not be the same as y or z.
	 * @return z
	 */
	Vector multAdd(
		double alpha,
		Matrix A,
		Vector x,
		double beta,
		Vector y,
		Vector z);

	/**
	 * <code>z = A*x + beta*y</code>. x can not be the same as y or z.
	 * @return z
	 */
	Vector multAdd(Matrix A, Vector x, double beta, Vector y, Vector z);

	/**
	 * <code>z = alpha*A*x + y</code>. x can not be the same as y or z.
	 * @return z
	 */
	Vector multAdd(double alpha, Matrix A, Vector x, Vector y, Vector z);

	/**
	 * <code>z = A*x + y</code>. x can not be the same as y or z.
	 * @return z
	 */
	Vector multAdd(Matrix A, Vector x, Vector y, Vector z);

	/**
	 * <code>y = alpha*A*x + beta*y</code>. x can not be the same as y.
	 * @return y
	 */
	Vector multAdd(double alpha, Matrix A, Vector x, double beta, Vector y);

	/**
	 * <code>y = alpha*A*x + y</code>. x can not be the same as y.
	 * @return y
	 */
	Vector multAdd(double alpha, Matrix A, Vector x, Vector y);

	/**
	 * <code>y = A*x + beta*y</code>. x can not be the same as y.
	 * @return y
	 */
	Vector multAdd(Matrix A, Vector x, double beta, Vector y);

	/**
	 * <code>y = A*x + y</code>. x can not be the same as y.
	 * @return y
	 */
	Vector multAdd(Matrix A, Vector x, Vector y);

	/**
	 * <code>y = alpha*A*x</code>. x can not be the same as y
	 * @return y
	 */
	Vector mult(double alpha, Matrix A, Vector x, Vector y);

	/**
	 * <code>y = A*x</code>. x can not be the same as y
	 * @return y
	 */
	Vector mult(Matrix A, Vector x, Vector y);

	/**
	 * <code>z = alpha*A<sup>T</sup>*x + beta*y</code>. x can not be the same as
	 * y or z.
	 * @return z
	 */
	Vector transMultAdd(
		double alpha,
		Matrix A,
		Vector x,
		double beta,
		Vector y,
		Vector z);

	/**
	 * <code>z = A<sup>T</sup>*x + beta*y</code>. x can not be the same as y or
	 * z.
	 * @return z
	 */
	Vector transMultAdd(Matrix A, Vector x, double beta, Vector y, Vector z);

	/**
	 * <code>z = alpha*A<sup>T</sup>*x + y</code>. x can not be the same as y or
	 * z.
	 * @return z
	 */
	Vector transMultAdd(double alpha, Matrix A, Vector x, Vector y, Vector z);

	/**
	 * <code>z = A<sup>T</sup>*x + y</code>. x can not be the same as y or z.
	 * @return z
	 */
	Vector transMultAdd(Matrix A, Vector x, Vector y, Vector z);

	/**
	 * <code>y = alpha*A<sup>T</sup>*x + beta*y</code>. x can not be the same as
	 * y.
	 * @return y
	 */
	Vector transMultAdd(
		double alpha,
		Matrix A,
		Vector x,
		double beta,
		Vector y);

	/**
	 * <code>y = alpha*A<sup>T</sup>*x + y</code>. x can not be the same as y.
	 * @return y
	 */
	Vector transMultAdd(double alpha, Matrix A, Vector x, Vector y);

	/**
	 * <code>y = A<sup>T</sup>*x + beta*y</code>. x can not be the same as y.
	 * @return y
	 */
	Vector transMultAdd(Matrix A, Vector x, double beta, Vector y);

	/**
	 * <code>y = A<sup>T</sup>*x + y</code>. x can not be the same as y.
	 * @return y
	 */
	Vector transMultAdd(Matrix A, Vector x, Vector y);

	/**
	 * <code>y = alpha*A<sup>T</sup>*x</code>. x can not be the same as y
	 * @return y
	 */
	Vector transMult(double alpha, Matrix A, Vector x, Vector y);

	/**
	 * <code>y = A<sup>T</sup>*x</code>. x can not be the same as y
	 * @return y
	 */
	Vector transMult(Matrix A, Vector x, Vector y);

	/**
	 * <code>A = alpha*x*y<sup>T</sup> + A</code>.
	 * @return A
	 */
	Matrix rank1(double alpha, Vector x, Vector y, Matrix A);

	/**
	 * <code>A = x*y<sup>T</sup> + A</code>.
	 * @return A
	 */
	Matrix rank1(Vector x, Vector y, Matrix A);

	/**
	 * <code>A = alpha*x*x<sup>T</sup> + A</code>.
	 * @return A
	 */
	Matrix rank1(double alpha, Vector x, Matrix A);

	/**
	 * <code>A = x*x<sup>T</sup> + A</code>.
	 * @return A
	 */
	Matrix rank1(Vector x, Matrix A);

	//==========================================================================
	//==== Level 1, Vector operations ==========================================
	//==========================================================================

	/**
	 * <code>z = alpha*x + beta*y</code>.
	 * @return z
	 */
	Vector add(double alpha, Vector x, double beta, Vector y, Vector z);

	/**
	 * <code>z = alpha*x + y</code>.
	 * @return z
	 */
	Vector add(double alpha, Vector x, Vector y, Vector z);

	/**
	 * <code>z = x + beta*y</code>.
	 * @return z
	 */
	Vector add(Vector x, double beta, Vector y, Vector z);

	/**
	 * <code>z = x + y</code>.
	 * @return z
	 */
	Vector add(Vector x, Vector y, Vector z);

	/**
	 * <code>y = alpha*x + beta*y</code>.
	 * @return y
	 */
	Vector add(double alpha, Vector x, double beta, Vector y);

	/**
	 * <code>y = alpha*x + y</code>.
	 * @return y
	 */
	Vector add(double alpha, Vector x, Vector y);

	/**
	 * <code>y = x + beta*y</code>.
	 * @return y
	 */
	Vector add(Vector x, double beta, Vector y);

	/**
	 * <code>y = x + y</code>.
	 * @return y
	 */
	Vector add(Vector x, Vector y);

	/**
	 * <code>x<sup>T</sup>*y</code>.
	 */
	double dot(Vector x, Vector y);

	/**
	 * Computes the specified norm of the vector.
	 * @param normtype One of BLAS.NORM1, BLAS.NORM2 or BLAS.NORMINF
	 */
	double norm(Vector x, int normtype);

	/**
	 * <code>y = alpha*x</code>.
	 * @return y
	 */
	Vector scaleCopy(double alpha, Vector x, Vector y);

	/**
	 * <code>x = alpha*x</code>.
	 * @return x
	 */
	Vector scale(double alpha, Vector x);

	/**
	 * <code>x = alpha</code>.
	 * @return x
	 */
	Vector set(double alpha, Vector x);

	/**
	 * <code>y = x</code>.
	 * @return y
	 */
	Vector copy(Vector x, Vector y);

	/**
	 * <code>y = x(start:stop)</code>. start is inclusive, stop is exclusive.
	 * @return y
	 */
	Vector copy(Vector x, Vector y, int start, int stop);

	/**
	 * <code>x = 0</code>. The non-zero structure is preserved for sparse
	 * vectors, speeding up later reconstructions.
	 * @return x
	 */
	Vector zero(Vector x);

	/**
	 * Number of non-zero entries of x
	 */
	int cardinality(Vector x);

	//==========================================================================
	//==== Miscellaneous =======================================================
	//==========================================================================

	/**
	 * Scatter x into the dense array y.
	 * @return y
	 */
	double[] scatter(IntDoubleVectorPair x, double[] y);

	/**
	 * Gathers the non-zero entries of the dense array x.
	 */
	IntDoubleVectorPair gather(double[] x);

	/**
	 * Gathers the given indices of xData.
	 */
	IntDoubleVectorPair gather(int[] xIndex, double[] xData);

	/**
	 * Exports the matrix into a dense structure, row-oriented.
	 */
	double[][] getArrayCopy(Matrix A);

	/**
	 * Exports the vector into a dense structure.
	 */
	double[] getArrayCopy(Vector x);

	/**
	 * <code>y = x</code>.
	 * @return x
	 */
	Vector setArray(double[] x, Vector y);

	/**
	 * 1-norm, for matrices and vectors.
	 */
	int NORM1 = 1;

	/**
	 * 2-norm, for vectors.
	 */
	int NORM2 = 2;

	/**
	 * Frobenius norm, for matrices.
	 */
	int NORMF = 2;

	/**
	 * Infinity norm, for matrices and vectors.
	 */
	int NORMINF = 3;

}
