/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp;

/**
 * Coordinate storage matrix. The data is not kept sorted.
 */
public class CoordinateMatrix
	extends AbstractMatrix
	implements Matrix, ElementalAccessMatrix, CoordinateAccessMatrix {

	/**
	 * The rows
	 */
	private int[] row;

	/**
	 * The columns
	 */
	private int[] column;

	/**
	 * The data
	 */
	private double[] data;

	/**
	 * Current insertion offset
	 */
	private int offset;

	/**
	 * Constructor for CoordinateMatrix. No pre-allocation
	 */
	public CoordinateMatrix(int numRows, int numColumns) {
		this(numRows, numColumns, 0);
	}

	/**
	 * Constructor for CoordinateMatrix.
	 * @param numEntries Initial number of non-zero entries
	 */
	public CoordinateMatrix(int numRows, int numColumns, int numEntries) {
		super(numRows, numColumns);
		column = new int[numEntries];
		row = new int[numEntries];
		data = new double[numEntries];
	}

	/**
	 * Constructor for CoordinateMatrix.
	 */
	public CoordinateMatrix(
		int numRows,
		int numColumns,
		int[] row,
		int[] column,
		double[] data) {
		super(numRows, numColumns);
		this.row = row;
		this.column = column;
		this.data = data;
		offset = data.length;
	}

	/**
	 * @see jmp.ElementalAccessMatrix#addValue(int, int, double)
	 */
	public synchronized void addValue(int row, int column, double value) {
		int index = getIndex(row, column);
		data[index] += value;
	}

	/**
	 * @see jmp.ElementalAccessMatrix#setValue(int, int, double)
	 */
	public synchronized void setValue(int row, int column, double value) {
		int index = getIndex(row, column);
		data[index] = value;
	}

	/**
	 * @see jmp.ElementalAccessMatrix#getValue(int, int)
	 */
	public double getValue(int row, int column) {
		int index = search(row, column);
		if (index >= 0)
			return data[index];
		else if (
			row >= 0 && row < numRows && column >= 0 && column < numColumns)
			return 0.;
		else
			throw new ArrayIndexOutOfBoundsException(
				"Row " + row + " Column " + column);
	}

	/**
	 * @see jmp.ElementalAccessMatrix#addValues(int[], int[], double[][])
	 */
	public void addValues(int[] row, int[] column, double[][] values) {
		for (int i = 0; i < row.length; ++i)
			for (int j = 0; j < column.length; ++j)
				addValue(row[i], column[j], values[i][j]);
	}

	/**
	 * @see jmp.ElementalAccessMatrix#setValues(int[], int[], double[][])
	 */
	public void setValues(int[] row, int[] column, double[][] values) {
		for (int i = 0; i < row.length; ++i)
			for (int j = 0; j < column.length; ++j)
				setValue(row[i], column[j], values[i][j]);
	}

	/**
	 * @see jmp.ElementalAccessMatrix#getValues(int[], int[])
	 */
	public double[][] getValues(int[] row, int[] column) {
		double[][] sub = new double[row.length][column.length];
		for (int i = 0; i < row.length; ++i)
			for (int j = 0; j < column.length; ++j)
				sub[i][j] = getValue(row[i], column[j]);
		return sub;
	}

	/**
	 * @see jmp.VectorAccess#getData()
	 */
	public double[] getData() {
		compact();
		return data;
	}

	/**
	 * @see jmp.CoordinateAccessMatrix#getRowIndices()
	 */
	public int[] getRowIndices() {
		compact();
		return row;
	}

	/**
	 * @see jmp.CoordinateAccessMatrix#getColumnIndices()
	 */
	public int[] getColumnIndices() {
		compact();
		return column;
	}

	/**
	 * @see jmp.CoordinateAccessMatrix#setRowIndices(int[])
	 */
	public synchronized void setRowIndices(int[] row) {
		this.row = row;
		offset = row.length;
	}

	/**
	 * @see jmp.CoordinateAccessMatrix#setColumnIndices(int[])
	 */
	public synchronized void setColumnIndices(int[] column) {
		this.column = column;
		offset = column.length;
	}

	/**
	 * @see jmp.CoordinateAccessMatrix#setData(double[])
	 */
	public synchronized void setData(double[] data) {
		this.data = data;
		offset = data.length;
	}

	/**
	 * Tries to get the index, or allocates more space if needed
	 */
	private int getIndex(int rowI, int columnI) {
		int index = search(rowI, columnI);
		if (index >= 0) // Found
			return index;
		else { // Not found

			// Need to allocate more space
			if (offset >= data.length)
				if (offset == 0)
					realloc(1);
				else
					realloc(2 * offset);

			// Put in the new indices
			row[offset] = rowI;
			column[offset] = columnI;
			return offset++;
		}
	}

	/**
	 * Reallocates the internal arrays to a new size
	 */
	private void realloc(int newSize) {
		int[] newRow = new int[newSize], newColumn = new int[newSize];
		double[] newData = new double[newSize];

		System.arraycopy(row, 0, newRow, 0, offset);
		System.arraycopy(column, 0, newColumn, 0, offset);
		System.arraycopy(data, 0, newData, 0, offset);

		row = newRow;
		column = newColumn;
		data = newData;
	}

	/**
	 * Searches for the given entry. Returns -1 if it is not found
	 */
	private int search(int rowI, int columnI) {
		for (int i = 0; i < offset; ++i)
			if (row[i] == rowI && column[i] == columnI)
				return i;
		return -1;
	}

	/**
	 * @see jmp.Sparse#compact()
	 */
	public synchronized void compact() {
		if (offset < row.length)
			realloc(offset);
	}

}
