/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp;

import java.util.Arrays;

/**
 * Dense matrix stored as a 2D array, column major.
 */
public class DenseColumnMatrix
	extends AbstractMatrix
	implements
		DenseColumnAccessMatrix,
		ElementalAccessZeroColumnMatrix,
		MatrixAccess,
		Matrix {

	/**
	 * Internal data representation
	 */
	private double[][] data;

	/**
	 * For thread safe operation. One lock for each column
	 */
	private Lock[] lock;

	/**
	 * Constructor for DenseColumnMatrix.
	 */
	public DenseColumnMatrix(int numRows, int numColumns) {
		super(numRows, numColumns);
		data = new double[numColumns][numRows];
		lock = Lock.createLocks(numColumns);
	}

	/**
	 * Constructor for DenseColumnMatrix, and copies the contents from the
	 * supplied matrix.
	 */
	public DenseColumnMatrix(Matrix A) {
		this(A.numRows(), A.numColumns());
		State.BLAS.copy(A, this);
	}

	/**
	 * @see jmp.DenseColumnAccessMatrix#getColumn(int)
	 */
	public double[] getColumn(int i) {
		return data[i];
	}

	/**
	 * @see jmp.DenseColumnAccessMatrix#setColumn(int, double[])
	 */
	public void setColumn(int i, double[] x) {
		synchronized (lock[i]) {
			data[i] = x;
		}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#addValue(int, int, double)
	 */
	public void addValue(int row, int column, double value) {
		synchronized (lock[column]) {
			data[column][row] += value;
		}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#setValue(int, int, double)
	 */
	public void setValue(int row, int column, double value) {
		synchronized (lock[column]) {
			data[column][row] = value;
		}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#getValue(int, int)
	 */
	public double getValue(int row, int column) {
		return data[column][row];
	}

	/**
	 * @see jmp.ElementalAccessMatrix#addValues(int[], int[], double[][])
	 */
	public void addValues(int[] row, int[] column, double[][] values) {
		for (int j = 0; j < column.length; ++j)
			synchronized (lock[column[j]]) {
				for (int i = 0; i < row.length; ++i)
					data[column[j]][row[i]] += values[i][j];
			}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#setValues(int[], int[], double[][])
	 */
	public void setValues(int[] row, int[] column, double[][] values) {
		for (int j = 0; j < column.length; ++j)
			synchronized (lock[column[j]]) {
				for (int i = 0; i < row.length; ++i)
					data[column[j]][row[i]] = values[i][j];
			}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#getValues(int[], int[])
	 */
	public double[][] getValues(int[] row, int[] column) {
		double[][] ret = new double[row.length][column.length];
		for (int j = 0; j < column.length; ++j)
			for (int i = 0; i < row.length; ++i)
				ret[i][j] = data[column[j]][row[i]];
		return ret;
	}

	/**
	 * @see jmp.ZeroColumnMatrix#zeroColumns(int[], double)
	 */
	public void zeroColumns(int[] column, double diagonal) {
		for (int i = 0; i < column.length; ++i)
			synchronized (lock[column[i]]) {
				Arrays.fill(data[column[i]], 0.);
				if (column[i] < numRows)
					data[column[i]][column[i]] = diagonal;
			}
	}

	/**
	 * @see jmp.MatrixAccess#getData()
	 */
	public double[][] getData() {
		return data;
	}

}
