/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp;

/**
 * Dense matrix stored as one long vector, column major.
 */
public class DenseColumnRowMatrix
	extends AbstractMatrix
	implements
		DenseColumnRowAccessMatrix,
		ElementalAccessZeroColumnMatrix,
		VectorAccess,
		Matrix {

	/**
	 * Internal data representation
	 */
	private double[] data;

	/**
	 * Indices to the start of each column
	 */
	private int[] columnIndex;

	/**
	 * For thread safe operation. One lock for each row
	 */
	private Lock[] lock;

	/**
	 * Constructor for DenseColumnRowMatrix, and copies the contents from the
	 * supplied matrix.
	 */
	public DenseColumnRowMatrix(Matrix A) {
		this(A.numRows(), A.numColumns());
		State.BLAS.copy(A, this);
	}

	/**
	 * Constructor for DenseColumnRowMatrix.
	 */
	public DenseColumnRowMatrix(int numRows, int numColumns) {
		super(numRows, numColumns);
		data = new double[numRows * numColumns];
		columnIndex = new int[numColumns + 1];
		lock = Lock.createLocks(numColumns);
		columnIndex[0] = 0;
		for (int i = 1; i < numColumns + 1; ++i)
			columnIndex[i] = columnIndex[i - 1] + numRows;
	}

	/**
	 * @see jmp.ElementalAccessMatrix#addValue(int, int, double)
	 */
	public void addValue(int row, int column, double value) {
		synchronized (lock[column]) {
			data[row + column * numRows] += value;
		}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#setValue(int, int, double)
	 */
	public void setValue(int row, int column, double value) {
		synchronized (lock[column]) {
			data[row + column * numRows] = value;
		}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#getValue(int, int)
	 */
	public double getValue(int row, int column) {
		return data[row + column * numRows];
	}

	/**
	 * @see jmp.ElementalAccessMatrix#addValues(int[], int[], double[][])
	 */
	public void addValues(int[] row, int[] column, double[][] value) {
		for (int j = 0; j < column.length; ++j)
			synchronized (lock[column[j]]) {
				for (int i = 0; i < row.length; ++i)
					data[row[i] + column[j] * numRows] += value[i][j];
			}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#setValues(int[], int[], double[][])
	 */
	public void setValues(int[] row, int[] column, double[][] value) {
		for (int j = 0; j < column.length; ++j)
			synchronized (lock[column[j]]) {
				for (int i = 0; i < row.length; ++i)
					data[row[i] + column[j] * numRows] = value[i][j];
			}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#getValues(int[], int[])
	 */
	public double[][] getValues(int[] row, int[] column) {
		double[][] ret = new double[row.length][column.length];
		for (int j = 0; j < column.length; ++j)
			for (int i = 0; i < row.length; ++i)
				ret[i][j] = data[row[i] + column[j] * numRows];
		return ret;
	}

	/**
	 * @see jmp.ZeroColumnMatrix#zeroColumns(int[], double)
	 */
	public void zeroColumns(int[] column, double diagonal) {
		for (int i = 0; i < column.length; ++i) {
			java.util.Arrays.fill(
				data,
				columnIndex[column[i]],
				columnIndex[column[i] + 1],
				0.);
			if (column[i] < numRows)
				data[column[i] + column[i] * numRows] = diagonal;
		}
	}

	/**
	 * @see jmp.DenseColumnRowAccessMatrix#getMatrix()
	 */
	public IntDoubleVectorPair getMatrix() {
		return new IntDoubleVectorPair(columnIndex, data);
	}

	/**
	 * @see jmp.DenseColumnRowAccessMatrix#setMatrix(jmp.IntDoubleVectorPair)
	 */
	public synchronized void setMatrix(IntDoubleVectorPair x) {
		data = x.data;
		columnIndex = x.index;
	}

	/**
	 * @see jmp.VectorAccess#getData()
	 */
	public double[] getData() {
		return data;
	}

}
