/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp;

import java.util.Arrays;

/**
 * Dense matrix stored as one long vector, row major.
 */
public class DenseRowColumnMatrix
	extends AbstractMatrix
	implements
		DenseRowColumnAccessMatrix,
		ElementalAccessZeroRowMatrix,
		VectorAccess,
		Matrix {

	/**
	 * Internal data representation
	 */
	private double[] data;

	/**
	 * Indices to the start of each row
	 */
	private int[] rowIndex;

	/**
	 * For thread safe operation. One lock for each row
	 */
	private Lock[] lock;

	/**
	 * Constructor for DenseRowColumnMatrix, and copies the contents from the
	 * supplied matrix.
	 */
	public DenseRowColumnMatrix(Matrix A) {
		this(A.numRows(), A.numColumns());
		State.BLAS.copy(A, this);
	}

	/**
	 * Constructor for DenseRowColumnMatrix.
	 */
	public DenseRowColumnMatrix(int numRows, int numColumns) {
		super(numRows, numColumns);
		data = new double[numRows * numColumns];
		rowIndex = new int[numRows + 1];
		lock = Lock.createLocks(numRows);
		rowIndex[0] = 0;
		for (int i = 1; i < numRows + 1; ++i)
			rowIndex[i] = rowIndex[i - 1] + numColumns;
	}

	/**
	 * @see jmp.ElementalAccessMatrix#addValue(int, int, double)
	 */
	public void addValue(int row, int column, double value) {
		synchronized (lock[row]) {
			data[row * numColumns + column] += value;
		}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#setValue(int, int, double)
	 */
	public void setValue(int row, int column, double value) {
		synchronized (lock[row]) {
			data[row * numColumns + column] = value;
		}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#getValue(int, int)
	 */
	public double getValue(int row, int column) {
		return data[row * numColumns + column];
	}

	/**
	 * @see jmp.ElementalAccessMatrix#addValues(int[], int[], double[][])
	 */
	public void addValues(int[] row, int[] column, double[][] value) {
		for (int i = 0; i < row.length; ++i)
			synchronized (lock[row[i]]) {
				for (int j = 0; j < column.length; ++j)
					data[row[i] * numColumns + column[j]] += value[i][j];
			}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#setValues(int[], int[], double[][])
	 */
	public void setValues(int[] row, int[] column, double[][] value) {
		for (int i = 0; i < row.length; ++i)
			synchronized (lock[row[i]]) {
				for (int j = 0; j < column.length; ++j)
					data[row[i] * numColumns + column[j]] = value[i][j];
			}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#getValues(int[], int[])
	 */
	public double[][] getValues(int[] row, int[] column) {
		double[][] ret = new double[row.length][column.length];
		for (int i = 0; i < row.length; ++i)
			for (int j = 0; j < column.length; ++j)
				ret[i][j] = data[row[i] * numColumns + column[j]];
		return ret;
	}

	/**
	 * @see jmp.ZeroRowMatrix#zeroRows(int[], double)
	 */
	public void zeroRows(int[] row, double diagonal) {
		for (int i = 0; i < row.length; ++i)
			synchronized (lock[row[i]]) {
				Arrays.fill(data, rowIndex[row[i]], rowIndex[row[i] + 1], 0.);
				if (row[i] < numColumns)
					data[row[i] * numColumns + row[i]] = diagonal;
			}
	}

	/**
	 * @see jmp.DenseRowColumnAccessMatrix#getMatrix()
	 */
	public IntDoubleVectorPair getMatrix() {
		return new IntDoubleVectorPair(rowIndex, data);
	}

	/**
	 * @see jmp.DenseRowColumnAccessMatrix#setMatrix(jmp.IntDoubleVectorPair)
	 */
	public synchronized void setMatrix(IntDoubleVectorPair x) {
		data = x.data;
		rowIndex = x.index;
	}

	/**
	 * @see jmp.VectorAccess#getData()
	 */
	public double[] getData() {
		return data;
	}

}
