/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp;

/**
 * Dense vector.
 */
public class DenseVector
	extends AbstractVector
	implements DenseAccessVector, ElementalAccessVector, VectorAccess, Vector {

	/**
	 * Internal representation
	 */
	private double[] data;

	/**
	 * For thread safe operations
	 */
	private Lock[] lock;

	/**
	 * Constructor for DenseVector.
	 */
	public DenseVector(int size) {
		super(size);
		data = new double[size];
		lock = Lock.createLocks(size);
	}

	/**
	 * Constructor for DenseVector, and copies the contents from the
	 * supplied vector.
	 */
	public DenseVector(Vector x) {
		this(x.size());
		State.BLAS.copy(x, this);
	}

	/**
	 * @see jmp.DenseAccessVector#getVector()
	 */
	public double[] getVector() {
		return data;
	}

	/**
	 * @see jmp.DenseAccessVector#setVector(double[])
	 */
	public synchronized void setVector(double[] x) {
		data = x;
	}

	/**
	 * @see jmp.ElementalAccessVector#addValue(int, double)
	 */
	public void addValue(int index, double value) {
		synchronized (lock[index]) {
			data[index] += value;
		}
	}

	/**
	 * @see jmp.ElementalAccessVector#setValue(int, double)
	 */
	public void setValue(int index, double value) {
		synchronized (lock[index]) {
			data[index] = value;
		}
	}

	/**
	 * @see jmp.ElementalAccessVector#getValue(int)
	 */
	public double getValue(int index) {
		return data[index];
	}

	/**
	 * @see jmp.ElementalAccessVector#addValues(int[], double[])
	 */
	public void addValues(int[] index, double[] values) {
		for (int i = 0; i < index.length; ++i)
			synchronized (lock[index[i]]) {
				data[index[i]] += values[i];
			}
	}

	/**
	 * @see jmp.ElementalAccessVector#setValues(int[], double[])
	 */
	public void setValues(int[] index, double[] values) {
		for (int i = 0; i < index.length; ++i)
			synchronized (lock[index[i]]) {
				data[index[i]] = values[i];
			}
	}

	/**
	 * @see jmp.ElementalAccessVector#getValues(int[])
	 */
	public double[] getValues(int[] index) {
		double[] ret = new double[index.length];
		for (int i = 0; i < index.length; ++i)
			ret[i] = data[index[i]];
		return ret;
	}

	/**
	 * @see jmp.VectorAccess#getData()
	 */
	public double[] getData() {
		return data;
	}

}
