/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp;

import jmp.util.ParallelWorker;

/**
 * Shared memory parallel BLAS implementation. The following operations are
 * parallelized:
 * <ul>
 * <li>add</li>
 * <li>multAdd</li>
 * <li>transMultAdd</li>
 * <li>transpose</li>
 * </ul>
 * Parallelizing other operations generally results in a slow-down.
 * <p>
 * Set parallel operations in State.
 * </p>
 * @see jmp.State
 */
public class ParallelBLAS extends SequentialBLAS implements BLAS {

	/**
	 * @see jmp.AbstractBLAS#transposeI(jmp.Matrix, jmp.Matrix)
	 */
	protected Matrix transposeI(Matrix A, Matrix B) {
		if (A instanceof SparseRowAccessMatrix
			&& B instanceof SparseColumnAccessMatrix) {
			int[] part = ParallelWorker.partition(A.numRows());
			ParallelWorker[] threads =
				new SparseRowSparseColumnTranspose[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new SparseRowSparseColumnTranspose(
						(SparseRowAccessMatrix) A,
						(SparseColumnAccessMatrix) B,
						part[i],
						part[i + 1]);
			ParallelWorker.runThreads(threads);
		} else if (
			A instanceof SparseRowColumnAccessMatrix
				&& B instanceof SparseColumnAccessMatrix) {
			int[] part = ParallelWorker.partition(A.numRows());
			ParallelWorker[] threads =
				new SparseRowColumnSparseColumnTranspose[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new SparseRowColumnSparseColumnTranspose(
						(SparseRowColumnAccessMatrix) A,
						(SparseColumnAccessMatrix) B,
						part[i],
						part[i + 1]);
			ParallelWorker.runThreads(threads);
		} else if (
			A instanceof SparseColumnAccessMatrix
				&& B instanceof SparseRowAccessMatrix) {
			int[] part = ParallelWorker.partition(A.numColumns());
			ParallelWorker[] threads =
				new SparseColumnSparseRowTranspose[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new SparseColumnSparseRowTranspose(
						(SparseColumnAccessMatrix) A,
						(SparseRowAccessMatrix) B,
						part[i],
						part[i + 1]);
			ParallelWorker.runThreads(threads);
		} else if (
			A instanceof SparseColumnRowAccessMatrix
				&& B instanceof SparseRowAccessMatrix) {
			int[] part = ParallelWorker.partition(A.numColumns());
			ParallelWorker[] threads =
				new SparseColumnRowSparseRowTranspose[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new SparseColumnRowSparseRowTranspose(
						(SparseColumnRowAccessMatrix) A,
						(SparseRowAccessMatrix) B,
						part[i],
						part[i + 1]);
			ParallelWorker.runThreads(threads);
		} else if (
			A instanceof DenseRowAccessMatrix
				&& B instanceof DenseColumnAccessMatrix) {
			int[] part = ParallelWorker.partition(A.numRows());
			ParallelWorker[] threads =
				new DenseRowDenseColumnTranspose[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new DenseRowDenseColumnTranspose(
						(DenseRowAccessMatrix) A,
						(DenseColumnAccessMatrix) B,
						part[i],
						part[i + 1]);
			ParallelWorker.runThreads(threads);
		} else if (
			A instanceof DenseRowColumnAccessMatrix
				&& B instanceof DenseColumnAccessMatrix) {
			int[] part = ParallelWorker.partition(A.numRows());
			ParallelWorker[] threads =
				new DenseRowColumnDenseColumnTranspose[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new DenseRowColumnDenseColumnTranspose(
						(DenseRowColumnAccessMatrix) A,
						(DenseColumnAccessMatrix) B,
						part[i],
						part[i + 1]);
			ParallelWorker.runThreads(threads);
		} else if (
			A instanceof DenseColumnAccessMatrix
				&& B instanceof DenseRowAccessMatrix) {
			int[] part = ParallelWorker.partition(A.numColumns());
			ParallelWorker[] threads =
				new DenseColumnDenseRowTranspose[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new DenseColumnDenseRowTranspose(
						(DenseColumnAccessMatrix) A,
						(DenseRowAccessMatrix) B,
						part[i],
						part[i + 1]);
			ParallelWorker.runThreads(threads);
		} else if (
			A instanceof DenseColumnRowAccessMatrix
				&& B instanceof DenseRowAccessMatrix) {
			int[] part = ParallelWorker.partition(A.numColumns());
			ParallelWorker[] threads =
				new DenseColumnRowDenseRowTranspose[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new DenseColumnRowDenseRowTranspose(
						(DenseColumnRowAccessMatrix) A,
						(DenseRowAccessMatrix) B,
						part[i],
						part[i + 1]);
			ParallelWorker.runThreads(threads);
		} else if (
			A instanceof ElementalAccessMatrix
				&& B instanceof ElementalAccessMatrix) {
			zero(B);
			int[] part = ParallelWorker.partition(A.numRows());
			ParallelWorker[] threads = new ElementalTranspose[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new ElementalTranspose(
						(ElementalAccessMatrix) A,
						(ElementalAccessMatrix) B,
						part[i],
						part[i + 1]);
			ParallelWorker.runThreads(threads);
		} else
			throw new UnsupportedOperationException();
		return B;
	}

	/**
	 * @see jmp.AbstractBLAS#addI(double, jmp.Vector, double, jmp.Vector, jmp.Vector)
	 */
	protected Vector addI(
		double alpha,
		Vector x,
		double beta,
		Vector y,
		Vector z) {
		if (x instanceof DenseAccessVector
			&& y instanceof DenseAccessVector
			&& z instanceof DenseAccessVector) {

			int[] part = ParallelWorker.partition(z.size());
			ParallelWorker[] threads = new DenseAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new DenseAdd(
						part[i],
						part[i + 1],
						alpha,
						((DenseAccessVector) x).getVector(),
						beta,
						((DenseAccessVector) y).getVector(),
						((DenseAccessVector) z).getVector());
			ParallelWorker.runThreads(threads);

		} else if (
			x instanceof BlockAccessVector
				&& z instanceof BlockAccessVector
				&& z instanceof BlockAccessVector) {

			int[] part =
				ParallelWorker.partition(((BlockAccessVector) z).numBlocks());
			ParallelWorker[] threads = new BlockAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new BlockAdd(
						part[i],
						part[i + 1],
						alpha,
						(BlockAccessVector) x,
						beta,
						(BlockAccessVector) y,
						(BlockAccessVector) z);
			ParallelWorker.runThreads(threads);

		} else
			throw new UnsupportedOperationException();

		return z;
	}

	private class DenseAdd extends ParallelWorker {

		private double alpha, beta;
		private double[] x, y, z;

		public DenseAdd(
			int start,
			int stop,
			double alpha,
			double[] x,
			double beta,
			double[] y,
			double[] z) {
			super(start, stop);
			this.alpha = alpha;
			this.x = x;
			this.beta = beta;
			this.y = y;
			this.z = z;
		}

		public void run() {
			addI(alpha, x, beta, y, z, start, stop);
		}
	}

	private class BlockAdd extends ParallelWorker {

		private double alpha, beta;
		private BlockAccessVector x, y, z;

		public BlockAdd(
			int start,
			int stop,
			double alpha,
			BlockAccessVector x,
			double beta,
			BlockAccessVector y,
			BlockAccessVector z) {
			super(start, stop);
			this.alpha = alpha;
			this.x = x;
			this.beta = beta;
			this.y = y;
			this.z = z;
		}

		public void run() {
			for (int i = start; i < stop; ++i)
				addI(alpha, x.getBlock(i), beta, y.getBlock(i), z.getBlock(i));
		}
	}

	/**
	 * @see jmp.AbstractBLAS#multAddI(double, jmp.Matrix, jmp.Vector, double, jmp.Vector, jmp.Vector)
	 */
	protected Vector multAddI(
		double alpha,
		Matrix A,
		Vector x,
		double beta,
		Vector y,
		Vector z) {
		if (A instanceof BlockAccessMatrix
			&& x instanceof BlockAccessVector
			&& y instanceof BlockAccessVector
			&& z instanceof BlockAccessVector) {
			BlockAccessMatrix Ab = (BlockAccessMatrix) A;
			BlockAccessVector xb = (BlockAccessVector) x,
				yb = (BlockAccessVector) y,
				zb = (BlockAccessVector) z;

			int[] part = ParallelWorker.partition(Ab.numBlocks());
			ParallelWorker[] threads = new BlockMultAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new BlockMultAdd(
						part[i],
						part[i + 1],
						alpha,
						Ab,
						xb,
						beta,
						yb,
						zb);
			ParallelWorker.runThreads(threads);

			return zb;
		} else if (A instanceof ShellMatrix)
			return ((ShellMatrix) A).multAdd(alpha, x, beta, y, z);

		// The following requires dense access to the vectors
		if (!(x instanceof DenseAccessVector)
			|| !(y instanceof DenseAccessVector)
			|| !(z instanceof DenseAccessVector))
			throw new UnsupportedOperationException();
		double[] xdata = ((DenseAccessVector) x).getVector(),
			ydata = ((DenseAccessVector) y).getVector(),
			zdata = ((DenseAccessVector) z).getVector();
		if (A instanceof SparseRowColumnAccessMatrix) {
			int[] part = ParallelWorker.partition(A.numRows());
			ParallelWorker[] threads =
				new SparseRowColumnMultAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new SparseRowColumnMultAdd(
						part[i],
						part[i + 1],
						alpha,
						(SparseRowColumnAccessMatrix) A,
						xdata,
						beta,
						ydata,
						zdata);
			ParallelWorker.runThreads(threads);
		} else if (A instanceof DenseRowColumnAccessMatrix) {
			int[] part = ParallelWorker.partition(A.numRows());
			ParallelWorker[] threads =
				new DenseRowColumnMultAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new DenseRowColumnMultAdd(
						part[i],
						part[i + 1],
						alpha,
						(DenseRowColumnAccessMatrix) A,
						xdata,
						beta,
						ydata,
						zdata);
			ParallelWorker.runThreads(threads);
		} else if (A instanceof SparseRowAccessMatrix) {
			int[] part = ParallelWorker.partition(A.numRows());
			ParallelWorker[] threads = new SparseRowMultAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new SparseRowMultAdd(
						part[i],
						part[i + 1],
						alpha,
						(SparseRowAccessMatrix) A,
						xdata,
						beta,
						ydata,
						zdata);
			ParallelWorker.runThreads(threads);
		} else if (A instanceof DenseRowAccessMatrix) {
			int[] part = ParallelWorker.partition(A.numRows());
			ParallelWorker[] threads = new DenseRowMultAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new DenseRowMultAdd(
						part[i],
						part[i + 1],
						alpha,
						(DenseRowAccessMatrix) A,
						xdata,
						beta,
						ydata,
						zdata);
			ParallelWorker.runThreads(threads);
		} else if (A instanceof DenseColumnAccessMatrix) {
			Lock[] lock = Lock.createLocks(A.numRows());

			if (y != z)
				zero(z);
			else
				scale(beta / alpha, z);

			int[] part = ParallelWorker.partition(A.numColumns());
			ParallelWorker[] threads = new DenseColumnMultAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new DenseColumnMultAdd(
						part[i],
						part[i + 1],
						(DenseColumnAccessMatrix) A,
						xdata,
						zdata,
						lock);
			ParallelWorker.runThreads(threads);

			if (y != z)
				add(beta, y, alpha, z);
			else
				scale(alpha, z);
		} else if (A instanceof DenseColumnRowAccessMatrix) {
			Lock[] lock = Lock.createLocks(A.numRows());

			if (y != z)
				zero(z);
			else
				scale(beta / alpha, z);

			int[] part = ParallelWorker.partition(A.numColumns());
			ParallelWorker[] threads =
				new DenseColumnRowMultAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new DenseColumnRowMultAdd(
						part[i],
						part[i + 1],
						(DenseColumnRowAccessMatrix) A,
						xdata,
						zdata,
						lock);
			ParallelWorker.runThreads(threads);

			if (y != z)
				add(beta, y, alpha, z);
			else
				scale(alpha, z);
		} else if (A instanceof SparseColumnAccessMatrix) {
			Lock[] lock = Lock.createLocks(A.numRows());

			if (y != z)
				zero(z);
			else
				scale(beta / alpha, z);

			int[] part = ParallelWorker.partition(A.numColumns());
			ParallelWorker[] threads =
				new SparseColumnMultAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new SparseColumnMultAdd(
						part[i],
						part[i + 1],
						(SparseColumnAccessMatrix) A,
						xdata,
						zdata,
						lock);
			ParallelWorker.runThreads(threads);

			if (y != z)
				add(beta, y, alpha, z);
			else
				scale(alpha, z);
		} else if (A instanceof SparseColumnRowAccessMatrix) {
			Lock[] lock = Lock.createLocks(A.numRows());

			if (y != z)
				zero(z);
			else
				scale(beta / alpha, z);

			int[] part = ParallelWorker.partition(A.numColumns());
			ParallelWorker[] threads =
				new SparseColumnRowMultAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new SparseColumnRowMultAdd(
						part[i],
						part[i + 1],
						(SparseColumnRowAccessMatrix) A,
						xdata,
						zdata,
						lock);
			ParallelWorker.runThreads(threads);

			if (y != z)
				add(beta, y, alpha, z);
			else
				scale(alpha, z);
		} else
			throw new UnsupportedOperationException();

		return z;
	}

	private class BlockMultAdd extends ParallelWorker {

		private double alpha, beta;
		private BlockAccessMatrix A;
		private BlockAccessVector x, y, z;

		public BlockMultAdd(
			int start,
			int stop,
			double alpha,
			BlockAccessMatrix A,
			BlockAccessVector x,
			double beta,
			BlockAccessVector y,
			BlockAccessVector z) {
			super(start, stop);
			this.alpha = alpha;
			this.A = A;
			this.x = x;
			this.beta = beta;
			this.y = y;
			this.z = z;
		}

		public void run() {
			multAddI(alpha, A, x, beta, y, z, start, stop);
		}
	}

	private class SparseRowMultAdd extends ParallelWorker {

		private double alpha, beta;
		private SparseRowAccessMatrix A;
		private double[] xdata, ydata, zdata;

		public SparseRowMultAdd(
			int start,
			int stop,
			double alpha,
			SparseRowAccessMatrix A,
			double[] xdata,
			double beta,
			double[] ydata,
			double[] zdata) {
			super(start, stop);
			this.alpha = alpha;
			this.A = A;
			this.xdata = xdata;
			this.beta = beta;
			this.ydata = ydata;
			this.zdata = zdata;
		}

		public void run() {
			multAddI(alpha, A, xdata, beta, ydata, zdata, start, stop);
		}
	}

	private class SparseRowColumnMultAdd extends ParallelWorker {

		private double alpha, beta;
		private SparseRowColumnAccessMatrix A;
		private double[] xdata, ydata, zdata;

		public SparseRowColumnMultAdd(
			int start,
			int stop,
			double alpha,
			SparseRowColumnAccessMatrix A,
			double[] xdata,
			double beta,
			double[] ydata,
			double[] zdata) {
			super(start, stop);
			this.alpha = alpha;
			this.A = A;
			this.xdata = xdata;
			this.beta = beta;
			this.ydata = ydata;
			this.zdata = zdata;
		}

		public void run() {
			multAddI(alpha, A, xdata, beta, ydata, zdata, start, stop);
		}
	}

	private class DenseRowMultAdd extends ParallelWorker {

		private double alpha, beta;
		private DenseRowAccessMatrix A;
		private double[] xdata, ydata, zdata;

		public DenseRowMultAdd(
			int start,
			int stop,
			double alpha,
			DenseRowAccessMatrix A,
			double[] xdata,
			double beta,
			double[] ydata,
			double[] zdata) {
			super(start, stop);
			this.alpha = alpha;
			this.A = A;
			this.xdata = xdata;
			this.beta = beta;
			this.ydata = ydata;
			this.zdata = zdata;
		}

		public void run() {
			multAddI(alpha, A, xdata, beta, ydata, zdata, start, stop);
		}
	}

	private class DenseRowColumnMultAdd extends ParallelWorker {

		private double alpha, beta;
		private DenseRowColumnAccessMatrix A;
		private double[] xdata, ydata, zdata;

		public DenseRowColumnMultAdd(
			int start,
			int stop,
			double alpha,
			DenseRowColumnAccessMatrix A,
			double[] xdata,
			double beta,
			double[] ydata,
			double[] zdata) {
			super(start, stop);
			this.alpha = alpha;
			this.A = A;
			this.xdata = xdata;
			this.beta = beta;
			this.ydata = ydata;
			this.zdata = zdata;
		}

		public void run() {
			multAddI(alpha, A, xdata, beta, ydata, zdata, start, stop);
		}
	}

	private class DenseColumnMultAdd extends ParallelWorker {

		private DenseColumnAccessMatrix A;
		private double[] xdata, zdata;
		private Lock[] lock;

		public DenseColumnMultAdd(
			int start,
			int stop,
			DenseColumnAccessMatrix A,
			double[] xdata,
			double[] zdata,
			Lock[] lock) {
			super(start, stop);
			this.A = A;
			this.xdata = xdata;
			this.zdata = zdata;
			this.lock = lock;
		}

		public void run() {
			multAddI(A, xdata, zdata, start, stop, lock);
		}
	}

	private class SparseColumnMultAdd extends ParallelWorker {

		private SparseColumnAccessMatrix A;
		private double[] xdata, zdata;
		private Lock[] lock;

		public SparseColumnMultAdd(
			int start,
			int stop,
			SparseColumnAccessMatrix A,
			double[] xdata,
			double[] zdata,
			Lock[] lock) {
			super(start, stop);
			this.A = A;
			this.xdata = xdata;
			this.zdata = zdata;
			this.lock = lock;
		}

		public void run() {
			multAddI(A, xdata, zdata, start, stop, lock);
		}
	}

	private class DenseColumnRowMultAdd extends ParallelWorker {

		private DenseColumnRowAccessMatrix A;
		private double[] xdata, zdata;
		private Lock[] lock;

		public DenseColumnRowMultAdd(
			int start,
			int stop,
			DenseColumnRowAccessMatrix A,
			double[] xdata,
			double[] zdata,
			Lock[] lock) {
			super(start, stop);
			this.A = A;
			this.xdata = xdata;
			this.zdata = zdata;
			this.lock = lock;
		}

		public void run() {
			multAddI(A, xdata, zdata, start, stop, lock);
		}
	}

	private class SparseColumnRowMultAdd extends ParallelWorker {

		private SparseColumnRowAccessMatrix A;
		private double[] xdata, zdata;
		private Lock[] lock;

		public SparseColumnRowMultAdd(
			int start,
			int stop,
			SparseColumnRowAccessMatrix A,
			double[] xdata,
			double[] zdata,
			Lock[] lock) {
			super(start, stop);
			this.A = A;
			this.xdata = xdata;
			this.zdata = zdata;
			this.lock = lock;
		}

		public void run() {
			multAddI(A, xdata, zdata, start, stop, lock);
		}
	}

	/**
	 * @see jmp.AbstractBLAS#transMultAddI(double, jmp.Matrix, jmp.Vector, double, jmp.Vector, jmp.Vector)
	 */
	protected Vector transMultAddI(
		double alpha,
		Matrix A,
		Vector x,
		double beta,
		Vector y,
		Vector z) {
		if (A instanceof BlockAccessMatrix
			&& x instanceof BlockAccessVector
			&& y instanceof BlockAccessVector
			&& z instanceof BlockAccessVector) {
			BlockAccessMatrix Ab = (BlockAccessMatrix) A;
			BlockAccessVector xb = (BlockAccessVector) x,
				yb = (BlockAccessVector) y,
				zb = (BlockAccessVector) z;

			int[] part = ParallelWorker.partition(Ab.numBlocks());
			ParallelWorker[] threads = new BlockTransMultAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new BlockTransMultAdd(
						part[i],
						part[i + 1],
						alpha,
						Ab,
						xb,
						beta,
						yb,
						zb);
			ParallelWorker.runThreads(threads);

			return zb;
		} else if (A instanceof ShellMatrix)
			return ((ShellMatrix) A).transMultAdd(alpha, x, beta, y, z);

		// The following requires dense access to the vectors
		if (!(x instanceof DenseAccessVector)
			|| !(y instanceof DenseAccessVector)
			|| !(z instanceof DenseAccessVector))
			throw new UnsupportedOperationException();
		double[] xdata = ((DenseAccessVector) x).getVector(),
			ydata = ((DenseAccessVector) y).getVector(),
			zdata = ((DenseAccessVector) z).getVector();
		if (A instanceof SparseRowColumnAccessMatrix) {
			Lock[] lock = Lock.createLocks(A.numColumns());

			if (y != z)
				zero(z);
			else
				scale(beta / alpha, z);

			int[] part = ParallelWorker.partition(A.numRows());
			ParallelWorker[] threads =
				new SparseRowColumnTransMultAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new SparseRowColumnTransMultAdd(
						part[i],
						part[i + 1],
						(SparseRowColumnAccessMatrix) A,
						xdata,
						zdata,
						lock);
			ParallelWorker.runThreads(threads);

			if (y != z)
				add(beta, y, alpha, z);
			else
				scale(alpha, z);
		} else if (A instanceof SparseRowAccessMatrix) {
			Lock[] lock = Lock.createLocks(A.numColumns());

			if (y != z)
				zero(z);
			else
				scale(beta / alpha, z);

			int[] part = ParallelWorker.partition(A.numRows());
			ParallelWorker[] threads =
				new SparseRowTransMultAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new SparseRowTransMultAdd(
						part[i],
						part[i + 1],
						(SparseRowAccessMatrix) A,
						xdata,
						zdata,
						lock);
			ParallelWorker.runThreads(threads);

			if (y != z)
				add(beta, y, alpha, z);
			else
				scale(alpha, z);
		} else if (A instanceof DenseRowColumnAccessMatrix) {
			Lock[] lock = Lock.createLocks(A.numColumns());

			if (y != z)
				zero(z);
			else
				scale(beta / alpha, z);

			int[] part = ParallelWorker.partition(A.numRows());
			ParallelWorker[] threads =
				new DenseRowColumnTransMultAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new DenseRowColumnTransMultAdd(
						part[i],
						part[i + 1],
						(DenseRowColumnAccessMatrix) A,
						xdata,
						zdata,
						lock);
			ParallelWorker.runThreads(threads);

			if (y != z)
				add(beta, y, alpha, z);
			else
				scale(alpha, z);
		} else if (A instanceof DenseRowAccessMatrix) {
			Lock[] lock = Lock.createLocks(A.numColumns());

			if (y != z)
				zero(z);
			else
				scale(beta / alpha, z);

			int[] part = ParallelWorker.partition(A.numRows());
			ParallelWorker[] threads =
				new DenseRowTransMultAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new DenseRowTransMultAdd(
						part[i],
						part[i + 1],
						(DenseRowAccessMatrix) A,
						xdata,
						zdata,
						lock);
			ParallelWorker.runThreads(threads);

			if (y != z)
				add(beta, y, alpha, z);
			else
				scale(alpha, z);
		} else if (A instanceof SparseColumnAccessMatrix) {
			int[] part = ParallelWorker.partition(A.numColumns());
			ParallelWorker[] threads =
				new SparseColumnTransMultAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new SparseColumnTransMultAdd(
						part[i],
						part[i + 1],
						alpha,
						(SparseColumnAccessMatrix) A,
						xdata,
						beta,
						ydata,
						zdata);
			ParallelWorker.runThreads(threads);
		} else if (A instanceof SparseColumnRowAccessMatrix) {
			int[] part = ParallelWorker.partition(A.numColumns());
			ParallelWorker[] threads =
				new SparseColumnRowTransMultAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new SparseColumnRowTransMultAdd(
						part[i],
						part[i + 1],
						alpha,
						(SparseColumnRowAccessMatrix) A,
						xdata,
						beta,
						ydata,
						zdata);
			ParallelWorker.runThreads(threads);
		} else if (A instanceof DenseColumnRowAccessMatrix) {
			int[] part = ParallelWorker.partition(A.numColumns());
			ParallelWorker[] threads =
				new DenseColumnRowTransMultAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new DenseColumnRowTransMultAdd(
						part[i],
						part[i + 1],
						alpha,
						(DenseColumnRowAccessMatrix) A,
						xdata,
						beta,
						ydata,
						zdata);
			ParallelWorker.runThreads(threads);
		} else if (A instanceof DenseColumnAccessMatrix) {
			int[] part = ParallelWorker.partition(A.numColumns());
			ParallelWorker[] threads =
				new DenseColumnTransMultAdd[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				threads[i] =
					new DenseColumnTransMultAdd(
						part[i],
						part[i + 1],
						alpha,
						(DenseColumnAccessMatrix) A,
						xdata,
						beta,
						ydata,
						zdata);
			ParallelWorker.runThreads(threads);
		} else
			throw new UnsupportedOperationException();

		return z;
	}

	private class BlockTransMultAdd extends ParallelWorker {

		private double alpha, beta;
		private BlockAccessMatrix A;
		private BlockAccessVector x, y, z;

		public BlockTransMultAdd(
			int start,
			int stop,
			double alpha,
			BlockAccessMatrix A,
			BlockAccessVector x,
			double beta,
			BlockAccessVector y,
			BlockAccessVector z) {
			super(start, stop);
			this.alpha = alpha;
			this.A = A;
			this.x = x;
			this.beta = beta;
			this.y = y;
			this.z = z;
		}

		public void run() {
			transMultAddI(alpha, A, x, beta, y, z, start, stop);
		}
	}

	private class SparseRowTransMultAdd extends ParallelWorker {

		private SparseRowAccessMatrix A;
		private double[] xdata, zdata;
		private Lock[] lock;

		public SparseRowTransMultAdd(
			int start,
			int stop,
			SparseRowAccessMatrix A,
			double[] xdata,
			double[] zdata,
			Lock[] lock) {
			super(start, stop);
			this.A = A;
			this.xdata = xdata;
			this.zdata = zdata;
			this.lock = lock;
		}

		public void run() {
			transMultAddI(A, xdata, zdata, start, stop, lock);
		}
	}

	private class SparseRowColumnTransMultAdd extends ParallelWorker {

		private SparseRowColumnAccessMatrix A;
		private double[] xdata, zdata;
		private Lock[] lock;

		public SparseRowColumnTransMultAdd(
			int start,
			int stop,
			SparseRowColumnAccessMatrix A,
			double[] xdata,
			double[] zdata,
			Lock[] lock) {
			super(start, stop);
			this.A = A;
			this.xdata = xdata;
			this.zdata = zdata;
			this.lock = lock;
		}

		public void run() {
			transMultAddI(A, xdata, zdata, start, stop, lock);
		}
	}

	private class DenseRowTransMultAdd extends ParallelWorker {

		private DenseRowAccessMatrix A;
		private double[] xdata, zdata;
		private Lock[] lock;

		public DenseRowTransMultAdd(
			int start,
			int stop,
			DenseRowAccessMatrix A,
			double[] xdata,
			double[] zdata,
			Lock[] lock) {
			super(start, stop);
			this.A = A;
			this.xdata = xdata;
			this.zdata = zdata;
			this.lock = lock;
		}

		public void run() {
			transMultAddI(A, xdata, zdata, start, stop, lock);
		}
	}

	private class DenseRowColumnTransMultAdd extends ParallelWorker {

		private DenseRowColumnAccessMatrix A;
		private double[] xdata, zdata;
		private Lock[] lock;

		public DenseRowColumnTransMultAdd(
			int start,
			int stop,
			DenseRowColumnAccessMatrix A,
			double[] xdata,
			double[] zdata,
			Lock[] lock) {
			super(start, stop);
			this.A = A;
			this.xdata = xdata;
			this.zdata = zdata;
			this.lock = lock;
		}

		public void run() {
			transMultAddI(A, xdata, zdata, start, stop, lock);
		}
	}

	private class DenseColumnTransMultAdd extends ParallelWorker {

		private double alpha, beta;
		private DenseColumnAccessMatrix A;
		private double[] xdata, ydata, zdata;

		public DenseColumnTransMultAdd(
			int start,
			int stop,
			double alpha,
			DenseColumnAccessMatrix A,
			double[] xdata,
			double beta,
			double[] ydata,
			double[] zdata) {
			super(start, stop);
			this.alpha = alpha;
			this.A = A;
			this.xdata = xdata;
			this.beta = beta;
			this.ydata = ydata;
			this.zdata = zdata;
		}

		public void run() {
			transMultAddI(alpha, A, xdata, beta, ydata, zdata, start, stop);
		}
	}

	private class SparseColumnTransMultAdd extends ParallelWorker {

		private double alpha, beta;
		private SparseColumnAccessMatrix A;
		private double[] xdata, ydata, zdata;

		public SparseColumnTransMultAdd(
			int start,
			int stop,
			double alpha,
			SparseColumnAccessMatrix A,
			double[] xdata,
			double beta,
			double[] ydata,
			double[] zdata) {
			super(start, stop);
			this.alpha = alpha;
			this.A = A;
			this.xdata = xdata;
			this.beta = beta;
			this.ydata = ydata;
			this.zdata = zdata;
		}

		public void run() {
			transMultAddI(alpha, A, xdata, beta, ydata, zdata, start, stop);
		}
	}

	private class DenseColumnRowTransMultAdd extends ParallelWorker {

		private double alpha, beta;
		private DenseColumnRowAccessMatrix A;
		private double[] xdata, ydata, zdata;

		public DenseColumnRowTransMultAdd(
			int start,
			int stop,
			double alpha,
			DenseColumnRowAccessMatrix A,
			double[] xdata,
			double beta,
			double[] ydata,
			double[] zdata) {
			super(start, stop);
			this.alpha = alpha;
			this.A = A;
			this.xdata = xdata;
			this.beta = beta;
			this.ydata = ydata;
			this.zdata = zdata;
		}

		public void run() {
			transMultAddI(alpha, A, xdata, beta, ydata, zdata, start, stop);
		}
	}

	private class SparseColumnRowTransMultAdd extends ParallelWorker {

		private double alpha, beta;
		private SparseColumnRowAccessMatrix A;
		private double[] xdata, ydata, zdata;

		public SparseColumnRowTransMultAdd(
			int start,
			int stop,
			double alpha,
			SparseColumnRowAccessMatrix A,
			double[] xdata,
			double beta,
			double[] ydata,
			double[] zdata) {
			super(start, stop);
			this.alpha = alpha;
			this.A = A;
			this.xdata = xdata;
			this.beta = beta;
			this.ydata = ydata;
			this.zdata = zdata;
		}

		public void run() {
			transMultAddI(alpha, A, xdata, beta, ydata, zdata, start, stop);
		}
	}

	private class SparseRowSparseColumnTranspose extends ParallelWorker {

		private SparseRowAccessMatrix A;
		private SparseColumnAccessMatrix B;

		public SparseRowSparseColumnTranspose(
			SparseRowAccessMatrix A,
			SparseColumnAccessMatrix B,
			int start,
			int stop) {
			super(start, stop);
			this.A = A;
			this.B = B;
		}

		public void run() {
			transposeI(A, B, start, stop);
		}
	}

	private class SparseRowColumnSparseColumnTranspose extends ParallelWorker {

		private SparseRowColumnAccessMatrix A;
		private SparseColumnAccessMatrix B;

		public SparseRowColumnSparseColumnTranspose(
			SparseRowColumnAccessMatrix A,
			SparseColumnAccessMatrix B,
			int start,
			int stop) {
			super(start, stop);
			this.A = A;
			this.B = B;
		}

		public void run() {
			transposeI(A, B, start, stop);
		}
	}

	private class SparseColumnSparseRowTranspose extends ParallelWorker {

		private SparseColumnAccessMatrix A;
		private SparseRowAccessMatrix B;

		public SparseColumnSparseRowTranspose(
			SparseColumnAccessMatrix A,
			SparseRowAccessMatrix B,
			int start,
			int stop) {
			super(start, stop);
			this.A = A;
			this.B = B;
		}

		public void run() {
			transposeI(A, B, start, stop);
		}
	}

	private class SparseColumnRowSparseRowTranspose extends ParallelWorker {

		private SparseColumnRowAccessMatrix A;
		private SparseRowAccessMatrix B;

		public SparseColumnRowSparseRowTranspose(
			SparseColumnRowAccessMatrix A,
			SparseRowAccessMatrix B,
			int start,
			int stop) {
			super(start, stop);
			this.A = A;
			this.B = B;
		}

		public void run() {
			transposeI(A, B, start, stop);
		}
	}

	private class DenseRowDenseColumnTranspose extends ParallelWorker {

		private DenseRowAccessMatrix A;
		private DenseColumnAccessMatrix B;

		public DenseRowDenseColumnTranspose(
			DenseRowAccessMatrix A,
			DenseColumnAccessMatrix B,
			int start,
			int stop) {
			super(start, stop);
			this.A = A;
			this.B = B;
		}

		public void run() {
			transposeI(A, B, start, stop);
		}
	}

	private class DenseRowColumnDenseColumnTranspose extends ParallelWorker {

		private DenseRowColumnAccessMatrix A;
		private DenseColumnAccessMatrix B;

		public DenseRowColumnDenseColumnTranspose(
			DenseRowColumnAccessMatrix A,
			DenseColumnAccessMatrix B,
			int start,
			int stop) {
			super(start, stop);
			this.A = A;
			this.B = B;
		}

		public void run() {
			transposeI(A, B, start, stop);
		}
	}

	private class DenseColumnDenseRowTranspose extends ParallelWorker {

		private DenseColumnAccessMatrix A;
		private DenseRowAccessMatrix B;

		public DenseColumnDenseRowTranspose(
			DenseColumnAccessMatrix A,
			DenseRowAccessMatrix B,
			int start,
			int stop) {
			super(start, stop);
			this.A = A;
			this.B = B;
		}

		public void run() {
			transposeI(A, B, start, stop);
		}
	}

	private class DenseColumnRowDenseRowTranspose extends ParallelWorker {

		private DenseColumnRowAccessMatrix A;
		private DenseRowAccessMatrix B;

		public DenseColumnRowDenseRowTranspose(
			DenseColumnRowAccessMatrix A,
			DenseRowAccessMatrix B,
			int start,
			int stop) {
			super(start, stop);
			this.A = A;
			this.B = B;
		}

		public void run() {
			transposeI(A, B, start, stop);
		}
	}

	private class ElementalTranspose extends ParallelWorker {

		private ElementalAccessMatrix A;
		private ElementalAccessMatrix B;

		public ElementalTranspose(
			ElementalAccessMatrix A,
			ElementalAccessMatrix B,
			int start,
			int stop) {
			super(start, stop);
			this.A = A;
			this.B = B;
		}

		public void run() {
			transposeI(A, B, start, stop);
		}
	}

}
