/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp;

import java.util.Arrays;

/**
 * Sequential BLAS implementation
 */
public class SequentialBLAS extends AbstractBLAS implements BLAS {

	/**
	 * @see jmp.AbstractBLAS#addDiagonalI(double, jmp.Matrix)
	 */
	protected Matrix addDiagonalI(double alpha, Matrix A) {
		if (A instanceof ElementalAccessMatrix)
			addDiagonalI(alpha, (ElementalAccessMatrix) A, 0, A.numRows());
		else
			throw new UnsupportedOperationException();
		return A;
	}

	protected void addDiagonalI(
		double alpha,
		ElementalAccessMatrix A,
		int start,
		int stop) {
		for (int i = start; i < stop && i < A.numColumns(); ++i)
			A.addValue(i, i, alpha);
	}

	/**
	 * @see jmp.AbstractBLAS#addI(double, jmp.Vector, double, jmp.Vector, jmp.Vector)
	 */
	protected Vector addI(
		double alpha,
		Vector x,
		double beta,
		Vector y,
		Vector z) {
		if (x instanceof DenseAccessVector
			&& y instanceof DenseAccessVector
			&& z instanceof DenseAccessVector)
			addI(
				alpha,
				((DenseAccessVector) x).getVector(),
				beta,
				((DenseAccessVector) y).getVector(),
				((DenseAccessVector) z).getVector(),
				0,
				z.size());
		else if (
			x instanceof BlockAccessVector
				&& z instanceof BlockAccessVector
				&& z instanceof BlockAccessVector)
			addI(
				alpha,
				(BlockAccessVector) x,
				beta,
				(BlockAccessVector) y,
				(BlockAccessVector) z,
				0,
				((BlockAccessVector) x).numBlocks());
		else
			throw new UnsupportedOperationException();

		return z;
	}

	protected void addI(
		double alpha,
		BlockAccessVector x,
		double beta,
		BlockAccessVector y,
		BlockAccessVector z,
		int start,
		int stop) {
		for (int i = start; i < stop; ++i)
			add(alpha, x.getBlock(i), beta, y.getBlock(i), z.getBlock(i));
	}

	protected void addI(
		double alpha,
		double[] x,
		double beta,
		double[] y,
		double[] z,
		int start,
		int stop) {
		for (int i = start; i < stop; ++i)
			z[i] = alpha * x[i] + beta * y[i];
	}

	/**
	 * @see jmp.AbstractBLAS#copyI(jmp.Matrix, jmp.Matrix, int, int, int, int)
	 */
	protected Matrix copyI(
		Matrix A,
		Matrix B,
		int startRow,
		int stopRow,
		int startColumn,
		int stopColumn) {
		zero(B);

		if (A instanceof SparseRowColumnAccessMatrix
			&& B instanceof ElementalAccessMatrix)
			copyI(
				(SparseRowColumnAccessMatrix) A,
				(ElementalAccessMatrix) B,
				startRow,
				stopRow,
				startColumn,
				stopColumn);
		else if (
			A instanceof SparseRowAccessMatrix
				&& B instanceof ElementalAccessMatrix)
			copyI(
				(SparseRowAccessMatrix) A,
				(ElementalAccessMatrix) B,
				startRow,
				stopRow,
				startColumn,
				stopColumn);
		else if (
			A instanceof SparseColumnAccessMatrix
				&& B instanceof ElementalAccessMatrix)
			copyI(
				(SparseColumnAccessMatrix) A,
				(ElementalAccessMatrix) B,
				startRow,
				stopRow,
				startColumn,
				stopColumn);
		else if (
			A instanceof SparseColumnRowAccessMatrix
				&& B instanceof ElementalAccessMatrix)
			copyI(
				(SparseColumnRowAccessMatrix) A,
				(ElementalAccessMatrix) B,
				startRow,
				stopRow,
				startColumn,
				stopColumn);
		else if (
			A instanceof DenseRowColumnAccessMatrix
				&& B instanceof ElementalAccessMatrix)
			copyI(
				(DenseRowColumnAccessMatrix) A,
				(ElementalAccessMatrix) B,
				startRow,
				stopRow,
				startColumn,
				stopColumn);
		else if (
			A instanceof DenseRowAccessMatrix
				&& B instanceof ElementalAccessMatrix)
			copyI(
				(DenseRowAccessMatrix) A,
				(ElementalAccessMatrix) B,
				startRow,
				stopRow,
				startColumn,
				stopColumn);
		else if (
			A instanceof DenseColumnRowAccessMatrix
				&& B instanceof ElementalAccessMatrix)
			copyI(
				(DenseColumnRowAccessMatrix) A,
				(ElementalAccessMatrix) B,
				startRow,
				stopRow,
				startColumn,
				stopColumn);
		else if (
			A instanceof DenseColumnAccessMatrix
				&& B instanceof ElementalAccessMatrix)
			copyI(
				(DenseColumnAccessMatrix) A,
				(ElementalAccessMatrix) B,
				startRow,
				stopRow,
				startColumn,
				stopColumn);
		else if (
			A instanceof CoordinateAccessMatrix
				&& B instanceof ElementalAccessMatrix)
			copyI(
				(CoordinateAccessMatrix) A,
				(ElementalAccessMatrix) B,
				startRow,
				stopRow,
				startColumn,
				stopColumn);
		else if (
			A instanceof ElementalAccessMatrix
				&& B instanceof ElementalAccessMatrix)
			copyI(
				(ElementalAccessMatrix) A,
				(ElementalAccessMatrix) B,
				startRow,
				stopRow,
				startColumn,
				stopColumn);
		else
			throw new UnsupportedOperationException();

		return B;
	}

	protected void copyI(
		CoordinateAccessMatrix A,
		ElementalAccessMatrix B,
		int startRow,
		int stopRow,
		int startColumn,
		int stopColumn) {
		int[] row = A.getRowIndices(), column = A.getColumnIndices();
		double[] data = A.getData();

		for (int i = 0; i < data.length; ++i)
			if (row[i] >= startRow
				&& row[i] < stopRow
				&& column[i] >= startColumn
				&& column[i] < stopColumn)
				B.setValue(row[i] - startRow, column[i] - startColumn, data[i]);
	}

	protected void copyI(
		SparseRowColumnAccessMatrix A,
		ElementalAccessMatrix B,
		int startRow,
		int stopRow,
		int startColumn,
		int stopColumn) {
		IntIntDoubleVectorTriple Amat = A.getMatrix();
		for (int i = startRow; i < stopRow; ++i)
			for (int j = Amat.minor[i]; j < Amat.minor[i + 1]; ++j)
				if (Amat.major[j] >= startColumn && Amat.major[j] < stopColumn)
					B.setValue(
						i - startRow,
						Amat.major[j] - startColumn,
						Amat.data[j]);
	}

	protected void copyI(
		SparseRowAccessMatrix A,
		ElementalAccessMatrix B,
		int startRow,
		int stopRow,
		int startColumn,
		int stopColumn) {
		for (int i = startRow; i < stopRow; ++i) {
			IntDoubleVectorPair Arow = A.getRow(i);
			for (int j = 0; j < Arow.index.length; ++j)
				if (Arow.index[j] >= startColumn && Arow.index[j] < stopColumn)
					B.setValue(
						i - startRow,
						Arow.index[j] - startColumn,
						Arow.data[j]);
		}
	}

	protected void copyI(
		SparseColumnRowAccessMatrix A,
		ElementalAccessMatrix B,
		int startRow,
		int stopRow,
		int startColumn,
		int stopColumn) {
		IntIntDoubleVectorTriple Amat = A.getMatrix();
		for (int i = startColumn; i < stopColumn; ++i)
			for (int j = Amat.minor[i]; j < Amat.minor[i + 1]; ++j)
				if (Amat.major[j] >= startRow && Amat.major[j] < stopRow)
					B.setValue(
						Amat.major[j] - startRow,
						i - startColumn,
						Amat.data[j]);
	}

	protected void copyI(
		SparseColumnAccessMatrix A,
		ElementalAccessMatrix B,
		int startRow,
		int stopRow,
		int startColumn,
		int stopColumn) {
		for (int i = startColumn; i < stopColumn; ++i) {
			IntDoubleVectorPair Acol = A.getColumn(i);
			for (int j = 0; j < Acol.index.length; ++j)
				if (Acol.index[j] >= startRow && Acol.index[j] < stopRow)
					B.setValue(
						Acol.index[j] - startRow,
						i - startColumn,
						Acol.data[j]);
		}
	}

	protected void copyI(
		DenseRowColumnAccessMatrix A,
		ElementalAccessMatrix B,
		int startRow,
		int stopRow,
		int startColumn,
		int stopColumn) {
		IntDoubleVectorPair Amat = A.getMatrix();
		for (int i = startRow; i < stopRow; ++i)
			for (int j = Amat.index[i] + startColumn;
				j < Amat.index[i] + stopColumn;
				++j)
				if (Amat.data[j] != 0.)
					B.setValue(
						i - startRow,
						j - Amat.index[i] - startColumn,
						Amat.data[j]);
	}

	protected void copyI(
		DenseRowAccessMatrix A,
		ElementalAccessMatrix B,
		int startRow,
		int stopRow,
		int startColumn,
		int stopColumn) {
		for (int i = startRow; i < stopRow; ++i) {
			double[] Arow = A.getRow(i);
			for (int j = startColumn; j < stopColumn; ++j)
				if (Arow[j] != 0.)
					B.setValue(i - startRow, j - startColumn, Arow[j]);
		}
	}

	protected void copyI(
		DenseColumnRowAccessMatrix A,
		ElementalAccessMatrix B,
		int startRow,
		int stopRow,
		int startColumn,
		int stopColumn) {
		IntDoubleVectorPair Amat = A.getMatrix();
		for (int i = startColumn; i < stopColumn; ++i)
			for (int j = Amat.index[i] + startRow;
				j < Amat.index[i] + stopRow;
				++j)
				if (Amat.data[j] != 0.)
					B.setValue(
						j - Amat.index[j] - startRow,
						i - startColumn,
						Amat.data[j]);
	}

	protected void copyI(
		DenseColumnAccessMatrix A,
		ElementalAccessMatrix B,
		int startRow,
		int stopRow,
		int startColumn,
		int stopColumn) {
		for (int i = startColumn; i < stopColumn; ++i) {
			double[] Acol = A.getColumn(i);
			for (int j = startRow; j < stopRow; ++j)
				if (Acol[j] != 0.)
					B.setValue(j - startRow, i - startColumn, Acol[j]);
		}
	}

	protected void copyI(
		ElementalAccessMatrix A,
		ElementalAccessMatrix B,
		int startRow,
		int stopRow,
		int startColumn,
		int stopColumn) {
		for (int i = startRow; i < stopRow; ++i)
			for (int j = startColumn; j < stopColumn; ++j) {
				double val = A.getValue(i, j);
				if (val != 0.)
					B.setValue(i - startRow, j - startColumn, val);
			}
	}

	/**
	 * @see jmp.AbstractBLAS#copyI(jmp.Matrix, jmp.Matrix)
	 */
	protected Matrix copyI(Matrix A, Matrix B) {
		int stopRow = A.numRows(), stopColumn = A.numColumns();
		zero(B);

		if (A instanceof SparseRowColumnAccessMatrix
			&& B instanceof SparseRowColumnAccessMatrix)
			copyI(
				(SparseRowColumnAccessMatrix) A,
				(SparseRowColumnAccessMatrix) B);
		else if (
			A instanceof SparseRowColumnAccessMatrix
				&& B instanceof ElementalAccessMatrix)
			copyI(
				(SparseRowColumnAccessMatrix) A,
				(ElementalAccessMatrix) B,
				0,
				stopRow);
		else if (
			A instanceof SparseRowAccessMatrix
				&& B instanceof SparseRowAccessMatrix)
			copyI(
				(SparseRowAccessMatrix) A,
				(SparseRowAccessMatrix) B,
				0,
				stopRow);
		else if (
			A instanceof SparseRowAccessMatrix
				&& B instanceof ElementalAccessMatrix)
			copyI(
				(SparseRowAccessMatrix) A,
				(ElementalAccessMatrix) B,
				0,
				stopRow);
		else if (
			A instanceof SparseColumnRowAccessMatrix
				&& B instanceof SparseColumnRowAccessMatrix)
			copyI(
				(SparseColumnRowAccessMatrix) A,
				(SparseColumnRowAccessMatrix) B);
		else if (
			A instanceof SparseColumnRowAccessMatrix
				&& B instanceof ElementalAccessMatrix)
			copyI(
				(SparseColumnRowAccessMatrix) A,
				(ElementalAccessMatrix) B,
				0,
				stopColumn);
		else if (
			A instanceof SparseColumnAccessMatrix
				&& B instanceof SparseColumnAccessMatrix)
			copyI(
				(SparseColumnAccessMatrix) A,
				(SparseColumnAccessMatrix) B,
				0,
				stopColumn);
		else if (
			A instanceof SparseColumnAccessMatrix
				&& B instanceof ElementalAccessMatrix)
			copyI(
				(SparseColumnAccessMatrix) A,
				(ElementalAccessMatrix) B,
				0,
				stopColumn);
		else if (
			A instanceof DenseRowColumnAccessMatrix
				&& B instanceof DenseRowColumnAccessMatrix)
			copyI(
				(DenseRowColumnAccessMatrix) A,
				(DenseRowColumnAccessMatrix) B);
		else if (
			A instanceof DenseRowColumnAccessMatrix
				&& B instanceof ElementalAccessMatrix)
			copyI(
				(DenseRowColumnAccessMatrix) A,
				(ElementalAccessMatrix) B,
				0,
				stopRow);
		else if (
			A instanceof DenseRowAccessMatrix
				&& B instanceof DenseRowAccessMatrix)
			copyI(
				(DenseRowAccessMatrix) A,
				(DenseRowAccessMatrix) B,
				0,
				stopRow);
		else if (
			A instanceof DenseRowAccessMatrix
				&& B instanceof ElementalAccessMatrix)
			copyI(
				(DenseRowAccessMatrix) A,
				(ElementalAccessMatrix) B,
				0,
				stopRow);
		else if (
			A instanceof DenseColumnRowAccessMatrix
				&& B instanceof DenseColumnRowAccessMatrix)
			copyI(
				(DenseColumnRowAccessMatrix) A,
				(DenseColumnRowAccessMatrix) B);
		else if (
			A instanceof DenseColumnRowAccessMatrix
				&& B instanceof ElementalAccessMatrix)
			copyI(
				(DenseColumnRowAccessMatrix) A,
				(ElementalAccessMatrix) B,
				0,
				stopColumn);
		else if (
			A instanceof DenseColumnAccessMatrix
				&& B instanceof DenseColumnAccessMatrix)
			copyI(
				(DenseColumnAccessMatrix) A,
				(DenseColumnAccessMatrix) B,
				0,
				stopColumn);
		else if (
			A instanceof DenseColumnAccessMatrix
				&& B instanceof ElementalAccessMatrix)
			copyI(
				(DenseColumnAccessMatrix) A,
				(ElementalAccessMatrix) B,
				0,
				stopColumn);
		else if (
			A instanceof CoordinateAccessMatrix
				&& B instanceof ElementalAccessMatrix) {
			copyI((CoordinateAccessMatrix) A, (ElementalAccessMatrix) B);
		} else if (
			A instanceof ElementalAccessMatrix
				&& B instanceof ElementalAccessMatrix)
			copyI(
				(ElementalAccessMatrix) A,
				(ElementalAccessMatrix) B,
				0,
				stopColumn);
		else
			throw new UnsupportedOperationException();

		return B;
	}

	protected void copyI(CoordinateAccessMatrix A, ElementalAccessMatrix B) {
		int[] row = A.getRowIndices(), column = A.getColumnIndices();
		double[] data = A.getData();

		for (int i = 0; i < data.length; ++i)
			B.setValue(row[i], column[i], data[i]);
	}

	protected void copyI(
		SparseRowColumnAccessMatrix A,
		SparseRowColumnAccessMatrix B) {
		B.setMatrix(A.getMatrix().copy());
	}

	protected void copyI(
		SparseRowColumnAccessMatrix A,
		ElementalAccessMatrix B,
		int startRow,
		int stopRow) {
		IntIntDoubleVectorTriple Amat = A.getMatrix();
		for (int i = startRow; i < stopRow; ++i)
			for (int j = Amat.minor[i]; j < Amat.minor[i + 1]; ++j)
				B.setValue(i, Amat.major[j], Amat.data[j]);
	}

	protected void copyI(
		SparseRowAccessMatrix A,
		SparseRowAccessMatrix B,
		int startRow,
		int stopRow) {
		for (int i = startRow; i < stopRow; ++i)
			B.setRow(i, A.getRow(i).copy());
	}

	protected void copyI(
		SparseRowAccessMatrix A,
		ElementalAccessMatrix B,
		int startRow,
		int stopRow) {
		for (int i = startRow; i < stopRow; ++i) {
			IntDoubleVectorPair Arow = A.getRow(i);
			for (int j = 0; j < Arow.index.length; ++j)
				B.setValue(i, Arow.index[j], Arow.data[j]);
		}
	}

	protected void copyI(
		SparseColumnRowAccessMatrix A,
		SparseColumnRowAccessMatrix B) {
		B.setMatrix(A.getMatrix().copy());
	}

	protected void copyI(
		SparseColumnRowAccessMatrix A,
		ElementalAccessMatrix B,
		int startColumn,
		int stopColumn) {
		IntIntDoubleVectorTriple Amat = A.getMatrix();
		for (int i = startColumn; i < stopColumn; ++i)
			for (int j = Amat.minor[i]; j < Amat.minor[i + 1]; ++j)
				B.setValue(Amat.major[j], i, Amat.data[j]);
	}

	protected void copyI(
		SparseColumnAccessMatrix A,
		SparseColumnAccessMatrix B,
		int startColumn,
		int stopColumn) {
		for (int i = startColumn; i < stopColumn; ++i)
			B.setColumn(i, A.getColumn(i).copy());
	}

	protected void copyI(
		SparseColumnAccessMatrix A,
		ElementalAccessMatrix B,
		int startColumn,
		int stopColumn) {
		for (int i = startColumn; i < stopColumn; ++i) {
			IntDoubleVectorPair Acol = A.getColumn(i);
			for (int j = 0; j < Acol.index.length; ++j)
				B.setValue(Acol.index[j], i, Acol.data[j]);
		}
	}

	protected void copyI(
		DenseRowColumnAccessMatrix A,
		DenseRowColumnAccessMatrix B) {
		B.setMatrix(A.getMatrix().copy());
	}

	protected void copyI(
		DenseRowColumnAccessMatrix A,
		ElementalAccessMatrix B,
		int startRow,
		int stopRow) {
		IntDoubleVectorPair Amat = A.getMatrix();
		for (int i = startRow; i < stopRow; ++i)
			for (int j = Amat.index[i], k = 0; j < Amat.index[i + 1]; ++j, ++k)
				if (Amat.data[j] != 0.)
					B.setValue(i, k, Amat.data[j]);
	}

	protected void copyI(
		DenseRowAccessMatrix A,
		DenseRowAccessMatrix B,
		int startRow,
		int stopRow) {
		for (int i = startRow; i < stopRow; ++i)
			System.arraycopy(A.getRow(i), 0, B.getRow(i), 0, A.numColumns());
	}

	protected void copyI(
		DenseRowAccessMatrix A,
		ElementalAccessMatrix B,
		int startRow,
		int stopRow) {
		for (int i = startRow; i < stopRow; ++i) {
			double[] Arow = A.getRow(i);
			for (int j = 0; j < Arow.length; ++j)
				if (Arow[j] != 0.)
					B.setValue(i, j, Arow[j]);
		}
	}

	protected void copyI(
		DenseColumnRowAccessMatrix A,
		DenseColumnRowAccessMatrix B) {
		B.setMatrix(A.getMatrix().copy());
	}

	protected void copyI(
		DenseColumnRowAccessMatrix A,
		ElementalAccessMatrix B,
		int startColumn,
		int stopColumn) {
		IntDoubleVectorPair Amat = A.getMatrix();
		for (int i = startColumn; i < stopColumn; ++i)
			for (int j = Amat.index[i]; j < Amat.index[i + 1]; ++j)
				if (Amat.data[j] != 0.)
					B.setValue(j - Amat.index[i], i, Amat.data[j]);
	}

	protected void copyI(
		DenseColumnAccessMatrix A,
		DenseColumnAccessMatrix B,
		int startColumn,
		int stopColumn) {
		for (int i = startColumn; i < stopColumn; ++i)
			System.arraycopy(A.getColumn(i), 0, B.getColumn(i), 0, A.numRows());
	}

	protected void copyI(
		DenseColumnAccessMatrix A,
		ElementalAccessMatrix B,
		int startColumn,
		int stopColumn) {
		for (int i = startColumn; i < stopColumn; ++i) {
			double[] Acol = A.getColumn(i);
			for (int j = 0; j < Acol.length; ++j)
				if (Acol[j] != 0.)
					B.setValue(j, i, Acol[j]);
		}
	}

	protected void copyI(
		ElementalAccessMatrix A,
		ElementalAccessMatrix B,
		int startRow,
		int stopRow) {
		for (int i = startRow; i < stopRow; ++i)
			for (int j = 0; j < A.numRows(); ++j) {
				double val = A.getValue(i, j);
				if (val != 0.)
					B.setValue(i, j, val);
			}
	}

	/**
	 * @see jmp.AbstractBLAS#copyI(jmp.Vector, jmp.Vector, int, int)
	 */
	protected Vector copyI(Vector x, Vector y, int start, int stop) {
		zero(y);

		if (x instanceof DenseAccessVector && y instanceof DenseAccessVector)
			copyI((DenseAccessVector) x, (DenseAccessVector) y, start, stop);
		else if (
			x instanceof DenseAccessVector
				&& y instanceof ElementalAccessVector)
			copyI(
				(DenseAccessVector) x,
				(ElementalAccessVector) y,
				start,
				stop);
		else if (
			x instanceof SparseAccessVector
				&& y instanceof ElementalAccessVector)
			copyI(
				(SparseAccessVector) x,
				(ElementalAccessVector) y,
				start,
				stop);
		else if (
			x instanceof ElementalAccessVector
				&& y instanceof ElementalAccessVector)
			copyI(
				(ElementalAccessVector) x,
				(ElementalAccessVector) y,
				start,
				stop);
		else
			throw new UnsupportedOperationException();

		return y;
	}

	protected void copyI(
		DenseAccessVector x,
		DenseAccessVector y,
		int start,
		int stop) {
		double[] ydata = new double[stop - start];
		System.arraycopy(x.getVector(), start, ydata, 0, ydata.length);
		y.setVector(ydata);
	}

	protected void copyI(
		DenseAccessVector x,
		ElementalAccessVector y,
		int start,
		int stop) {
		double[] xdata = x.getVector();
		for (int i = start; i < stop; ++i)
			if (xdata[i] != 0.)
				y.setValue(i - start, xdata[i]);
	}

	protected void copyI(
		SparseAccessVector x,
		ElementalAccessVector y,
		int start,
		int stop) {
		IntDoubleVectorPair xdata = x.getVector();
		for (int i = 0; i < xdata.index.length; ++i)
			if (xdata.index[i] >= start && xdata.index[i] < stop)
				y.setValue(xdata.index[i] - start, xdata.data[i]);
	}

	protected void copyI(
		ElementalAccessVector x,
		ElementalAccessVector y,
		int start,
		int stop) {
		for (int i = start; i < stop; ++i) {
			double val = x.getValue(i);
			if (val != 0.)
				y.setValue(i - start, val);
		}
	}

	/**
	 * @see jmp.AbstractBLAS#copyI(jmp.Vector, jmp.Vector)
	 */
	protected Vector copyI(Vector x, Vector y) {
		zero(y);

		if (x instanceof DenseAccessVector && y instanceof DenseAccessVector)
			copyI((DenseAccessVector) x, (DenseAccessVector) y);
		else if (
			x instanceof DenseAccessVector
				&& y instanceof ElementalAccessVector)
			copyI((DenseAccessVector) x, (ElementalAccessVector) y);
		else if (
			x instanceof SparseAccessVector && y instanceof SparseAccessVector)
			copyI((SparseAccessVector) x, (SparseAccessVector) y);
		else if (
			x instanceof SparseAccessVector
				&& y instanceof ElementalAccessVector)
			copyI((SparseAccessVector) x, (ElementalAccessVector) y);
		else if (
			x instanceof BlockAccessVector && y instanceof BlockAccessVector)
			copyI(
				(BlockAccessVector) x,
				(BlockAccessVector) y,
				0,
				((BlockAccessVector) x).numBlocks());
		else if (
			x instanceof ElementalAccessVector
				&& y instanceof ElementalAccessVector)
			copyI((ElementalAccessVector) x, (ElementalAccessVector) y);
		else
			throw new UnsupportedOperationException();

		return y;
	}

	protected void copyI(
		BlockAccessVector x,
		BlockAccessVector y,
		int start,
		int stop) {
		for (int i = start; i < stop; ++i)
			copy(x.getBlock(i), y.getBlock(i));
	}

	protected void copyI(DenseAccessVector x, DenseAccessVector y) {
		double[] ydata = new double[x.size()];
		System.arraycopy(x.getVector(), 0, ydata, 0, x.size());
		y.setVector(ydata);
	}

	protected void copyI(DenseAccessVector x, ElementalAccessVector y) {
		double[] xdata = x.getVector();
		for (int i = 0; i < xdata.length; ++i)
			if (xdata[i] != 0.)
				y.setValue(i, xdata[i]);
	}

	protected void copyI(SparseAccessVector x, SparseAccessVector y) {
		IntDoubleVectorPair xdata = x.getVector();
		int[] yindex = new int[xdata.index.length];
		double[] ydata = new double[xdata.data.length];
		System.arraycopy(xdata.index, 0, yindex, 0, yindex.length);
		System.arraycopy(xdata.data, 0, ydata, 0, ydata.length);
		y.setVector(new IntDoubleVectorPair(yindex, ydata));
	}

	protected void copyI(SparseAccessVector x, ElementalAccessVector y) {
		IntDoubleVectorPair xdata = x.getVector();
		for (int i = 0; i < xdata.index.length; ++i)
			y.setValue(xdata.index[i], xdata.data[i]);
	}

	protected void copyI(ElementalAccessVector x, ElementalAccessVector y) {
		for (int i = 0; i < x.size(); ++i) {
			double val = x.getValue(i);
			if (val != 0.)
				y.setValue(i, val);
		}
	}

	/**
	 * @see jmp.AbstractBLAS#dotI(jmp.Vector, jmp.Vector)
	 */
	protected double dotI(Vector x, Vector y) {
		if (x instanceof DenseAccessVector && y instanceof DenseAccessVector)
			return dotI((DenseAccessVector) x, (DenseAccessVector) y);
		else if (
			x instanceof DenseAccessVector && y instanceof SparseAccessVector)
			return dotI((DenseAccessVector) x, (SparseAccessVector) y);
		else if (
			x instanceof SparseAccessVector && y instanceof DenseAccessVector)
			return dotI((DenseAccessVector) y, (SparseAccessVector) x);
		else if (
			x instanceof BlockAccessVector && y instanceof BlockAccessVector)
			return dotI(
				block2array((BlockAccessVector) x),
				block2array((BlockAccessVector) y));
		else
			throw new UnsupportedOperationException();
	}

	protected double dotI(double[] xdata, double[] ydata) {
		double dot = 0.;
		for (int i = 0; i < xdata.length; ++i)
			dot += xdata[i] * ydata[i];
		return dot;
	}

	protected double dotI(DenseAccessVector x, DenseAccessVector y) {
		double[] xdata = x.getVector(), ydata = y.getVector();
		return dotI(xdata, ydata);
	}

	protected double dotI(DenseAccessVector x, SparseAccessVector y) {
		double dot = 0.;
		double[] xdata = x.getVector();
		IntDoubleVectorPair ydata = y.getVector();
		for (int i = 0; i < ydata.index.length; ++i)
			dot += xdata[ydata.index[i]] * ydata.data[i];
		return dot;
	}

	/**
	 * @see jmp.AbstractBLAS#multAddI(double, jmp.Matrix, jmp.Vector, double, jmp.Vector, jmp.Vector)
	 */
	protected Vector multAddI(
		double alpha,
		Matrix A,
		Vector x,
		double beta,
		Vector y,
		Vector z) {
		if (A instanceof BlockAccessMatrix
			&& x instanceof BlockAccessVector
			&& y instanceof BlockAccessVector
			&& z instanceof BlockAccessVector) {
			multAddI(
				alpha,
				(BlockAccessMatrix) A,
				(BlockAccessVector) x,
				beta,
				(BlockAccessVector) y,
				(BlockAccessVector) z,
				0,
				((BlockAccessMatrix) A).numBlocks());
			return z;
		} else if (A instanceof ShellMatrix)
			return ((ShellMatrix) A).multAdd(alpha, x, beta, y, z);

		// The following requires dense access to the vectors
		if (!(x instanceof DenseAccessVector)
			|| !(y instanceof DenseAccessVector)
			|| !(z instanceof DenseAccessVector))
			throw new UnsupportedOperationException();
		double[] xdata = ((DenseAccessVector) x).getVector(),
			ydata = ((DenseAccessVector) y).getVector(),
			zdata = ((DenseAccessVector) z).getVector();
		if (A instanceof SparseRowColumnAccessMatrix)
			multAddI(
				alpha,
				(SparseRowColumnAccessMatrix) A,
				xdata,
				beta,
				ydata,
				zdata,
				0,
				A.numRows());
		else if (A instanceof SparseRowAccessMatrix)
			multAddI(
				alpha,
				(SparseRowAccessMatrix) A,
				xdata,
				beta,
				ydata,
				zdata,
				0,
				A.numRows());
		else if (A instanceof DenseRowColumnAccessMatrix)
			multAddI(
				alpha,
				(DenseRowColumnAccessMatrix) A,
				xdata,
				beta,
				ydata,
				zdata,
				0,
				A.numRows());
		else if (A instanceof DenseRowAccessMatrix)
			multAddI(
				alpha,
				(DenseRowAccessMatrix) A,
				xdata,
				beta,
				ydata,
				zdata,
				0,
				A.numRows());
		else if (A instanceof DenseColumnAccessMatrix) {
			Lock[] lock = Lock.createLocks(A.numRows());
			if (y != z)
				zero(z);
			else
				scale(beta / alpha, z);
			multAddI(
				(DenseColumnAccessMatrix) A,
				xdata,
				zdata,
				0,
				A.numColumns(),
				lock);
			if (y != z)
				add(beta, y, alpha, z);
			else
				scale(alpha, z);
		} else if (A instanceof DenseColumnRowAccessMatrix) {
			Lock[] lock = Lock.createLocks(A.numRows());
			if (y != z)
				zero(z);
			else
				scale(beta / alpha, z);
			multAddI(
				(DenseColumnRowAccessMatrix) A,
				xdata,
				zdata,
				0,
				A.numColumns(),
				lock);
			if (y != z)
				add(beta, y, alpha, z);
			else
				scale(alpha, z);
		} else if (A instanceof SparseColumnAccessMatrix) {
			Lock[] lock = Lock.createLocks(A.numRows());
			if (y != z)
				zero(z);
			else
				scale(beta / alpha, z);
			multAddI(
				(SparseColumnAccessMatrix) A,
				xdata,
				zdata,
				0,
				A.numColumns(),
				lock);
			if (y != z)
				add(beta, y, alpha, z);
			else
				scale(alpha, z);
		} else if (A instanceof SparseColumnRowAccessMatrix) {
			Lock[] lock = Lock.createLocks(A.numRows());
			if (y != z)
				zero(z);
			else
				scale(beta / alpha, z);
			multAddI(
				(SparseColumnRowAccessMatrix) A,
				xdata,
				zdata,
				0,
				A.numColumns(),
				lock);
			if (y != z)
				add(beta, y, alpha, z);
			else
				scale(alpha, z);
		} else
			throw new UnsupportedOperationException();

		return z;
	}

	protected void multAddI(
		double alpha,
		BlockAccessMatrix A,
		BlockAccessVector x,
		double beta,
		BlockAccessVector y,
		BlockAccessVector z,
		int start,
		int stop) {
		for (int i = start; i < stop; ++i)
			multAdd(
				alpha,
				A.getBlock(i),
				x.getBlock(i),
				beta,
				y.getBlock(i),
				z.getBlock(i));
	}

	protected void multAddI(
		SparseColumnAccessMatrix A,
		double[] xdata,
		double[] zdata,
		int start,
		int stop,
		Lock[] lock) {
		for (int i = start; i < stop; ++i) {
			IntDoubleVectorPair Acol = A.getColumn(i);
			for (int j = 0; j < Acol.index.length; ++j)
				synchronized (lock[j]) {
					zdata[Acol.index[j]] += Acol.data[j] * xdata[i];
				}
		}
	}

	protected void multAddI(
		SparseColumnRowAccessMatrix A,
		double[] xdata,
		double[] zdata,
		int start,
		int stop,
		Lock[] lock) {
		IntIntDoubleVectorTriple Amat = A.getMatrix();
		for (int i = start; i < stop; ++i)
			for (int j = Amat.minor[i]; j < Amat.minor[i + 1]; ++j)
				synchronized (lock[Amat.major[j]]) {
					zdata[Amat.major[j]] += Amat.data[j] * xdata[i];
				}
	}

	protected void multAddI(
		DenseColumnAccessMatrix A,
		double[] xdata,
		double[] zdata,
		int start,
		int stop,
		Lock[] lock) {
		for (int i = start; i < stop; ++i) {
			double[] Acol = A.getColumn(i);
			for (int j = 0; j < Acol.length; ++j)
				synchronized (lock[j]) {
					zdata[j] += Acol[j] * xdata[i];
				}
		}
	}

	protected void multAddI(
		DenseColumnRowAccessMatrix A,
		double[] xdata,
		double[] zdata,
		int start,
		int stop,
		Lock[] lock) {
		IntDoubleVectorPair Amat = A.getMatrix();
		for (int i = start; i < stop; ++i)
			for (int j = Amat.index[i], k = 0;
				j < Amat.index[i + 1];
				++j, ++k) {
				synchronized (lock[k]) {
					zdata[k] += Amat.data[j] * xdata[i];
				}
			}
	}

	protected void multAddI(
		double alpha,
		DenseRowColumnAccessMatrix A,
		double[] xdata,
		double beta,
		double[] ydata,
		double[] zdata,
		int start,
		int stop) {
		IntDoubleVectorPair Amat = A.getMatrix();
		int columns = A.numColumns();
		for (int i = start; i < stop; ++i) {
			double val = 0.;

			int j = Amat.index[i], l = 0;
			for (int k = columns % 4; --k >= 0;)
				val += Amat.data[j++] * xdata[l++];
			for (int k = columns / 4; --k >= 0;)
				val += Amat.data[j++] * xdata[l++]
					+ Amat.data[j++] * xdata[l++]
					+ Amat.data[j++] * xdata[l++]
					+ Amat.data[j++] * xdata[l++];

			//for (int j = Amat.index[i], k = 0; j < Amat.index[i + 1];)
			//	val += Amat.data[j++] * xdata[k++];

			zdata[i] = alpha * val + beta * ydata[i];
		}
	}

	protected void multAddI(
		double alpha,
		DenseRowAccessMatrix A,
		double[] xdata,
		double beta,
		double[] ydata,
		double[] zdata,
		int start,
		int stop) {
		int columns = A.numColumns();
		for (int i = start; i < stop; ++i) {
			double[] Arow = A.getRow(i);
			double val = 0.;

			int j = 0, l = 0;
			for (int k = columns % 4; --k >= 0;)
				val += Arow[j++] * xdata[l++];
			for (int k = columns / 4; --k >= 0;)
				val += Arow[j++] * xdata[l++]
					+ Arow[j++] * xdata[l++]
					+ Arow[j++] * xdata[l++]
					+ Arow[j++] * xdata[l++];

			//for (int j = 0; j < Arow.length; ++j)
			//	val += Arow[j] * xdata[j];

			zdata[i] = alpha * val + beta * ydata[i];
		}
	}

	protected void multAddI(
		double alpha,
		SparseRowColumnAccessMatrix A,
		double[] xdata,
		double beta,
		double[] ydata,
		double[] zdata,
		int start,
		int stop) {
		IntIntDoubleVectorTriple Amat = A.getMatrix();
		for (int i = start; i < stop; ++i) {
			double val = 0.;
			for (int j = Amat.minor[i]; j < Amat.minor[i + 1]; ++j)
				val += Amat.data[j] * xdata[Amat.major[j]];
			zdata[i] = alpha * val + beta * ydata[i];
		}
	}

	protected void multAddI(
		double alpha,
		SparseRowAccessMatrix A,
		double[] xdata,
		double beta,
		double[] ydata,
		double[] zdata,
		int start,
		int stop) {
		for (int i = start; i < stop; ++i) {
			IntDoubleVectorPair Amat = A.getRow(i);
			double val = 0.;
			for (int j = 0; j < Amat.index.length; ++j)
				val += Amat.data[j] * xdata[Amat.index[j]];
			zdata[i] = alpha * val + beta * ydata[i];
		}
	}

	/**
	 * @see jmp.AbstractBLAS#norm1(jmp.Matrix)
	 */
	protected double norm1(Matrix A) {
		if (A instanceof DenseRowColumnAccessMatrix)
			return norm1((DenseRowColumnAccessMatrix) A);
		else if (A instanceof DenseRowAccessMatrix)
			return norm1((DenseRowAccessMatrix) A);
		else if (A instanceof SparseRowAccessMatrix)
			return norm1((SparseRowAccessMatrix) A);
		else if (A instanceof SparseRowColumnAccessMatrix)
			return norm1((SparseRowColumnAccessMatrix) A);
		else
			throw new UnsupportedOperationException();
	}

	protected double norm1(DenseRowColumnAccessMatrix A) {
		IntDoubleVectorPair Amat = A.getMatrix();
		double norm = 0;
		for (int i = 0; i < Amat.index.length; ++i)
			norm =
				Math.max(
					norm,
					norm1(Amat.data, Amat.index[i], Amat.index[i + 1]));
		return norm;
	}

	protected double norm1(DenseRowAccessMatrix A) {
		double norm = 0;
		for (int i = 0; i < A.numRows(); ++i)
			norm = Math.max(norm, norm1(A.getRow(i)));
		return norm;
	}

	protected double norm1(SparseRowAccessMatrix A) {
		double norm = 0;
		for (int i = 0; i < A.numRows(); ++i)
			norm = Math.max(norm, norm1(A.getRow(i).data));
		return norm;
	}

	protected double norm1(SparseRowColumnAccessMatrix A) {
		IntIntDoubleVectorTriple Amat = A.getMatrix();
		double norm = 0;
		for (int i = 0; i < Amat.minor.length; ++i)
			norm =
				Math.max(
					norm,
					norm1(Amat.data, Amat.minor[i], Amat.minor[i + 1]));
		return norm;
	}

	/**
	 * @see jmp.AbstractBLAS#norm1(jmp.Vector)
	 */
	protected double norm1(Vector x) {
		if (x instanceof VectorAccess)
			return norm1(((VectorAccess) x).getData());
		else if (x instanceof BlockAccessVector)
			return norm1(block2array((BlockAccessVector) x));
		else
			throw new UnsupportedOperationException();
	}

	protected double norm1(double[] data) {
		return norm1(data, 0, data.length);
	}

	protected double norm1(double[] data, int start, int stop) {
		double norm = 0.;
		for (int i = start; i < stop; ++i)
			norm += Math.abs(data[i]);
		return norm;
	}

	/**
	 * @see jmp.AbstractBLAS#norm2(jmp.Vector)
	 */
	protected double norm2(Vector x) {
		if (x instanceof VectorAccess)
			return norm2((VectorAccess) x);
		else if (x instanceof BlockAccessVector)
			return norm2(block2array((BlockAccessVector) x));
		else
			throw new UnsupportedOperationException();
	}

	protected double norm2(double[] xdata) {
		double scale = 0, ssq = 1;
		for (int i = 0; i < xdata.length; ++i) {
			if (xdata[i] != 0) {
				double absxi = Math.abs(xdata[i]);
				if (scale < absxi) {
					ssq = 1 + ssq * Math.pow(scale / absxi, 2);
					scale = absxi;
				} else
					ssq = ssq + Math.pow(absxi / scale, 2);
			}
		}
		return scale * Math.sqrt(ssq);
	}

	protected double norm2(VectorAccess x) {
		double[] xdata = x.getData();
		return norm2(xdata);
	}

	/**
	 * @see jmp.AbstractBLAS#normF(jmp.Matrix)
	 */
	protected double normF(Matrix A) {
		if (A instanceof VectorAccess)
			return norm2((VectorAccess) A);
		else if (A instanceof MatrixAccess)
			return normF((MatrixAccess) A);
		else
			throw new UnsupportedOperationException();
	}

	protected double normF(MatrixAccess A) {
		double[][] Adata = A.getData();
		double scale = 0, ssq = 1;
		for (int i = 0; i < Adata.length; ++i)
			for (int j = 0; j < Adata[i].length; ++j) {
				if (Adata[i][j] != 0) {
					double absxi = Math.abs(Adata[i][j]);
					if (scale < absxi) {
						ssq = 1 + ssq * Math.pow(scale / absxi, 2);
						scale = absxi;
					} else
						ssq = ssq + Math.pow(absxi / scale, 2);
				}
			}
		return scale * Math.sqrt(ssq);
	}

	/**
	 * @see jmp.AbstractBLAS#normInf(jmp.Matrix)
	 */
	protected double normInf(Matrix A) {
		if (A instanceof DenseColumnRowAccessMatrix)
			return normInf((DenseColumnRowAccessMatrix) A);
		else if (A instanceof DenseColumnAccessMatrix)
			return normInf((DenseColumnAccessMatrix) A);
		else if (A instanceof SparseColumnRowAccessMatrix)
			return normInf((SparseColumnRowAccessMatrix) A);
		else if (A instanceof SparseColumnAccessMatrix)
			return normInf((SparseColumnAccessMatrix) A);
		else
			throw new UnsupportedOperationException();
	}

	protected double normInf(DenseColumnAccessMatrix A) {
		double norm = 0;
		for (int i = 0; i < A.numColumns(); ++i)
			norm = Math.max(norm, norm1(A.getColumn(i)));
		return norm;
	}

	protected double normInf(DenseColumnRowAccessMatrix A) {
		IntDoubleVectorPair Amat = A.getMatrix();
		double norm = 0;
		for (int i = 0; i < Amat.index.length; ++i)
			norm =
				Math.max(
					norm,
					norm1(Amat.data, Amat.index[i], Amat.index[i + 1]));
		return norm;
	}

	protected double norm1(SparseColumnAccessMatrix A) {
		double norm = 0;
		for (int i = 0; i < A.numRows(); ++i)
			norm = Math.max(norm, norm1(A.getColumn(i).data));
		return norm;
	}

	protected double norm1(SparseColumnRowAccessMatrix A) {
		IntIntDoubleVectorTriple Amat = A.getMatrix();
		double norm = 0;
		for (int i = 0; i < Amat.minor.length; ++i)
			norm =
				Math.max(
					norm,
					norm1(Amat.data, Amat.minor[i], Amat.minor[i + 1]));
		return norm;
	}

	/**
	 * @see jmp.AbstractBLAS#normInf(jmp.Vector)
	 */
	protected double normInf(Vector x) {
		if (x instanceof VectorAccess)
			return normInf(((VectorAccess) x).getData());
		else if (x instanceof BlockAccessVector)
			return normInf(block2array((BlockAccessVector) x));
		else
			throw new UnsupportedOperationException();
	}

	protected double normInf(double[] data) {
		double norm = 0.;
		for (int i = 0; i < data.length; ++i)
			if (Math.abs(data[i]) > norm)
				norm = Math.abs(data[i]);
		return norm;
	}

	/**
	 * @see jmp.AbstractBLAS#rank1I(double, jmp.Vector, jmp.Vector, jmp.Matrix)
	 */
	protected Matrix rank1I(double alpha, Vector x, Vector y, Matrix A) {
		try {
			rank1I(
				alpha,
				(DenseVector) x,
				(DenseVector) y,
				(ElementalAccessMatrix) A,
				0,
				x.size());
		} catch (ClassCastException e) {
			throw new UnsupportedOperationException();
		}

		return A;
	}

	protected void rank1I(
		double alpha,
		DenseVector x,
		DenseVector y,
		ElementalAccessMatrix A,
		int start,
		int stop) {
		double[] xdata = x.getData(), ydata = y.getData();

		for (int i = start; i < stop; ++i) {
			double xval = xdata[i];
			if (xval != 0.)
				for (int j = 0; j < ydata.length; ++j) {
					double yval = ydata[j];
					if (yval != 0.)
						A.addValue(i, j, alpha * xval * yval);
				}
		}
	}

	/**
	 * @see jmp.AbstractBLAS#scaleI(double, jmp.Matrix)
	 */
	protected Matrix scaleI(double alpha, Matrix A) {
		if (A instanceof VectorAccess)
			scaleI(alpha, (VectorAccess) A);
		else if (A instanceof MatrixAccess)
			scaleI(alpha, (MatrixAccess) A, 0, A.numRows());
		else
			throw new UnsupportedOperationException();
		return A;
	}

	/**
	 * @see jmp.AbstractBLAS#scaleI(double, jmp.Vector)
	 */
	protected Vector scaleI(double alpha, Vector x) {
		if (x instanceof VectorAccess)
			scaleI(alpha, (VectorAccess) x);
		else if (x instanceof BlockAccessVector)
			scaleI(
				alpha,
				(BlockAccessVector) x,
				0,
				((BlockAccessVector) x).numBlocks());
		else
			throw new UnsupportedOperationException();
		return x;
	}

	protected void scaleI(
		double alpha,
		BlockAccessVector x,
		int start,
		int stop) {
		for (int i = start; i < stop; ++i)
			scale(alpha, x.getBlock(i));
	}

	protected void scaleI(double alpha, VectorAccess x) {
		double[] xdata = x.getData();
		for (int i = 0; i < xdata.length; ++i)
			xdata[i] *= alpha;
	}

	protected void scaleI(double alpha, MatrixAccess A, int start, int stop) {
		double[][] Adata = A.getData();
		for (int i = start; i < stop; ++i)
			for (int j = 0; j < Adata[i].length; ++j)
				Adata[i][j] *= alpha;
	}

	/**
	 * @see jmp.AbstractBLAS#setI(double, jmp.Vector)
	 */
	protected Vector setI(double alpha, Vector x) {
		if (x instanceof DenseAccessVector)
			return setI(alpha, (DenseAccessVector) x);
		else if (x instanceof ElementalAccessMatrix)
			return setI(alpha, (ElementalAccessVector) x);
		else
			throw new UnsupportedOperationException();
	}

	protected Vector setI(double alpha, DenseAccessVector x) {
		double[] xdata = x.getVector();
		Arrays.fill(xdata, alpha);
		return x;
	}

	protected Vector setI(double alpha, ElementalAccessVector x) {
		for (int i = 0; i < x.size(); ++i)
			x.setValue(i, alpha);
		return x;
	}

	/**
	 * @see jmp.AbstractBLAS#transMultAddI(double, jmp.Matrix, jmp.Vector, double, jmp.Vector, jmp.Vector)
	 */
	protected Vector transMultAddI(
		double alpha,
		Matrix A,
		Vector x,
		double beta,
		Vector y,
		Vector z) {
		if (A instanceof BlockAccessMatrix
			&& x instanceof BlockAccessVector
			&& y instanceof BlockAccessVector
			&& z instanceof BlockAccessVector) {
			transMultAddI(
				alpha,
				(BlockAccessMatrix) A,
				(BlockAccessVector) x,
				beta,
				(BlockAccessVector) y,
				(BlockAccessVector) z,
				0,
				((BlockAccessMatrix) A).numBlocks());
			return z;
		} else if (A instanceof ShellMatrix)
			return ((ShellMatrix) A).transMultAdd(alpha, x, beta, y, z);

		// The following requires dense access to the vectors
		if (!(x instanceof DenseAccessVector)
			|| !(y instanceof DenseAccessVector)
			|| !(z instanceof DenseAccessVector))
			throw new UnsupportedOperationException();
		double[] xdata = ((DenseAccessVector) x).getVector(),
			ydata = ((DenseAccessVector) y).getVector(),
			zdata = ((DenseAccessVector) z).getVector();
		if (A instanceof SparseRowColumnAccessMatrix) {
			Lock[] lock = Lock.createLocks(A.numColumns());
			if (y != z)
				zero(z);
			else
				scale(beta / alpha, z);
			transMultAddI(
				(SparseRowColumnAccessMatrix) A,
				xdata,
				zdata,
				0,
				A.numRows(),
				lock);
			if (y != z)
				add(beta, y, alpha, z);
			else
				scale(alpha, z);
		} else if (A instanceof SparseRowAccessMatrix) {
			Lock[] lock = Lock.createLocks(A.numColumns());
			if (y != z)
				zero(z);
			else
				scale(beta / alpha, z);
			transMultAddI(
				(SparseRowAccessMatrix) A,
				xdata,
				zdata,
				0,
				A.numRows(),
				lock);
			if (y != z)
				add(beta, y, alpha, z);
			else
				scale(alpha, z);
		} else if (A instanceof DenseRowColumnAccessMatrix) {
			Lock[] lock = Lock.createLocks(A.numColumns());
			if (y != z)
				zero(z);
			else
				scale(beta / alpha, z);
			transMultAddI(
				(DenseRowColumnAccessMatrix) A,
				xdata,
				zdata,
				0,
				A.numRows(),
				lock);
			if (y != z)
				add(beta, y, alpha, z);
			else
				scale(alpha, z);
		} else if (A instanceof DenseRowAccessMatrix) {
			Lock[] lock = Lock.createLocks(A.numColumns());

			if (y != z)
				zero(z);
			else
				scale(beta / alpha, z);

			transMultAddI(
				(DenseRowAccessMatrix) A,
				xdata,
				zdata,
				0,
				A.numRows(),
				lock);

			if (y != z)
				add(beta, y, alpha, z);
			else
				scale(alpha, z);
		} else if (A instanceof SparseColumnAccessMatrix)
			transMultAddI(
				alpha,
				(SparseColumnAccessMatrix) A,
				xdata,
				beta,
				ydata,
				zdata,
				0,
				A.numColumns());
		else if (A instanceof SparseColumnRowAccessMatrix)
			transMultAddI(
				alpha,
				(SparseColumnRowAccessMatrix) A,
				xdata,
				beta,
				ydata,
				zdata,
				0,
				A.numColumns());
		else if (A instanceof DenseColumnRowAccessMatrix)
			transMultAddI(
				alpha,
				(DenseColumnRowAccessMatrix) A,
				xdata,
				beta,
				ydata,
				zdata,
				0,
				A.numColumns());
		else if (A instanceof DenseColumnAccessMatrix)
			transMultAddI(
				alpha,
				(DenseColumnAccessMatrix) A,
				xdata,
				beta,
				ydata,
				zdata,
				0,
				A.numColumns());
		else
			throw new UnsupportedOperationException();

		return z;
	}

	protected void transMultAddI(
		double alpha,
		BlockAccessMatrix A,
		BlockAccessVector x,
		double beta,
		BlockAccessVector y,
		BlockAccessVector z,
		int start,
		int stop) {
		for (int i = start; i < stop; ++i)
			transMultAdd(
				alpha,
				A.getBlock(i),
				x.getBlock(i),
				beta,
				y.getBlock(i),
				z.getBlock(i));
	}

	protected void transMultAddI(
		double alpha,
		SparseColumnRowAccessMatrix A,
		double[] xdata,
		double beta,
		double[] ydata,
		double[] zdata,
		int start,
		int stop) {
		IntIntDoubleVectorTriple Amat = A.getMatrix();
		for (int i = start; i < stop; ++i) {
			double val = 0.;
			for (int j = Amat.minor[i]; j < Amat.minor[i + 1]; ++j)
				val += Amat.data[j] * xdata[Amat.major[j]];
			zdata[i] = alpha * val + beta * ydata[i];
		}
	}

	protected void transMultAddI(
		double alpha,
		SparseColumnAccessMatrix A,
		double[] xdata,
		double beta,
		double[] ydata,
		double[] zdata,
		int start,
		int stop) {
		for (int i = start; i < stop; ++i) {
			IntDoubleVectorPair Acol = A.getColumn(i);
			double val = 0.;
			for (int j = 0; j < Acol.index.length; ++j)
				val += Acol.data[j] * xdata[Acol.index[j]];
			zdata[i] = alpha * val + beta * ydata[i];
		}
	}

	protected void transMultAddI(
		double alpha,
		DenseColumnRowAccessMatrix A,
		double[] xdata,
		double beta,
		double[] ydata,
		double[] zdata,
		int start,
		int stop) {
		IntDoubleVectorPair Amat = A.getMatrix();
		int rows = A.numRows();
		for (int i = start; i < stop; ++i) {
			double val = 0.;

			int j = Amat.index[i], l = 0;
			for (int k = rows % 4; --k >= 0;)
				val += Amat.data[j++] * xdata[l++];
			for (int k = rows / 4; --k >= 0;)
				val += Amat.data[j++] * xdata[l++]
					+ Amat.data[j++] * xdata[l++]
					+ Amat.data[j++] * xdata[l++]
					+ Amat.data[j++] * xdata[l++];

			//for (int j = Amat.index[i], k = 0; j < Amat.index[i + 1]; ++j, ++k)
			//	val += Amat.data[j] * xdata[k];

			zdata[i] = alpha * val + beta * ydata[i];
		}
	}

	protected void transMultAddI(
		double alpha,
		DenseColumnAccessMatrix A,
		double[] xdata,
		double beta,
		double[] ydata,
		double[] zdata,
		int start,
		int stop) {
		int rows = A.numRows();
		for (int i = start; i < stop; ++i) {
			double[] Acol = A.getColumn(i);
			double val = 0.;

			int j = 0, l = 0;
			for (int k = rows % 4; --k >= 0;)
				val += Acol[j++] * xdata[l++];
			for (int k = rows / 4; --k >= 0;)
				val += Acol[j++] * xdata[l++]
					+ Acol[j++] * xdata[l++]
					+ Acol[j++] * xdata[l++]
					+ Acol[j++] * xdata[l++];

			//for (int j = 0; j < Arow.length; ++j)
			//	val += Arow[j] * xdata[j];

			zdata[i] = alpha * val + beta * ydata[i];
		}
	}

	protected void transMultAddI(
		SparseRowColumnAccessMatrix A,
		double[] xdata,
		double[] zdata,
		int start,
		int stop,
		Lock[] lock) {
		IntIntDoubleVectorTriple Amat = A.getMatrix();
		for (int i = start; i < stop; ++i) {
			for (int j = Amat.minor[i]; j < Amat.minor[i + 1]; ++j)
				synchronized (lock[Amat.major[j]]) {
					zdata[Amat.major[j]] += Amat.data[j] * xdata[i];
				}
		}
	}

	protected void transMultAddI(
		SparseRowAccessMatrix A,
		double[] xdata,
		double[] zdata,
		int start,
		int stop,
		Lock[] lock) {
		for (int i = start; i < stop; ++i) {
			IntDoubleVectorPair Arow = A.getRow(i);
			for (int j = 0; j < Arow.index.length; ++j)
				synchronized (lock[Arow.index[j]]) {
					zdata[Arow.index[j]] += Arow.data[j] * xdata[i];
				}
		}
	}

	protected void transMultAddI(
		DenseRowColumnAccessMatrix A,
		double[] xdata,
		double[] zdata,
		int start,
		int stop,
		Lock[] lock) {
		IntDoubleVectorPair Amat = A.getMatrix();
		for (int i = start; i < stop; ++i) {
			for (int j = Amat.index[i]; j < Amat.index[i + 1]; ++j)
				synchronized (lock[j - Amat.index[i]]) {
					zdata[j - Amat.index[i]] += Amat.data[j] * xdata[i];
				}
		}
	}

	protected void transMultAddI(
		DenseRowAccessMatrix A,
		double[] xdata,
		double[] zdata,
		int start,
		int stop,
		Lock[] lock) {
		for (int i = start; i < stop; ++i) {
			double[] Arow = A.getRow(i);
			for (int j = 0; j < Arow.length; ++j)
				synchronized (lock[j]) {
					zdata[j] += Arow[j] * xdata[i];
				}
		}
	}

	/**
	 * @see jmp.AbstractBLAS#transposeI(jmp.Matrix, jmp.Matrix)
	 */
	protected Matrix transposeI(Matrix A, Matrix B) {
		if (A instanceof SparseRowAccessMatrix
			&& B instanceof SparseColumnAccessMatrix)
			transposeI(
				(SparseRowAccessMatrix) A,
				(SparseColumnAccessMatrix) B,
				0,
				A.numRows());
		else if (
			A instanceof SparseRowColumnAccessMatrix
				&& B instanceof SparseColumnAccessMatrix)
			transposeI(
				(SparseRowColumnAccessMatrix) A,
				(SparseColumnAccessMatrix) B,
				0,
				A.numRows());
		else if (
			A instanceof SparseColumnAccessMatrix
				&& B instanceof SparseRowAccessMatrix)
			transposeI(
				(SparseColumnAccessMatrix) A,
				(SparseRowAccessMatrix) B,
				0,
				A.numColumns());
		else if (
			A instanceof SparseColumnRowAccessMatrix
				&& B instanceof SparseRowAccessMatrix)
			transposeI(
				(SparseColumnRowAccessMatrix) A,
				(SparseRowAccessMatrix) B,
				0,
				A.numColumns());
		else if (
			A instanceof DenseRowAccessMatrix
				&& B instanceof DenseColumnAccessMatrix)
			transposeI(
				(DenseRowAccessMatrix) A,
				(DenseColumnAccessMatrix) B,
				0,
				A.numRows());
		else if (
			A instanceof DenseRowColumnAccessMatrix
				&& B instanceof DenseColumnAccessMatrix)
			transposeI(
				(DenseRowColumnAccessMatrix) A,
				(DenseColumnAccessMatrix) B,
				0,
				A.numRows());
		else if (
			A instanceof DenseColumnAccessMatrix
				&& B instanceof DenseRowAccessMatrix)
			transposeI(
				(DenseColumnAccessMatrix) A,
				(DenseRowAccessMatrix) B,
				0,
				A.numColumns());
		else if (
			A instanceof DenseColumnRowAccessMatrix
				&& B instanceof DenseRowAccessMatrix)
			transposeI(
				(DenseColumnRowAccessMatrix) A,
				(DenseRowAccessMatrix) B,
				0,
				A.numColumns());
		else if (
			A instanceof ElementalAccessMatrix
				&& B instanceof ElementalAccessMatrix) {
			zero(B);
			transposeI(
				(ElementalAccessMatrix) A,
				(ElementalAccessMatrix) B,
				0,
				A.numRows());
		} else
			throw new UnsupportedOperationException();
		return B;
	}

	protected void transposeI(
		ElementalAccessMatrix A,
		ElementalAccessMatrix B,
		int start,
		int stop) {
		for (int i = start; i < stop; ++i)
			for (int j = 0; j < A.numColumns(); ++j) {
				double val = A.getValue(i, j);
				if (val != 0.)
					B.setValue(j, i, val);
			}
	}

	protected void transposeI(
		SparseRowAccessMatrix A,
		SparseColumnAccessMatrix B,
		int start,
		int stop) {
		for (int i = start; i < stop; ++i)
			B.setColumn(i, A.getRow(i).copy());
	}

	protected void transposeI(
		SparseRowColumnAccessMatrix A,
		SparseColumnAccessMatrix B,
		int start,
		int stop) {
		IntIntDoubleVectorTriple Amat = A.getMatrix();
		for (int i = start; i < stop; ++i) {
			B.setColumn(i, Amat.copy(i));
		}
	}

	protected void transposeI(
		SparseColumnAccessMatrix A,
		SparseRowAccessMatrix B,
		int start,
		int stop) {
		for (int i = start; i < stop; ++i)
			B.setRow(i, A.getColumn(i).copy());
	}

	protected void transposeI(
		SparseColumnRowAccessMatrix A,
		SparseRowAccessMatrix B,
		int start,
		int stop) {
		IntIntDoubleVectorTriple Amat = A.getMatrix();
		for (int i = start; i < stop; ++i) {
			B.setRow(i, Amat.copy(i));
		}
	}

	protected void transposeI(
		DenseRowAccessMatrix A,
		DenseColumnAccessMatrix B,
		int start,
		int stop) {
		for (int i = start; i < stop; ++i) {
			double[] Bcol = B.getColumn(i);
			System.arraycopy(A.getRow(i), 0, Bcol, 0, Bcol.length);
		}
	}

	protected void transposeI(
		DenseRowColumnAccessMatrix A,
		DenseColumnAccessMatrix B,
		int start,
		int stop) {
		IntDoubleVectorPair Amat = A.getMatrix();
		for (int i = start; i < stop; ++i)
			B.setColumn(i, Amat.copy(i));
	}

	protected void transposeI(
		DenseColumnAccessMatrix A,
		DenseRowAccessMatrix B,
		int start,
		int stop) {
		for (int i = start; i < stop; ++i) {
			double[] Brow = B.getRow(i);
			System.arraycopy(A.getColumn(i), 0, Brow, 0, Brow.length);
		}
	}

	protected void transposeI(
		DenseColumnRowAccessMatrix A,
		DenseRowAccessMatrix B,
		int start,
		int stop) {
		IntDoubleVectorPair Amat = A.getMatrix();
		for (int i = start; i < stop; ++i)
			B.setRow(i, Amat.copy(i));
	}

	/**
	 * @see jmp.BLAS#zero(jmp.Matrix)
	 */
	public Matrix zero(Matrix A) {
		if (A instanceof VectorAccess)
			zeroI(((VectorAccess) A).getData());
		else if (A instanceof MatrixAccess)
			zeroI(((MatrixAccess) A).getData());
		else if (A instanceof BlockAccessMatrix)
			zeroI(
				(BlockAccessMatrix) A,
				0,
				((BlockAccessMatrix) A).numBlocks());
		else
			throw new UnsupportedOperationException();
		return A;
	}

	protected void zeroI(BlockAccessMatrix A, int start, int stop) {
		for (int i = start; i < stop; ++i)
			zero(A.getBlock(i));
	}

	protected void zeroI(double[] data) {
		Arrays.fill(data, 0.);
	}

	protected void zeroI(double[][] data) {
		for (int i = 0; i < data.length; ++i)
			Arrays.fill(data[i], 0.);
	}

	/**
	 * @see jmp.BLAS#zero(jmp.Vector)
	 */
	public Vector zero(Vector x) {
		if (x instanceof VectorAccess)
			zeroI(((VectorAccess) x).getData());
		else if (x instanceof BlockAccessVector)
			zeroI(
				(BlockAccessVector) x,
				0,
				((BlockAccessVector) x).numBlocks());
		else
			throw new UnsupportedOperationException();

		return x;
	}

	protected void zeroI(BlockAccessVector x, int start, int stop) {
		for (int i = start; i < stop; ++i)
			zero(x.getBlock(i));
	}

	/**
	 * @see jmp.BLAS#cardinality(jmp.Matrix)
	 */
	public int cardinality(Matrix A) {
		if (A instanceof VectorAccess)
			return cardinalityI(((VectorAccess) A).getData());
		else if (A instanceof MatrixAccess)
			return cardinalityI(((MatrixAccess) A).getData());
		else
			return A.numRows() * A.numColumns();
	}

	protected int cardinalityI(double[] data) {
		return data.length;
	}

	protected int cardinalityI(double[][] data) {
		int nnz = 0;
		for (int i = 0; i < data.length; ++i)
			nnz += data[i].length;
		return nnz;
	}

	/**
	 * @see jmp.BLAS#cardinality(jmp.Vector)
	 */
	public int cardinality(Vector x) {
		if (x instanceof VectorAccess)
			return cardinalityI(((VectorAccess) x).getData());
		else
			return x.size();
	}

	/**
	 * @see jmp.BLAS#gather(double[])
	 */
	public IntDoubleVectorPair gather(double[] x) {
		// Number of non-zeros
		int nnz = 0;
		for (int i = 0; i < x.length; ++i)
			if (x[i] == 0.)
				nnz++;

		// Locations of the non-zeros
		int[] nz = new int[nnz];
		int j = 0;
		for (int i = 0; i < x.length; ++i)
			if (x[i] == 0.)
				nz[j++] = i;

		return gather(nz, x);
	}

	/**
	 * @see jmp.BLAS#gather(int[], double[])
	 */
	public IntDoubleVectorPair gather(int[] xIndex, double[] xData) {
		int[] index = new int[xIndex.length];
		System.arraycopy(xIndex, 0, index, 0, index.length);

		double[] data = new double[xIndex.length];
		for (int i = 0; i < xIndex.length; ++i)
			data[i] = xData[xIndex[i]];

		return new IntDoubleVectorPair(index, data);
	}

	/**
	 * @see jmp.BLAS#scatter(jmp.IntDoubleVectorPair, double[])
	 */
	public double[] scatter(IntDoubleVectorPair x, double[] y) {
		Arrays.fill(y, 0.);
		for (int i = 0; i < x.index.length; ++i)
			y[x.index[i]] = x.data[i];
		return y;
	}

	/**
	 * @see jmp.BLAS#getArrayCopy(jmp.Matrix)
	 */
	public double[][] getArrayCopy(Matrix A) {
		if (A instanceof ElementalAccessMatrix) {
			int[] row = new int[A.numRows()], column = new int[A.numColumns()];
			for (int i = 0; i < row.length; row[i] = i++);
			for (int i = 0; i < column.length; column[i] = i++);
			return ((ElementalAccessMatrix) A).getValues(row, column);
		} else
			throw new UnsupportedOperationException();
	}

	/**
	 * @see jmp.BLAS#getArrayCopy(jmp.Vector)
	 */
	public double[] getArrayCopy(Vector x) {
		double[] ret = new double[x.size()];
		if (x instanceof DenseAccessVector)
			System.arraycopy(
				((DenseAccessVector) x).getVector(),
				0,
				ret,
				0,
				ret.length);
		else if (x instanceof SparseAccessVector)
			ret = scatter(((SparseAccessVector) x).getVector(), ret);
		else if (x instanceof BlockAccessVector)
			System.arraycopy(
				((BlockAccessVector) x).getArray(),
				0,
				ret,
				0,
				ret.length);
		else
			throw new UnsupportedOperationException();
		return ret;
	}

	/**
	 * @see jmp.BLAS#setArray(double[], jmp.Vector)
	 */
	public Vector setArray(double[] x, Vector y) {
		if (y instanceof DenseAccessVector)
			 ((DenseAccessVector) y).setVector(x);
		else if (y instanceof SparseAccessVector)
			scatter(((SparseAccessVector) y).getVector(), x);
		else
			throw new UnsupportedOperationException();
		return y;
	}

	/**
	 * Populates the dense array of x with its data, and returns it
	 */
	protected double[] block2array(BlockAccessVector x) {
		double[] y = ((BlockAccessVector) x).getArray();
		for (int i = 0; i < x.numBlocks(); ++i) {
			int[] index = x.getBlockIndices(i);
			double[] data = ((DenseAccessVector) x.getBlock(i)).getVector();
			for (int j = 0; j < data.length; ++j)
				y[index[j]] += data[j];
		}
		return y;
	}

}
