/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp;

import jmp.util.Arrays;

/**
 * Sparse matrix stored as 2D ragged columns. For best performance during
 * assembly, ensure that enough memory is allocated at construction time,
 * as re-allocation may be time-consuming.
 */
public class SparseColumnMatrix
	extends AbstractMatrix
	implements
		SparseColumnAccessMatrix,
		ElementalAccessZeroColumnMatrix,
		MatrixAccess,
		Matrix {

	/**
	 * Matrix data
	 */
	private double[][] data;

	/**
	 * Row indices. These are kept sorted within each column.
	 */
	private int[][] rowIndex;

	/**
	 * Number of indices in use on each column.
	 */
	private int[] used;

	/**
	 * For thread safe operation. One lock for each column
	 */
	private Lock[] lock;

	/**
	 * Is the matrix compacted?
	 */
	private boolean compact;

	/**
	 * Constructor for SparseColumnMatrix.
	 * @param nz Initial number of non-zeros on each column
	 */
	public SparseColumnMatrix(int numRows, int numColumns, int[] nz) {
		super(numRows, numColumns);
		data = new double[numColumns][];
		rowIndex = new int[numColumns][];
		lock = Lock.createLocks(numColumns);
		for (int i = 0; i < numColumns; ++i) {
			data[i] = new double[nz[i]];
			rowIndex[i] = new int[nz[i]];
		}
		used = new int[numColumns];
	}

	/**
	 * Constructor for SparseColumnMatrix.
	 * @param nz Initial number of non-zeros on each column
	 */
	public SparseColumnMatrix(int numRows, int numColumns, int nz) {
		super(numRows, numColumns);
		data = new double[numColumns][];
		rowIndex = new int[numColumns][];
		lock = Lock.createLocks(numColumns);
		for (int i = 0; i < numColumns; ++i) {
			data[i] = new double[nz];
			rowIndex[i] = new int[nz];
		}
		used = new int[numColumns];
	}

	/**
	 * Constructor for SparseColumnMatrix. Zero initial pre-allocation
	 */
	public SparseColumnMatrix(int numRows, int numColumns) {
		this(numRows, numColumns, 0);
	}

	/**
	 * Constructor for SparseColumnMatrix, and copies the contents from the
	 * supplied matrix.
	 * @param nz Initial number of non-zeros on each column
	 */
	public SparseColumnMatrix(Matrix A, int[] nz) {
		this(A.numRows(), A.numColumns(), nz);
		State.BLAS.copy(A, this);
	}

	/**
	 * Constructor for SparseColumnMatrix, and copies the contents from the
	 * supplied matrix.
	 * @param nz Initial number of non-zeros on each column
	 */
	public SparseColumnMatrix(Matrix A, int nz) {
		this(A.numRows(), A.numColumns(), nz);
		State.BLAS.copy(A, this);
	}

	/**
	 * Constructor for SparseColumnMatrix, and copies the contents from the
	 * supplied matrix.
	 */
	public SparseColumnMatrix(Matrix A) {
		this(A, 0);
	}

	/**
	 * @see jmp.SparseColumnAccessMatrix#getColumn(int)
	 */
	public IntDoubleVectorPair getColumn(int i) {
		synchronized (lock[i]) {
			compact(i);
		}
		return new IntDoubleVectorPair(rowIndex[i], data[i]);
	}

	/**
	 * @see jmp.SparseColumnAccessMatrix#setColumn(int, jmp.IntDoubleVectorPair)
	 */
	public void setColumn(int i, IntDoubleVectorPair x) {
		synchronized (lock[i]) {
			rowIndex[i] = x.index;
			data[i] = x.data;
			used[i] = x.data.length;
			compact = false;
		}
	}

	/**
	 * @see jmp.ZeroColumnMatrix#zeroColumns(int[], double)
	 */
	public void zeroColumns(int[] column, double diagonal) {
		for (int i = 0; i < column.length; ++i)
			synchronized (lock[column[i]]) {
				int col = column[i];

				if (col < numRows) {
					data[col][0] = diagonal;
					rowIndex[col][0] = col;
					used[col] = 1;
				} else
					used[col] = 0;
			}

		compact = false;
	}

	/**
	 * @see jmp.ElementalAccessMatrix#addValue(int, int, double)
	 */
	public void addValue(int row, int column, double value) {
		synchronized (lock[column]) {
			int index = getRowIndex(row, column);
			data[column][index] += value;
		}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#setValue(int, int, double)
	 */
	public void setValue(int row, int column, double value) {
		synchronized (lock[column]) {
			int index = getRowIndex(row, column);
			data[column][index] = value;
		}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#getValue(int, int)
	 */
	public double getValue(int row, int column) {
		int ind = Arrays.binarySearch(rowIndex[column], row, 0, used[column]);
		if (ind != -1)
			return data[column][ind];
		else if (row < numRows && row >= 0)
			return 0.;
		else
			throw new ArrayIndexOutOfBoundsException(
				"Row " + row + " Column " + column);
	}

	/**
	 * @see jmp.ElementalAccessMatrix#addValues(int[], int[], double[][])
	 */
	public void addValues(int[] row, int[] column, double[][] values) {
		for (int i = 0; i < column.length; ++i)
			synchronized (lock[column[i]]) {
				for (int j = 0; j < row.length; ++j) {
					int index = getRowIndex(row[j], column[i]);
					data[column[i]][index] += values[j][i];
				}
			}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#setValues(int[], int[], double[][])
	 */
	public void setValues(int[] row, int[] column, double[][] values) {
		for (int i = 0; i < column.length; ++i)
			synchronized (lock[column[i]]) {
				for (int j = 0; j < row.length; ++j) {
					int index = getRowIndex(row[j], column[i]);
					data[column[i]][index] = values[j][i];
				}
			}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#getValues(int[], int[])
	 */
	public double[][] getValues(int[] row, int[] column) {
		double[][] sub = new double[row.length][column.length];
		for (int i = 0; i < row.length; ++i)
			for (int j = 0; j < column.length; ++j)
				sub[i][j] = getValue(row[i], column[j]);
		return sub;
	}

	/**
	 * @see jmp.MatrixAccess#getData()
	 */
	public double[][] getData() {
		compact();
		return data;
	}

	/**
	 * @see jmp.Sparse#compact()
	 */
	public synchronized void compact() {
		if (compact)
			return;
		else {
			for (int i = 0; i < numColumns; ++i)
				compact(i);
			compact = true;
		}
	}

	/**
	 * Compacts the column-indices and entries.
	 */
	private void compact(int column) {
		if (used[column] < data[column].length) {
			double[] newData = new double[used[column]];
			System.arraycopy(data[column], 0, newData, 0, used[column]);
			int[] newInd = new int[used[column]];
			System.arraycopy(rowIndex[column], 0, newInd, 0, used[column]);
			data[column] = newData;
			rowIndex[column] = newInd;
		}
	}

	/**
	 * Tries to find the row-index in the given column. If it is not found,
	 * a reallocation is done, and a new index is returned.
	 */
	private int getRowIndex(int row, int col) {

		int[] curRow = rowIndex[col];
		double[] curDat = data[col];

		// Try to find column index
		int ind =
			jmp.util.Arrays.binarySearchGreater(curRow, row, 0, used[col]);

		// Found
		if (ind < used[col] && curRow[ind] == row)
			return ind;

		// Not found, try to make room
		if (row < 0 || row >= numRows)
			throw new ArrayIndexOutOfBoundsException(
				" Row " + row + " Column " + col);
		used[col]++;

		// Check available memory
		if (used[col] > curDat.length) {

			// If zero-length, use new length of 1, else double the bandwidth
			int newLength = 1;
			if (curDat.length != 0)
				newLength = 2 * curDat.length;

			// Copy existing data into new arrays
			int[] newRow = new int[newLength];
			double[] newDat = new double[newLength];
			System.arraycopy(curRow, 0, newRow, 0, curDat.length);
			System.arraycopy(curDat, 0, newDat, 0, curDat.length);

			// Update pointers
			rowIndex[col] = newRow;
			data[col] = newDat;
			curRow = newRow;
			curDat = newDat;
		}

		// All ok, make room for insertion
		for (int i = used[col] - 1; i >= ind + 1; --i) {
			curRow[i] = curRow[i - 1];
			curDat[i] = curDat[i - 1];
		}

		// Put in new structure
		curRow[ind] = row;
		curDat[ind] = 0.;
		compact = false;

		return ind;
	}

}
