/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp;

import jmp.util.Arrays;

/**
 * Sparse matrix stored as a vector, column major
 */
public class SparseColumnRowMatrix
	extends AbstractMatrix
	implements
		SparseColumnRowAccessMatrix,
		ElementalAccessZeroColumnMatrix,
		VectorAccess,
		Matrix {

	/**
	 * Matrix data
	 */
	private double[] data;

	/**
	 * Row indices. These are kept sorted within each column.
	 */
	private int[] rowIndex;

	/**
	 * Indices to the start of each column
	 */
	private int[] columnIndex;

	/**
	 * Number of indices in use on each column.
	 */
	private int[] used;

	/**
	 * For thread safe operation. One lock for each column
	 */
	private Lock[] lock;

	/**
	 * Is the matrix compacted?
	 */
	private boolean compact;

	/**
	 * Constructor for SparseColumnRowMatrix.
	 * @param nz Maximum number of non-zeros on each row
	 */
	public SparseColumnRowMatrix(int numRows, int numColumns, int[] nz) {
		super(numRows, numColumns);
		int nnz = 0;
		for (int i = 0; i < nz.length; ++i)
			nnz += nz[i];

		data = new double[nnz];
		rowIndex = new int[nnz];
		columnIndex = new int[numColumns + 1];
		used = new int[numColumns];
		lock = Lock.createLocks(numColumns);
		for (int i = 1; i < numColumns + 1; ++i)
			columnIndex[i] = nz[i - 1] + columnIndex[i - 1];
	}

	/**
	 * Constructor for SparseColumnRowMatrix.
	 * @param nz Maximum number of non-zeros on each row
	 */
	public SparseColumnRowMatrix(int numRows, int numColumns, int nz) {
		super(numRows, numColumns);
		int nnz = nz * numColumns;

		data = new double[nnz];
		rowIndex = new int[nnz];
		columnIndex = new int[numColumns + 1];
		used = new int[numColumns];
		lock = Lock.createLocks(numColumns);
		for (int i = 1; i < numColumns + 1; ++i)
			columnIndex[i] = nz + columnIndex[i - 1];
	}

	/**
	 * Constructor for SparseColumnRowMatrix, and copies the contents from the
	 * supplied matrix.
	 * @param nz Maximum number of non-zeros on each row
	 */
	public SparseColumnRowMatrix(Matrix A, int nz) {
		this(A.numRows(), A.numColumns(), nz);
		State.BLAS.copy(A, this);
	}

	/**
	 * Constructor for SparseColumnRowMatrix, and copies the contents from the
	 * supplied matrix.
	 * @param nz Maximum number of non-zeros on each row
	 */
	public SparseColumnRowMatrix(Matrix A, int[] nz) {
		this(A.numRows(), A.numColumns(), nz);
		State.BLAS.copy(A, this);
	}

	/**
	 * @see jmp.SparseRowColumnAccessMatrix#getMatrix()
	 */
	public IntIntDoubleVectorTriple getMatrix() {
		compact();
		return new IntIntDoubleVectorTriple(rowIndex, columnIndex, data);
	}

	/**
	 * @see jmp.SparseRowColumnAccessMatrix#setMatrix(jmp.IntIntDoubleVectorTriple)
	 */
	public synchronized void setMatrix(IntIntDoubleVectorTriple x) {
		data = x.data;
		rowIndex = x.major;
		columnIndex = x.minor;
	}

	/**
	 * @see jmp.ZeroColumnMatrix#zeroColumns(int[], double)
	 */
	public void zeroColumns(int[] column, double diagonal) {
		for (int i = 0; i < column.length; ++i)
			synchronized (lock[column[i]]) {
				int col = column[i];

				if (col < numRows) {
					data[columnIndex[col]] = diagonal;
					rowIndex[columnIndex[col]] = col;
					used[col] = 1;
				} else
					used[col] = 0;
			}

		compact = false;
	}

	/**
	 * @see jmp.ElementalAccessMatrix#addValue(int, int, double)
	 */
	public void addValue(int row, int column, double value) {
		synchronized (lock[column]) {
			int index = getRowIndex(row, column);
			data[columnIndex[column] + index] += value;
		}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#setValue(int, int, double)
	 */
	public void setValue(int row, int column, double value) {
		synchronized (lock[column]) {
			int index = getRowIndex(row, column);
			data[columnIndex[column] + index] = value;
		}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#getValue(int, int)
	 */
	public double getValue(int row, int column) {
		int ind =
			Arrays.binarySearch(
				rowIndex,
				row,
				columnIndex[column],
				columnIndex[column] + used[column]);
		if (ind >= 0)
			return data[ind];
		else if (row < numRows && row >= 0)
			return 0;
		else
			throw new ArrayIndexOutOfBoundsException(
				"Row " + row + " Column " + column);
	}

	/**
	 * @see jmp.ElementalAccessMatrix#addValues(int[], int[], double[][])
	 */
	public void addValues(int[] row, int[] column, double[][] values) {
		for (int i = 0; i < column.length; ++i)
			synchronized (lock[column[i]]) {
				for (int j = 0; j < row.length; ++j) {
					int index = getRowIndex(row[j], column[i]);
					data[columnIndex[column[i]] + index] += values[j][i];
				}
			}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#setValues(int[], int[], double[][])
	 */
	public void setValues(int[] row, int[] column, double[][] values) {
		for (int i = 0; i < column.length; ++i)
			synchronized (lock[column[i]]) {
				for (int j = 0; j < row.length; ++j) {
					int index = getRowIndex(row[j], column[i]);
					data[columnIndex[column[i]] + index] = values[j][i];
				}
			}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#getValues(int[], int[])
	 */
	public double[][] getValues(int[] row, int[] column) {
		double[][] sub = new double[row.length][column.length];
		for (int i = 0; i < row.length; ++i)
			for (int j = 0; j < column.length; ++j)
				sub[i][j] = getValue(row[i], column[j]);
		return sub;
	}

	/**
	 * @see jmp.VectorAccess#getData()
	 */
	public double[] getData() {
		compact();
		return data;
	}

	/**
	 * @see jmp.Sparse#compact()
	 */
	public synchronized void compact() {
		if (compact)
			return;

		int nnz = 0;
		for (int i = 0; i < numColumns; ++i)
			nnz += used[i];

		if (nnz < data.length) {
			double[] newData = new double[nnz];
			int[] newRowIndex = new int[nnz];
			int[] newColumnIndex = new int[numColumns + 1];

			newColumnIndex[0] = columnIndex[0];
			for (int i = 0; i < numColumns; ++i) {
				newColumnIndex[i + 1] = newColumnIndex[i] + used[i];
				System.arraycopy(
					data,
					columnIndex[i],
					newData,
					newColumnIndex[i],
					used[i]);
				System.arraycopy(
					rowIndex,
					columnIndex[i],
					newRowIndex,
					newColumnIndex[i],
					used[i]);
			}

			data = newData;
			rowIndex = newRowIndex;
			columnIndex = newColumnIndex;
		}

		compact = true;
	}

	/**
	 * Tries to find the col-index in the given row. If it is not found,
	 * a reallocation is done, and a new index is returned. If there is no more
	 * space for further allocation, an exception is thrown.
	 */
	private int getRowIndex(int row, int col) {
		int columnOffset = columnIndex[col];

		int ind =
			Arrays.binarySearchGreater(
				rowIndex,
				row,
				columnOffset,
				columnOffset + used[col])
				- columnOffset;

		// Found
		if (ind < used[col] && rowIndex[columnOffset + ind] == row)
			return ind;

		// Need to insert

		// Check row index
		if (row < 0 || row >= numRows)
			throw new ArrayIndexOutOfBoundsException(
				"Row " + row + " Column " + col);

		// Is the column full?
		if (columnOffset + used[col] >= data.length
			|| ((col + 1) < columnIndex.length
				&& columnOffset + used[col] >= columnIndex[col + 1]))
			throw new ArrayIndexOutOfBoundsException(
				"Too many non-zeros on column " + col);

		// Make room for insertion
		for (int i = used[col]; i >= ind + 1; --i) {
			rowIndex[columnOffset + i] = rowIndex[columnOffset + i - 1];
			data[columnOffset + i] = data[columnOffset + i - 1];
		}

		// Put in new structure
		used[col]++;
		rowIndex[columnOffset + ind] = row;
		data[columnOffset + ind] = 0;
		compact = false;

		return ind;
	}

}
