/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp;

import jmp.util.Arrays;

/**
 * Sparse matrix stored as one long vector, row major.
 */
public class SparseRowColumnMatrix
	extends AbstractMatrix
	implements
		ElementalAccessZeroRowMatrix,
		SparseRowColumnAccessMatrix,
		VectorAccess,
		Matrix {

	/**
	 * Matrix data
	 */
	private double[] data;

	/**
	 * Column indices. These are kept sorted within each row.
	 */
	private int[] columnIndex;

	/**
	 * Indices to the start of each row
	 */
	private int[] rowIndex;

	/**
	 * Number of indices in use on each row.
	 */
	private int[] used;

	/**
	 * For thread safe operation. One lock for each row
	 */
	private Lock[] lock;

	/**
	 * Is the matrix compacted?
	 */
	private boolean compact;

	/**
	 * Constructor for SparseRowColumnMatrix, and copies the contents from the
	 * supplied matrix.
	 * @param nz Maximum number of non-zeros on each row
	 */
	public SparseRowColumnMatrix(Matrix A, int[] nz) {
		this(A.numRows(), A.numColumns(), nz);
		State.BLAS.copy(A, this);
	}

	/**
	 * Constructor for SparseRowColumnMatrix, and copies the contents from the
	 * supplied matrix.
	 * @param nz Maximum number of non-zeros on each row
	 */
	public SparseRowColumnMatrix(Matrix A, int nz) {
		this(A.numRows(), A.numColumns(), nz);
		State.BLAS.copy(A, this);
	}

	/**
	 * Constructor for SparseRowColumnMatrix.
	 * @param nz Maximum number of non-zeros on each row
	 */
	public SparseRowColumnMatrix(int numRows, int numColumns, int[] nz) {
		super(numRows, numColumns);

		int nnz = 0;
		for (int i = 0; i < nz.length; ++i)
			nnz += nz[i];

		rowIndex = new int[numRows + 1];
		columnIndex = new int[nnz];
		data = new double[nnz];
		used = new int[numRows];
		lock = Lock.createLocks(numRows);
		for (int i = 1; i < numRows + 1; ++i)
			rowIndex[i] = nz[i - 1] + rowIndex[i - 1];
	}

	/**
	 * Constructor for SparseRowColumnMatrix.
	 * @param nz Maximum number of non-zeros on each row
	 */
	public SparseRowColumnMatrix(int numRows, int numColumns, int nz) {
		super(numRows, numColumns);

		int nnz = nz * numRows;

		rowIndex = new int[numRows + 1];
		columnIndex = new int[nnz];
		data = new double[nnz];
		used = new int[numRows];
		lock = Lock.createLocks(numRows);
		for (int i = 1; i < numRows + 1; ++i)
			rowIndex[i] = nz + rowIndex[i - 1];
	}

	/**
	 * @see jmp.ElementalAccessMatrix#addValue(int, int, double)
	 */
	public void addValue(int row, int column, double value) {
		synchronized (lock[row]) {
			int index = getColumnIndex(column, row);
			data[rowIndex[row] + index] += value;
		}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#setValue(int, int, double)
	 */
	public void setValue(int row, int column, double value) {
		synchronized (lock[row]) {
			int index = getColumnIndex(column, row);
			data[rowIndex[row] + index] = value;
		}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#getValue(int, int)
	 */
	public double getValue(int row, int column) {
		int ind =
			Arrays.binarySearch(
				columnIndex,
				column,
				rowIndex[row],
				rowIndex[row] + used[row]);
		if (ind >= 0)
			return data[ind];
		else if (column < numColumns && column >= 0)
			return 0;
		else
			throw new ArrayIndexOutOfBoundsException(
				"Row " + row + " Column " + column);
	}

	/**
	 * @see jmp.ElementalAccessMatrix#addValues(int[], int[], double[][])
	 */
	public void addValues(int[] row, int[] column, double[][] values) {
		for (int i = 0; i < row.length; ++i)
			synchronized (lock[row[i]]) {
				for (int j = 0; j < column.length; ++j) {
					int index = getColumnIndex(column[j], row[i]);
					data[rowIndex[row[i]] + index] += values[i][j];
				}
			}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#setValues(int[], int[], double[][])
	 */
	public void setValues(int[] row, int[] column, double[][] values) {
		for (int i = 0; i < row.length; ++i)
			synchronized (lock[row[i]]) {
				for (int j = 0; j < column.length; ++j) {
					int index = getColumnIndex(column[j], row[i]);
					data[rowIndex[row[i]] + index] = values[i][j];
				}
			}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#getValues(int[], int[])
	 */
	public double[][] getValues(int[] row, int[] column) {
		double[][] sub = new double[row.length][column.length];
		for (int i = 0; i < row.length; ++i)
			for (int j = 0; j < column.length; ++j)
				sub[i][j] = getValue(row[i], column[j]);
		return sub;
	}

	/**
	 * @see jmp.SparseRowColumnAccessMatrix#getMatrix()
	 */
	public IntIntDoubleVectorTriple getMatrix() {
		compact();
		return new IntIntDoubleVectorTriple(columnIndex, rowIndex, data);
	}

	/**
	 * @see jmp.Sparse#compact()
	 */
	public synchronized void compact() {
		if (compact)
			return;

		int nnz = 0;
		for (int i = 0; i < numRows; ++i)
			nnz += used[i];

		if (nnz < data.length) {
			int[] newRowIndex = new int[numRows + 1];
			int[] newColumnIndex = new int[nnz];
			double[] newData = new double[nnz];

			newRowIndex[0] = rowIndex[0];
			for (int i = 0; i < numRows; ++i) {
				newRowIndex[i + 1] = newRowIndex[i] + used[i];
				System.arraycopy(
					data,
					rowIndex[i],
					newData,
					newRowIndex[i],
					used[i]);
				System.arraycopy(
					columnIndex,
					rowIndex[i],
					newColumnIndex,
					newRowIndex[i],
					used[i]);
			}

			rowIndex = newRowIndex;
			columnIndex = newColumnIndex;
			data = newData;
		}

		compact = true;
	}

	/**
	 * @see jmp.SparseRowColumnAccessMatrix#setMatrix(jmp.IntIntDoubleVectorTriple)
	 */
	public synchronized void setMatrix(IntIntDoubleVectorTriple x) {
		data = x.data;
		columnIndex = x.major;
		rowIndex = x.minor;
	}

	/**
	 * @see jmp.ZeroRowMatrix#zeroRows(int[], double)
	 */
	public void zeroRows(int[] row, double diagonal) {
		for (int i = 0; i < row.length; ++i)
			synchronized (lock[row[i]]) {
				int rowI = row[i];

				if (rowI < numColumns) {
					data[rowIndex[rowI]] = diagonal;
					columnIndex[rowIndex[rowI]] = rowI;
					used[rowI] = 1;
				} else
					used[rowI] = 0;
			}

		compact = false;
	}

	/**
	 * Tries to find the col-index in the given row. If it is not found,
	 * a reallocation is done, and a new index is returned. If there is no more
	 * space for further allocation, an exception is thrown.
	 */
	private int getColumnIndex(int col, int row) {
		int rowOffset = rowIndex[row];

		int ind =
			Arrays.binarySearchGreater(
				columnIndex,
				col,
				rowOffset,
				rowOffset + used[row])
				- rowOffset;

		// Found
		if (ind < used[row] && columnIndex[rowOffset + ind] == col)
			return ind;

		// Need to insert

		// Check column index
		if (col < 0 || col >= numColumns)
			throw new ArrayIndexOutOfBoundsException(
				"Row " + row + " Column " + col);

		// Is the row full?
		if (rowOffset + used[row] >= data.length
			|| ((row + 1) < rowIndex.length
				&& rowOffset + used[row] >= rowIndex[row + 1]))
			throw new ArrayIndexOutOfBoundsException(
				"Too many non-zeros on row " + row);

		// Make room for insertion
		for (int i = used[row]; i >= ind + 1; --i) {
			columnIndex[rowOffset + i] = columnIndex[rowOffset + i - 1];
			data[rowOffset + i] = data[rowOffset + i - 1];
		}

		// Put in new structure
		used[row]++;
		columnIndex[rowOffset + ind] = col;
		data[rowOffset + ind] = 0;
		compact = false;

		return ind;
	}

	/**
	 * @see jmp.VectorAccess#getData()
	 */
	public double[] getData() {
		compact();
		return data;
	}

}
