/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp;

import jmp.util.Arrays;

/**
 * Sparse matrix stored as 2D ragged rows. For best performance during
 * assembly, ensure that enough memory is allocated at construction time,
 * as re-allocation may be time-consuming.
 */
public class SparseRowMatrix
	extends AbstractMatrix
	implements SparseRowAccessMatrix, ElementalAccessZeroRowMatrix, MatrixAccess, Matrix {

	/**
	 * Matrix data
	 */
	private double[][] data;

	/**
	 * Column indices. These are kept sorted within each row.
	 */
	private int[][] columnIndex;

	/**
	 * Number of indices in use on each row.
	 */
	private int[] used;

	/**
	 * For thread safe operation. One lock for each row
	 */
	private Lock[] lock;

	/**
	 * Is the matrix compacted?
	 */
	private boolean compact;

	/**
	 * Constructor for SparseRowMatrix, and copies the contents from the
	 * supplied matrix.
	 * @param nz Initial number of non-zeros on each row
	 */
	public SparseRowMatrix(Matrix A, int[] nz) {
		this(A.numRows(), A.numColumns(), nz);
		State.BLAS.copy(A, this);
	}

	/**
	 * Constructor for SparseRowMatrix, and copies the contents from the
	 * supplied matrix.
	 * @param nz Initial number of non-zeros on each row
	 */
	public SparseRowMatrix(Matrix A, int nz) {
		this(A.numRows(), A.numColumns(), nz);
		State.BLAS.copy(A, this);
	}

	/**
	 * Constructor for SparseRowMatrix, and copies the contents from the
	 * supplied matrix. Zero initial pre-allocation
	 */
	public SparseRowMatrix(Matrix A) {
		this(A, 0);
	}

	/**
	 * Constructor for SparseRowMatrix.
	 * @param nz Initial number of non-zeros on each row
	 */
	public SparseRowMatrix(int numRows, int numColumns, int[] nz) {
		super(numRows, numColumns);
		data = new double[numRows][];
		columnIndex = new int[numRows][];
		lock = Lock.createLocks(numRows);
		used = new int[numRows];
		for (int i = 0; i < numRows; ++i) {
			data[i] = new double[nz[i]];
			columnIndex[i] = new int[nz[i]];
		}
	}

	/**
	 * Constructor for SparseRowMatrix.
	 * @param nz Initial number of non-zeros on each row
	 */
	public SparseRowMatrix(int numRows, int numColumns, int nz) {
		super(numRows, numColumns);
		data = new double[numRows][nz];
		columnIndex = new int[numRows][nz];
		lock = Lock.createLocks(numRows);
		used = new int[numRows];
	}

	/**
	 * Constructor for SparseRowMatrix. Zero initial pre-allocation
	 */
	public SparseRowMatrix(int numRows, int numColumns) {
		this(numRows, numColumns, 0);
	}

	/**
	 * @see jmp.ZeroRowMatrix#zeroRows(int[], double)
	 */
	public void zeroRows(int[] row, double diagonal) {
		for (int i = 0; i < row.length; ++i)
			synchronized (lock[row[i]]) {
				int rowI = row[i];

				if (rowI < numColumns) {
					data[rowI][0] = diagonal;
					columnIndex[rowI][0] = rowI;
					used[rowI] = 1;
				} else
					used[rowI] = 0;
			}
		compact = false;
	}

	/**
	 * @see jmp.ElementalAccessMatrix#addValue(int, int, double)
	 */
	public void addValue(int i, int j, double value) {
		synchronized (lock[i]) {
			int index = getColumnIndex(j, i);
			data[i][index] += value;
		}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#setValue(int, int, double)
	 */
	public void setValue(int i, int j, double value) {
		synchronized (lock[i]) {
			int index = getColumnIndex(j, i);
			data[i][index] = value;
		}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#getValue(int, int)
	 */
	public double getValue(int row, int column) {
		int ind = Arrays.binarySearch(columnIndex[row], column, 0, used[row]);
		if (ind != -1)
			return data[row][ind];
		else if (column < numColumns && column >= 0)
			return 0.;
		else
			throw new ArrayIndexOutOfBoundsException(
				"Row " + row + " Column " + column);
	}

	/**
	 * @see jmp.ElementalAccessMatrix#addValues(int[], int[], double[][])
	 */
	public void addValues(int[] row, int[] column, double[][] values) {
		for (int i = 0; i < row.length; ++i)
			synchronized (lock[row[i]]) {
				for (int j = 0; j < column.length; ++j) {
					int index = getColumnIndex(column[j], row[i]);
					data[row[i]][index] += values[i][j];
				}
			}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#setValues(int[], int[], double[][])
	 */
	public void setValues(int[] row, int[] column, double[][] values) {
		for (int i = 0; i < row.length; ++i)
			synchronized (lock[row[i]]) {
				for (int j = 0; j < column.length; ++j) {
					int index = getColumnIndex(column[j], row[i]);
					data[row[i]][index] = values[i][j];
				}
			}
	}

	/**
	 * @see jmp.ElementalAccessMatrix#getValues(int[], int[])
	 */
	public double[][] getValues(int[] row, int[] column) {
		double[][] sub = new double[row.length][column.length];
		for (int i = 0; i < row.length; ++i)
			for (int j = 0; j < column.length; ++j)
				sub[i][j] = getValue(row[i], column[j]);
		return sub;
	}

	/**
	 * Tries to find the col-index in the given row. If it is not found,
	 * a reallocation is done, and a new index is returned.
	 */
	private int getColumnIndex(int col, int row) {

		int[] curCol = columnIndex[row];
		double[] curDat = data[row];

		// Try to find column index
		int ind =
			jmp.util.Arrays.binarySearchGreater(curCol, col, 0, used[row]);

		// Found
		if (ind < used[row] && curCol[ind] == col)
			return ind;

		// Not found, try to make room
		if (col < 0 || col >= numColumns)
			throw new ArrayIndexOutOfBoundsException(
				"Row " + row + " Column " + col);
		used[row]++;

		// Check available memory
		if (used[row] > curDat.length) {

			// If zero-length, use new length of 1, else double the bandwidth
			int newLength = 1;
			if (curDat.length != 0)
				newLength = 2 * curDat.length;

			// Copy existing data into new arrays
			int[] newCol = new int[newLength];
			double[] newDat = new double[newLength];
			System.arraycopy(curCol, 0, newCol, 0, curDat.length);
			System.arraycopy(curDat, 0, newDat, 0, curDat.length);

			// Update pointers
			columnIndex[row] = newCol;
			data[row] = newDat;
			curCol = newCol;
			curDat = newDat;
		}

		// All ok, make room for insertion
		for (int i = used[row] - 1; i >= ind + 1; --i) {
			curCol[i] = curCol[i - 1];
			curDat[i] = curDat[i - 1];
		}

		// Put in new structure
		curCol[ind] = col;
		curDat[ind] = 0.;
		return ind;
	}

	/**
	 * @see jmp.SparseRowAccessMatrix#setRow(int, jmp.IntDoubleVectorPair)
	 */
	public void setRow(int i, IntDoubleVectorPair x) {
		synchronized (lock[i]) {
			columnIndex[i] = x.index;
			data[i] = x.data;
			used[i] = x.data.length;
			compact = false;
		}
	}

	/**
	 * @see jmp.SparseRowAccessMatrix#getRow(int)
	 */
	public IntDoubleVectorPair getRow(int i) {
		synchronized (lock[i]) {
			compact(i);
		}
		return new IntDoubleVectorPair(columnIndex[i], data[i]);
	}

	/**
	 * Compacts the row-indices and entries.
	 */
	private void compact(int row) {
		if (used[row] < data[row].length) {
			double[] newDat = new double[used[row]];
			System.arraycopy(data[row], 0, newDat, 0, used[row]);
			int[] newInd = new int[used[row]];
			System.arraycopy(columnIndex[row], 0, newInd, 0, used[row]);
			data[row] = newDat;
			columnIndex[row] = newInd;
		}
	}

	/**
	 * @see jmp.MatrixAccess#getData()
	 */
	public double[][] getData() {
		compact();
		return data;
	}

	/**
	 * @see jmp.Sparse#compact()
	 */
	public synchronized void compact() {
		if (compact)
			return;
		else {
			for (int i = 0; i < numRows; ++i)
				compact(i);
			compact = true;
		}
	}

}
