/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp;

import jmp.util.Arrays;

/**
 * Sparse vector.
 */
public class SparseVector
	extends AbstractVector
	implements ElementalAccessVector, SparseAccessVector, VectorAccess, Vector {

	/**
	 * Data
	 */
	private double[] data;

	/**
	 * Indices to data
	 */
	private int[] ind;

	/**
	 * How much has been used
	 */
	private int used;

	/**
	 * Constructor for SparseVector.
	 * @param nz Initial number of non-zeros
	 */
	public SparseVector(int size, int nz) {
		super(size);
		data = new double[nz];
		ind = new int[nz];
	}

	/**
	 * Constructor for SparseVector, and copies the contents from the
	 * supplied vector.
	 * @param nz Initial number of non-zeros
	 */
	public SparseVector(Vector x, int nz) {
		this(x.size(), nz);
		State.BLAS.copy(x, this);
	}

	/**
	 * Constructor for SparseVector, and copies the contents from the
	 * supplied vector. Zero initial pre-allocation
	 */
	public SparseVector(Vector x) {
		this(x.size());
		State.BLAS.copy(x, this);
	}

	/**
	 * Constructor for SparseVector. Zero initial pre-allocation
	 */
	public SparseVector(int size) {
		this(size, 0);
	}

	/**
	 * @see jmp.ElementalAccessVector#addValue(int, double)
	 */
	public synchronized void addValue(int index, double value) {
		int i = getIndex(index);
		data[i] += value;
	}

	/**
	 * @see jmp.ElementalAccessVector#setValue(int, double)
	 */
	public synchronized void setValue(int index, double value) {
		int i = getIndex(index);
		data[i] = value;
	}

	/**
	 * @see jmp.ElementalAccessVector#getValue(int)
	 */
	public double getValue(int index) {
		int in = jmp.util.Arrays.binarySearch(ind, index, 0, used);
		if (in != -1)
			return data[in];
		else if (index < size && index >= 0)
			return 0.;
		else
			throw new ArrayIndexOutOfBoundsException("Index " + index);
	}

	/**
	 * @see jmp.ElementalAccessVector#addValues(int[], double[])
	 */
	public synchronized void addValues(int[] index, double[] values) {
		for (int i = 0; i < index.length; ++i)
			addValue(index[i], values[i]);
	}

	/**
	 * @see jmp.ElementalAccessVector#setValues(int[], double[])
	 */
	public synchronized void setValues(int[] index, double[] values) {
		for (int i = 0; i < index.length; ++i)
			setValue(index[i], values[i]);
	}

	/**
	 * @see jmp.ElementalAccessVector#getValues(int[])
	 */
	public double[] getValues(int[] index) {
		double[] ret = new double[index.length];
		for (int i = 0; i < index.length; ++i)
			ret[i] = getValue(index[i]);
		return ret;
	}

	/**
	 * Tries to find the index. If it is not found, a reallocation is done, and
	 * a new index is returned.
	 */
	private int getIndex(int index) {

		// Try to find column index
		int i = Arrays.binarySearchGreater(ind, index, 0, used);

		// Found
		if (i < used && ind[i] == index)
			return i;

		// Not found, try to make room
		if (index < 0 || index >= size)
			throw new ArrayIndexOutOfBoundsException("Index " + index);
		used++;

		// Check available memory
		if (used > data.length) {

			// If zero-length, use new length of 1, else double the bandwidth
			int newLength = 1;
			if (data.length != 0)
				newLength = 2 * data.length;

			// Copy existing data into new arrays
			int[] newInd = new int[newLength];
			double[] newDat = new double[newLength];
			System.arraycopy(ind, 0, newInd, 0, data.length);
			System.arraycopy(data, 0, newDat, 0, data.length);

			// Update pointers
			ind = newInd;
			data = newDat;
		}

		// All ok, make room for insertion
		for (int j = used - 1; j >= i + 1; --j) {
			ind[j] = ind[j - 1];
			data[j] = data[j - 1];
		}

		// Put in new structure
		ind[i] = index;
		data[i] = 0.;
		return i;
	}

	/**
	 * @see jmp.SparseAccessVector#getVector()
	 */
	public synchronized IntDoubleVectorPair getVector() {
		compact();
		return new IntDoubleVectorPair(ind, data);
	}

	/**
	 * @see jmp.SparseAccessVector#setVector(jmp.IntDoubleVectorPair)
	 */
	public synchronized void setVector(IntDoubleVectorPair x) {
		data = x.data;
		ind = x.index;
	}

	/**
	 * @see jmp.VectorAccess#getData()
	 */
	public synchronized double[] getData() {
		compact();
		return data;
	}

	/**
	 * @see jmp.Sparse#compact()
	 */
	public synchronized void compact() {
		if (used < ind.length) {
			double[] newData = new double[used];
			System.arraycopy(data, 0, newData, 0, used);
			int[] newIndex = new int[used];
			System.arraycopy(ind, 0, newIndex, 0, used);
			data = newData;
			ind = newIndex;
		}
	}

}
