/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.io;

import java.io.BufferedReader;
import java.io.FilterReader;
import java.io.IOException;
import java.io.Reader;

import jmp.IntDoubleVectorPair;
import jmp.Matrix;
import jmp.SparseRowColumnMatrix;
import jmp.SparseVector;
import jmp.Vector;

/**
 * Reads matrices and vectors stored in coordinate format. All data must be
 * in real format (no complex entries). The format is explained in more details
 * <a href="http://math.nist.gov/MatrixMarket/formats.html#coord">here</a>.
 * <p>
 * Matrices are return as a SparseRowColumnMatrix, and vectors as
 * SparseVector.
 * </p>
 */
public class CoordinateReader
	extends FilterReader
	implements MatrixVectorReader {

	/**
	 * Constructor for CoordinateReader.
	 */
	public CoordinateReader(Reader in) {
		super(in);
	}

	/**
	 * @see jmp.io.MatrixVectorReader#readMatrix()
	 */
	public Matrix readMatrix() throws IOException {
		synchronized (lock) {
			BufferedReader r = new BufferedReader(in, 1025);
			String[] component = r.readLine().trim().split(" +");

			// Matrix dimensions
			int numRows = Integer.parseInt(component[0]);
			int numColumns = Integer.parseInt(component[1]);
			int numEntries = Integer.parseInt(component[2]);

			// Matrix contents
			int[] rows = new int[numEntries], columns = new int[numEntries];
			double[] data = new double[numEntries];
			for (int i = 0; i < numEntries; ++i) {
				component = r.readLine().trim().split(" +");

				rows[i] = Integer.parseInt(component[0]) - 1;
				columns[i] = Integer.parseInt(component[1]) - 1;

				data[i] = Double.parseDouble(component[2]);
			}

			// Get bandwidth on each row
			int[] nz = new int[numRows];
			for (int i = 0; i < rows.length; ++i)
				nz[rows[i]]++;

			// Copy into an efficient structure
			SparseRowColumnMatrix A =
				new SparseRowColumnMatrix(numRows, numColumns, nz);

			for (int i = 0; i < data.length; ++i)
				A.setValue(rows[i], columns[i], data[i]);

			return A;
		}
	}

	/**
	 * @see jmp.io.MatrixVectorReader#readVector()
	 */
	public Vector readVector() throws IOException {
		synchronized (lock) {
			BufferedReader r = new BufferedReader(in, 1025);
			String[] component = r.readLine().trim().split(" +");

			// Vector size
			int size = Integer.parseInt(component[0]);
			int nz = Integer.parseInt(component[1]);
			SparseVector x = new SparseVector(size, nz);

			// Vector contents
			int[] index = new int[nz];
			double[] data = new double[nz];
			for (int i = 0; i < nz; ++i) {
				component = r.readLine().trim().split(" +");

				index[i] = Integer.parseInt(component[0]) - 1;
				data[i] = Double.parseDouble(component[1]);
			}

			IntDoubleVectorPair xdata = x.getVector();
			System.arraycopy(index, 0, xdata.index, 0, nz);
			System.arraycopy(data, 0, xdata.data, 0, nz);

			return x;
		}
	}

}
