/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.io;

import java.io.BufferedReader;
import java.io.FilterReader;
import java.io.IOException;
import java.io.Reader;

import jmp.CoordinateMatrix;
import jmp.DenseRowColumnMatrix;
import jmp.DenseVector;
import jmp.ElementalAccessMatrix;
import jmp.ElementalAccessVector;
import jmp.Matrix;
import jmp.SparseRowColumnMatrix;
import jmp.SparseVector;
import jmp.Vector;

/**
 * Reads matrices and vectors stored in the Matrix Market exchange format.
 * Both sparse and dense representations are supported, however complex numbers
 * are not supported. For information on this format, see
 * <a href="http://math.nist.gov/MatrixMarket/formats.html#MMformat">this
 * page</a>.
 * <p>
 * Matrices are return as a SparseRowColumnMatrix or DenseRowColumnMatrix,
 * depending on wheter a sparse or dense matrix was read. Vectors are returned
 * as DenseVector or SparseVector, again depending on the underlying format.
 * </p>
 */
public class MatrixMarketReader
	extends FilterReader
	implements MatrixVectorReader {

	/**
	 * Constructor for MatrixMarketReader.
	 */
	public MatrixMarketReader(Reader in) {
		super(in);
	}

	/**
	 * Filelayout
	 */
	private final static int COORDINATE = 1, ARRAY = 2;

	/**
	 * Dataformats for Matrix Market
	 */
	private final static int REAL = 1, INTEGER = 2, PATTERN = 3;

	/**
	 * Algebraic matrix formats
	 */
	private final static int GENERAL = 1, SYMMETRIC = 2, SKEWSYMMETRIC = 3;

	/**
	 * @see jmp.io.MatrixVectorReader#readMatrix()
	 */
	public Matrix readMatrix() throws IOException {
		synchronized (lock) {
			BufferedReader r = new BufferedReader(in, 1025);
			String[] component = r.readLine().trim().split(" +");

			// Read header
			if (!component[0].equals("%%MatrixMarket"))
				throw new IOException("Not in Matrix Market exchange format");

			// This will always be "matrix"
			if (!component[1].equals("matrix"))
				throw new IOException(
					"Expected \"matrix\", got " + component[1]);

			// Sparse or dense?
			int layout = -1;
			if (component[2].equals("coordinate"))
				layout = COORDINATE;
			else if (component[2].equals("array"))
				layout = ARRAY;
			else
				throw new IOException("Unknown layout " + component[2]);

			// Dataformat
			int data = -1;
			if (component[3].equals("real"))
				data = REAL;
			else if (component[3].equals("integer"))
				data = INTEGER;
			else if (component[3].equals("pattern"))
				data = PATTERN;
			else if (component[3].equals("complex"))
				throw new IOException("Cannot read complex data");
			else
				throw new IOException("Unknown dataformat " + component[3]);

			// Matrix pattern
			int pattern = -1;
			if (component[4].equals("general"))
				pattern = GENERAL;
			else if (component[4].equals("symmetric"))
				pattern = SYMMETRIC;
			else if (component[4].equals("skew-symmetric"))
				pattern = SKEWSYMMETRIC;
			else
				throw new IOException("Unknown pattern " + component[4]);

			// Skip initial comments
			while (true) {
				r.mark(1024); // Line length equal 1024 at most
				if (r.readLine().charAt(0) != '%') {
					r.reset();
					break;
				}
			}

			// Call appropriate parser
			if (layout == COORDINATE
				&& (data == REAL || data == INTEGER)
				&& pattern == GENERAL)
				return readCoordinateRealGeneralMatrix(r);
			else if (
				layout == COORDINATE
					&& (data == REAL || data == INTEGER)
					&& pattern == SYMMETRIC)
				return readCoordinateRealSymmetricMatrix(r);
			else if (
				layout == COORDINATE
					&& (data == REAL || data == INTEGER)
					&& pattern == SKEWSYMMETRIC)
				return readCoordinateRealSkewSymmetricMatrix(r);

			else if (
				layout == ARRAY
					&& (data == REAL || data == INTEGER)
					&& pattern == GENERAL)
				return readArrayRealGeneralMatrix(r);
			else if (
				layout == ARRAY
					&& (data == REAL || data == INTEGER)
					&& pattern == SYMMETRIC)
				return readArrayRealSymmetricMatrix(r);
			else if (
				layout == ARRAY
					&& (data == REAL || data == INTEGER)
					&& pattern == SKEWSYMMETRIC)
				return readArrayRealSkewSymmetricMatrix(r);

			else if (
				layout == COORDINATE && data == PATTERN && pattern == GENERAL)
				return readCoordinatePatternGeneralMatrix(r);
			else if (
				layout == COORDINATE
					&& data == PATTERN
					&& pattern == SYMMETRIC)
				return readCoordinatePatternSymmetricMatrix(r);

			else
				throw new IOException("Cannot parse specified format");
		}
	}

	private Matrix readCoordinateRealGeneralMatrix(BufferedReader r)
		throws IOException {
		return convertCoordinateMatrix2Matrix(readCoordinate(r));
	}

	private Matrix readCoordinateRealSymmetricMatrix(BufferedReader r)
		throws IOException {
		return convertSymmetricalCoordinateMatrix2Matrix(readCoordinate(r));
	}

	private Matrix readCoordinateRealSkewSymmetricMatrix(BufferedReader r)
		throws IOException {
		return convertSkewSymmetricalCoordinateMatrix2Matrix(readCoordinate(r));
	}

	private Matrix readCoordinatePatternGeneralMatrix(BufferedReader r)
		throws IOException {
		return convertCoordinateMatrix2Matrix(readCoordinatePattern(r));
	}

	private Matrix readCoordinatePatternSymmetricMatrix(BufferedReader r)
		throws IOException {
		return convertSymmetricalCoordinateMatrix2Matrix(
			readCoordinatePattern(r));
	}

	private CoordinateMatrix readCoordinate(BufferedReader r)
		throws IOException {
		String[] component = r.readLine().trim().split(" +");

		// Matrix dimensions
		int numRows = Integer.parseInt(component[0]);
		int numColumns = Integer.parseInt(component[1]);
		int numEntries = Integer.parseInt(component[2]);

		// Matrix entries
		int[] rows = new int[numEntries], columns = new int[numEntries];
		double[] data = new double[numEntries];
		for (int i = 0; i < numEntries; ++i) {
			component = r.readLine().trim().split(" +");

			rows[i] = Integer.parseInt(component[0]) - 1;
			columns[i] = Integer.parseInt(component[1]) - 1;

			data[i] = Double.parseDouble(component[2]);
		}

		return new CoordinateMatrix(numRows, numColumns, rows, columns, data);
	}

	private CoordinateMatrix readCoordinatePattern(BufferedReader r)
		throws IOException {
		String[] component = r.readLine().trim().split(" +");

		// Matrix dimensions
		int numRows = Integer.parseInt(component[0]);
		int numColumns = Integer.parseInt(component[1]);
		int numEntries = Integer.parseInt(component[2]);

		// Matrix entries
		int[] rows = new int[numEntries], columns = new int[numEntries];
		double[] data = new double[numEntries];
		for (int i = 0; i < numEntries; ++i) {
			component = r.readLine().trim().split(" +");

			rows[i] = Integer.parseInt(component[0]) - 1;
			columns[i] = Integer.parseInt(component[1]) - 1;
		}

		return new CoordinateMatrix(numRows, numColumns, rows, columns, data);
	}

	/**
	 * Converts the coordinate storage into a row-compressed sparse matrix
	 */
	private Matrix convertCoordinateMatrix2Matrix(CoordinateMatrix A) {
		int[] row = A.getRowIndices(), column = A.getColumnIndices();
		double[] data = A.getData();

		// Get bandwidth on each row
		int[] nz = new int[A.numRows()];
		for (int i = 0; i < row.length; ++i)
			nz[row[i]]++;

		SparseRowColumnMatrix B =
			new SparseRowColumnMatrix(A.numRows(), A.numColumns(), nz);

		for (int i = 0; i < data.length; ++i)
			B.setValue(row[i], column[i], data[i]);

		return B;
	}

	/**
	 * Converts the coordinate storage into a row-compressed sparse matrix.
	 * Original data is symmetrical.
	 */
	private Matrix convertSymmetricalCoordinateMatrix2Matrix(CoordinateMatrix A) {
		int[] row = A.getRowIndices(), column = A.getColumnIndices();
		double[] data = A.getData();

		// Get bandwidth on each row
		int[] nz = new int[A.numRows()];
		for (int i = 0; i < row.length; ++i) {
			nz[row[i]]++;
			nz[column[i]]++;
		}

		SparseRowColumnMatrix B =
			new SparseRowColumnMatrix(A.numRows(), A.numColumns(), nz);

		for (int i = 0; i < data.length; ++i) {
			B.setValue(row[i], column[i], data[i]);
			B.setValue(column[i], row[i], data[i]);
		}

		B.compact();

		return B;
	}

	/**
	 * Converts the coordinate storage into a row-compressed sparse matrix.
	 * Original data is skew-symmetrical.
	 */
	private Matrix convertSkewSymmetricalCoordinateMatrix2Matrix(CoordinateMatrix A) {
		int[] row = A.getRowIndices(), column = A.getColumnIndices();
		double[] data = A.getData();

		// Get bandwidth on each row
		int[] nz = new int[A.numRows()];
		for (int i = 0; i < row.length; ++i) {
			nz[row[i]]++;
			nz[column[i]]++;
		}

		SparseRowColumnMatrix B =
			new SparseRowColumnMatrix(A.numRows(), A.numColumns(), nz);

		for (int i = 0; i < data.length; ++i) {
			B.setValue(row[i], column[i], data[i]);
			B.setValue(column[i], row[i], -data[i]);
		}

		B.compact();

		return B;
	}

	private Matrix readArrayRealGeneralMatrix(BufferedReader r)
		throws IOException {

		// Matrix dimensions
		ElementalAccessMatrix A = readArrayMatrix(r);

		// Matrix entries
		for (int j = 0; j < A.numColumns(); ++j)
			for (int i = 0; i < A.numRows(); ++i) {
				String[] component = r.readLine().trim().split(" +");

				double value = Double.parseDouble(component[0]);
				A.setValue(i, j, value);
			}

		return A;
	}

	private Matrix readArrayRealSymmetricMatrix(BufferedReader r)
		throws IOException {

		// Matrix dimensions
		ElementalAccessMatrix A = readArrayMatrix(r);

		// Matrix entries
		for (int j = 0; j < A.numColumns(); ++j)
			for (int i = j; i < A.numRows(); ++i) {
				String[] component = r.readLine().trim().split(" +");

				double value = Double.parseDouble(component[0]);
				A.setValue(i, j, value);
				A.setValue(j, i, value);
			}

		return A;
	}

	private Matrix readArrayRealSkewSymmetricMatrix(BufferedReader r)
		throws IOException {

		// Matrix dimensions
		ElementalAccessMatrix A = readArrayMatrix(r);

		// Matrix entries
		for (int j = 0; j < A.numColumns(); ++j)
			for (int i = j + 1; i < A.numRows(); ++i) {
				String[] component = r.readLine().trim().split(" +");

				double value = Double.parseDouble(component[0]);
				A.setValue(i, j, value);
				A.setValue(j, i, -value);
			}

		return A;
	}

	/**
	 * Reads matrix dimensions, return new dense matrix
	 */
	private ElementalAccessMatrix readArrayMatrix(BufferedReader r)
		throws IOException {
		String[] component = r.readLine().trim().split(" +");

		int numRows = Integer.parseInt(component[0]);
		int numColumns = Integer.parseInt(component[1]);
		return new DenseRowColumnMatrix(numRows, numColumns);
	}

	/**
	 * @see jmp.io.MatrixVectorReader#readVector()
	 */
	public Vector readVector() throws IOException {
		synchronized (lock) {
			BufferedReader r = new BufferedReader(in, 1025);
			String[] component = r.readLine().trim().split(" +");

			// Read header
			if (!component[0].equals("%%MatrixMarket"))
				throw new IOException("Not in Matrix Market exchange format");

			// This will always be "matrix"
			if (!component[1].equals("matrix"))
				throw new IOException(
					"Expected \"matrix\", got " + component[1]);

			// Sparse or dense?
			int layout = -1;
			if (component[2].equals("coordinate"))
				layout = COORDINATE;
			else if (component[2].equals("array"))
				layout = ARRAY;
			else
				throw new IOException("Unknown layout " + component[2]);

			// Dataformat
			int data = -1;
			if (component[3].equals("real"))
				data = REAL;
			else if (component[3].equals("integer"))
				data = INTEGER;
			else if (component[3].equals("pattern"))
				data = PATTERN;
			else if (component[3].equals("complex"))
				throw new IOException("Cannot read complex data");
			else
				throw new IOException("Unknown dataformat " + component[3]);

			// Vector pattern
			if (!component[4].equals("general"))
				throw new IOException(
					"Expected \"general\", got " + component[4]);

			// Call appropriate parser
			if (layout == COORDINATE && (data == REAL || data == INTEGER))
				return readCoordinateRealGeneralVector(r);
			else if (layout == ARRAY && (data == REAL || data == INTEGER))
				return readArrayRealGeneralVector(r);
			else if (layout == COORDINATE && data == PATTERN)
				return readCoordinatePatternGeneralVector(r);
			else
				throw new IOException("Cannot parse specified format");
		}
	}

	private Vector readCoordinateRealGeneralVector(BufferedReader r)
		throws IOException {
		String[] component = r.readLine().trim().split(" +");

		// Size
		ElementalAccessVector x =
			new SparseVector(
				Integer.parseInt(component[0]),
				Integer.parseInt(component[2]));

		// Entries
		for (int i = 0; i < x.size(); ++i) {
			component = r.readLine().trim().split(" +");

			int index = Integer.parseInt(component[0]) - 1;

			double data = Double.parseDouble(component[2]);

			x.setValue(index, data);
		}

		return x;
	}

	private Vector readCoordinatePatternGeneralVector(BufferedReader r)
		throws IOException {
		String[] component = r.readLine().trim().split(" +");

		// Size
		ElementalAccessVector x =
			new SparseVector(
				Integer.parseInt(component[0]),
				Integer.parseInt(component[2]));

		// Entries
		for (int i = 0; i < x.size(); ++i) {
			component = r.readLine().trim().split(" +");

			int index = Integer.parseInt(component[0]) - 1;

			x.setValue(index, 0.);
		}

		return x;
	}

	private Vector readArrayRealGeneralVector(BufferedReader r)
		throws IOException {
		String[] component = r.readLine().trim().split(" +");

		// Size
		ElementalAccessVector x =
			new DenseVector(Integer.parseInt(component[0]));

		// Entries
		for (int i = 0; i < x.size(); ++i) {
			component = r.readLine().trim().split(" +");

			double data = Double.parseDouble(component[0]);

			x.setValue(i, data);
		}

		return x;
	}

}
