/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.io;

import java.io.BufferedWriter;
import java.io.FilterWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Writer;

import jmp.CoordinateAccessMatrix;
import jmp.Dense;
import jmp.DenseAccessVector;
import jmp.DenseColumnAccessMatrix;
import jmp.DenseColumnRowAccessMatrix;
import jmp.ElementalAccessMatrix;
import jmp.ElementalAccessVector;
import jmp.IntDoubleVectorPair;
import jmp.IntIntDoubleVectorTriple;
import jmp.Matrix;
import jmp.Sparse;
import jmp.SparseAccessVector;
import jmp.SparseColumnAccessMatrix;
import jmp.SparseColumnRowAccessMatrix;
import jmp.SparseRowAccessMatrix;
import jmp.SparseRowColumnAccessMatrix;
import jmp.State;
import jmp.Vector;

/**
 * Writes matrices and vectors in the Matrix Market exchange format. The format
 * specification is
 * <a href="http://math.nist.gov/MatrixMarket/formats.html#MMformat">here</a>.
 */
public class MatrixMarketWriter
	extends FilterWriter
	implements MatrixVectorWriter {

	/**
	 * Constructor for MatrixMarketWriter.
	 */
	public MatrixMarketWriter(Writer out) {
		super(out);
	}

	/**
	 * @see jmp.io.MatrixVectorWriter#writeMatrix(jmp.Matrix)
	 */
	public void writeMatrix(Matrix A) throws IOException {
		synchronized (lock) {
			PrintWriter p = new PrintWriter(new BufferedWriter(out));

			// Header
			if (A instanceof Sparse) {
				p.println("%%MatrixMarket matrix coordinate real general");
				p.println(
					A.numRows()
						+ " "
						+ A.numColumns()
						+ " "
						+ State.BLAS.cardinality(A));
			} else if (
				A instanceof Dense || A instanceof ElementalAccessMatrix) {
				p.println("%%MatrixMarket matrix array real general");
				p.println(A.numRows() + " " + A.numColumns());
			} else
				throw new UnsupportedOperationException();

			// Entries
			if (A instanceof SparseRowColumnAccessMatrix)
				writeMatrix((SparseRowColumnAccessMatrix) A, p);
			else if (A instanceof SparseColumnRowAccessMatrix)
				writeMatrix((SparseColumnRowAccessMatrix) A, p);
			else if (A instanceof DenseColumnRowAccessMatrix)
				writeMatrix((DenseColumnRowAccessMatrix) A, p);
			else if (A instanceof SparseRowAccessMatrix)
				writeMatrix((SparseRowAccessMatrix) A, p);
			else if (A instanceof SparseColumnAccessMatrix)
				writeMatrix((SparseColumnAccessMatrix) A, p);
			else if (A instanceof DenseColumnAccessMatrix)
				writeMatrix((DenseColumnAccessMatrix) A, p);
			else if (A instanceof CoordinateAccessMatrix)
				writeMatrix((CoordinateAccessMatrix) A, p);
			else if (A instanceof ElementalAccessMatrix)
				writeMatrix((ElementalAccessMatrix) A, p);
			else
				throw new UnsupportedOperationException();

			p.flush();
		}
	}

	private void writeMatrix(CoordinateAccessMatrix A, PrintWriter p) {
		int[] row = A.getRowIndices(), column = A.getColumnIndices();
		double[] data = A.getData();
		for (int i = 0; i < data.length; ++i)
			p.println((row[i] + 1) + " " + (column[i] + 1) + " " + data[i]);
	}

	private void writeMatrix(SparseRowColumnAccessMatrix A, PrintWriter p) {
		IntIntDoubleVectorTriple Amat = A.getMatrix();
		for (int i = 0; i < A.numRows(); ++i)
			for (int j = Amat.minor[i]; j < Amat.minor[i + 1]; ++j)
				p.println(
					(i + 1) + " " + (Amat.major[j] + 1) + " " + Amat.data[j]);
	}

	private void writeMatrix(SparseColumnRowAccessMatrix A, PrintWriter p) {
		IntIntDoubleVectorTriple Amat = A.getMatrix();
		for (int i = 0; i < A.numColumns(); ++i)
			for (int j = Amat.minor[i]; j < Amat.minor[i + 1]; ++j)
				p.println(
					(Amat.major[j] + 1) + " " + (i + 1) + " " + Amat.data[j]);
	}

	private void writeMatrix(DenseColumnRowAccessMatrix A, PrintWriter p) {
		IntDoubleVectorPair Amat = A.getMatrix();
		for (int i = 0; i < Amat.index.length; ++i)
			p.println(Amat.data[i]);
	}

	private void writeMatrix(SparseRowAccessMatrix A, PrintWriter p) {
		for (int i = 0; i < A.numRows(); ++i) {
			IntDoubleVectorPair Arow = A.getRow(i);
			for (int j = 0; j < Arow.index.length; ++j)
				p.println(
					(i + 1) + " " + (Arow.index[j] + 1) + " " + Arow.data[j]);
		}
	}

	private void writeMatrix(SparseColumnAccessMatrix A, PrintWriter p) {
		for (int j = 0; j < A.numColumns(); ++j) {
			IntDoubleVectorPair Acol = A.getColumn(j);
			for (int i = 0; i < Acol.index.length; ++i)
				p.println(
					(Acol.index[i] + 1) + " " + (j + 1) + " " + Acol.data[i]);
		}
	}

	private void writeMatrix(DenseColumnAccessMatrix A, PrintWriter p) {
		for (int i = 0; i < A.numColumns(); ++i) {
			double[] Acol = A.getColumn(i);
			for (int j = 0; j < Acol.length; ++j)
				p.println(Acol[j]);
		}
	}

	private void writeMatrix(ElementalAccessMatrix A, PrintWriter p) {
		for (int j = 0; j < A.numColumns(); ++j)
			for (int i = 0; i < A.numRows(); ++i)
				p.println(A.getValue(i, j));
	}

	/**
	 * @see jmp.io.MatrixVectorWriter#writeVector(jmp.Vector)
	 */
	public void writeVector(Vector x) throws IOException {
		synchronized (lock) {
			PrintWriter p = new PrintWriter(new BufferedWriter(out));

			// Header
			if (x instanceof SparseAccessVector) {
				p.println("%%MatrixMarket matrix coordinate real general");
				p.println(x.size() + " 1 " + State.BLAS.cardinality(x));
			} else if (
				x instanceof DenseAccessVector
					|| x instanceof ElementalAccessVector) {
				p.println("%%MatrixMarket matrix array real general");
				p.println(x.size() + " 1");
			} else
				throw new UnsupportedOperationException();

			// Entries
			if (x instanceof SparseAccessVector)
				writeVector((SparseAccessVector) x, p);
			else if (x instanceof DenseAccessVector)
				writeVector((DenseAccessVector) x, p);
			else if (x instanceof ElementalAccessVector)
				writeVector((ElementalAccessVector) x, p);
			else
				throw new UnsupportedOperationException();

			p.flush();
		}
	}

	private void writeVector(DenseAccessVector x, PrintWriter p) {
		double[] data = x.getVector();
		for (int i = 0; i < data.length; ++i)
			p.println((i + 1) + " " + data[i]);
	}

	private void writeVector(SparseAccessVector x, PrintWriter p) {
		IntDoubleVectorPair data = x.getVector();
		for (int i = 0; i < data.index.length; ++i)
			p.println((data.index[i] + 1) + " " + data.data[i]);
	}

	private void writeVector(ElementalAccessVector x, PrintWriter p) {
		for (int i = 0; i < x.size(); ++i)
			p.println((i + 1) + " " + x.getValue(i));
	}

}
