/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter;

import java.lang.reflect.Constructor;

import jmp.BlockAccessVector;
import jmp.Vector;

/**
 * Convience factory methods for the iterative methods. Creates empty
 * work-vectors.
 */
public class Factory {

	/**
	 * No need to create an instance
	 */
	private Factory() {
	}

	/**
	 * Creates a new, empty vector based upon the given template
	 */
	public static Vector createVector(Vector template) {
		return createVector(template, template.size());
	}

	/**
	 * Creates a new, empty vector based upon the given template with the
	 * given size
	 */
	public static Vector createVector(Vector template, int size) {
		return createVectors(template, size, 1)[0];
	}

	/**
	 * Creates new, empty vectors based upon the given templates
	 */
	public static Vector[] createVectors(Vector[] template) {
		Vector[] ret = new Vector[template.length];
		for (int i = 0; i < ret.length; ++i)
			ret[i] = createVector(template[i]);
		return ret;
	}

	/**
	 * Creates an array of new, empty vectors based upon the given template
	 */
	public static Vector[] createVectors(Vector template, int num) {
		// Special handling of block-vectors
		if (template instanceof BlockAccessVector) {
			BlockAccessVector templateB = (BlockAccessVector) template;
			int numBlocks = templateB.numBlocks();

			Constructor[] cons = template.getClass().getConstructors();
			Integer[] args =
				new Integer[] {
					new Integer(template.size()),
					new Integer(numBlocks)};

			// Find a constructor which takes two integer parameter
			try {
				for (int i = 0; i < cons.length; ++i) {
					Class[] c = cons[i].getParameterTypes();
					if (c.length == 2
						&& c[0].getName().equals("int")
						&& c[1].getName().equals("int")) {
						BlockAccessVector[] ret = new BlockAccessVector[num];
						for (int j = 0; j < num; ++j) {
							ret[j] =
								(BlockAccessVector) cons[i].newInstance(args);

							// Copy indices and create empty internal vector
							for (int k = 0; k < numBlocks; ++k) {
								int[] index = templateB.getBlockIndices(k),
									newIndex = new int[index.length];
								System.arraycopy(
									index,
									0,
									newIndex,
									0,
									index.length);
								Vector newVector =
									createVector(templateB.getBlock(k));

								ret[j].setBlock(k, newIndex, newVector);
							}
						}
						return ret;
					}
				}
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		}

		// Not a block-vector
		return createVectors(template, template.size(), num);
	}

	/**
	 * Creates an array of new, empty vectors based upon the given template
	 * with the given size of the output vectors
	 */
	public static Vector[] createVectors(Vector template, int size, int num) {
		Constructor[] cons = template.getClass().getConstructors();
		Integer[] args = new Integer[] { new Integer(size)};

		// Find a constructor which only takes an integer parameter (size)
		try {
			for (int i = 0; i < cons.length; ++i) {
				Class[] c = cons[i].getParameterTypes();
				if (c.length == 1 && c[0].getName().equals("int")) {
					Vector[] ret = new Vector[num];
					for (int j = 0; j < num; ++j)
						ret[j] = (Vector) cons[i].newInstance(args);
					return ret;
				}
			}
		} catch (Exception e) {
			throw new RuntimeException(e);
		}

		throw new UnsupportedOperationException();
	}

}
