/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter;

/**
 * Signals the lack of convergence of an iterative solver.
 * It could be due to too stringent convergence criteria, too few iterations
 * used, or simply due to a malformed problem. The number of iterations are
 * stored, along with the reason (BREAKDOWN, DIVERGENCE, or something else).
 */
public class NotConvergedException extends RuntimeException {

	/**
	 * Required too many iterations to converge
	 */
	public static final int ITERATIONS = 1;

	/**
	 * Residual increased to divergence tolerance
	 */
	public static final int DIVERGENCE = 2;

	/**
	 * Breakdown detected in iteration
	 */
	public static final int BREAKDOWN = 3;

	/**
	 * Reason for the exception
	 */
	private int reason;

	/**
	 * Iteration count when this exception was thrown
	 */
	private int iterations;

	/**
	 * Message describing the exception.
	 */
	private String message;

	/**
	 * Constructor for NotConvergedException.
	 * @param reason Reason for exception
	 * @param iterations Iteration count when this exception was thrown
	 */
	public NotConvergedException(int reason, int iterations) {
		this(reason, iterations, "");
	}

	/**
	 * Constructor for NotConvergedException.
	 * @param reason Reason for exception
	 * @param iterations Iteration count when this exception was thrown
	 * @param message Message describing the exception
	 */
	public NotConvergedException(
		int reason,
		int iterations,
		String message) {
		super();
		this.reason = reason;
		this.iterations = iterations;
		this.message = message;
	}

	/**
	 * Gets the reason for this exception
	 * @return Either BREAKDOWN, DIVERGENCE or ITERATIONS.
	 */
	public int getReason() {
		return reason;
	}

	/**
	 * @see java.lang.Throwable#getMessage()
	 */
	public String getMessage() {
		return message;
	}

	/**
	 * Gets the number of iterations used when this exception was thrown
	 */
	public int getIterations() {
		return iterations;
	}

}
