/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.eig;

import jmp.BLAS;
import jmp.State;
import jmp.Vector;
import jmp.iter.AbstractIteration;
import jmp.iter.NotConvergedException;

/**
 * Partial implementation of EigenvalueIteration
 */
public abstract class AbstractEigenvalueIteration
	extends AbstractIteration
	implements EigenvalueIteration {

	/**
	 * Local reference
	 */
	protected BLAS BLAS;

	/**
	 * Iteration monitor
	 */
	protected EigenvalueIterationMonitor monitor;

	/**
	 * Vector-norm
	 */
	protected int normType;

	/**
	 * Constructor for AbstractEigenvalueIteration.
	 */
	public AbstractEigenvalueIteration() {
		this.BLAS = State.BLAS;
		monitor = new NullEigenvalueIterationMonitor();
		normType = jmp.BLAS.NORM2;
	}

	/**
	 * @see jmp.iter.eig.EigenvalueIteration#converged(jmp.Vector[], double[], jmp.Vector[])
	 */
	public boolean converged(Vector[] r, double[] eig, Vector[] x)
		throws NotConvergedException {
		return converged(computeNorms(r), eig, x);
	}

	/**
	 * @see jmp.iter.eig.EigenvalueIteration#converged(jmp.Vector[], double[])
	 */
	public boolean converged(Vector[] r, double[] eig)
		throws NotConvergedException {
		return converged(computeNorms(r), eig);
	}

	/**
	 * @see jmp.iter.eig.EigenvalueIteration#converged(jmp.Vector[])
	 */
	public boolean converged(Vector[] r) throws NotConvergedException {
		return converged(computeNorms(r));
	}

	/**
	 * @see jmp.iter.eig.EigenvalueIteration#converged(double[], double[], jmp.Vector[])
	 */
	public boolean converged(double[] r, double[] eig, Vector[] x)
		throws NotConvergedException {
		monitor.monitor(r, eig, x, iter);
		return convergedI(r);
	}

	/**
	 * @see jmp.iter.eig.EigenvalueIteration#converged(double[], double[])
	 */
	public boolean converged(double[] r, double[] eig)
		throws NotConvergedException {
		monitor.monitor(r, eig, iter);
		return convergedI(r);
	}

	/**
	 * @see jmp.iter.eig.EigenvalueIteration#converged(double[])
	 */
	public boolean converged(double[] r) throws NotConvergedException {
		monitor.monitor(r, iter);
		return convergedI(r);
	}

	protected abstract boolean convergedI(double[] r)
		throws NotConvergedException;

	/**
	 * @see jmp.iter.eig.EigenvalueIteration#getIterationMonitor()
	 */
	public EigenvalueIterationMonitor getIterationMonitor() {
		return monitor;
	}

	/**
	 * @see jmp.iter.eig.EigenvalueIteration#getNormType()
	 */
	public int getNormType() {
		return normType;
	}

	/**
	 * @see jmp.iter.eig.EigenvalueIteration#setIterationMonitor(jmp.iter.eig.EigenvalueIterationMonitor)
	 */
	public void setIterationMonitor(EigenvalueIterationMonitor monitor) {
		this.monitor = monitor;
	}

	/**
	 * @see jmp.iter.eig.EigenvalueIteration#setNormType(int)
	 */
	public void setNormType(int normType) {
		this.normType = normType;
	}

	private double[] computeNorms(Vector[] r) {
		double[] rs = new double[r.length];
		for (int i = 0; i < r.length; ++i)
			rs[i] = BLAS.norm(r[i], normType);
		return rs;
	}

}
