/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.eig;

import java.util.Arrays;

import jmp.BLAS;
import jmp.DenseAccessVector;
import jmp.DenseColumnRowMatrix;
import jmp.ElementalAccessVector;
import jmp.IntDoubleVectorPair;
import jmp.Matrix;
import jmp.State;
import jmp.Vector;
import jmp.iter.Factory;
import jmp.iter.NotConvergedException;
import jmp.util.Maths;

/**
 * Partial implementation of EigenvalueSolver.
 */
public abstract class AbstractEigenvalueSolver implements EigenvalueSolver {

	/**
	 * Local reference
	 */
	protected BLAS BLAS;

	/**
	 * Iteration object
	 */
	protected EigenvalueIteration iter;

	/**
	 * Spectrum transformation
	 */
	protected EigenvalueTransformation et;

	/**
	 * Constructor for AbstractEigenvalueSolver.
	 */
	public AbstractEigenvalueSolver() {
		BLAS = State.BLAS;
		iter = new DefaultEigenvalueIteration();
		et = new NormalEigenvalueTransformation();
	}

	/**
	 * @see jmp.iter.eig.EigenvalueSolver#solve(jmp.Matrix, double[], jmp.Vector[])
	 */
	public void solve(Matrix A, double[] eig, Vector[] x)
		throws NotConvergedException {
		checkSolveArguments(A, eig, x);
		solveI(A, eig, x);
	}

	private void checkSolveArguments(Matrix A, double[] eig, Vector[] x) {
		if (!A.isSquare())
			throw new ArrayIndexOutOfBoundsException("!A.isSquare()");
		else if (eig.length != x.length)
			throw new ArrayIndexOutOfBoundsException("eig.length != x.length");
		for (int i = 0; i < x.length; ++i)
			if (x[i].size() != A.numRows())
				throw new ArrayIndexOutOfBoundsException(
					"x[" + i + "].size() != A.numRows()");
	}

	protected abstract void solveI(Matrix A, double[] eig, Vector[] x)
		throws NotConvergedException;

	/**
	 * @see jmp.iter.eig.EigenvalueSolver#setIteration(jmp.iter.eig.EigenvalueIteration)
	 */
	public void setIteration(EigenvalueIteration iter) {
		this.iter = iter;
	}

	/**
	 * @see jmp.iter.eig.EigenvalueSolver#getIteration()
	 */
	public EigenvalueIteration getIteration() {
		return iter;
	}

	/**
	 * Packs the vectors into a dense column matrix
	 */
	protected DenseColumnRowMatrix pack(Vector[] x) {
		int n = x[0].size();
		DenseColumnRowMatrix A = new DenseColumnRowMatrix(n, n);

		IntDoubleVectorPair Amat = A.getMatrix();
		for (int i = 0; i < x.length; ++i) {
			if (!(x[i] instanceof DenseAccessVector))
				throw new UnsupportedOperationException();
			double[] xdata = ((DenseAccessVector) x[i]).getVector();
			System.arraycopy(xdata, 0, Amat.data, Amat.index[i], xdata.length);
		}

		return A;
	}

	/**
	 * Uses the Rayleigh Quotient to compute real eigenvalues for the given
	 * eigenvectors.
	 */
	protected void eigenvalues(
		Matrix A,
		double[] eigR,
		double[] eigI,
		Vector[] x) {
		Arrays.fill(eigI, 0.);
		Vector temp = Factory.createVector(x[0]);
		for (int i = 0; i < x.length; ++i)
			eigR[i] =
				BLAS.dot(x[i], BLAS.mult(A, x[i], temp)) / BLAS.dot(x[i], x[i]);
	}

	/**
	 * Runs normalize on each array-component
	 */
	protected Vector[] normalize(Vector[] x) {
		for (int i = 0; i < x.length; ++i)
			normalize(x[i]);
		return x;
	}

	/**
	 * Normalizes the given vector using the 2-norm. If it is a zero-vector,
	 * its entries will be populated from a uniform distribution, then
	 * normalized.
	 */
	protected Vector normalize(Vector x) {
		double norm = BLAS.norm(x, jmp.BLAS.NORM2);
		if (!Maths.isZero(norm))
			return BLAS.scale(1. / norm, x);
		else
			return random(x);
	}

	/**
	 * Normalizes the given vector using the 2-norm. If it is a zero-vector,
	 * nothing is done.
	 */
	protected Vector normalize0(Vector x) {
		double norm = BLAS.norm(x, jmp.BLAS.NORM2);
		if (!Maths.isZero(norm))
			return BLAS.scale(1. / norm, x);
		else
			return x;
	}

	/**
	 * Each entry of q is populated from a uniform distribution. The vector will
	 * have a 2-norm of 1.
	 */
	protected Vector random(Vector q) {
		if (!(q instanceof ElementalAccessVector))
			throw new UnsupportedOperationException();
		ElementalAccessVector qe = (ElementalAccessVector) q;
		for (int i = 0; i < qe.size(); ++i)
			qe.setValue(i, Math.random());
		return BLAS.scale(1. / BLAS.norm(q, jmp.BLAS.NORM2), q);
	}

	/**
	 * @see jmp.iter.eig.EigenvalueSolver#getEigenvalueTransformation()
	 */
	public EigenvalueTransformation getEigenvalueTransformation() {
		return et;
	}

	/**
	 * @see jmp.iter.eig.EigenvalueSolver#setEigenvalueTransformation(jmp.iter.eig.EigenvalueTransformation)
	 */
	public void setEigenvalueTransformation(EigenvalueTransformation et) {
		this.et = et;
	}

}
