/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.eig;

import jmp.iter.NotConvergedException;

/**
 * Default eigenvalue iteration object
 */
public class DefaultEigenvalueIteration
	extends AbstractEigenvalueIteration
	implements EigenvalueIteration {

	private double ctol, dtol;

	private int maxIter;

	/**
	 * Constructor for DefaultEigenvalueIteration. Default is 100000 iterations
	 * at most, convergence tolerance of 1e-3 and divergence tolerance of 1e+5.
	 */
	public DefaultEigenvalueIteration() {
		super();
		this.maxIter = 100000;
		this.ctol = 1e-5;
		this.dtol = 1e+5;
	}

	/**
	 * @param ctol New convergence tolerance
	 * @param dtol New divergence tolerance
	 * @param maxIter Maximum number of iterations
	 */
	public void setParameters(double ctol, double dtol, int maxIter) {
		this.maxIter = maxIter;
		this.ctol = ctol;
		this.dtol = dtol;
	}

	/**
	 * @see jmp.iter.eig.AbstractEigenvalueIteration#convergedI(double[])
	 */
	protected boolean convergedI(double[] r) throws NotConvergedException {
		boolean converged = true;

		// Too many iterations?
		if (iter > maxIter)
			throw new NotConvergedException(
				NotConvergedException.ITERATIONS,
				iter);

		for (int i=0; i<r.length; ++i) {

			// Check for convergence
			converged &= r[i] < ctol;

			// Check for divergence
			if (r[i] > dtol)
				throw new NotConvergedException(
					NotConvergedException.DIVERGENCE,
					iter);
			if (Double.isNaN(r[i]))
				throw new NotConvergedException(
					NotConvergedException.DIVERGENCE,
					iter);
		}

		return converged;
	}

}
