/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.eig;

import java.util.Arrays;

import jmp.DenseAccessVector;
import jmp.Matrix;
import jmp.Vector;
import jmp.decomp.EigenvalueDecomposition;
import jmp.iter.Factory;
import jmp.iter.NotConvergedException;

/**
 * Lanczos method. Note that the current implementation is not very advanced
 */
public class Lanczos extends AbstractEigenvalueSolver {

	private boolean full;

	/**
	 * Constructor for Lanczos
	 * @param full True for full orthogonalization, false for partial
	 */
	public Lanczos(boolean full) {
		this.full = full;
	}

	/**
	 * Constructor for Lanczos, using partial orthogonalization
	 */
	public Lanczos() {
		full = false;
	}

	/**
	 * @see jmp.iter.eig.AbstractEigenvalueSolver#solveI(jmp.Matrix, double[], jmp.Vector[])
	 */
	protected void solveI(Matrix A, double[] eig, Vector[] x)
		throws NotConvergedException {

		// Number of eigenvalues wanted and the maximum factorization size
		int n = eig.length, m = A.numRows();

		// Current factorization size, increment, and new size
		int nv = n, inc = 2, nvn = Math.min(nv * inc, m);

		// These holds the matrix factorization
		Vector[] q = Factory.createVectors(x);
		for (int i = 0; i < nv; ++i)
			q[i] = BLAS.copy(x[i], q[i]);
		double[] alpha = new double[nv], beta = new double[nv];

		// Work-vector
		Vector z = Factory.createVector(q[0]);

		// Initial factorization
		lanczos(A, q, alpha, beta, z, 0);

		// Compute the eigenvalues, and estimate residuals
		double[] r = new double[n];
		int[] ind = new int[n];
		EigenvalueDecomposition evd = new EigenvalueDecomposition(alpha, beta);
		estimateResidual(nv, beta, evd, r, ind);

		// Continue while not converged or until the factorization is complete
		// Increase the size of the factorization at each iteration
		for (iter.setFirst();
			!iter.converged(r, eig, x) && nv < m;
			iter.next(), nv = nvn, nvn = Math.min(nv * inc, m)) {

			// Allocate more memory for larger factorization
			Vector[] qn = new Vector[nvn];
			double[] alphan = new double[nvn], betan = new double[nvn];
			expand(q, alpha, beta, qn, alphan, betan);
			q = qn;
			alpha = alphan;
			beta = betan;

			// Compute larger factorization, starting at previous offset
			lanczos(A, q, alpha, beta, z, nv - 1);

			// Get eigenvalues and the residuals
			evd = new EigenvalueDecomposition(alpha, beta);
			estimateResidual(nv, beta, evd, r, ind);

			System.err.println(nvn);
		}

		// Get the eigenvalues (only real)
		double[] eigLoc = evd.getRealEigenvalues();
		for (int i = 0; i < eig.length; ++i)
			eig[i] = eigLoc[ind[i]];

		// Compute the eigenvectors (x=q*V)
		DenseAccessVector[] qa = new DenseAccessVector[x.length],
			xa = new DenseAccessVector[x.length];
		for (int i = 0; i < x.length; ++i) {
			qa[i] = (DenseAccessVector) q[ind[i]];
			xa[i] = (DenseAccessVector) x[i];
		}
		prod(qa, evd.getV(), xa);

		// Sort by the eigenvalues
		IntDoubleVectorPair[] s = new IntDoubleVectorPair[eig.length];
		for (int i = 0; i < s.length; ++i)
			s[i] = new IntDoubleVectorPair(eig[i], i, xa[i]);
		Arrays.sort(s);

		// Extract and return
		for (int i = 0; i < s.length; ++i) {
			eig[i] = s[i].d;
			x[i] = s[i].v;
		}
		et.eigenvalue(eig);
	}

	/**
	 * @return r = q*v, for constructing eigenvectors
	 */
	private Vector[] prod(
		DenseAccessVector[] q,
		double[][] V,
		DenseAccessVector[] r) {
		int m = q[0].size(), n = q.length;

		// Simple multiplication. Not the most efficient, but it'll do
		for (int i = 0; i < m; ++i)
			for (int j = 0; j < n; ++j) {
				double val = 0.;
				for (int l = 0; l < n; ++l)
					val += q[l].getVector()[i] * V[l][j];
				r[j].getVector()[i] = val;
			}

		return r;
	}

	/**
	 * Computes a Lanczos decomposition
	 * @param A Matrix
	 * @param q q[offset] is the initial vector. The rest are overwritten
	 * @param alpha Diagonal of the factorization
	 * @param beta Sub/super diagonals of the factorization
	 * @param z Work vector
	 * @param offset For continued factorization
	 */
	private void lanczos(
		Matrix A,
		Vector[] q,
		double[] alpha,
		double[] beta,
		Vector z,
		int offset) {

		int k = q.length;

		// Normalize the initial vector
		if (offset == 0)
			q[0] = normalize(q[0]);

		for (int i = offset; i < k; ++i) {
			z = et.apply(A, q[i], z); // Either z=A*q[i] or z=(A-sI)\q[i]
			alpha[i] = BLAS.dot(q[i], z);

			// Reorthogonalization
			if (full) {
				z = gramSchmidt(z, q, i - 1);
				z = gramSchmidt(z, q, i - 1);
			} else
				z = lanczosOrthogonalization(z, q, alpha, beta, i);

			// Expand factorization somehow
			if (i + 1 < q.length) {
				beta[i] = BLAS.norm(z, jmp.BLAS.NORM2);
				if (beta[i] != 0.0)
					q[i + 1] = BLAS.scaleCopy(1. / beta[i], z, q[i + 1]);
				else
					q[i + 1] = random(q[i + 1]);
			}
		}
	}

	/**
	 * Applies the Gram-Schmidt procedure to orthogonalize z against the
	 * set of vectors q. Stops at index ql, exclusive
	 */
	private Vector gramSchmidt(Vector z, Vector[] q, int ql) {
		if (ql < 0)
			return z;
		double[] r = new double[ql];
		for (int i = 0; i < ql; ++i)
			r[i] = BLAS.dot(z, q[i]);
		for (int i = 0; i < ql; ++i)
			z = BLAS.add(-r[i], q[i], z);
		return z;
	}

	/**
	 * The usual Lanczos orthogonalization
	 */
	private Vector lanczosOrthogonalization(
		Vector z,
		Vector[] q,
		double[] alpha,
		double[] beta,
		int i) {
		z = BLAS.add(-alpha[i], q[i], z);
		if (i > 0)
			z = BLAS.add(-beta[i - 1], q[i - 1], z);
		return z;
	}

	/**
	 * Eigenvalue residual estimation
	 */
	private void estimateResidual(
		int nv,
		double[] beta,
		EigenvalueDecomposition evd,
		double[] r,
		int[] ind) {
		double[][] V = evd.getV();
		double[] e = evd.getRealEigenvalues();

		// Get all the residuals
		double[] rl = new double[beta.length];
		for (int i = 0; i < beta.length; ++i)
			rl[i] = Math.abs(beta[i] * V[nv - 1][i]) / Math.abs(e[i]);

		// Get the smallest residuals and their indices
		sort(rl, r, ind);
	}

	/**
	 * Sorts a, buts the smallest values into outD and the indices into outI
	 */
	private void sort(double[] a, double[] outD, int[] outI) {
		// Sort both data and indices
		IntDoublePair[] ra = new IntDoublePair[a.length];
		for (int i = 0; i < a.length; ++i)
			ra[i] = new IntDoublePair(a[i], i);
		Arrays.sort(ra);

		// Extract the smallest
		for (int i = 0; i < outD.length; ++i) {
			outD[i] = ra[i].d;
			outI[i] = ra[i].i;
		}
	}

	/**
	 * Pair of an int and a double, comparisons by the double value
	 */
	private class IntDoublePair implements Comparable {

		double d;
		int i;

		public IntDoublePair(double d, int i) {
			this.d = d;
			this.i = i;
		}

		public int compareTo(Object o) {
			IntDoublePair ob = (IntDoublePair) o;
			if (this.d > ob.d)
				return 1;
			else if (this.d < ob.d)
				return -1;
			else
				return 0;
		}

	}

	/**
	 * Pairing of an int, a double and a vector, comparisons by the double value
	 */
	private class IntDoubleVectorPair extends IntDoublePair {

		Vector v;

		public IntDoubleVectorPair(double d, int i, Vector v) {
			super(d, i);
			this.v = v;
		}

	}

	/**
	 * Copies q, alpha and beta over to qn, alphan and betan.
	 */
	private void expand(
		Vector[] q,
		double[] alpha,
		double[] beta,
		Vector[] qn,
		double[] alphan,
		double[] betan) {
		for (int i = 0; i < q.length; ++i)
			qn[i] = q[i];
		for (int i = q.length; i < qn.length; ++i)
			qn[i] = Factory.createVector(q[0]);
		System.arraycopy(alpha, 0, alphan, 0, q.length);
		System.arraycopy(beta, 0, betan, 0, q.length - 1);
	}

}
