/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.eig;

import jmp.Matrix;
import jmp.Vector;
import jmp.iter.Factory;
import jmp.iter.NotConvergedException;

/**
 * Generic power iteration eigenvalue solver. Using
 * NormalEigenvalueTransformation gives the Power method,
 * LockedShiftInvertEigenvalueTransformation will produce the Inverse iteration,
 * while ShiftInvertEigenvalueTransformation gives the Rayleigh quotient
 * iteration.
 */
public class PowerIteration
	extends AbstractEigenvalueSolver
	implements EigenvalueSolver {

	/**
	 * @see jmp.iter.eig.AbstractEigenvalueSolver#solveI(jmp.Matrix, double[], jmp.Vector[])
	 */
	protected void solveI(Matrix A, double[] eig, Vector[] x)
		throws NotConvergedException {
		int n = x.length;
		Vector Ax = Factory.createVector(x[0]);

		// Residual vectors
		Vector[] r = Factory.createVectors(x);
		for (int i = 0; i < n; ++i)
			r[i] = BLAS.set(1., r[i]);

		x = normalize(x);

		for (iter.setFirst(); !iter.converged(r, eig, x); iter.next()) {
			// One step for each eigenvector
			for (int i = 0; i < n; ++i) {
				// Improve on the eigenvector, and compute the eigenvalue
				x[i] = normalize0(x[i]);
				et.setShift(eig[i]);
				Ax = et.apply(A, x[i], Ax);
				eig[i] = BLAS.dot(x[i], Ax);

				// Compute residual, and cycle vectors
				r[i] = BLAS.add(Ax, -eig[i], x[i], r[i]); // r = A*x-eig*x
				x[i] = BLAS.copy(Ax, x[i]); // x = A*x
			}
		}

		// Untransform the eigenvalues
		et.eigenvalue(eig);
	}

}
