/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
/*
 * Derived from public domain software at http://www.netlib.org/templates
 */

package jmp.iter.linear;

import jmp.Matrix;
import jmp.Vector;
import jmp.iter.Factory;
import jmp.iter.NotConvergedException;

/**
 * BiCG solver.
 * BiCG solves the unsymmetric linear system <code>Ax = b</code> using the
 * Preconditioned BiConjugate Gradient method.
 * @author Templates
 */
public class BiCG extends AbstractLinearSolver implements LinearSolver {

	/**
	 * @see jmp.iter.linear.AbstractLinearSolver#solveI(jmp.Matrix, jmp.Vector, jmp.Vector)
	 */
	protected void solveI(Matrix A, Vector b, Vector x)
		throws LinearNotConvergedException {
		double rho_1 = 1., rho_2 = 1., alpha = 1., beta = 1.;
		Vector[] temp = Factory.createVectors(b, 8);
		Vector z = temp[0],
			p = temp[1],
			q = temp[2],
			r = temp[3],
			ztilde = temp[4],
			ptilde = temp[5],
			qtilde = temp[6],
			rtilde = temp[7];

		r = BLAS.multAdd(-1., A, x, b, r);
		rtilde = BLAS.copy(r, rtilde);

		for (iter.setFirst(); !iter.converged(r, x); iter.next()) {
			z = M.apply(A, r, z);
			ztilde = M.transApply(A, rtilde, ztilde);
			rho_1 = BLAS.dot(z, rtilde);

			if (rho_1 == 0.)
				throw new LinearNotConvergedException(
					NotConvergedException.BREAKDOWN,
					iter.iterations(),
					"rho",
					BLAS.norm(r, iter.getNormType()));

			if (iter.isFirst()) {
				p = BLAS.copy(z, p);
				ptilde = BLAS.copy(ztilde, ptilde);
			} else {
				beta = rho_1 / rho_2;
				p = BLAS.add(z, beta, p);
				ptilde = BLAS.add(ztilde, beta, ptilde);
			}

			q = BLAS.mult(A, p, q);
			qtilde = BLAS.transMult(A, ptilde, qtilde);

			alpha = rho_1 / BLAS.dot(ptilde, q);
			x = BLAS.add(alpha, p, x);
			r = BLAS.add(-alpha, q, r);
			rtilde = BLAS.add(-alpha, qtilde, rtilde);

			rho_2 = rho_1;
		}
	}

}
