/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
/*
 * Derived from public domain software at http://www.netlib.org/templates
 */

package jmp.iter.linear;

import jmp.Matrix;
import jmp.Vector;
import jmp.iter.Factory;
import jmp.iter.NotConvergedException;

/**
 * BiCG stablized solver.
 * <p>
 * BiCGstab solves the unsymmetric linear system <code>Ax = b</code>
 * using the Preconditioned BiConjugate Gradient Stabilized method
 * </p>
 * @author Templates
 */
public class BiCGstab extends AbstractLinearSolver implements LinearSolver {

	/**
	 * @see jmp.iter.linear.AbstractLinearSolver#solveI(jmp.Matrix, jmp.Vector, jmp.Vector)
	 */
	protected void solveI(Matrix A, Vector b, Vector x)
		throws LinearNotConvergedException {
		double rho_1 = 1., rho_2 = 1., alpha = 1., beta = 1., omega = 1.;
		Vector[] tempv = Factory.createVectors(b, 9);
		Vector p = tempv[0],
			s = tempv[1],
			phat = tempv[2],
			shat = tempv[3],
			t = tempv[4],
			v = tempv[5],
			temp = tempv[6],
			r = tempv[7],
			rtilde = tempv[8];

		r = BLAS.multAdd(-1., A, x, b, r);
		rtilde = BLAS.copy(r, rtilde);

		for (iter.setFirst(); !iter.converged(r, x); iter.next()) {
			rho_1 = BLAS.dot(rtilde, r);
			if (rho_1 == 0)
				throw new LinearNotConvergedException(
					NotConvergedException.BREAKDOWN,
					iter.iterations(),
					"rho",
					BLAS.norm(r, iter.getNormType()));
			if (omega == 0)
				throw new LinearNotConvergedException(
					NotConvergedException.BREAKDOWN,
					iter.iterations(),
					"omega",
					BLAS.norm(r, iter.getNormType()));

			if (iter.isFirst())
				p = BLAS.copy(r, p);
			else {
				beta = (rho_1 / rho_2) * (alpha / omega);
				temp = BLAS.add(-omega, v, p, temp);
				p = BLAS.add(beta, temp, r, p);
			}
			phat = M.apply(A, p, phat);
			v = BLAS.mult(A, phat, v);
			alpha = rho_1 / BLAS.dot(rtilde, v);
			s = BLAS.add(-alpha, v, r, s);
			if (iter.converged(s, x)) {
				x = BLAS.add(alpha, phat, x);
				return;
			}
			shat = M.apply(A, s, shat);
			t = BLAS.mult(A, shat, t);
			omega = BLAS.dot(t, s) / BLAS.dot(t, t);
			temp = BLAS.add(alpha, phat, omega, shat, temp);
			x = BLAS.add(temp, x);
			r = BLAS.add(-omega, t, s, r);

			rho_2 = rho_1;
		}
	}

}
