/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.linear;

import jmp.DenseAccessVector;
import jmp.ElementalAccessVector;
import jmp.Matrix;
import jmp.State;
import jmp.Vector;
import jmp.iter.Factory;
import jmp.util.ParallelWorker;

/**
 * Block diagonal preconditioner.
 * Like DiagonalPreconditioner, but it uses blocks on the diagonal instead.
 * On each block, a preconditioner is applied, and the user must supply these
 * other preconditioners. The number of diagonal blocks is equal to the number
 * of preconditioners, but the user can specify where the blocks start/stop.
 * The preconditioners supplied are setup once this preconditioner is setup.
 */
public class BlockDiagonalPreconditioner
	extends AbstractPreconditioner
	implements Preconditioner {

	/**
	 * Number of blocks to use.
	 */
	private int numBlocks;

	/**
	 * Parallelization partition
	 */
	private int part[];

	/**
	 * Block diagonal parts of A
	 */
	private Matrix[] Aii;

	/**
	 * The preconditioners which will be applied to the blocks.
	 */
	private Preconditioner[] localPreconditioner;

	/**
	 * The partition of the blocks
	 */
	private int blockPart[];

	/**
	 * Indices internal in the blocks (row/column indices)
	 */
	private int blockIndex[][];

	/**
	 * Constructs a block Jacobi preconditioner with local solvers given as
	 * preconditioners. The number of blocks equal the number of given
	 * preconditioners, and the user specifies the alignment of the blocks.
	 * @param localPreconditioner Local preconditioners
	 * @param Aii Empty matrices to contain the diagonal blocks
	 * @param blockPart User defined block-partition
	 */
	public BlockDiagonalPreconditioner(
		Preconditioner[] localPreconditioner,
		Matrix[] Aii,
		int[] blockPart) {

		this.localPreconditioner = localPreconditioner;
		this.Aii = Aii;
		this.blockPart = blockPart;

		numBlocks = localPreconditioner.length;
		blockIndex = new int[numBlocks][];

		part = ParallelWorker.partition(numBlocks);
	}

	/**
	 * Constructs a block Jacobi preconditioner with local solvers given as
	 * preconditioners. The number of blocks equal the number of given
	 * preconditioners. The blocks are aligned automatically.
	 * @param localPreconditioner Local preconditioners
	 * @param Aii Empty matrices to contain the diagonal blocks
	 */
	public BlockDiagonalPreconditioner(
		Preconditioner[] localPreconditioner,
		Matrix[] Aii) {

		this.localPreconditioner = localPreconditioner;
		this.Aii = Aii;

		numBlocks = localPreconditioner.length;
		blockIndex = new int[numBlocks][];

		int size = 0;
		for (int i = 0; i < Aii.length; ++i)
			size += Aii[i].numRows();
		part = ParallelWorker.partition(numBlocks);
		blockPart = ParallelWorker.partition(size);
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#apply(Matrix, Vector, Vector)
	 */
	public Vector apply(Matrix A, Vector b, Vector x) {
		ParallelWorker[] thread = new BlockApplyThread[State.numThreads];
		for (int i = 0; i < State.numThreads; ++i)
			thread[i] =
				new BlockApplyThread(
					part[i],
					part[i + 1],
					Aii,
					b,
					x,
					localPreconditioner,
					blockIndex,
					false);
		ParallelWorker.runThreads(thread);

		return x;
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#transApply(Matrix, Vector, Vector)
	 */
	public Vector transApply(Matrix A, Vector b, Vector x) {
		ParallelWorker[] thread = new BlockApplyThread[State.numThreads];
		for (int i = 0; i < State.numThreads; ++i)
			thread[i] =
				new BlockApplyThread(
					part[i],
					part[i + 1],
					Aii,
					b,
					x,
					localPreconditioner,
					blockIndex,
					true);
		ParallelWorker.runThreads(thread);

		return x;
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#setup(jmp.Matrix)
	 */
	public void setup(Matrix A) {
		// Dice the matrix
		ParallelWorker[] thread = new BlockExtractThread[State.numThreads];
		for (int i = 0; i < State.numThreads; ++i)
			thread[i] =
				new BlockExtractThread(
					part[i],
					part[i + 1],
					Aii,
					blockPart,
					localPreconditioner,
					blockIndex,
					A);
		ParallelWorker.runThreads(thread);
	}

	/**
	 * Extracts diagonal blocks from the matrix in parallel.
	 */
	private class BlockExtractThread extends ParallelWorker {

		private Matrix A;
		private Matrix[] Aii;
		private int blockPart[];
		private int[][] blockIndex;
		private Preconditioner[] localPreconditioner;

		public BlockExtractThread(
			int start,
			int stop,
			Matrix[] Aii,
			int[] blockPart,
			Preconditioner[] localPreconditioner,
			int[][] blockIndex,
			Matrix A) {
			super(start, stop);
			this.A = A;
			this.Aii = Aii;
			this.blockPart = blockPart;
			this.blockIndex = blockIndex;
			this.localPreconditioner = localPreconditioner;
		}

		public void run() {
			for (int i = start; i < stop; ++i) {
				Aii[i] =
					BLAS.copy(
						A,
						Aii[i],
						blockPart[i],
						blockPart[i + 1],
						blockPart[i],
						blockPart[i + 1]);

				localPreconditioner[i].setup(Aii[i]);

				blockIndex[i] = new int[blockPart[i + 1] - blockPart[i]];
				for (int j = 0; j < blockIndex[i].length; ++j)
					blockIndex[i][j] = j + blockPart[i];
			}
		}

	}

	/**
	 * Applies local preconditioners to their respective blocks.
	 */
	private class BlockApplyThread extends ParallelWorker {
		private Matrix[] Aii;
		private Vector b, x;
		private Preconditioner[] local;
		private int[][] blockIndex;
		private boolean transpose;

		public BlockApplyThread(
			int start,
			int stop,
			Matrix[] Aii,
			Vector b,
			Vector x,
			Preconditioner[] local,
			int[][] blockIndex,
			boolean transpose) {
			super(start, stop);
			this.Aii = Aii;
			this.b = b;
			this.x = x;
			this.local = local;
			this.blockIndex = blockIndex;
			this.transpose = transpose;
		}

		public void run() {
			for (int i = start; i < stop; ++i) {

				// Extract local vectors
				int end = blockIndex[i].length - 1;
				int locsize = blockIndex[i][end] + 1 - blockIndex[i][0];
				Vector bi =
					BLAS.copy(
						b,
						Factory.createVector(b, locsize),
						blockIndex[i][0],
						blockIndex[i][end] + 1);
				Vector xi =
					BLAS.copy(
						x,
						Factory.createVector(b, locsize),
						blockIndex[i][0],
						blockIndex[i][end] + 1);

				// Solve local problem
				if (!transpose)
					xi = local[i].apply(Aii[i], bi, xi);
				else
					xi = local[i].transApply(Aii[i], bi, xi);

				// Put updates back into x
				((ElementalAccessVector) x).setValues(
					blockIndex[i],
					((DenseAccessVector) xi).getVector());
			}
		}

	}

}
