/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
/*
 * Derived from public domain software at http://www.netlib.org/templates
 */

package jmp.iter.linear;

import jmp.Matrix;
import jmp.Vector;
import jmp.iter.Factory;

/**
 * Chebyshev solver.
 * Solves the symmetric positive definite linear system <code>Ax = b</code>
 * using the Preconditioned Chebyshev Method. Chebyshev requires an acurate
 * estimate on the bounds of the spectrum of the matrix.
 * @author Templates
 */
public class Chebyshev extends AbstractLinearSolver implements LinearSolver {

	/**
	 * Estimates for the eigenvalue of the matrix
	 */
	private double eigmin, eigmax;

	/**
	 * @see jmp.iter.linear.AbstractLinearSolver#solveI(jmp.Matrix, jmp.Vector, jmp.Vector)
	 */
	protected void solveI(Matrix A, Vector b, Vector x)
		throws LinearNotConvergedException {
		double alpha = 0, beta = 0., c = 0., d = 0.;
		Vector[] temp = Factory.createVectors(b, 3);
		Vector p = temp[0], z = temp[1], r = temp[2];

		r = BLAS.multAdd(-1., A, x, b, r);

		c = (eigmax - eigmin) / 2.0;
		d = (eigmax + eigmin) / 2.0;

		for (iter.setFirst(); !iter.converged(r, x); iter.next()) {
			z = M.apply(A, r, z); // apply preconditioner

			if (iter.isFirst()) {
				p = BLAS.copy(z, p);
				alpha = 2.0 / d;
			} else {
				beta = alpha * (c * c / 4); // calculate new beta
				alpha = 1.0 / (d - beta); // calculate new alpha
				p = BLAS.add(z, beta, p); // update search direction
			}

			x = BLAS.add(alpha, p, x); // update approximation vector
			r = BLAS.multAdd(-1., A, x, b, r); // compute residual
		}
	}

	/**
	 * Sets eigenvalue estimates
	 * @param eigmin Smallest eigenvalue
	 * @param eigmax Largest eigenvalue
	 */
	public void setEigenvalues(double eigmin, double eigmax) {
		this.eigmax = eigmax;
		this.eigmin = eigmin;
	}

}
