/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.linear;

import jmp.Matrix;
import jmp.Vector;

/**
 * Composite preconditioner.
 * Applies several preconditioners in sequence serially. If the preconditioners
 * have already been set up, it is not necessary to call the setup method
 * in this class, as it simply calls setup() on every preconditioner.
 */
public class CompositePreconditioner
	extends AbstractPreconditioner
	implements Preconditioner {

	/**
	 * The preconditioners which will be applied
	 */
	private Preconditioner[] preconditioner;

	/**
	 * Constructor for CompositePreconditioner.
	 * @param preconditioner The preconditioners to apply
	 */
	public CompositePreconditioner(Preconditioner[] preconditioner) {
		this.preconditioner = preconditioner;
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#apply(Matrix, Vector, Vector)
	 */
	public Vector apply(Matrix A, Vector b, Vector x) {
		for (int i = 0; i < preconditioner.length; ++i)
			x = preconditioner[i].apply(A, b, x);
		return x;
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#transApply(Matrix, Vector, Vector)
	 */
	public Vector transApply(Matrix A, Vector b, Vector x) {
		for (int i = 0; i < preconditioner.length; ++i)
			x = preconditioner[i].transApply(A, b, x);
		return x;
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#setup(jmp.Matrix)
	 */
	public void setup(Matrix A) {
		for (int i = 0; i < preconditioner.length; ++i)
			preconditioner[i].setup(A);
	}

}
