/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.linear;

import java.util.Arrays;

import jmp.IntDoubleVectorPair;
import jmp.Matrix;
import jmp.SparseRowAccessMatrix;

/**
 * Partial preconditioner for decomposition based methods.
 * Contains some methods to make incomplete decomposition based preconditioners
 * easier to implement, such as ILU and ICC.
 */
public abstract class DecompositionPreconditioner
	extends AbstractPreconditioner
	implements Preconditioner {

	/**
	 * The decomposition matrix
	 */
	protected SparseRowAccessMatrix F;

	/**
	 * Original matrix reference
	 */
	protected Matrix A;

	/**
	 * Diagonal indices to the decomposition matrix F.
	 */
	protected int[] diagInd;

	/**
	 * Diagonal shift
	 */
	protected double shift;

	/**
	 * Initial diagonal shift
	 */
	protected double initShift;

	/**
	 * Wheter or not to allow fill
	 */
	protected boolean fill;

	/**
	 * Constructor for DecompositionPreconditioner. Uses a default value of
	 * 0.1 as shift, and disallows fill.
	 * @param F Matrix to contain the decomposition
	 */
	public DecompositionPreconditioner(SparseRowAccessMatrix F) {
		this.F = F;
		shift = 0.1;
	}

	/**
	 * Sets preconditioner parameters
	 * @param initShift Initial diagonal shift. Default is 0
	 * @param shift Diagonal shift to apply in case of zero-pivot. Default is
	 * 0.1
	 * @param fill Allow fill-in? Default is false
	 */
	public void setParamters(double initShift, double shift, boolean fill) {
		this.initShift = initShift;
		this.shift = shift;
		this.fill = fill;
	}

	/**
	 * Sets decomposition matrix
	 */
	public void setDecompositionMatrix(SparseRowAccessMatrix F) {
		this.F = F;
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#setup(jmp.Matrix)
	 */
	public void setup(Matrix A) {
		this.A = A;
		factor();
		for (int i = 0; i < F.numRows(); ++i)
			getDiagInd(i);
	}

	/**
	 * This method should create the factor F of A. When it is called, A is set,
	 * but not A. After the call, diagInd will be set.
	 */
	protected abstract void factor();

	/**
	 * F is lower-triangular, and it is solved for.
	 * @param data Initially the right hand side. Overwritten with solution
	 * @param diagDiv True if the diagonal will be divided with
	 */
	protected void solveL(double[] data, boolean diagDiv) {
		for (int i = 0; i < F.numRows(); ++i) {

			IntDoubleVectorPair cur = F.getRow(i);
			int[] curInd = cur.index;
			double[] curDat = cur.data;

			double sum = 0;
			int j = 0;
			for (; curInd[j] < i; ++j)
				sum += curDat[j] * data[curInd[j]];

			data[i] -= sum;

			// Divide by diagonal. The factorization guarantees its existence
			if (diagDiv)
				data[i] /= curDat[j];
		}
	}

	/**
	 * F is upper-triangular, and it is solved for.
	 * @param data Initially the right hand side. Overwritten with solution
	 * @param diagDiv True if the diagonal will be divided with
	 */
	protected void solveU(double[] data, boolean diagDiv) {
		for (int i = F.numRows() - 1; i >= 0; --i) {

			IntDoubleVectorPair cur = F.getRow(i);
			int[] curInd = cur.index;
			double[] curDat = cur.data;

			double sum = 0;
			for (int j = diagInd[i] + 1; j < curInd.length; ++j)
				sum += curDat[j] * data[curInd[j]];

			data[i] -= sum;

			// Divide by diagonal. The factorization guarantees its existence
			if (diagDiv)
				data[i] /= curDat[diagInd[i]];
		}
	}

	/**
	 * F is lower-triangular, and F<sup>T</sup> is solved for.
	 * @param data Initially the right hand side. Overwritten with solution
	 * @param diagDiv True if the diagonal will be divided with
	 */
	protected void solveLT(double[] data, boolean diagDiv) {
		for (int i = F.numRows() - 1; i >= 0; --i) {

			IntDoubleVectorPair cur = F.getRow(i);
			int[] curInd = cur.index;
			double[] curDat = cur.data;

			// Solve at current position
			if (diagDiv)
				data[i] /= curDat[diagInd[i]];
			double val = data[i];

			// Move over to right hand side
			for (int j = 0; curInd[j] < i; ++j)
				data[curInd[j]] -= curDat[j] * val;
		}
	}

	/**
	 * F is upper-triangular, and F<sup>T</sup> is solved for.
	 * @param data Initially the right hand side. Overwritten with solution
	 * @param diagDiv True if the diagonal will be divided with
	 */
	protected void solveUT(double[] data, boolean diagDiv) {
		for (int i = 0; i < F.numRows(); ++i) {

			IntDoubleVectorPair cur = F.getRow(i);
			int[] curInd = cur.index;
			double[] curDat = cur.data;

			// Solve at current position
			if (diagDiv)
				data[i] /= curDat[diagInd[i]];
			double val = data[i];

			// Move over to right hand side
			for (int j = diagInd[i] + 1; j < curInd.length; ++j)
				data[curInd[j]] -= curDat[j] * val;
		}
	}

	/**
	 * Gets diagonal index from cache (if it exists there), or updates the
	 * cache with diagonal index.
	 * @param row Row to get diagonal index for
	 * @return diagInd[row]
	 */
	protected int getDiagInd(int row) {
		return getDiagInd(row, F.getRow(row).index);
	}

	/**
	 * Gets diagonal index from cache (if it exists there), or updates the
	 * cache with diagonal index.
	 * @param row Row to get diagonal index for
	 * @param rowInd The row indices to search
	 * @return diagInd[row]
	 */
	protected int getDiagInd(int row, int[] rowInd) {
		if (diagInd[row] < 0)
			diagInd[row] = Arrays.binarySearch(rowInd, row);
		return diagInd[row];
	}

}
