/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.linear;

import jmp.Vector;
import jmp.iter.Iteration;
import jmp.iter.NotConvergedException;

/**
 * Default linear iteration object.
 * This tester checks declares convergence if the absolute value of the residual
 * norm is sufficiently small, or if the relative decrease is small.
 * Divergence is decleared if too many iterations are spent, or the
 * residual has grown too much. NaNs will also cause divergence to be flagged.
 */
public class DefaultLinearIteration
	extends AbstractLinearIteration
	implements Iteration, LinearIteration {

	/**
	 * Initial residual
	 */
	protected double initR;

	/**
	 * Relative tolerance
	 */
	protected double rtol;
	
	/**
	 * Absolute tolerance
	 */
	protected double atol;
	
	/**
	 * Divergence tolerance
	 */
	protected double dtol;

	/**
	 * Maximum number of iterations
	 */
	protected int maxIter;

	/**
	 * Constructor for DefaultLinearIteration. Default is 100000 iterations at
	 * most, relative tolerance of 1e-5, absolute tolerance of 1e-50 and a
	 * divergence tolerance of 1e+5.
	 */
	public DefaultLinearIteration() {
		super();
		this.maxIter = 100000;
		this.rtol = 1e-5;
		this.atol = 1e-50;
		this.dtol = 1e+5;
	}

	/**
	 * @param rtol New relative tolerance
	 * @param atol New absolute tolerance
	 * @param dtol New divergence tolerance
	 * @param maxIter Maximum number of iterations
	 */
	public void setParameters(double rtol, double atol, double dtol, int maxIter) {
		this.maxIter = maxIter;
		this.rtol = rtol;
		this.atol = atol;
		this.dtol = dtol;
	}

	/**
	 * @see jmp.iter.linear.AbstractLinearIteration#convergedI(double)
	 */
	protected boolean convergedI(double r) throws NotConvergedException {
		// Store initial residual
		if (isFirst())
			initR = r;

		// Check for convergence
		if (r < Math.max(rtol * initR, atol))
			return true;

		// Check for divergence
		if (r > dtol * initR)
			throw new LinearNotConvergedException(
				NotConvergedException.DIVERGENCE,
				iter,
				r);
		if (iter > maxIter)
			throw new LinearNotConvergedException(
				NotConvergedException.ITERATIONS,
				iter,
				r);
		if (Double.isNaN(r))
			throw new LinearNotConvergedException(
				NotConvergedException.DIVERGENCE,
				iter,
				r);

		// Neither convergence nor divergence
		return false;
	}

	/**
	 * @see jmp.iter.linear.AbstractLinearIteration#convergedI(double, jmp.Vector)
	 */
	protected boolean convergedI(double r, Vector x)
		throws NotConvergedException {
		return convergedI(r);
	}

}
