/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.linear;

import jmp.DenseAccessVector;
import jmp.ElementalAccessMatrix;
import jmp.Matrix;
import jmp.State;
import jmp.Vector;
import jmp.util.ParallelWorker;

/**
 * Diagonal preconditioner.
 * Uses the inverse of the diagonal as preconditioner. Runs in parallel,
 * but is typically not the best preconditioner.
 */
public class DiagonalPreconditioner
	extends AbstractPreconditioner
	implements Preconditioner {

	/**
	 * This contains the inverse of the diagonal
	 */
	private double[] invdiag;

	/**
	 * For parallelization
	 */
	private int part[];

	/**
	 * @see jmp.iter.linear.Preconditioner#apply(Matrix, Vector, Vector)
	 */
	public Vector apply(Matrix A, Vector b, Vector x) {
		checkSizes(A, b, x);

		if (!(b instanceof DenseAccessVector)
			|| !(x instanceof DenseAccessVector))
			throw new UnsupportedOperationException();

		double[] bdat = ((DenseAccessVector) b).getVector(),
			xdat = ((DenseAccessVector) x).getVector();

		ParallelWorker[] thread = new DiagonalWorker[State.numThreads];
		for (int i = 0; i < State.numThreads; ++i)
			thread[i] =
				new DiagonalWorker(part[i], part[i + 1], bdat, xdat, invdiag);
		ParallelWorker.runThreads(thread);

		return x;
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#transApply(Matrix, Vector, Vector)
	 */
	public Vector transApply(Matrix A, Vector b, Vector x) {
		return apply(A, b, x);
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#setup(jmp.Matrix)
	 */
	public void setup(Matrix A) {
		if (!(A instanceof ElementalAccessMatrix))
			throw new UnsupportedOperationException();

		invdiag = new double[A.numRows()];
		part = ParallelWorker.partition(invdiag.length);

		ParallelWorker[] thread = new InvdiagWorker[State.numThreads];
		for (int i = 0; i < State.numThreads; ++i)
			thread[i] =
				new InvdiagWorker(
					part[i],
					part[i + 1],
					(ElementalAccessMatrix) A,
					invdiag);
		ParallelWorker.runThreads(thread);
	}

	/**
	 * Applies the preconditioner to part of the vector
	 */
	private class DiagonalWorker extends ParallelWorker {
		double x[], b[], invdiag[];

		public DiagonalWorker(
			int start,
			int stop,
			double[] b,
			double[] x,
			double[] invdiag) {
			super(start, stop);
			this.b = b;
			this.x = x;
			this.invdiag = invdiag;
		}

		public void run() {
			for (int i = start; i < stop; ++i)
				x[i] = b[i] * invdiag[i];
		}
	}

	/**
	 * Computes the inverse of the diagonal in setup()
	 */
	private class InvdiagWorker extends ParallelWorker {
		ElementalAccessMatrix diag;
		double invdiag[];

		public InvdiagWorker(
			int start,
			int stop,
			ElementalAccessMatrix diag,
			double invdiag[]) {
			super(start, stop);
			this.diag = diag;
			this.invdiag = invdiag;
		}

		public void run() {
			for (int i = start; i < stop; ++i) {
				invdiag[i] = diag.getValue(i, i);
				if (Math.abs(invdiag[i]) == 0.) // Avoid zero-division
					invdiag[i] = 1.;
				else
					invdiag[i] = 1. / invdiag[i];
			}
		}
	}

}
