/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
/*
 * Derived from public domain software at http://www.netlib.org/templates
 */

package jmp.iter.linear;

import java.util.Arrays;

import jmp.Matrix;
import jmp.Vector;
import jmp.iter.Factory;

/**
 * GMRES solver.
 * GMRES solves the unsymmetric linear system <code>Ax = b</code> using the
 * Generalized Minimum Residual method. The GMRES iteration is restarted after
 * a given number of iterations. By default it is restarted after 30 iterations.
 * @author Templates
 */
public class GMRES extends AbstractLinearSolver implements LinearSolver {

	/**
	 * After this many iterations, the GMRES will be restarted.
	 */
	private int restart;

	/**
	 * Default restart of 30
	 * @see jmp.iter.linear.AbstractLinearSolver#AbstractLinearSolver()
	 */
	public GMRES() {
		super();
		restart = 30;
	}

	/**
	 * Sets restart to use
	 * @param restart GMRES iteration is restarted after this number of
	 * iterations
	 */
	public void setRestart(int restart) {
		this.restart = restart;
	}

	/**
	 * @see jmp.iter.linear.AbstractLinearSolver#solveI(jmp.Matrix, jmp.Vector, jmp.Vector)
	 */
	protected void solveI(Matrix A, Vector b, Vector x)
		throws LinearNotConvergedException {
		Vector[] tempv = Factory.createVectors(b, 3);
		Vector w = tempv[0], temp = tempv[1], r = tempv[2];

		double[] s = new double[restart + 1],
			cs = new double[restart + 1],
			sn = new double[restart + 1];

		double[][] H = new double[restart + 1][restart];

		temp = BLAS.multAdd(-1., A, x, b, temp);
		r = M.apply(A, temp, r);
		double normr = BLAS.norm(r, jmp.BLAS.NORM2);
		temp = M.apply(A, b, temp);

		Vector[] v = Factory.createVectors(b, restart + 1);

		for (iter.setFirst(); !iter.converged(r, x); iter.next()) {

			v[0] = BLAS.scaleCopy(1. / normr, r, v[0]);
			Arrays.fill(s, 0.);
			s[0] = normr;

			for (int i = 0; i < restart; i++, iter.next()) {
				temp = BLAS.mult(A, v[i], temp);
				w = M.apply(A, temp, w);
				for (int k = 0; k <= i; k++) {
					H[k][i] = BLAS.dot(w, v[k]);
					w = BLAS.add(-H[k][i], v[k], w);
				}
				H[i + 1][i] = BLAS.norm(w, jmp.BLAS.NORM2);
				v[i + 1] = BLAS.scaleCopy(1. / H[i + 1][i], w, v[i + 1]);

				for (int k = 0; k < i; k++) {
					RotationData RD =
						new RotationData(H[k][i], H[k + 1][i], cs[k], sn[k]);
					applyPlaneRotation(RD);
					H[k][i] = RD.dx;
					H[k + 1][i] = RD.dy;
					cs[k] = RD.cs;
					sn[k] = RD.sn;
				}

				RotationData RD =
					new RotationData(H[i][i], H[i + 1][i], cs[i], sn[i]);
				generatePlaneRotation(RD);
				H[i][i] = RD.dx;
				H[i + 1][i] = RD.dy;
				cs[i] = RD.cs;
				sn[i] = RD.sn;

				applyPlaneRotation(RD);
				H[i][i] = RD.dx;
				H[i + 1][i] = RD.dy;
				cs[i] = RD.cs;
				sn[i] = RD.sn;

				RD.dx = s[i];
				RD.dy = s[i + 1];
				applyPlaneRotation(RD);
				s[i] = RD.dx;
				s[i + 1] = RD.dy;
				cs[i] = RD.cs;
				sn[i] = RD.sn;

				if (iter.converged(Math.abs(s[i + 1]), x)) {
					update(x, i, s, v, H);
					return;
				}
			}

			update(x, restart - 1, s, v, H);
			temp = BLAS.multAdd(-1., A, x, b, temp);
			r = M.apply(A, temp, r);
			normr = BLAS.norm(r, jmp.BLAS.NORM2);
		}
	}

	private void update(
		Vector x,
		int k,
		double[] s,
		Vector v[],
		double[][] H) {
		double[] yy = new double[s.length];
		System.arraycopy(s, 0, yy, 0, s.length);

		// Backsolve
		for (int i = k; i >= 0; i--) {
			yy[i] /= H[i][i];
			for (int j = i - 1; j >= 0; j--)
				yy[j] -= H[j][i] * yy[i];
		}

		for (int j = 0; j <= k; j++)
			x = BLAS.add(yy[j], v[j], x);
	}

	/**
	 * Constructs a Givens rotation, to be used by applyPlaneRotation
	 */
	private void generatePlaneRotation(RotationData RD) {
		if (RD.dy == 0.0) {
			RD.cs = 1.0;
			RD.sn = 0.0;
		} else if (Math.abs(RD.dy) > Math.abs(RD.dx)) {
			double temp = RD.dx / RD.dy;
			RD.sn = 1.0 / Math.sqrt(1.0 + temp * temp);
			RD.cs = temp * RD.sn;
		} else {
			double temp = RD.dy / RD.dx;
			RD.cs = 1.0 / Math.sqrt(1.0 + temp * temp);
			RD.sn = temp * RD.cs;
		}
	}

	/**
	 * Applies a Givens plane rotation
	 */
	private void applyPlaneRotation(RotationData RD) {
		double temp = RD.cs * RD.dx + RD.sn * RD.dy;
		RD.dy = -RD.sn * RD.dx + RD.cs * RD.dy;
		RD.dx = temp;
	}

	/**
	 * Data for a Givens rotation in GMRES.
	 */
	private class RotationData {
		public double dx, dy, cs, sn;

		public RotationData(double dx, double dy, double cs, double sn) {
			this.dx = dx;
			this.dy = dy;
			this.cs = cs;
			this.sn = sn;
		}

		public RotationData() {
			this.dx = 0.;
			this.dy = 0.;
			this.cs = 0.;
			this.sn = 0.;
		}
	}

}
