/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.linear;

import java.util.Arrays;

import jmp.DenseAccessVector;
import jmp.IntDoubleVectorPair;
import jmp.Matrix;
import jmp.SparseRowAccessMatrix;
import jmp.Vector;

/**
 * An ICC preconditioner.
 * Uses an incomplete Cholesky decomposition as preconditioner. Applicable to
 * symmetrical, positive definite problems.
 */
public class ICC
	extends DecompositionPreconditioner
	implements Preconditioner {

	/**
	 * Constructor for ICC.
	 */
	public ICC(SparseRowAccessMatrix F) {
		super(F);
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#apply(Matrix, Vector, Vector)
	 */
	public Vector apply(Matrix A, Vector b, Vector x) {
		checkSizes(A, b, x);

		// Copy b into x, and get the array
		x = BLAS.copy(b, x);
		double[] xdat = ((DenseAccessVector) x).getVector();

		// First solve F^T y=b, then Fx=y
		solveUT(xdat, true);
		solveU(xdat, true);

		// Put xdat back into x
		 ((DenseAccessVector) x).setVector(xdat);

		return x;
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#transApply(Matrix, Vector, Vector)
	 */
	public Vector transApply(Matrix A, Vector b, Vector x) {
		return apply(A, b, x);
	}

	/**
	 * @see jmp.iter.linear.DecompositionPreconditioner#factor()
	 */
	protected void factor() {
		double diagMod = initShift;
		int n = F.numRows();
		diagInd = new int[n];
		double[] actRow = new double[n];

		outer : while (true) {

			// Copy A into F, and ensure non-zero diagonal
			BLAS.addDiagonal(diagMod, BLAS.copy(A, F));

			// Extract the upper triangular part
			for (int i = 0; i < n; ++i) {
				IntDoubleVectorPair curRow = F.getRow(i);
				int diagInd = Arrays.binarySearch(curRow.index, i);
				if (diagInd < 0) {
					diagMod += shift;
					continue outer;
				}
				int[] index = new int[curRow.index.length - diagInd];
				double[] data = new double[curRow.data.length - diagInd];
				System.arraycopy(curRow.index, diagInd, index, 0, index.length);
				System.arraycopy(curRow.data, diagInd, data, 0, data.length);
				F.setRow(i, new IntDoubleVectorPair(index, data));
			}

			// Factorize
			for (int i = 0; i < n; ++i) {

				// Get current row
				IntDoubleVectorPair curRow = F.getRow(i);

				// We must have positive entries on the diagonal
				if (curRow.data.length == 0 || curRow.data[0] <= 0.) {
					diagMod += shift;
					continue outer;
				}

				// Divide by rooted diagonal
				double diag = curRow.data[0] = Math.sqrt(curRow.data[0]);
				for (int j = 1; j < curRow.data.length; ++j)
					curRow.data[j] /= diag;
				F.setRow(i, curRow);

				// Reduce submatrix using current row
				for (int j = 1; j < curRow.index.length; ++j) {

					// Active row in dense format
					IntDoubleVectorPair actRowI = F.getRow(curRow.index[j]);
					actRow = BLAS.scatter(actRowI, actRow);

					// Reduce active row
					double factor = curRow.data[j];
					for (int k = j; k < curRow.data.length; ++k)
						actRow[curRow.index[k]] -= factor * curRow.data[k];

					// Put updated row back into the factor matrix
					// We either keep or discard filled in values
					if (fill)
						F.setRow(i, BLAS.gather(actRow));
					else
						F.setRow(i, BLAS.gather(actRowI.index, actRow));
				}

			}

			// No errors during factorization
			break;
		}
	}

}
