/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.linear;

import jmp.DenseAccessVector;
import jmp.IntDoubleVectorPair;
import jmp.Matrix;
import jmp.SparseRowAccessMatrix;
import jmp.Vector;

/**
 * ILU preconditioner.
 * Performs an incomplete LU decomposition as a preconditioner. Applicable to
 * unsymmetrical problems, however it is strictly sequential.
 */
public class ILU
	extends DecompositionPreconditioner
	implements Preconditioner {

	/**
	 * Constructor for ILU.
	 */
	public ILU(SparseRowAccessMatrix F) {
		super(F);
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#apply(Matrix, Vector, Vector)
	 */
	public Vector apply(Matrix A, Vector b, Vector x) {
		checkSizes(A, b, x);

		// Copy b into x, and get the array
		x = BLAS.copy(b, x);
		double[] xdat = ((DenseAccessVector) x).getVector();

		// Solve L*U
		solveL(xdat, false);
		solveU(xdat, true);

		// Put xdat back into x
		 ((DenseAccessVector) x).setVector(xdat);

		return x;
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#transApply(Matrix, Vector, Vector)
	 */
	public Vector transApply(Matrix A, Vector b, Vector x) {
		checkSizes(A, b, x);

		// Copy b into x, and get the array
		x = BLAS.copy(b, x);
		double[] xdat = ((DenseAccessVector) x).getVector();

		// Solve U'*L'
		solveUT(xdat, true);
		solveLT(xdat, false);

		// Put xdat back into x
		 ((DenseAccessVector) x).setVector(xdat);

		return x;
	}

	protected void factor() {
		double diagMod = initShift;
		int n = A.numRows();
		diagInd = new int[n];
		double[] curRow = new double[n];

		outer : while (true) {

			// Copy A into F, and ensure non-zero diagonal
			BLAS.addDiagonal(diagMod, BLAS.copy(A, F));

			// Clear cache of diagonal indices
			java.util.Arrays.fill(diagInd, -1);
			getDiagInd(0);

			// Row-based Gaussian elimination
			for (int i = 1; i < n; ++i) {

				// Get current row in dense format
				IntDoubleVectorPair curRowI = F.getRow(i);
				curRow = BLAS.scatter(curRowI, curRow);

				// Check for empty row
				if (curRowI.index.length == 0) {
					diagMod += shift;
					continue outer;
				}

				// Traverse all active rows
				for (int j = 0;
					j < curRowI.index.length && curRowI.index[j] < i;
					++j) {

					IntDoubleVectorPair actRow = F.getRow(curRowI.index[j]);

					// Check for missing or zero diagonal
					int diag = getDiagInd(curRowI.index[j], actRow.index);
					if (diag < 0 || actRow.data[diag] == 0) {
						diagMod += shift;
						continue outer;
					}

					// Get multiplier (to be stored in L)
					double mult = -curRow[curRowI.index[j]] / actRow.data[diag];
					curRow[curRowI.index[j]] = -mult;

					// Reduce everything on the upper diagonal of current row
					for (int k = diag + 1; k < actRow.index.length; ++k)
						curRow[actRow.index[k]] += mult * actRow.data[k];
				}

				// Put updated row back into the factor matrix
				// We either keep or discard filled in values
				if (fill)
					F.setRow(i, BLAS.gather(curRow));
				else
					F.setRow(i, BLAS.gather(curRowI.index, curRow));
			}

			// No errors during factorization
			break;
		}
	}

}
