/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.linear;

import jmp.DenseAccessVector;
import jmp.DenseRowAccessMatrix;
import jmp.DenseRowColumnAccessMatrix;
import jmp.IntDoubleVectorPair;
import jmp.IntIntDoubleVectorTriple;
import jmp.Matrix;
import jmp.SparseRowAccessMatrix;
import jmp.SparseRowColumnAccessMatrix;
import jmp.State;
import jmp.Vector;
import jmp.util.ParallelWorker;

/**
 * Jacobi preconditioner.
 * Applies one sweep of Jacobi to the system matrix. Fully parallel. Does not
 * support transpose operations.
 */
public class Jacobi extends AbstractPreconditioner implements Preconditioner {

	/**
	 * For parallelization
	 */
	private int part[];

	/**
	 * @see jmp.iter.linear.Preconditioner#apply(jmp.Matrix, jmp.Vector, jmp.Vector)
	 */
	public Vector apply(Matrix A, Vector b, Vector x) {
		checkSizes(A, b, x);

		if (!(b instanceof DenseAccessVector)
			|| !(x instanceof DenseAccessVector))
			throw new UnsupportedOperationException();

		double[] bdata = ((DenseAccessVector) b).getVector(),
			xdata = ((DenseAccessVector) x).getVector(),
			xprev = new double[xdata.length];
		System.arraycopy(xdata, 0, xprev, 0, xdata.length);

		if (A instanceof SparseRowAccessMatrix) {
			ParallelWorker[] thread =
				new JacobiSparseRowWorker[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				thread[i] =
					new JacobiSparseRowWorker(
						part[i],
						part[i + 1],
						(SparseRowAccessMatrix) A,
						bdata,
						xdata,
						xprev);
			ParallelWorker.runThreads(thread);
		} else if (A instanceof SparseRowColumnAccessMatrix) {
			ParallelWorker[] thread =
				new JacobiSparseRowColumnWorker[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				thread[i] =
					new JacobiSparseRowColumnWorker(
						part[i],
						part[i + 1],
						(SparseRowColumnAccessMatrix) A,
						bdata,
						xdata,
						xprev);
			ParallelWorker.runThreads(thread);
		} else if (A instanceof DenseRowAccessMatrix) {
			ParallelWorker[] thread =
				new JacobiDenseRowWorker[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				thread[i] =
					new JacobiDenseRowWorker(
						part[i],
						part[i + 1],
						(DenseRowAccessMatrix) A,
						bdata,
						xdata,
						xprev);
			ParallelWorker.runThreads(thread);
		} else if (A instanceof DenseRowColumnAccessMatrix) {
			ParallelWorker[] thread =
				new JacobiDenseRowColumnWorker[State.numThreads];
			for (int i = 0; i < State.numThreads; ++i)
				thread[i] =
					new JacobiDenseRowColumnWorker(
						part[i],
						part[i + 1],
						(DenseRowColumnAccessMatrix) A,
						bdata,
						xdata,
						xprev);
			ParallelWorker.runThreads(thread);
		} else
			throw new UnsupportedOperationException();

		return x;
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#transApply(jmp.Matrix, jmp.Vector, jmp.Vector)
	 */
	public Vector transApply(Matrix A, Vector b, Vector x) {
		throw new UnsupportedOperationException();
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#setup(jmp.Matrix)
	 */
	public void setup(Matrix A) {
		part = ParallelWorker.partition(A.numRows());
	}

	private class JacobiSparseRowWorker extends ParallelWorker {

		private SparseRowAccessMatrix A;
		private double bdata[], xdata[], xprev[];

		public JacobiSparseRowWorker(
			int start,
			int stop,
			SparseRowAccessMatrix A,
			double[] bdata,
			double[] xdata,
			double[] xprev) {
			super(start, stop);
			this.A = A;
			this.bdata = bdata;
			this.xdata = xdata;
			this.xprev = xprev;
		}

		public void run() {
			for (int i = start; i < stop; ++i) {
				IntDoubleVectorPair Arow = A.getRow(i);

				xdata[i] = 0;
				double diagVal = 0.;
				for (int j = 0; j < Arow.index.length; ++j)
					if (Arow.index[j] != i)
						xdata[i] += Arow.data[j] * xprev[Arow.index[j]];
					else
						diagVal = Arow.data[j];

				xdata[i] = (bdata[i] - xdata[i]) / diagVal;
			}
		}
	}

	private class JacobiSparseRowColumnWorker extends ParallelWorker {

		private SparseRowColumnAccessMatrix A;
		private double bdata[], xdata[], xprev[];

		public JacobiSparseRowColumnWorker(
			int start,
			int stop,
			SparseRowColumnAccessMatrix A,
			double[] bdata,
			double[] xdata,
			double[] xprev) {
			super(start, stop);
			this.A = A;
			this.bdata = bdata;
			this.xdata = xdata;
			this.xprev = xprev;
		}

		public void run() {
			IntIntDoubleVectorTriple Amat = A.getMatrix();

			for (int i = start; i < stop; ++i) {

				xdata[i] = 0;
				double diagVal = 0.;
				for (int j = Amat.minor[i]; j < Amat.minor[i + 1]; ++j)
					if (Amat.major[j] != i)
						xdata[i] += Amat.data[j] * xprev[Amat.major[j]];
					else
						diagVal = Amat.data[j];

				xdata[i] = (bdata[i] - xdata[i]) / diagVal;
			}
		}
	}

	private class JacobiDenseRowWorker extends ParallelWorker {

		private DenseRowAccessMatrix A;
		private double bdata[], xdata[], xprev[];

		public JacobiDenseRowWorker(
			int start,
			int stop,
			DenseRowAccessMatrix A,
			double[] bdata,
			double[] xdata,
			double[] xprev) {
			super(start, stop);
			this.A = A;
			this.bdata = bdata;
			this.xdata = xdata;
			this.xprev = xprev;
		}

		public void run() {
			for (int i = start; i < stop; ++i) {
				double[] Arow = A.getRow(i);

				xdata[i] = 0;
				double diagVal = 0.;
				for (int j = 0; j < Arow.length; ++j)
					if (j != i)
						xdata[i] += Arow[j] * xprev[j];
					else
						diagVal = Arow[j];

				xdata[i] = (bdata[i] - xdata[i]) / diagVal;
			}
		}
	}

	private class JacobiDenseRowColumnWorker extends ParallelWorker {

		private DenseRowColumnAccessMatrix A;
		private double bdata[], xdata[], xprev[];

		public JacobiDenseRowColumnWorker(
			int start,
			int stop,
			DenseRowColumnAccessMatrix A,
			double[] bdata,
			double[] xdata,
			double[] xprev) {
			super(start, stop);
			this.A = A;
			this.bdata = bdata;
			this.xdata = xdata;
			this.xprev = xprev;
		}

		public void run() {
			IntDoubleVectorPair Amat = A.getMatrix();
			for (int i = start; i < stop; ++i) {

				xdata[i] = 0;
				double diagVal = 0.;
				for (int j = Amat.index[i], k = 0;
					j < Amat.index[i + 1];
					++j, ++k)
					if (k != i)
						xdata[i] += Amat.data[j] * xprev[k];
					else
						diagVal = Amat.data[j];

				xdata[i] = (bdata[i] - xdata[i]) / diagVal;
			}
		}
	}

}
