/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.linear;

import jmp.Vector;
import jmp.iter.Iteration;
import jmp.iter.NotConvergedException;

/**
 * Linear iteration object based on matrix norms.
 * Extends the default linear iteration object to compare with the norm of the
 * system matrix and the right hand side. Can often be a better convergence
 * criteria than the default, but requires the computation of the matrix norm.
 */
public class MatrixLinearIteration
	extends DefaultLinearIteration
	implements Iteration, LinearIteration {

	/**
	 * Norm of the system matrix
	 */
	protected double normA;

	/**
	 * Norm of the right hand side
	 */
	protected double normb;

	/**
	 * Constructor for MatrixLinearIteration. Default is 100000 iterations at
	 * most, relative tolerance of 1e-5, absolute tolerance of 1e-50 and a
	 * divergence tolerance of 1e+5.
	 */
	public MatrixLinearIteration() {
		super();
	}

	/**
	 * Sets the norms
	 * @param normA Norm of the system matrix
	 * @param normb Norm of the right hand side
	 */
	public void setNorms(double normA, double normb) {
		this.normA = normA;
		this.normb = normb;
	}

	/**
	 * @see jmp.iter.linear.AbstractLinearIteration#convergedI(double, jmp.Vector)
	 */
	protected boolean convergedI(double r, Vector x)
		throws NotConvergedException {
		// Store initial residual
		if (isFirst())
			initR = r;

		// Check for convergence
		if (r
			< Math.max(rtol * (normA * BLAS.norm(x, normType) + normb), atol))
			return true;

		// Check for divergence
		if (r > dtol * initR)
			throw new LinearNotConvergedException(
				NotConvergedException.DIVERGENCE,
				iter,
				r);
		if (iter > maxIter)
			throw new LinearNotConvergedException(
				NotConvergedException.ITERATIONS,
				iter,
				r);
		if (Double.isNaN(r))
			throw new LinearNotConvergedException(
				NotConvergedException.DIVERGENCE,
				iter,
				r);

		// Neither convergence nor divergence
		return false;
	}

	/**
	 * @see jmp.iter.linear.AbstractLinearIteration#convergedI(double)
	 */
	protected boolean convergedI(double r) throws NotConvergedException {
		throw new UnsupportedOperationException();
	}

}
