/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.linear;

import jmp.Matrix;
import jmp.Vector;
import jmp.iter.Factory;

/**
 * Multigrid preconditioner.
 * Uses a geometric multigrid, thus the user needs to create matrix operators
 * and restriction operators. The preconditioner supports different kinds
 * of cycles (V, W etc), and the smoothers can be chosen by the user.
 */
public class Multigrid
	extends AbstractPreconditioner
	implements Preconditioner {

	/**
	 * Restriction operators
	 */
	private Matrix[] restrict;

	/**
	 * Matrices on each grid (finer to coarser)
	 */
	private Matrix[] subA;

	/**
	 * Solvers on each grid
	 */
	private Preconditioner[] subSolver;

	/**
	 * Number of pre-, mid, and postsmoothings
	 */
	private int pre, mid, post;

	/**
	 * Number of sublevels
	 */
	private int numLevels;

	/**
	 * Constructor for Multigrid.
	 * @param subA Coefficient matrix for each level (including the finest)
	 * @param restrict Restriction operators
	 * @param subSolver Solvers at each level
	 * @param pre Number of pre-smoothings
	 * @param mid Number of descents (1=V-cycle, 2=W-cycle etc)
	 * @param post Number of post-smoothings
	 */
	public Multigrid(
		Matrix[] subA,
		Matrix[] restrict,
		Preconditioner[] subSolver,
		int pre,
		int mid,
		int post) {
		this.subA = subA;
		this.restrict = restrict;
		this.subSolver = subSolver;
		this.pre = pre;
		this.mid = mid;
		this.post = post;
		numLevels = subA.length;

		check();
	}

	/**
	 * Constructor for Multigrid. Uses pre=mid=post=1.
	 * @param subA Coefficient matrix for each level (including the finest)
	 * @param restrict Restriction operators
	 * @param subSolver Solvers at each level
	 */
	public Multigrid(
		Matrix[] subA,
		Matrix[] restrict,
		Preconditioner[] subSolver) {
		this(subA, restrict, subSolver, 1, 1, 1);
	}

	/**
	 * Constructor for Multigrid. Uses pre=mid=post=1 and GaussSeidel
	 * subsolvers.
	 * @param subA Coefficient matrix for each level (including the finest)
	 * @param restrict Restriction operators
	 */
	public Multigrid(Matrix[] subA, Matrix[] restrict) {
		this.subA = subA;
		this.restrict = restrict;
		pre = mid = post = 1;
		numLevels = subA.length;
		subSolver = new GaussSeidel[numLevels];
		for (int i = 0; i < numLevels; ++i)
			subSolver[i] = new GaussSeidel();

		check();
	}

	/**
	 * Checks for internal consistency
	 */
	private void check() {
		// Check that we have the correct number of everything
		if (subA.length != numLevels)
			throw new ArrayIndexOutOfBoundsException("subA.length != numLevels");
		if (restrict.length != numLevels - 1)
			throw new ArrayIndexOutOfBoundsException("restrict.length != numLevels-1");
		if (subSolver.length != numLevels)
			throw new ArrayIndexOutOfBoundsException("subSolver.length != numLevels");

		// Every subA matrix must be square
		for (int i = 0; i < numLevels; ++i)
			if (!subA[i].isSquare())
				throw new ArrayIndexOutOfBoundsException("subA[i].isSquare() != true");

		// Check that the sizes of the restriction operators are correct
		for (int i = 0; i < numLevels - 1; ++i)
			if (restrict[i].numRows() != subA[i + 1].numRows())
				throw new ArrayIndexOutOfBoundsException("restrict[i].numRows() != subA[i + 1].numRows()");
			else if (restrict[i].numColumns() != subA[i].numColumns())
				throw new ArrayIndexOutOfBoundsException("restrict[i].numColumns() != subA[i - 1].numColumns()");

		// Check that the matrix sizes decrease
		for (int i = 0; i < numLevels - 1; ++i)
			if (subA[i].numRows() < subA[i + 1].numRows())
				throw new ArrayIndexOutOfBoundsException("subA[i].numRows() < subA[i + 1].numRows()");
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#apply(jmp.Matrix, jmp.Vector, jmp.Vector)
	 */
	public Vector apply(Matrix A, Vector b, Vector x) {
		checkSizes(A, b, x);
		cycle(0, b, x);
		return x;
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#transApply(jmp.Matrix, jmp.Vector, jmp.Vector)
	 */
	public Vector transApply(Matrix A, Vector b, Vector x) {
		checkSizes(A, b, x);
		transCycle(0, b, x);
		return x;
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#setup(jmp.Matrix)
	 */
	public void setup(Matrix A) {
		for (int i = 0; i < numLevels; ++i)
			subSolver[i].setup(subA[i]);
	}

	/**
	 * Performs recursive multigrid cycle
	 */
	private void cycle(int level, Vector b, Vector x) {
		// Pre-smoothings
		for (int i = 0; i < pre; ++i)
			subSolver[level].apply(subA[level], b, x);

		// Decend if possible
		if (level < numLevels - 1) {
			Vector r = Factory.createVector(b),
				rs = Factory.createVector(b, restrict[level].numRows()),
				xs = Factory.createVector(b, restrict[level].numRows());

			// Compute residual, and restrict it
			r = BLAS.multAdd(-1., subA[level], x, b, r);
			rs = BLAS.mult(restrict[level], r, rs);

			// Smooth residual
			for (int i = 0; i < mid; ++i)
				cycle(level + 1, rs, xs);

			// Prolongate and apply correction
			x = BLAS.transMultAdd(restrict[level], xs, x);
		}

		// Post-smoothings
		for (int i = 0; i < post; ++i)
			subSolver[level].apply(subA[level], b, x);
	}

	/**
	 * Performs recursive multigrid cycle for transpose problem
	 */
	private void transCycle(int level, Vector b, Vector x) {
		// Pre-smoothings
		for (int i = 0; i < pre; ++i)
			subSolver[level].transApply(subA[level], b, x);

		// Decend if possible
		if (level < numLevels) {
			Vector r = Factory.createVector(b),
				rs = Factory.createVector(b, restrict[level].numRows()),
				xs = Factory.createVector(b, restrict[level].numRows());

			// Compute residual, and restrict it
			r = BLAS.transMultAdd(-1., subA[level], x, b, r);
			rs = BLAS.mult(restrict[level], r, rs);

			// Smooth residual
			for (int i = 0; i < mid; ++i)
				transCycle(level + 1, rs, xs);

			// Prolongate and apply correction
			x = BLAS.transMultAdd(restrict[level], xs, x);
		}

		// Post-smoothings
		for (int i = 0; i < post; ++i)
			subSolver[level].transApply(subA[level], b, x);
	}

}
