/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.linear;

import jmp.Matrix;
import jmp.Vector;
import jmp.iter.Factory;

/**
 * Neumann series preconditioner.
 * Expands the inverse of the matrix into a Neumann series, ie.
 * <code>A<sup>-1</sup>=I + (I-A) + (I-A)<sup>2</sup> + ...</code>. This
 * series only converge if the norm of A is less than 1. If the norm is greater,
 * a scaling factor may be supplied.
 */
public class PolynomialPreconditioner
	extends AbstractPreconditioner
	implements Preconditioner {

	/**
	 * Number of terms to use in the expansion
	 */
	private int terms;

	/**
	 * The matrix I-wA
	 */
	private Matrix Am;

	/**
	 * Scaling factor
	 */
	private double w;

	/**
	 * Creates preconditioner with a default of two terms and scaling equal 1.
	 */
	public PolynomialPreconditioner() {
		setParameters(1., 2);
	}

	/**
	 * Sets parameters
	 * @param w Matrix-scaling
	 * @param terms Number of terms in the expansion
	 */
	public void setParameters(double w, int terms) {
		this.w = w;
		this.terms = terms;
	}

	/**
	 * Sets matrix to contain I-wA
	 */
	public void setMatrix(Matrix Am) {
		this.Am = Am;
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#apply(Matrix, Vector, Vector)
	 */
	public Vector apply(Matrix A, Vector b, Vector x) {
		if (terms < 1)
			return x;

		checkSizes(A, b, x);

		Vector temp = Factory.createVector(b), acc = Factory.createVector(b);

		// The first term
		x = BLAS.copy(b, x);

		// Terms involving multiplication
		for (int i = 1; i < terms; ++i) {

			// Compute (I-wA)^j b
			for (int j = 0; j < i; ++j) {
				temp = BLAS.mult(A, acc, temp);
				acc = BLAS.copy(temp, acc); // acc *= A'
			}

			// Add scaled result into x
			x = BLAS.add(1. / w, acc, x);
		}

		return x;
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#transApply(Matrix, Vector, Vector)
	 */
	public Vector transApply(Matrix A, Vector b, Vector x) {
		if (terms < 1)
			return x;

		checkSizes(A, b, x);

		Vector temp = Factory.createVector(b), acc = Factory.createVector(b);

		// The first term
		x = BLAS.copy(b, x);

		// Terms involving multiplication
		for (int i = 1; i < terms; ++i) {

			// Compute (I-wA)^j b
			acc = BLAS.copy(b, acc);
			for (int j = 0; j < i; ++j) {
				temp = BLAS.transMult(A, acc, temp);
				acc = BLAS.copy(temp, acc); // acc *= A'
			}

			// Add scaled result into x
			x = BLAS.add(1. / w, acc, x);
		}

		return x;
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#setup(jmp.Matrix)
	 */
	public void setup(Matrix A) {
		// Am = I-wA
		if (terms > 1)
			Am = BLAS.addDiagonal(1., BLAS.scale(-w, BLAS.copy(A, Am)));
	}

}
