/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.linear;

import jmp.Matrix;
import jmp.Vector;

/**
 * Preconditioner interface.
 * Preconditioners are approximate solvers to some problem. They may be used
 * in iterative solution procedures to speed up convergence.
 * <p>
 * The only method the user needs to call from this interface is setup.
 * Before using the preconditioner in a LinearSolver, setup must have been
 * called.
 * The advantage of this is that the same internal structure of the
 * preconditioner can be reused for repeated solves with the same matrix.
 * This is particularly important for complicated preconditioners such as
 * ILU or BlockDiagonalPreconditioner.
 * </p>
 */
public interface Preconditioner {

	/**
	 * Solves the approximate problem with the given right hand side.
	 * Result is stored in given vector.
	 * @param A Matrix to precondition
	 * @param b Right hand side of problem
	 * @param x Result is stored here
	 * @return x
	 */
	Vector apply(Matrix A, Vector b, Vector x);

	/**
	 * Solves the approximate transpose problem with the given right hand side.
	 * Result is stored in given vector.
	 * @param A Matrix to precondition
	 * @param b Right hand side of problem
	 * @param x Result is stored here
	 * @return x
	 */
	Vector transApply(Matrix A, Vector b, Vector x);

	/**
	 * Sets up the preconditioner on the given matrix. Must be called before
	 * a system preconditioning is applied.
	 * @param A Matrix to work on
	 */
	void setup(Matrix A);

}
