/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
/*
 * Derived from public domain software at http://www.netlib.org/templates
 */

package jmp.iter.linear;

import jmp.Matrix;
import jmp.Vector;
import jmp.iter.Factory;
import jmp.iter.NotConvergedException;

/**
 * Quasi-Minimal Residual method.
 * QMR solves the unsymmetric linear system <code>Ax = b</code> using the
 * Quasi-Minimal Residual method. QMR uses two preconditioners, and by default
 * these are the same preconditioner.
 * @author Templates
 */
public class QMR extends AbstractLinearSolver implements LinearSolver {

	/**
	 * Left preconditioner
	 */
	private Preconditioner M1;

	/**
	 * Right preconditioner
	 */
	private Preconditioner M2;

	/**
	 * @see jmp.iter.linear.AbstractLinearSolver#AbstractLinearSolver()
	 */
	public QMR() {
		super();
		M1 = M;
		M2 = M;
	}

	/**
	 * @see jmp.iter.linear.AbstractLinearSolver#solveI(jmp.Matrix, jmp.Vector, jmp.Vector)
	 */
	protected void solveI(Matrix A, Vector b, Vector x)
		throws LinearNotConvergedException {
		double rho = 0,
			rho_1 = 0,
			xi = 0,
			gamma = 0,
			gamma_1 = 0,
			theta = 0,
			theta_1 = 0,
			eta = 0,
			delta = 0,
			ep = 0,
			beta = 0;
		Vector[] temp = Factory.createVectors(b, 14);
		Vector r = temp[0],
			y = temp[1],
			z = temp[2],
			v = temp[3],
			w = temp[4],
			p = temp[5],
			q = temp[6],
			d = temp[7],
			s = temp[8],
			v_tld = temp[9],
			w_tld = temp[10],
			y_tld = temp[11],
			z_tld = temp[12],
			p_tld = temp[13];

		r = BLAS.multAdd(-1., A, x, b, r);
		v_tld = BLAS.copy(r, v_tld);
		y = M1.apply(A, v_tld, y);
		rho = BLAS.norm(y, jmp.BLAS.NORM2);
		w_tld = BLAS.copy(r, w_tld);
		z = M2.transApply(A, w_tld, z);
		xi = BLAS.norm(z, jmp.BLAS.NORM2);
		gamma = 1.0;
		eta = -1.0;
		theta = 0.0;

		for (iter.setFirst(); !iter.converged(r, x); iter.next()) {

			if (rho == 0.0)
				throw new LinearNotConvergedException(
					NotConvergedException.BREAKDOWN,
					iter.iterations(),
					"rho",
					BLAS.norm(r, iter.getNormType()));
			if (xi == 0.0)
				throw new LinearNotConvergedException(
					NotConvergedException.BREAKDOWN,
					iter.iterations(),
					"xi",
					BLAS.norm(r, iter.getNormType()));

			v = BLAS.scaleCopy(1. / rho, v_tld, v);
			y = BLAS.scale(1. / rho, y);
			w = BLAS.scaleCopy(1. / xi, w_tld, w);
			z = BLAS.scale(1. / xi, z);
			delta = BLAS.dot(z, y);
			if (delta == 0.0)
				throw new LinearNotConvergedException(
					NotConvergedException.BREAKDOWN,
					iter.iterations(),
					"delta",
					BLAS.norm(r, iter.getNormType()));

			M2.apply(A, y, y_tld);
			M1.transApply(A, z, z_tld);
			if (iter.isFirst()) {
				p = BLAS.copy(y_tld, p);
				q = BLAS.copy(z_tld, q);
			} else {
				p = BLAS.add(y_tld, -xi * delta / ep, p);
				q = BLAS.add(z_tld, -rho * delta / ep, q);
			}

			p_tld = BLAS.mult(A, p, p_tld);
			ep = BLAS.dot(q, p_tld);
			if (ep == 0.0)
				throw new LinearNotConvergedException(
					NotConvergedException.BREAKDOWN,
					iter.iterations(),
					"ep",
					BLAS.norm(r, iter.getNormType()));
			beta = ep / delta;
			if (beta == 0.0)
				throw new LinearNotConvergedException(
					NotConvergedException.BREAKDOWN,
					iter.iterations(),
					"beta",
					BLAS.norm(r, iter.getNormType()));
			v_tld = BLAS.add(-beta, v, p_tld, v_tld);
			M1.apply(A, v_tld, y);
			rho_1 = rho;
			rho = BLAS.norm(y, jmp.BLAS.NORM2);
			w_tld = BLAS.transMultAdd(A, q, -beta, w, w_tld);
			M2.transApply(A, w_tld, z);
			xi = BLAS.norm(z, jmp.BLAS.NORM2);
			gamma_1 = gamma;
			theta_1 = theta;
			theta = rho / (gamma_1 * beta);
			gamma = 1.0 / Math.sqrt(1.0 + theta * theta);
			if (gamma == 0.0)
				throw new LinearNotConvergedException(
					NotConvergedException.BREAKDOWN,
					iter.iterations(),
					"gamma",
					BLAS.norm(r, iter.getNormType()));
			eta = -eta * rho_1 * gamma * gamma / (beta * gamma_1 * gamma_1);
			if (iter.isFirst()) {
				d = BLAS.scaleCopy(eta, p, d);
				s = BLAS.scaleCopy(eta, p_tld, s);
			} else {
				double val = theta_1 * theta_1 * gamma * gamma;
				d = BLAS.add(eta, p, val, d);
				s = BLAS.add(eta, p_tld, val, s);
			}

			// update approximation vector
			x = BLAS.add(d, x);
			r = BLAS.add(-1., s, r);
		}
	}

	/**
	 * Sets left preconditioner
	 */
	public void setM1(Preconditioner preconditioner) {
		M1 = preconditioner;
	}

	/**
	 * Sets right preconditioner
	 */
	public void setM2(Preconditioner preconditioner) {
		M2 = preconditioner;
	}

}
