/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.linear;

import jmp.DenseAccessVector;
import jmp.DenseRowAccessMatrix;
import jmp.DenseRowColumnAccessMatrix;
import jmp.IntDoubleVectorPair;
import jmp.IntIntDoubleVectorTriple;
import jmp.Matrix;
import jmp.SparseRowAccessMatrix;
import jmp.SparseRowColumnAccessMatrix;
import jmp.Vector;

/**
 * SOR preconditioner.
 * Applies one sweep of SOR to the system matrix. Sequential. For best
 * performance, a good choice of the overrelaxation parameter omega must be
 * made (0 &lt; omega &lt; 2). Does not perform transpose operations.
 */
public class SOR extends AbstractPreconditioner implements Preconditioner {

	/**
	 * Overrelaxation parameter
	 */
	private double omega;

	/**
	 * Constructor for SOR. Uses omega=1 by default.
	 */
	public SOR() {
		omega = 1;
	}

	/**
	 * Sets over-relaxation parameter.
	 */
	public void setParameters(double omega) {
		this.omega = omega;
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#apply(jmp.Matrix, jmp.Vector, jmp.Vector)
	 */
	public Vector apply(Matrix A, Vector b, Vector x) {
		checkSizes(A, b, x);

		if (!(b instanceof DenseAccessVector)
			|| !(x instanceof DenseAccessVector))
			throw new UnsupportedOperationException();

		double[] bdata = ((DenseAccessVector) b).getVector(),
			xdata = ((DenseAccessVector) x).getVector();

		if (A instanceof SparseRowAccessMatrix)
			sor((SparseRowAccessMatrix) A, bdata, xdata);
		else if (A instanceof SparseRowColumnAccessMatrix)
			sor((SparseRowColumnAccessMatrix) A, bdata, xdata);
		else if (A instanceof DenseRowAccessMatrix)
			sor((DenseRowAccessMatrix) A, bdata, xdata);
		else if (A instanceof DenseRowColumnAccessMatrix)
			sor((DenseRowColumnAccessMatrix) A, bdata, xdata);
		else
			throw new UnsupportedOperationException();

		return x;
	}

	private void sor(SparseRowAccessMatrix A, double[] bdata, double[] xdata) {
		for (int i = 0; i < A.numRows(); ++i) {
			IntDoubleVectorPair Arow = A.getRow(i);

			double newVal = 0, diagVal = 0.;
			for (int j = 0; j < Arow.index.length; ++j)
				if (Arow.index[j] != i)
					newVal += Arow.data[j] * xdata[Arow.index[j]];
				else
					diagVal = Arow.data[j];
			newVal = (bdata[i] - newVal) / diagVal;
			xdata[i] += omega * (newVal - xdata[i]);
		}
	}

	private void sor(
		SparseRowColumnAccessMatrix A,
		double[] bdata,
		double[] xdata) {
		IntIntDoubleVectorTriple Amat = A.getMatrix();

		for (int i = 0; i < A.numRows(); ++i) {

			double newVal = 0, diagVal = 0.;
			for (int j = Amat.minor[i]; j < Amat.minor[i + 1]; ++j)
				if (Amat.major[j] != i)
					newVal += Amat.data[j] * xdata[Amat.major[j]];
				else
					diagVal = Amat.data[j];
			newVal = (bdata[i] - newVal) / diagVal;
			xdata[i] += omega * (newVal - xdata[i]);
		}
	}

	private void sor(DenseRowAccessMatrix A, double[] bdata, double[] xdata) {
		for (int i = 0; i < A.numRows(); ++i) {
			double[] Arow = A.getRow(i);

			double newVal = 0, diagVal = 0.;
			for (int j = 0; j < Arow.length; ++j)
				if (j != i)
					newVal += Arow[j] * xdata[j];
				else
					diagVal = Arow[j];
			newVal = (bdata[i] - newVal) / diagVal;
			xdata[i] += omega * (newVal - xdata[i]);
		}
	}

	private void sor(
		DenseRowColumnAccessMatrix A,
		double[] bdata,
		double[] xdata) {
		IntDoubleVectorPair Amat = A.getMatrix();
		for (int i = 0; i < A.numRows(); ++i) {

			double newVal = 0, diagVal = 0.;
			for (int j = Amat.index[i], k = 0; j < Amat.index[i + 1]; ++j, ++k)
				if (k != i)
					newVal += Amat.data[j] * xdata[k];
				else
					diagVal = Amat.data[j];
			newVal = (bdata[i] - newVal) / diagVal;
			xdata[i] += omega * (newVal - xdata[i]);
		}
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#transApply(jmp.Matrix, jmp.Vector, jmp.Vector)
	 */
	public Vector transApply(Matrix A, Vector b, Vector x) {
		throw new UnsupportedOperationException();
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#setup(jmp.Matrix)
	 */
	public void setup(Matrix A) {
	}

}
