/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.linear;

import java.util.Arrays;

import jmp.DenseAccessVector;
import jmp.DenseRowAccessMatrix;
import jmp.DenseRowColumnAccessMatrix;
import jmp.IntDoubleVectorPair;
import jmp.IntIntDoubleVectorTriple;
import jmp.Matrix;
import jmp.SparseRowAccessMatrix;
import jmp.SparseRowColumnAccessMatrix;
import jmp.Vector;

/**
 * SSOR preconditioner.
 * Uses a symmetrical SOR as a preconditioner. Meant for symmetrical matrices.
 * For best performance, omega must be carefully chosen (between 0 and 2).
 */
public class SSOR extends AbstractPreconditioner implements Preconditioner {

	/**
	 * Overrelaxation parameter
	 */
	private double omega;

	/**
	 * All diagonal entries of the matrix
	 */
	private double[] diag;

	/**
	 * Indices to the diagonal entries of the matrix
	 */
	private int[] diagInd;

	/**
	 * Constructor for SSOR. Uses omega=1 by default.
	 */
	public SSOR() {
		omega = 1;
	}

	/**
	 * Sets over-relaxation parameter.
	 */
	public void setParameters(double omega) {
		this.omega = omega;
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#apply(Matrix, Vector, Vector)
	 */
	public Vector apply(Matrix A, Vector b, Vector x) {
		checkSizes(A, b, x);

		if (!(x instanceof DenseAccessVector))
			throw new UnsupportedOperationException();

		// Copy b over to x, and get a dense array representation
		x = BLAS.copy(b, x);
		double[] xdata = ((DenseAccessVector) x).getVector();

		if (A instanceof SparseRowAccessMatrix)
			ssor((SparseRowAccessMatrix) A, xdata);
		else if (A instanceof SparseRowColumnAccessMatrix)
			ssor((SparseRowColumnAccessMatrix) A, xdata);
		else if (A instanceof DenseRowAccessMatrix)
			ssor((DenseRowAccessMatrix) A, xdata);
		else if (A instanceof DenseRowColumnAccessMatrix)
			ssor((DenseRowColumnAccessMatrix) A, xdata);
		else
			throw new UnsupportedOperationException();

		return x;
	}

	private void ssor(SparseRowAccessMatrix A, double[] xdata) {

		int n = A.numRows();

		// M = (D+L) D^{-1} (D+L)^T

		// Solves (1/omega)*(D+L) y = b
		for (int i = 0; i < n; ++i) {

			// Do nothing if we get a divide by zero
			if (Math.abs(diag[i]) == 0.0)
				continue;

			IntDoubleVectorPair Arow = A.getRow(i);
			double[] Adat = Arow.data;
			int[] Aind = Arow.index;

			double sum = 0;
			for (int j = 0; j < Aind.length && Aind[j] < i; ++j)
				sum += Adat[j] * xdata[Aind[j]];
			xdata[i] = (omega / diag[i]) * (xdata[i] - sum);
		}

		// Solves (omega/(2-omega))*D^{-1} z = y
		for (int i = 0; i < n; ++i) {

			// No need to do anything
			if (Math.abs(diag[i]) == 0.0)
				continue;

			xdata[i] = (2 - omega) / omega * diag[i] * xdata[i];
		}

		// Solves (1/omega)*(D+L)^T x = z
		for (int i = n - 1; i >= 0; --i) {

			// Do nothing if we get a divide by zero
			if (Math.abs(diag[i]) == 0.0)
				continue;

			IntDoubleVectorPair Arow = A.getRow(i);
			double[] Adat = Arow.data;
			int[] Aind = Arow.index;

			double sum = 0;
			for (int j = diagInd[i] + 1; j < Aind.length; ++j)
				sum += Adat[j] * xdata[Aind[j]];
			xdata[i] = (omega / diag[i]) * (xdata[i] - sum);
		}
	}

	private void ssor(SparseRowColumnAccessMatrix A, double[] xdata) {

		int n = A.numRows();

		IntIntDoubleVectorTriple Amat = A.getMatrix();

		// M = (D+L) D^{-1} (D+L)^T

		// Solves (1/omega)*(D+L) y = b
		for (int i = 0; i < n; ++i) {

			// Do nothing if we get a divide by zero
			if (Math.abs(diag[i]) == 0.0)
				continue;

			double sum = 0;
			for (int j = Amat.minor[i];
				j < Amat.minor[i + 1] && Amat.major[j] < i;
				++j)
				sum += Amat.data[j] * xdata[Amat.major[j]];
			xdata[i] = (omega / diag[i]) * (xdata[i] - sum);
		}

		// Solves (omega/(2-omega))*D^{-1} z = y
		for (int i = 0; i < n; ++i) {

			// No need to do anything
			if (Math.abs(diag[i]) == 0.0)
				continue;

			xdata[i] = (2 - omega) / omega * diag[i] * xdata[i];
		}

		// Solves (1/omega)*(D+L)^T x = z
		for (int i = n - 1; i >= 0; --i) {

			// Do nothing if we get a divide by zero
			if (Math.abs(diag[i]) == 0.0)
				continue;

			double sum = 0;
			for (int j = diagInd[i] + 1; j < Amat.minor[i + 1]; ++j)
				sum += Amat.data[j] * xdata[Amat.major[j]];
			xdata[i] = (omega / diag[i]) * (xdata[i] - sum);
		}
	}

	private void ssor(DenseRowAccessMatrix A, double[] xdata) {

		int n = A.numRows();

		// M = (D+L) D^{-1} (D+L)^T

		// Solves (1/omega)*(D+L) y = b
		for (int i = 0; i < n; ++i) {

			// Do nothing if we get a divide by zero
			if (Math.abs(diag[i]) == 0.0)
				continue;

			double[] Arow = A.getRow(i);

			double sum = 0;
			for (int j = 0; j < i; ++j)
				sum += Arow[j] * xdata[j];
			xdata[i] = (omega / diag[i]) * (xdata[i] - sum);
		}

		// Solves (omega/(2-omega))*D^{-1} z = y
		for (int i = 0; i < n; ++i) {

			// No need to do anything
			if (Math.abs(diag[i]) == 0.0)
				continue;

			xdata[i] = (2 - omega) / omega * diag[i] * xdata[i];
		}

		// Solves (1/omega)*(D+L)^T x = z
		for (int i = n - 1; i >= 0; --i) {

			// Do nothing if we get a divide by zero
			if (Math.abs(diag[i]) == 0.0)
				continue;

			double[] Arow = A.getRow(i);

			double sum = 0;
			for (int j = i + 1; j < n; ++j)
				sum += Arow[j] * xdata[j];
			xdata[i] = (omega / diag[i]) * (xdata[i] - sum);
		}
	}

	private void ssor(DenseRowColumnAccessMatrix A, double[] xdata) {

		int n = A.numRows();

		IntDoubleVectorPair Amat = A.getMatrix();

		// M = (D+L) D^{-1} (D+L)^T

		// Solves (1/omega)*(D+L) y = b
		for (int i = 0; i < n; ++i) {

			// Do nothing if we get a divide by zero
			if (Math.abs(diag[i]) == 0.0)
				continue;

			double sum = 0;
			for (int j = Amat.index[i], k = 0; j < diagInd[i]; ++j, ++k)
				sum += Amat.data[j] * xdata[k];
			xdata[i] = (omega / diag[i]) * (xdata[i] - sum);
		}

		// Solves (omega/(2-omega))*D^{-1} z = y
		for (int i = 0; i < n; ++i) {

			// No need to do anything
			if (Math.abs(diag[i]) == 0.0)
				continue;

			xdata[i] = (2 - omega) / omega * diag[i] * xdata[i];
		}

		// Solves (1/omega)*(D+L)^T x = z
		for (int i = n - 1; i >= 0; --i) {

			// Do nothing if we get a divide by zero
			if (Math.abs(diag[i]) == 0.0)
				continue;

			double sum = 0;
			for (int j = diagInd[i] + 1, k = i + 1; j < n; ++j, ++k)
				sum += Amat.data[j] * xdata[k];
			xdata[i] = (omega / diag[i]) * (xdata[i] - sum);
		}
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#transApply(Matrix, Vector, Vector)
	 */
	public Vector transApply(Matrix A, Vector b, Vector x) {
		return apply(A, b, x);
	}

	/**
	 * @see jmp.iter.linear.Preconditioner#setup(jmp.Matrix)
	 */
	public void setup(Matrix A) {
		int n = A.numRows();
		diag = new double[n];
		diagInd = new int[n];

		// Get the diagonal and its indices
		if (A instanceof SparseRowAccessMatrix) {
			for (int i = 0; i < n; ++i) {
				IntDoubleVectorPair Arow =
					((SparseRowAccessMatrix) A).getRow(i);
				int ind = Arrays.binarySearch(Arow.index, i);
				if (ind < 0)
					throw new IllegalArgumentException("Diagonal not present");
				diag[i] = Arow.data[ind];
				diagInd[i] = ind;
			}
		} else if (A instanceof SparseRowColumnAccessMatrix) {
			IntIntDoubleVectorTriple Amat =
				((SparseRowColumnAccessMatrix) A).getMatrix();
			for (int i = 0; i < n; ++i) {
				int ind =
					jmp.util.Arrays.binarySearch(
						Amat.major,
						i,
						Amat.minor[i],
						Amat.minor[i + 1]);
				if (ind < 0)
					throw new IllegalArgumentException("Diagonal not present");
				diag[i] = Amat.data[ind];
				diagInd[i] = ind;
			}
		} else if (A instanceof DenseRowColumnAccessMatrix) {
			IntDoubleVectorPair Amat =
				((DenseRowColumnAccessMatrix) A).getMatrix();
			for (int i = 0; i < n; ++i) {
				diagInd[i] = Amat.index[i] + i;
				diag[i] = Amat.data[diagInd[i]];
			}
		} else if (A instanceof DenseRowAccessMatrix) {
			for (int i = 0; i < n; ++i) {
				double[] Arow = ((DenseRowAccessMatrix) A).getRow(i);
				diagInd[i] = i;
				diag[i] = Arow[i];
			}
		} else
			throw new UnsupportedOperationException();

	}

}
