/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.mixed;

import jmp.BLAS;
import jmp.Matrix;
import jmp.State;
import jmp.Vector;
import jmp.iter.linear.CG;
import jmp.iter.linear.DefaultLinearIteration;
import jmp.iter.linear.LinearIteration;
import jmp.iter.linear.LinearNotConvergedException;
import jmp.iter.linear.LinearSolver;

/**
 * Partial implementation of MixedSolver
 */
public abstract class AbstractMixedSolver implements MixedSolver {

	/**
	 * Local reference
	 */
	protected BLAS BLAS;

	/**
	 * LinearSolver to use
	 */
	protected LinearSolver solver;

	/**
	 * Convergence tester
	 */
	protected LinearIteration iter;

	/**
	 * Constructor for AbstractMixedSolver, with CG as solver and default
	 * linear iteration object.
	 */
	public AbstractMixedSolver() {
		this.iter = new DefaultLinearIteration();
		this.solver = new CG();
		BLAS = State.BLAS;
	}

	/**
	 * @see jmp.iter.mixed.MixedSolver#setSubSolver(jmp.iter.linear.LinearSolver)
	 */
	public void setSubSolver(LinearSolver solver) {
		this.solver = solver;
	}

	/**
	 * @see jmp.iter.mixed.MixedSolver#getSubSolver()
	 */
	public LinearSolver getSubSolver() {
		return solver;
	}

	/**
	 * @see jmp.iter.mixed.MixedSolver#getIteration()
	 */
	public LinearIteration getIteration() {
		return iter;
	}

	/**
	 * @see jmp.iter.mixed.MixedSolver#setIteration(jmp.iter.linear.LinearIteration)
	 */
	public void setIteration(LinearIteration iter) {
		this.iter = iter;
	}

	/**
	 * @see jmp.iter.mixed.MixedSolver#solve(jmp.Matrix, jmp.Matrix, jmp.Matrix, jmp.Matrix, jmp.Vector, jmp.Vector, jmp.Vector, jmp.Vector)
	 */
	public void solve(
		Matrix A,
		Matrix B,
		Matrix Bt,
		Matrix C,
		Vector q,
		Vector u,
		Vector f,
		Vector g)
		throws LinearNotConvergedException {
		checkSolveArguments(A, B, Bt, C, q, u, f, g);
		solveI(A, B, Bt, C, q, u, f, g);
	}

	/**
	 * Checks that all sizes conform, or else throws an exception.
	 */
	private void checkSolveArguments(
		Matrix A,
		Matrix B,
		Matrix Bt,
		Matrix C,
		Vector q,
		Vector u,
		Vector f,
		Vector g) {

		int m = B.numRows();
		int n = B.numColumns();

		if (n != A.numColumns())
			throw new ArrayIndexOutOfBoundsException("B.numColumns() != A.numColumns()");
		if (q.size() != n)
			throw new ArrayIndexOutOfBoundsException("q.size() != B.numColumns()");
		if (u.size() != m)
			throw new ArrayIndexOutOfBoundsException("u.size() != B.numRows()");
		if (f.size() != n)
			throw new ArrayIndexOutOfBoundsException("f.size() != B.numColumns()");
		if (g.size() != m)
			throw new ArrayIndexOutOfBoundsException("g.size() != B.numRows()");
		if (Bt.numRows() != n || Bt.numColumns() != m)
			throw new ArrayIndexOutOfBoundsException("Bt.numRows() != B.numColumns() || Bt.numColumns() != B.numRows()");
		if (!A.isSquare())
			throw new ArrayIndexOutOfBoundsException("!A.isSquare()");
		if (!C.isSquare())
			throw new ArrayIndexOutOfBoundsException("!C.isSquare()");
		if (C.numRows() != m)
			throw new ArrayIndexOutOfBoundsException("C.numRows() != B.numRows()");
	}

	protected abstract void solveI(
		Matrix A,
		Matrix B,
		Matrix Bt,
		Matrix C,
		Vector q,
		Vector u,
		Vector f,
		Vector g)
		throws LinearNotConvergedException;

}
