/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.mixed;

import jmp.Matrix;
import jmp.Vector;
import jmp.iter.Factory;
import jmp.iter.linear.LinearNotConvergedException;

/**
 * Uzawa algorithm using the CG method
 */
public class CGUzawa extends AbstractMixedSolver {

	/**
	 * @see jmp.iter.mixed.AbstractMixedSolver#solveI(jmp.Matrix, jmp.Matrix, jmp.Matrix, jmp.Matrix, jmp.Vector, jmp.Vector, jmp.Vector, jmp.Vector)
	 */
	protected void solveI(
		Matrix A,
		Matrix B,
		Matrix Bt,
		Matrix C,
		Vector q,
		Vector u,
		Vector f,
		Vector g)
		throws LinearNotConvergedException {
		Vector[] tempq = Factory.createVectors(q, 3);
		Vector[] tempu = Factory.createVectors(u, 4);
		Vector ur = tempq[0], p = tempq[1], h = tempq[2];
		Vector r = tempu[0], rhat = tempu[1], d = tempu[2], t = tempu[3];

		ur = BLAS.multAdd(-1., Bt, u, f, ur); // ur = f - Bt*u
		q = solver.solve(A, ur, q); // q = A \ ur = A \ (f - Bt*u)

		t = BLAS.multAdd(-1., C, u, g, t); // t = g - C*u
		r = BLAS.multAdd(-1., B, q, t, r); // r = t - B*q = g - B*q - C*u
		d = BLAS.scaleCopy(-1., r, d); // d = -r

		for (iter.setFirst(); !iter.converged(r); iter.next()) {
			p = BLAS.mult(Bt, d, p); // p = Bt*d
			h = solver.solve(A, p, h); // h = A\p

			double alpha = BLAS.dot(r, r) / BLAS.dot(p, h);
			u = BLAS.add(alpha, d, u); // u = u + alpha*d
			q = BLAS.add(-alpha, h, q); // q = q - alpha*h

			rhat = BLAS.copy(r, rhat); // rhat = r
			t = BLAS.multAdd(-1., C, u, g, t); // t = g - C*u
			r = BLAS.multAdd(-1., B, q, t, r); // r = t - B*q = g - B*q - C*u
			double beta = BLAS.dot(r, r) / BLAS.dot(rhat, rhat);
			d = BLAS.add(-1., r, beta, d); // d = -r + beta*d
		}
	}

}
