/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.iter.mixed;

import jmp.Matrix;
import jmp.Vector;
import jmp.iter.Factory;
import jmp.iter.linear.LinearNotConvergedException;

/**
 * Uzawa algorithm for mixed problems. Requires a steplength, but uses length of
 * <code>1</code> by default.
 */
public class Uzawa extends AbstractMixedSolver {

	/**
	 * Steplength (fixed)
	 */
	private double alpha;

	/**
	 * Constructor for Uzawa. Uses steplength of 1 by default.
	 */
	public Uzawa() {
		alpha = 1.;
	}

	/**
	 * Sets steplength to use in the algorithm
	 * @param alpha
	 */
	public void setSteplength(double alpha) {
		this.alpha = alpha;
	}

	/**
	 * @see jmp.iter.mixed.AbstractMixedSolver#solveI(jmp.Matrix, jmp.Matrix, jmp.Matrix, jmp.Matrix, jmp.Vector, jmp.Vector, jmp.Vector, jmp.Vector)
	 */
	protected void solveI(
		Matrix A,
		Matrix B,
		Matrix Bt,
		Matrix C,
		Vector q,
		Vector u,
		Vector f,
		Vector g)
		throws LinearNotConvergedException {
		Vector[] tempq = Factory.createVectors(q, 1);
		Vector[] tempu = Factory.createVectors(u, 2);
		Vector qr = tempq[0], r = tempu[0], t = tempu[1];

		t = BLAS.multAdd(C, u, -1., g, t); // t = C*u - g
		r = BLAS.multAdd(B, q, t, r); // r = B*q + t

		for (iter.setFirst(); !iter.converged(r); iter.next()) {
			qr = BLAS.multAdd(-1., Bt, u, f, qr); // qr = f - Bt*u
			q = solver.solve(A, qr, q); // q = A\qr

			t = BLAS.multAdd(C, u, -1., g, t); // t = C*u - g
			r = BLAS.multAdd(B, q, t, r); // r = B*q + t
			u = BLAS.add(alpha, r, u); // u = u + alpha*r
		}
	}

}
