/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.test;

import java.util.Random;

import jmp.DenseVector;
import jmp.ElementalAccessMatrix;
import jmp.ElementalAccessVector;
import jmp.Matrix;
import jmp.SparseColumnMatrix;
import jmp.SparseColumnRowMatrix;
import jmp.SparseRowColumnMatrix;
import jmp.SparseRowMatrix;
import jmp.State;
import jmp.Vector;
import jmp.iter.linear.BiCG;
import jmp.iter.linear.BiCGstab;
import jmp.iter.linear.CG;
import jmp.iter.linear.CGS;
import jmp.iter.linear.DefaultLinearIteration;
import jmp.iter.linear.DiagonalPreconditioner;
import jmp.iter.linear.GMRES;
import jmp.iter.linear.ILU;
import jmp.iter.linear.IdentityPreconditioner;
import jmp.iter.linear.LinearIteration;
import jmp.iter.linear.LinearSolver;
import jmp.iter.linear.Preconditioner;
import jmp.iter.linear.QMR;
import junit.framework.TestCase;

/**
 * Test of the iterative linear solvers
 */
public class LinearSolverTest extends TestCase {

	private static int nmax = 100;

	/**
	 * Constructor for LinearSolverTest.
	 * @param name
	 */
	public LinearSolverTest(String name) {
		super(name);
	}

	public void testSparseColumnRowMatrix() {
		Random r = new Random();
		int n = Math.max(Utilities.getInt(nmax, r), 4);

		ElementalAccessMatrix A = new SparseColumnRowMatrix(n, n, 3);
		ElementalAccessVector b = new DenseVector(n), x = new DenseVector(n);

		test(A, b, x);
	}

	public void testSparseColumnMatrix() {
		Random r = new Random();
		int n = Math.max(Utilities.getInt(nmax, r), 4);

		ElementalAccessMatrix A = new SparseColumnMatrix(n, n, 3);
		ElementalAccessVector b = new DenseVector(n), x = new DenseVector(n);

		test(A, b, x);
	}

	public void testSparseRowColumnMatrix() {
		Random r = new Random();
		int n = Math.max(Utilities.getInt(nmax, r), 4);

		ElementalAccessMatrix A = new SparseRowColumnMatrix(n, n, 3);
		ElementalAccessVector b = new DenseVector(n), x = new DenseVector(n);

		test(A, b, x);
	}

	public void testSparseRowMatrix() {
		Random r = new Random();
		int n = Math.max(Utilities.getInt(nmax, r), 4);

		ElementalAccessMatrix A = new SparseRowMatrix(n, n, 3);
		ElementalAccessVector b = new DenseVector(n), x = new DenseVector(n);

		test(A, b, x);
	}

	private void test(
		ElementalAccessMatrix A,
		ElementalAccessVector b,
		ElementalAccessVector x) {

		// Test of usual solvers and preconditoners
		LinearSolver[] solver =
			new LinearSolver[] {
				new BiCG(),
				new BiCGstab(),
				new CG(),
				new CGS(),
				new GMRES(),
				new QMR()};

		Preconditioner[] M =
			new Preconditioner[] {
				new IdentityPreconditioner(),
				new DiagonalPreconditioner(),
				new ILU(new SparseRowMatrix(A.numRows(), A.numColumns()))};

		DefaultLinearIteration iter = new DefaultLinearIteration();
		iter.setParameters(1e-10, 1e-50, 1e+5, 1000000);

		for (int i = 0; i < solver.length; ++i)
			for (int j = 0; j < M.length; ++j)
				for (int k = 1; k <= 4; ++k) {
					State.setNumThreads(k);
					test(A, b, x, solver[i], M[j], iter);
				}
	}

	private void test(
		ElementalAccessMatrix A,
		ElementalAccessVector b,
		ElementalAccessVector x,
		LinearSolver solver,
		Preconditioner M,
		LinearIteration iter) {
		Random r = new Random();
		double K = r.nextDouble(),
			f = r.nextDouble(),
			gl = r.nextDouble(),
			gr = r.nextDouble();

		assemble(A, b, K, f);
		boundary(A, b, x, gl, gr);
		double[] ans = solve(A, b, x, solver, M, iter);
		double[] ref = reference(b.size() - 1, K, f, gl, gr);

		checkEqual(ans, ref);

		State.BLAS.zero(A);
		State.BLAS.zero(b);
		State.BLAS.zero(x);
	}

	private double[] reference(
		int n,
		double K,
		double f,
		double gl,
		double gr) {
		double[] ret = new double[n + 1];
		double x = 0, dx = 1. / ((double) n);

		ret[0] = gl;

		for (int i = 1; i < n; ++i) {
			x += dx;
			ret[i] =
				(-f / (2. * K)) * x * x + (gr - gl + f / (2. * K)) * x + gl;
		}

		ret[n] = gr;

		return ret;
	}

	private void checkEqual(double[] A, double[] B) {
		assertTrue(A.length == B.length);
		for (int i = 0; i < A.length; ++i)
			assertEquals(B[i], A[i], 1e-8);
	}

	private void assemble(
		ElementalAccessMatrix A,
		ElementalAccessVector b,
		double K,
		double f) {
		int n = b.size() - 1;
		double h = 1. / ((double) n);

		double Ae[][] = { { K / h, -K / h }, {
				-K / h, K / h }
		};
		double be[] = { f * h / 2., f * h / 2. };

		for (int i = 0; i < n; ++i) {
			int ind[] = { i, i + 1 };
			A.addValues(ind, ind, Ae);
			b.addValues(ind, be);
		}
	}

	private void boundary(
		ElementalAccessMatrix A,
		ElementalAccessVector b,
		ElementalAccessVector x,
		double gl,
		double gr) {
		int n = b.size() - 1;
		int boundary[] = { 0, n };
		double boundaryV[] = { gl, gr };

		x.setValues(boundary, boundaryV);
		b.setValues(boundary, boundaryV);

		A.setValue(0, 0, 1);
		A.setValue(0, 1, 0);
		A.setValue(n, n, 1);
		A.setValue(n, n - 1, 0);
	}

	private double[] solve(
		Matrix A,
		Vector b,
		Vector x,
		LinearSolver solver,
		Preconditioner M,
		LinearIteration iter) {
		M.setup(A);
		solver.setPreconditioner(M);
		solver.setIteration(iter);
		return State.BLAS.getArrayCopy(solver.solve(A, b, x));
	}

}
