/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.test;

import java.util.Random;

import jmp.CoordinateMatrix;
import jmp.DenseColumnMatrix;
import jmp.DenseColumnRowMatrix;
import jmp.DenseRowColumnMatrix;
import jmp.DenseRowMatrix;
import jmp.ElementalAccessMatrix;
import jmp.SparseColumnMatrix;
import jmp.SparseColumnRowMatrix;
import jmp.SparseRowColumnMatrix;
import jmp.SparseRowMatrix;
import jmp.State;
import junit.framework.TestCase;

/**
 * Test matrix assembly
 */
public class MatrixAssemblyTest extends TestCase {

	private static int nmax = 100, mmax = 100, numax = 10, repeat = 3;

	/**
	 * Constructor for MatrixAssemblyTest.
	 * @param arg0
	 */
	public MatrixAssemblyTest(String arg0) {
		super(arg0);
	}

	private int getNu(int n, int m, Random r) {
		return Math.min(n, Math.min(m, Utilities.getInt(numax, r)));
	}

	public void testSparseRowColumnMatrix() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new SparseRowColumnMatrix(n, m, nu);
			check(A, Utilities.setAssembleRowMatrix(A, nu));

			A = new SparseRowColumnMatrix(n, m, nu);
			check(A, Utilities.addAssembleRowMatrix(A, nu));
		}
	}

	public void testSparseRowMatrix() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		ElementalAccessMatrix A = new SparseRowMatrix(n, m);
		repeatedAssembly(A, nu);
	}

	public void testDenseRowColumnMatrix() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		ElementalAccessMatrix A = new DenseRowColumnMatrix(n, m);
		repeatedAssembly(A, nu);
	}

	public void testDenseRowMatrix() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		ElementalAccessMatrix A = new DenseRowMatrix(n, m);
		repeatedAssembly(A, nu);
	}

	public void testSparseColumnRowMatrix() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < 10; ++i) {
			ElementalAccessMatrix A = new SparseColumnRowMatrix(n, m, nu);
			check(A, Utilities.setAssembleColumnMatrix(A, nu));

			A = new SparseColumnRowMatrix(n, m, nu);
			check(A, Utilities.addAssembleColumnMatrix(A, nu));
		}
	}

	public void testSparseColumnMatrix() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		ElementalAccessMatrix A = new SparseColumnMatrix(n, m);
		repeatedAssembly(A, nu);
	}

	public void testDenseColumnRowMatrix() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		ElementalAccessMatrix A = new DenseColumnRowMatrix(n, m);
		repeatedAssembly(A, nu);
	}

	public void testDenseColumnMatrix() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		ElementalAccessMatrix A = new DenseColumnMatrix(n, m);
		repeatedAssembly(A, nu);
	}

	public void testCoordinateMatrix() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		ElementalAccessMatrix A = new CoordinateMatrix(n, m, nu);
		repeatedAssembly(A, nu);
	}

	private void checkEqual(double[][] A, double[][] B) {
		assertTrue(A.length == B.length);
		for (int i = 0; i < A.length; ++i) {
			assertTrue(A[i].length == B[i].length);
			for (int j = 0; j < A[i].length; ++j)
				assertEquals(B[i][j], A[i][j], 1e-12);
		}
	}

	private void check(ElementalAccessMatrix A, double[][] real) {
		checkEqual(Utilities.getElMatrix(A), real);
		checkEqual(Utilities.getMatrix(A), real);
		checkEqual(Utilities.getMatrixCopy(A), real);
	}

	private void repeatedAssembly(ElementalAccessMatrix A, int nu) {
		for (int i = 0; i < repeat; ++i) {
			check(A, Utilities.setAssembleRowMatrix(A, nu));
			State.BLAS.zero(A);

			check(A, Utilities.setsAssembleRowMatrix(A, nu));
			State.BLAS.zero(A);

			check(A, Utilities.addAssembleRowMatrix(A, nu));
			State.BLAS.zero(A);

			check(A, Utilities.addsAssembleRowMatrix(A, nu));
			State.BLAS.zero(A);

			check(A, Utilities.setAssembleColumnMatrix(A, nu));
			State.BLAS.zero(A);

			check(A, Utilities.setsAssembleColumnMatrix(A, nu));
			State.BLAS.zero(A);

			check(A, Utilities.addAssembleColumnMatrix(A, nu));
			State.BLAS.zero(A);

			check(A, Utilities.addsAssembleColumnMatrix(A, nu));
			State.BLAS.zero(A);
		}
	}

}
