/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.test;

import java.util.Random;

import jmp.DenseColumnMatrix;
import jmp.DenseColumnRowMatrix;
import jmp.DenseRowColumnMatrix;
import jmp.DenseRowMatrix;
import jmp.DenseVector;
import jmp.ElementalAccessMatrix;
import jmp.ElementalAccessVector;
import jmp.SparseColumnMatrix;
import jmp.SparseColumnRowMatrix;
import jmp.SparseRowColumnMatrix;
import jmp.SparseRowMatrix;
import jmp.State;
import junit.framework.TestCase;

/**
 * Test matrix multiplication. Both normal and transpose
 */
public class MatrixMultiplyTest extends TestCase {

	private static int nmax = 100,
		mmax = 100,
		numax = 10,
		repeat = 3,
		threads = 4;

	/**
	 * Constructor for MatrixMultiplyTest
	 * @param arg0
	 */
	public MatrixMultiplyTest(String arg0) {
		super(arg0);
	}

	public void testDenseRowColumnMult() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new DenseRowColumnMatrix(n, m);
			ElementalAccessVector x = new DenseVector(m),
				y = new DenseVector(n),
				z = new DenseVector(n);

			rowCheck(A, x, y, z, nu);
		}
	}

	public void testDenseRowMult() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new DenseRowMatrix(n, m);
			ElementalAccessVector x = new DenseVector(m),
				y = new DenseVector(n),
				z = new DenseVector(n);

			rowCheck(A, x, y, z, nu);
		}
	}

	public void testDenseColumnRowMult() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new DenseColumnRowMatrix(n, m);
			ElementalAccessVector x = new DenseVector(m),
				y = new DenseVector(n),
				z = new DenseVector(n);

			columnCheck(A, x, y, z, nu);
		}
	}

	public void testDenseColumnMult() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new DenseColumnMatrix(n, m);
			ElementalAccessVector x = new DenseVector(m),
				y = new DenseVector(n),
				z = new DenseVector(n);

			columnCheck(A, x, y, z, nu);
		}
	}

	public void testSparseRowColumnMult() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new SparseRowColumnMatrix(n, m, nu);
			ElementalAccessVector x = new DenseVector(m),
				y = new DenseVector(n),
				z = new DenseVector(n);

			rowCheck(A, x, y, z, nu);
		}
	}

	public void testSparseRowMult() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new SparseRowMatrix(n, m, nu);
			ElementalAccessVector x = new DenseVector(m),
				y = new DenseVector(n),
				z = new DenseVector(n);

			rowCheck(A, x, y, z, nu);
		}
	}

	public void testSparseColumnRowMult() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new SparseColumnRowMatrix(n, m, nu);
			ElementalAccessVector x = new DenseVector(m),
				y = new DenseVector(n),
				z = new DenseVector(n);

			columnCheck(A, x, y, z, nu);
		}
	}

	public void testSparseColumnMult() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new SparseColumnMatrix(n, m, nu);
			ElementalAccessVector x = new DenseVector(m),
				y = new DenseVector(n),
				z = new DenseVector(n);

			columnCheck(A, x, y, z, nu);
		}
	}

	public void testDenseRowColumnTransMult() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new DenseRowColumnMatrix(n, m);
			ElementalAccessVector x = new DenseVector(n),
				y = new DenseVector(m),
				z = new DenseVector(m);

			transRowCheck(A, x, y, z, nu);
		}
	}

	public void testDenseRowTransMult() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new DenseRowMatrix(n, m);
			ElementalAccessVector x = new DenseVector(n),
				y = new DenseVector(m),
				z = new DenseVector(m);

			transRowCheck(A, x, y, z, nu);
		}
	}

	public void testDenseColumnRowTransMult() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new DenseColumnRowMatrix(n, m);
			ElementalAccessVector x = new DenseVector(n),
				y = new DenseVector(m),
				z = new DenseVector(m);

			transColumnCheck(A, x, y, z, nu);
		}
	}

	public void testDenseColumnTransMult() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new DenseColumnMatrix(n, m);
			ElementalAccessVector x = new DenseVector(n),
				y = new DenseVector(m),
				z = new DenseVector(m);

			transColumnCheck(A, x, y, z, nu);
		}
	}

	public void testSparseRowColumnTransMult() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new SparseRowColumnMatrix(n, m, nu);
			ElementalAccessVector x = new DenseVector(n),
				y = new DenseVector(m),
				z = new DenseVector(m);

			transRowCheck(A, x, y, z, nu);
		}
	}

	public void testSparseRowTransMult() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new SparseRowMatrix(n, m, nu);
			ElementalAccessVector x = new DenseVector(n),
				y = new DenseVector(m),
				z = new DenseVector(m);

			transRowCheck(A, x, y, z, nu);
		}
	}

	public void testSparseColumnRowTransMult() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new SparseColumnRowMatrix(n, m, nu);
			ElementalAccessVector x = new DenseVector(n),
				y = new DenseVector(m),
				z = new DenseVector(m);

			transColumnCheck(A, x, y, z, nu);
		}
	}

	public void testSparseColumnTransMult() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new SparseColumnMatrix(n, m, nu);
			ElementalAccessVector x = new DenseVector(n),
				y = new DenseVector(m),
				z = new DenseVector(m);

			transColumnCheck(A, x, y, z, nu);
		}
	}

	private void rowCheck(
		ElementalAccessMatrix A,
		ElementalAccessVector x,
		ElementalAccessVector y,
		ElementalAccessVector z,
		int nu) {
		double[][] Ad = Utilities.setAssembleRowMatrix(A, nu);
		check(A, x, y, z, Ad);
	}

	private void columnCheck(
		ElementalAccessMatrix A,
		ElementalAccessVector x,
		ElementalAccessVector y,
		ElementalAccessVector z,
		int nu) {
		double[][] Ad = Utilities.setAssembleColumnMatrix(A, nu);
		check(A, x, y, z, Ad);
	}

	private void transRowCheck(
		ElementalAccessMatrix A,
		ElementalAccessVector x,
		ElementalAccessVector y,
		ElementalAccessVector z,
		int nu) {
		double[][] Ad = Utilities.setAssembleRowMatrix(A, nu);
		transCheck(A, x, y, z, Ad);
	}

	private void transColumnCheck(
		ElementalAccessMatrix A,
		ElementalAccessVector x,
		ElementalAccessVector y,
		ElementalAccessVector z,
		int nu) {
		double[][] Ad = Utilities.setAssembleColumnMatrix(A, nu);
		transCheck(A, x, y, z, Ad);
	}

	private void check(
		ElementalAccessMatrix A,
		ElementalAccessVector x,
		ElementalAccessVector y,
		ElementalAccessVector z,
		double[][] Ad) {
		double[] xd = Utilities.setAssembleVector(x),
			yd = Utilities.setAssembleVector(y),
			zd = Utilities.setAssembleVector(z);
		Random r = new Random();
		double alpha = r.nextDouble(), beta = r.nextDouble();

		for (int i = 1; i <= threads; ++i) {
			State.setNumThreads(i);
			State.BLAS.multAdd(alpha, A, x, beta, y, z);
			multAdd(alpha, Ad, xd, beta, yd, zd);

			checkEqual(z, zd);
		}
	}

	private void transCheck(
		ElementalAccessMatrix A,
		ElementalAccessVector x,
		ElementalAccessVector y,
		ElementalAccessVector z,
		double[][] Ad) {
		double[] xd = Utilities.setAssembleVector(x),
			yd = Utilities.setAssembleVector(y),
			zd = Utilities.setAssembleVector(z);
		Random r = new Random();
		double alpha = r.nextDouble(), beta = r.nextDouble();

		for (int i = 1; i <= threads; ++i) {
			State.setNumThreads(i);
			State.BLAS.transMultAdd(alpha, A, x, beta, y, z);
			transMultAdd(alpha, Ad, xd, beta, yd, zd);

			checkEqual(z, zd);
		}
	}

	private void multAdd(
		double alpha,
		double[][] A,
		double[] x,
		double beta,
		double[] y,
		double[] z) {
		for (int i = 0; i < A.length; ++i) {
			double[] Arow = A[i];
			double val = 0.;

			for (int j = 0; j < Arow.length; ++j)
				val += Arow[j] * x[j];

			z[i] = alpha * val + beta * y[i];
		}
	}

	private void transMultAdd(
		double alpha,
		double[][] A,
		double[] x,
		double beta,
		double[] y,
		double[] z) {
		int n = x.length, m = y.length;

		// Perform transposition
		double[][] At = new double[m][n];
		for (int i = 0; i < m; ++i)
			for (int j = 0; j < n; ++j)
				At[i][j] = A[j][i];

		multAdd(alpha, At, x, beta, y, z);
	}

	private void checkEqual(ElementalAccessVector x, double[] data) {
		assertTrue(x.size() == data.length);
		for (int i = 0; i < data.length; ++i)
			assertEquals(x.getValue(i), data[i], 1e-12);
	}

	private int getNu(int n, int m, Random r) {
		return Math.min(n, Math.min(m, Utilities.getInt(numax, r)));
	}

}
