/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.test;

import java.util.Random;

import jmp.CoordinateMatrix;
import jmp.DenseColumnMatrix;
import jmp.DenseColumnRowMatrix;
import jmp.DenseRowColumnMatrix;
import jmp.DenseRowMatrix;
import jmp.ElementalAccessMatrix;
import jmp.SparseColumnMatrix;
import jmp.SparseColumnRowMatrix;
import jmp.SparseRowColumnMatrix;
import jmp.SparseRowMatrix;
import jmp.State;
import junit.framework.TestCase;

/**
 * Tests matrix transposition
 */
public class TransposeTest extends TestCase {

	private static int nmax = 100,
		mmax = 100,
		numax = 10,
		repeat = 3,
		threads = 4;

	/**
	 * Constructor for TransposeTest.
	 * @param arg0
	 */
	public TransposeTest(String arg0) {
		super(arg0);
	}

	private int getNu(int n, int m, Random r) {
		return Math.min(n, Math.min(m, Utilities.getInt(numax, r)));
	}

	public void testSparseRowColumnTranspose() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new SparseRowColumnMatrix(n, m, nu);
			double[][] Am = Utilities.setAssembleRowMatrix(A, nu);
			check(A, Am);
		}
	}

	public void testSparseRowTranspose() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new SparseRowMatrix(n, m, nu);
			double[][] Am = Utilities.setAssembleRowMatrix(A, nu);
			check(A, Am);
		}
	}

	public void testSparseColumnTranspose() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new SparseColumnMatrix(n, m, nu);
			double[][] Am = Utilities.setAssembleRowMatrix(A, nu);
			check(A, Am);
		}
	}

	public void testSparseColumnRowTranspose() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new SparseColumnRowMatrix(n, m, nu);
			double[][] Am = Utilities.setAssembleColumnMatrix(A, nu);
			check(A, Am);
		}
	}

	public void testDenseRowColumnTranspose() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new DenseRowColumnMatrix(n, m);
			double[][] Am = Utilities.setAssembleRowMatrix(A, nu);
			check(A, Am);
		}
	}

	public void testDenseRowTranspose() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new DenseRowMatrix(n, m);
			double[][] Am = Utilities.setAssembleRowMatrix(A, nu);
			check(A, Am);
		}
	}

	public void testDenseColumnTranspose() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new DenseColumnMatrix(n, m);
			double[][] Am = Utilities.setAssembleRowMatrix(A, nu);
			check(A, Am);
		}
	}

	public void testDenseColumnRowTranspose() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new DenseColumnRowMatrix(n, m);
			double[][] Am = Utilities.setAssembleRowMatrix(A, nu);
			check(A, Am);
		}
	}

	public void testCoordinateTranspose() {
		Random r = new Random();
		int n = Utilities.getInt(nmax, r),
			m = Utilities.getInt(mmax, r),
			nu = getNu(n, m, r);
		for (int i = 0; i < repeat; ++i) {
			ElementalAccessMatrix A = new CoordinateMatrix(n, m);
			double[][] Am = Utilities.setAssembleRowMatrix(A, nu);
			check(A, Am);
		}
	}

	private void check(ElementalAccessMatrix A, double[][] Am) {
		int n = A.numRows(), m = A.numColumns();

		test(A, new SparseRowMatrix(m, n), Am);
		test(A, new SparseColumnMatrix(m, n), Am);
		test(A, new DenseRowMatrix(m, n), Am);
		test(A, new DenseColumnMatrix(m, n), Am);
		test(A, new DenseRowColumnMatrix(m, n), Am);
		test(A, new DenseColumnRowMatrix(m, n), Am);
		test(A, new CoordinateMatrix(m, n), Am);
	}

	private void test(
		ElementalAccessMatrix A,
		ElementalAccessMatrix B,
		double[][] Am) {
		for (int i = 0; i < threads; ++i) {
			State.setNumThreads(i);
			State.BLAS.transpose(A, B);
			double[][] Bm =
				transpose(Am, new double[A.numColumns()][A.numRows()]);
			checkEqual(State.BLAS.getArrayCopy(B), Bm);
		}
	}

	private double[][] transpose(double[][] A, double[][] B) {
		int n = A.length, m = B.length;
		for (int i = 0; i < n; ++i)
			for (int j = 0; j < m; ++j)
				B[j][i] = A[i][j];
		return B;
	}

	private void checkEqual(double[][] A, double[][] B) {
		assertTrue(A.length == B.length);
		for (int i = 0; i < A.length; ++i) {
			assertTrue(A[i].length == B[i].length);
			for (int j = 0; j < A[i].length; ++j)
				assertEquals(B[i][j], A[i][j], 1e-12);
		}
	}

}
