/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.test;

import java.util.Random;

import jmp.ElementalAccessMatrix;
import jmp.ElementalAccessVector;
import jmp.Matrix;
import jmp.State;
import jmp.Vector;

/**
 * Utilities for the testers
 */
public class Utilities {

	private Utilities() {
	}

	/**
	 * Creates a double array of the given size
	 */
	public static double[] getDoubleArray(int n, Random r) {
		double[] arr = new double[n];
		for (int i = 0; i < arr.length; ++i)
			arr[i] = r.nextDouble();
		return arr;
	}

	/**
	 * Creates a double array of the given size
	 */
	public static double[][] getDoubleArray(int n, int m, Random r) {
		double[][] arr = new double[n][m];
		for (int i = 0; i < arr.length; ++i)
			for (int j = 0; j < arr[i].length; ++j)
				arr[i][j] = r.nextDouble();
		return arr;
	}

	/**
	 * Creates an integer array of the given size and maximum amplitude
	 */
	public static int[] getIntArray(int max, Random r) {
		int[] arr = new int[getInt(max, r)];
		for (int k = 0; k < arr.length; ++k) {
			arr[k] = getInt(max, r);
			if (arr[k] >= max)
				throw new ArrayIndexOutOfBoundsException();
		}
		return arr;
	}

	/**
	 * Returns an integer between zero (inclusive) and max (exclusive)
	 */
	public static int getInt(int max, Random r) {
		return (int) (r.nextDouble() * ((double) max));
	}

	/**
	 * Dense representation of the matrix using getValue
	 */
	public static double[][] getElMatrix(ElementalAccessMatrix A) {
		double[][] ret = new double[A.numRows()][A.numColumns()];
		for (int i = 0; i < A.numRows(); ++i)
			for (int j = 0; j < A.numColumns(); ++j)
				ret[i][j] = A.getValue(i, j);
		return ret;
	}

	/**
	 * Dense representation of the matrix using getValues
	 */
	public static double[][] getMatrix(ElementalAccessMatrix A) {
		int[] row = new int[A.numRows()], col = new int[A.numColumns()];
		for (int i = 0; i < row.length; ++i)
			row[i] = i;
		for (int i = 0; i < col.length; ++i)
			col[i] = i;
		return A.getValues(row, col);
	}

	/**
	 * Dense representation of the matrix using getArrayCopy
	 */
	public static double[][] getMatrixCopy(Matrix A) {
		return State.BLAS.getArrayCopy(A);
	}

	/**
	 * Dense representation of the vector using getValue
	 */
	public static double[] getElVector(ElementalAccessVector x) {
		double[] ret = new double[x.size()];
		for (int i = 0; i < ret.length; ++i)
			ret[i] = x.getValue(i);
		return ret;
	}

	/**
	 * Dense representation of the vector using getValues
	 */
	public static double[] getVector(ElementalAccessVector x) {
		int[] ind = new int[x.size()];
		for (int i = 0; i < ind.length; ++i)
			ind[i] = i;
		return x.getValues(ind);
	}

	/**
	 * Dense representation of the vector using getArrayCopy
	 */
	public static double[] getVectorCopy(Vector x) {
		return State.BLAS.getArrayCopy(x);
	}

	/**
	 * Row-wise assembly using setValue
	 * @param nu Maximum bandwidth on each row
	 * @return Dense representation (not a direct copy)
	 */
	public static double[][] setAssembleRowMatrix(
		ElementalAccessMatrix A,
		int nu) {
		int n = A.numRows(), m = A.numColumns();
		double[][] data = new double[n][m];
		Random r = new Random();

		for (int i = 0; i < n; ++i)
			for (int j = 0; j < nu; ++j) {
				int ind = Utilities.getInt(m, r);
				double entry = r.nextDouble();
				data[i][ind] = entry;
				A.setValue(i, ind, entry);
			}

		return data;
	}

	/**
	 * Column-wise assembly using set
	 * @param nu Maximum bandwidth on each column
	 * @return Dense representation (not a direct copy)
	 */
	public static double[][] setAssembleColumnMatrix(
		ElementalAccessMatrix A,
		int nu) {
		int n = A.numRows(), m = A.numColumns();
		double[][] data = new double[n][m];
		Random r = new Random();

		// Construct the matrix
		for (int i = 0; i < m; ++i)
			for (int j = 0; j < nu; ++j) {
				int ind = Utilities.getInt(n, r);
				double entry = r.nextDouble();
				data[ind][i] = entry;
				A.setValue(ind, i, entry);
			}

		return data;
	}

	/**
	 * Assembly using set
	 * @return Dense representation (not a direct copy)
	 */
	public static double[] setAssembleVector(ElementalAccessVector x) {
		int n = x.size();
		double[] data = new double[n];
		Random r = new Random();

		for (int i = 0; i < n; ++i) {
			double entry = r.nextDouble();
			data[i] = entry;
			x.setValue(i, entry);
		}
		return data;
	}

	/**
	 * Row-wise assembly using setValues
	 * @param nu Maximum bandwidth on each row
	 * @return Dense representation (not a direct copy)
	 */
	public static double[][] setsAssembleRowMatrix(
		ElementalAccessMatrix A,
		int nu) {
		int n = A.numRows(), m = A.numColumns();
		double[][] data = new double[n][m];
		Random r = new Random();

		for (int i = 0; i < n; ++i) {
			int[] row = Utilities.getIntArray(n, r),
				col = Utilities.getIntArray(nu, r);
			double[][] entry =
				Utilities.getDoubleArray(row.length, col.length, r);
			for (int ii = 0; ii < row.length; ++ii)
				for (int jj = 0; jj < col.length; ++jj)
					data[row[ii]][col[jj]] = entry[ii][jj];
			A.setValues(row, col, entry);
		}

		return data;
	}

	/**
	 * Row-wise assembly using addValues
	 * @param nu Maximum bandwidth on each row
	 * @return Dense representation (not a direct copy)
	 */
	public static double[][] addsAssembleRowMatrix(
		ElementalAccessMatrix A,
		int nu) {
		int n = A.numRows(), m = A.numColumns();
		double[][] data = new double[n][m];
		Random r = new Random();

		for (int i = 0; i < n; ++i) {
			int[] row = Utilities.getIntArray(n, r),
				col = Utilities.getIntArray(nu, r);
			double[][] entry =
				Utilities.getDoubleArray(row.length, col.length, r);
			for (int ii = 0; ii < row.length; ++ii)
				for (int jj = 0; jj < col.length; ++jj)
					data[row[ii]][col[jj]] += entry[ii][jj];
			A.addValues(row, col, entry);
		}

		return data;
	}

	/**
	 * Row-wise assembly using add
	 * @param nu Maximum bandwidth on each row
	 * @return Dense representation (not a direct copy)
	 */
	public static double[][] addAssembleRowMatrix(
		ElementalAccessMatrix A,
		int nu) {
		int n = A.numRows(), m = A.numColumns();
		double[][] data = new double[n][m];
		Random r = new Random();

		for (int i = 0; i < n; ++i)
			for (int j = 0; j < nu; ++j) {
				int ind = Utilities.getInt(m, r);
				double entry = r.nextDouble();
				data[i][ind] += entry;
				A.addValue(i, ind, entry);
			}

		return data;
	}

	/**
	 * Column-wise assembly using addValues
	 * @param nu Maximum bandwidth on each column
	 * @return Dense representation (not a direct copy)
	 */
	public static double[][] addsAssembleColumnMatrix(
		ElementalAccessMatrix A,
		int nu) {
		int n = A.numRows(), m = A.numColumns();
		double[][] data = new double[n][m];
		Random r = new Random();

		for (int i = 0; i < m; ++i) {
			int[] row = Utilities.getIntArray(nu, r),
				col = Utilities.getIntArray(m, r);
			double[][] entry =
				Utilities.getDoubleArray(row.length, col.length, r);
			for (int ii = 0; ii < row.length; ++ii)
				for (int jj = 0; jj < col.length; ++jj)
					data[row[ii]][col[jj]] += entry[ii][jj];
			A.addValues(row, col, entry);
		}

		return data;
	}

	/**
	 * Column-wise assembly using setValues
	 * @param nu Maximum bandwidth on each column
	 * @return Dense representation (not a direct copy)
	 */
	public static double[][] setsAssembleColumnMatrix(
		ElementalAccessMatrix A,
		int nu) {
		int n = A.numRows(), m = A.numColumns();
		double[][] data = new double[n][m];
		Random r = new Random();

		for (int i = 0; i < m; ++i) {
			int[] row = Utilities.getIntArray(nu, r),
				col = Utilities.getIntArray(m, r);
			double[][] entry =
				Utilities.getDoubleArray(row.length, col.length, r);
			for (int ii = 0; ii < row.length; ++ii)
				for (int jj = 0; jj < col.length; ++jj)
					data[row[ii]][col[jj]] = entry[ii][jj];
			A.setValues(row, col, entry);
		}

		return data;
	}

	/**
	 * Column-wise assembly using add
	 * @param nu Maximum bandwidth on each column
	 * @return Dense representation (not a direct copy)
	 */
	public static double[][] addAssembleColumnMatrix(
		ElementalAccessMatrix A,
		int nu) {
		int n = A.numRows(), m = A.numColumns();
		double[][] data = new double[n][m];
		Random r = new Random();

		for (int i = 0; i < m; ++i)
			for (int j = 0; j < nu; ++j) {
				int ind = Utilities.getInt(n, r);
				double entry = r.nextDouble();
				data[ind][i] += entry;
				A.addValue(ind, i, entry);
			}

		return data;
	}

	/**
	 * Assembly using addValue
	 * @param nu Maximum number of entries
	 * @return Dense representation (not a direct copy)
	 */
	public static double[] addAssembleVector(ElementalAccessVector x, int nu) {
		int n = x.size();
		double[] data = new double[n];
		Random r = new Random();

		for (int i = 0; i < nu; ++i) {
			int ind = Utilities.getInt(n, r);
			double entry = r.nextDouble();
			data[ind] += entry;
			x.addValue(ind, entry);
		}

		return data;
	}

	/**
	 * Assembly using setValue
	 * @param nu Maximum number of entries
	 * @return Dense representation (not a direct copy)
	 */
	public static double[] setAssembleVector(ElementalAccessVector x, int nu) {
		int n = x.size();
		double[] data = new double[n];
		Random r = new Random();

		for (int i = 0; i < nu; ++i) {
			int ind = Utilities.getInt(n, r);
			double entry = r.nextDouble();
			data[ind] = entry;
			x.setValue(ind, entry);
		}

		return data;
	}

	/**
	 * Assembly using addValues
	 * @param nu Maximum number of entries
	 * @return Dense representation (not a direct copy)
	 */
	public static double[] addsAssembleVector(
		ElementalAccessVector x,
		int nu) {
		int n = x.size();
		double[] data = new double[n];
		Random r = new Random();

		int[] ind = Utilities.getIntArray(nu, r);
		double[] arr = Utilities.getDoubleArray(ind.length, r);
		for (int i = 0; i < ind.length; ++i)
			data[ind[i]] += arr[i];
		x.addValues(ind, arr);
		return data;
	}

	/**
	 * Assembly using setValues
	 * @param nu Maximum number of entries
	 * @return Dense representation (not a direct copy)
	 */
	public static double[] setsAssembleVector(
		ElementalAccessVector x,
		int nu) {
		int n = x.size();
		double[] data = new double[n];
		Random r = new Random();

		int[] ind = Utilities.getIntArray(nu, r);
		double[] arr = Utilities.getDoubleArray(ind.length, r);
		for (int i = 0; i < ind.length; ++i)
			data[ind[i]] = arr[i];
		x.setValues(ind, arr);
		return data;
	}

}
