/*
 * Copyright (C) 2003 Bjrn-Ove Heimsund
 * 
 * This file is part of JMP.
 * 
 * JMP is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * JMP is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JMP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package jmp.util;

import EDU.oswego.cs.dl.util.concurrent.FJTask;
import EDU.oswego.cs.dl.util.concurrent.FJTaskRunner;
import jmp.State;

/**
 * Supports parallel operations by multi-threading. It is based on the <a
 * href="http://gee.cs.oswego.edu/dl/classes/EDU/oswego/cs/dl/util/concurrent">util.concurrent</a>
 * package by Doug Lea.
 * <p>
 * The basic usage is best illustrated by an example. Suppose one wishes
 * to parallelize the method
 * <p>
 * <pre>
 * void compute() {
 *   for (int i = 0; i &lt; length; ++i)
 *     doComputation(i);
 * }</pre>
 * <p>
 * ParallelWorker can be used for three things, namely
 * <ol>
 * <li>Partition the loop-indices into disjoint pieces</li>
 * <li>Serve as a base-class to create worker-threads</li>
 * <li>Execute the threads, and await their termination</li>
 * </ol>
 * The preceding example is parallelized as follows:
 * <p>
 * <pre>
 * void compute(int numThreads) {
 *   // Set the number of threads to use
 *   State.setNumThreads(numThreads);
 * 
 *   // Create worker threads
 *   int[] part = ParallelWorker.partition(length);
 *   ParallelWorker[] thread = new ComputeThread[numThreads];
 *   for (int i = 0; i &lt; numThreads; ++i)
 *     thread[i] = new ComputeThread(part[i], part[i+1]);
 *
 *   // Execute
 *   ParallelWorker.runThreads(thread);
 * }
 * 
 * class ComputeThread extends ParallelWorker {
 *   public ComputeThread(int start, int stop) {
 *     super(start, stop);
 *   }
 *   public void run() {
 *     for (int i = start; i &lt; stop; ++i)
 *       doComputation(i);
 *   }
 * }</pre>
 * <p>
 * Of related interest is State which keeps track of the numer of threads.
 * @see jmp.State
 */
public abstract class ParallelWorker extends FJTask {

	/**
	 * Work indices, set by the constructor.
	 */
	protected int start, stop;

	/**
	 * Sets up this thread to work in the specified interval. The indices
	 * can be used in the run-method.
	 * @param start Work index to start from, inclusive
	 * @param stop Work index to stop at, exclusive
	 */
	public ParallelWorker(int start, int stop) {
		this.start = start;
		this.stop = stop;
	}

	/**
	 * Runs the given threads, and waits for them to finish. The tasks must be
	 * independent on each other, otherwise deadlocks may occur. However, it is
	 * possible to use nested parallelization, that is, one thread can create
	 * subthreads and await their termination.
	 * @param thread Tasks to execute
	 */
	public final static void runThreads(ParallelWorker[] thread) {
		try {
			// If the invoking thread is an FJTaskRunner, the new threads
			// are put on top of the queue (LIFO style) to avoid deadlocks.
			// Otherwise, the usual FIFO style is used
			if ((Runnable) Thread.currentThread() instanceof FJTaskRunner) {
				FJTask task = new FJTask.Par(thread);
				task.fork();
				task.join();
			} else
				State.runner.invoke(FJTask.par(thread));
		} catch (InterruptedException e) {
			throw new RuntimeException(e.toString());
		}
	}

	/**
	 * Creates a disjoint partition. Useful to setup parallel operations.
	 * @param start First index value, inclusive
	 * @param stop Last index value, exclusive
	 * @param size Size of partition
	 * @return Disjoint partition of length <code>size + 1</code>
	 */
	public final static int[] partition(int start, int stop, int size) {
		int[] part = new int[size + 1];

		int local =
			(int) Math.round(((double) (stop - start)) / ((double) size));

		part[0] = start;
		for (int i = 1; i < size; ++i)
			part[i] = Math.min(part[i - 1] + local, stop);
		part[size] = stop;

		return part;
	}

	/**
	 * Creates a disjoint partition. Useful to setup parallel operations.
	 * @param start First index value, inclusive
	 * @param stop Last index value, exclusive
	 * @return Disjoint partition of length <code>numThreads + 1</code>
	 */
	public final static int[] partition(int start, int stop) {
		return partition(start, stop, State.numThreads);
	}

	/**
	 * Creates a disjoint partition. Useful to setup parallel operations.
	 * @param length Number of elements to partition
	 * @return Disjoint partition of length <code>numThreads + 1</code>
	 */
	public final static int[] partition(int length) {
		return partition(0, length);
	}

}
